#----------------------------------------------------------
# cshelper.py
# Contains supporting functions for mission scripts.
#----------------------------------------------------------

import sys

if __name__ != '__main__':
    import js
    import ff
    import event
    import ai
    import goal
    import random
    import types
    import datfiles

    from js import *
    from ff import *
    from event import *
    from ai import *
    from goal import *
    from random import *
    from types import *
    from prestige import getPrestige

print 'loading cshelper ...'
#--------------------
# cutscenes
#--------------------

class CutscenePlayer:
    def __init__(self):
        self.csList = []
        self.stepIndex = 0
        self.playing = 0
        self.trivial = 0
        self.csQueue = []
        
    def onStepCB(self):
        # sanity checks to make sure we aren't already off the end
        if len(self.csList) == 0:
            return
        if self.stepIndex >= len(self.csList):
            return
        if not self.playing:
            return
        print "Step %d of %d" % (self.stepIndex+1, len(self.csList))
        lineIndex = 0
        for step in self.csList[self.stepIndex]:
            lineIndex = lineIndex + 1
            if step != "blank":
                try:
                    exec step
                except:
                    print "exception happened in cutscene playback of line %d in step %d" % (lineIndex,self.stepIndex+1)
                    print sys.exc_type, sys.exc_value
                    print step
                    endCS()
                    self.end
                        
        # catch going off the end of the list                        
        self.stepIndex = self.stepIndex + 1
        if (self.stepIndex >= self.csLength):
            self.end()
       
    def end(self):
        self.playing = 0
        print "cshelper: Cutscene [" + self.name + "] has ended"
        if (len(self.csQueue) > 0):
            nextCS = self.csQueue[0][0]
            nextName = self.csQueue[0][1]
            nextPriority = self.csQueue[0][2]
            nextTrivial = self.csQueue[0][3]
            del self.csQueue[0]
            self.play(nextCS, nextName, nextPriority, nextTrivial)
                
    def onEcho(self):
        print "onEcho"
        
    def play(self, csArg, name, priority, trivial = 0, interrupt = 0):
        if self.isPlaying():
            if name != "unknown" and self.isPlayingName(name):
                print "cshelper: cutscene [" + name + "] already playing"
                return
            while self.isQueued():
                print 'checking queue'
                cs = self.csQueue[-1]
                queuedPriority = cs[2]
                if (not interrupt and queuedPriority<priority) or (interrupt and queuedPriority<=priority):
                    print "cshelper: Bumping cutscene [" + cs[1] + "] from queue for [" + name + "]"
                    del self.csQueue[-1]
                else:
                    print 'nothing to bump from queue'
                    break
            if not self.isQueued() and ((not interrupt and self.priority < priority) or (interrupt and self.priority<=priority)):
                print "cshelper: Bumping playing cutscene [" + self.name + "] for [" + name + "]"
                js.Event_CancelSinks('onStepCB')
                self.end()
            else:
                print "cshelper: Queueing cutscene [" + name + "]"
                self.queue(csArg, name, priority, trivial)
                return
        print "cshelper: Now playing cutscene [" + name + "]"
        self.stepIndex = 0
        self.csLength = len(csArg)
        self.csList = csArg
        self.playing = 1
        self.name = name
        self.trivial = trivial
        self.priority = priority
        self.onStepCB()

    def queue(self, csArg, name, priority, trivial):
        self.csQueue.append([csArg, name, priority, trivial])
        
    def isQueued(self):
        return len(self.csQueue)>0
        
    def isPlaying(self):
        return self.playing

    def isPlayingName(self, name):
        if isPlaying() and self.name == name:
            return 1
        for cutscene in self.csQueue:
            if cutscene[1] == name:
                return 1
        return 0
        
    def getName(self):
        if isPlaying():
            return self.name
        return ''
    
    def isTrivial(self):
        return self.playing and self.trivial
    
csHelper = CutscenePlayer()

# note: trivial not implemented currently
def play(csArg, name = "unknown", priority = 1, trivial = 0, interrupt = 0):
    print "cshelper: Cutscene [" + name + "] submitted for playing"
    csHelper.play(csArg, name, priority, trivial, interrupt)

def isPlaying():
    return csHelper.isPlaying()

def null():
    js.Event_RegisterSink(js.EVENT_TIMER, "onStepCB", '', '', 0.01)

def onCancelledCS():
    print "cutscene has been cancelled"
    csHelper.end()

def onStepCB(e):
    csHelper.onStepCB()

def startCS():
    js.Camera_SetWidescreen(1)
    startCSNormalScreen()

def startCSNormalScreen():
    cancelHeroOrders()
    normalizeHeroes()
    js.UI_Enable(0)
    AI_EnableDynamicSpeech(0)

def endCS():
    js.Camera_SetWidescreen(0)
    js.UI_Enable(1)
    js.Camera_Release()
    AI_EnableDynamicSpeech(1)

def isCSPlaying():
    return js.Camera_IsWidescreen()

def endBriefingCS():
    endCS()
    js.UI_Briefing_Done()

#----------------
# effects
#----------------
def playEffect(effect, pos = (0,0,0), orientation = 0, fn = "", flags = 0, user = 0, key = '', keyFn = ''):
    if key == '' or keyFn == '':
        return Object_PlayEffect(Object_GetTerrain(), effect, fn, flags, user, "", pos, orientation)
    else:
        return Object_PlayEffect(Object_GetTerrain(), effect, fn, flags, user, "", pos, orientation, (0, 0, 1), key, keyFn)

def playEffectMarker(effect, marker, orientation = 0, fn = "", flags = 0, user = 0, key = '', keyFn = ''):
    if key == '' or keyFn == '':
        return Object_PlayEffect(Object_GetTerrain(), effect, fn, flags, user, "", Get_MarkerPos(marker), orientation)
    else:
        return Object_PlayEffect(Object_GetTerrain(), effect, fn, flags, user, "", Get_MarkerPos(marker), orientation, (0, 0, 1), key, keyFn)

#----------------
# lighting
#----------------

def sunRot(rotation):
    params = Mission_GetSunlight()
    Mission_SetSunlight(rotation, params[1], params[2], params[3], params[4])

def sunTilt(tilt):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], tilt, params[2], params[3], params[4])

def sunBright(bright):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], bright, params[3], params[4])

def sunAmbient(ambient):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], params[2], ambient, params[4])

def sunColor(r, g, b):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], params[2], params[3], (r,g,b))

def sunRed(r):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], params[2], params[3], (r, params[4][1], params[4][2]))
    
def sunGreen(g):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], params[2], params[3], (params[4][0], g, params[4][2]))

def sunBlue(b):
    params = Mission_GetSunlight()
    Mission_SetSunlight(params[0], params[1], params[2], params[3], (params[4][0], params[4][1], b))

def shadowRot(rotation):
    params = Mission_GetShadows()
    Mission_SetShadows(rotation, params[1], params[2])

def shadowTilt(tilt):
    params = Mission_GetShadows()
    Mission_SetShadows(params[0], tilt, params[2])

def shadowColor(r,g,b):
    params = Mission_GetShadows()
    Mission_SetShadows(params[0], params[1], (r,g,b))

def shadowRed(r):
    params = Mission_GetShadows()
    Mission_SetShadows(params[0], params[1], (r, params[2][1], params[2][2]))

def shadowGreen(g):
    params = Mission_GetShadows()
    Mission_SetShadows(params[0], params[1], (params[2][0], g, params[2][2]))

def shadowBlue(b):
    params = Mission_GetShadows()
    Mission_SetShadows(params[0], params[1], (params[2][0], params[2][1], b))
    
#----------------
# heroes
#----------------

def cancelHeroOrders():
    heroes = getAllHeroes()
    for hero in heroes:
        js.AI_NullController(hero)

def reviveHeroes():
    heroes = getAllHeroes()
    for hero in heroes:
        revive(hero)
        
def normalizeHeroes():
    heroes = getAllHeroes()
    for hero in heroes:
        normalize(hero)

#------------------
# is functions
#------------------

def isHero(object):
    return (Object_GetClass(object) & js.OC_CONTROLLABLE)

def isCivilian(object):
    return (Object_GetClass(object) & js.OC_CIVILIAN)

def isMaleCivilian(object):
    return Object_GetTemplate(object) == 'civilian_male'

def isFemaleCivilian(object):
    return Object_GetTemplate(object) == 'civilian_female'

def isPowerup(object):
    return Object_GetTemplate(object) == 'powerup_pp' or Object_GetTemplate(object) == 'powerup_ep' or Object_GetTemplate(object) == 'powerup_cp' or Object_GetTemplate(object) == 'powerup_hp'

def isHeroTemplate(template):
    return Template_Exists(template) and Template_GetClassString(template) == 'GAME_OBJ_HERO'

def isMinionTemplate(template):
    return Template_Exists(template) and Template_GetClassString(template) == 'GAME_OBJ_MINION'

def isVillainTemplate(template):
    return Template_Exists(template) and Template_GetClassString(template) == 'GAME_OBJ_VILLIAN'

def isPoliceTemplate(template):
    return Template_Exists(template) and Template_GetClassString(template) == 'GAME_OBJ_POLICE'

def isCivilianTemplate(template):
    return Template_Exists(template) and Template_GetClassString(template) == 'GAME_OBJ_CIVILIAN'

def isCharacterTemplate(template):
    return isHeroTemplate(template) or isVillainTemplate(template) or isMinionTemplate(template) or isPoliceTemplate(template) or isCivilianTemplate(template)

def isDead(object):
    if Object_Exists(object) and Object_GetAttr(object, 'health') == 0:
        return 1
    else:
        return 0

def getRandomLiveHeroName():
	hero_list = getAllHeroes()
	liveheroes = ['nullhero']
	for a_hero in hero_list:
		if Object_IsAlive(a_hero):
			liveheroes.append(Object_GetTemplate(a_hero))
	temp = randint(1,len(liveheroes)-1)
	return liveheroes[temp]
    

#----------------------
# get functions
#----------------------

# Note this only gets dynamic template matches
def getAll(template):
    objects = js.Mission_GetDynamicObjects()
    found = []
    for object in objects:
        if (js.Object_GetTemplate(object) == template):
            found.append(object)
    return found

def getAllByPrefix(prefix):
    objects = Mission_GetObjects()  # *** stumpy changed from original
    found = []
    for object in objects:
        if (object[:len(prefix)] == prefix):
            found.append(object)
    return found

# Note: name is a little misleading as it gets dynamic objects, too, if they are moving.
def getAllStatic(template):
    objects = js.Mission_GetObjects()
    found = []
    for object in objects:
        if (Object_GetTemplate(object) == template):
            found.append(object)
    return found
    
def getAllCivilians():
    objects = js.Mission_GetDynamicObjects()
    civilians = []
    for object in objects:
        if isCivilian(object):
            civilians.append(object)
    return civilians

def getAllHeroes():
    objects = js.Mission_GetDynamicObjects()
    heroes = []
    for object in objects:
        if isHero(object):
            heroes.append(object)
    return heroes

def isBaddie(object):
    return Object_GetClass(object) & (js.OC_MINION | js.OC_VILLIAN)
    
def getAllBaddies(checkLiving = 1, template = 'none'):
    objects = js.Mission_GetDynamicObjects()
    baddies = []
    for object in objects:
        if not isBaddie(object):
            continue
        if not ((template == 'none') or (js.Object_GetTemplate(object) == template)):
            continue
        if (checkLiving and not Object_IsAlive(object)):
            continue
        baddies.append(object)
    return baddies

def distance2D(pos1, pos2):
    vector = Vector_Sub(pos1, pos2)
    return sqrt(vector[0] * vector[0] + vector[1] * vector[1])

#  modified to get get baddy nearest a position if object is a length 3 tuple
def getNearBaddies(object, radius, checkLiving = 1, template = 'none'):
    if ( type(object) == TupleType ) and ( len(object) == 3 ):
        pos2 = object
    else:
        pos2 = Get_ObjectPos(object)
    baddies = getAllBaddies(checkLiving, template)
    results = []
    for baddie in baddies:
        pos1 = Get_ObjectPos(baddie)
        vector = Vector_Sub(pos1, pos2)
        distance = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2])
        if distance <= radius:
            results.append(baddie)
    return results

#  modified to get get hero nearest a position if villian is a length 3 tuple
def getNearestHero(villian):
    minDistance = pow(10,6)
    if (type(villian) == TupleType) and ( len(villian) == 3 ):
        pos1 = villian
    else:
        pos1 = Get_ObjectPos(villian)
    hero = ''
    objects = js.Mission_GetDynamicObjects()
    for object in objects:
        objclass = js.Object_GetClass(object)
        if (objclass & js.OC_CONTROLLABLE) and Object_IsAlive(object):
            pos2 = Get_ObjectPos(object)
            vector = Vector_Sub(pos1, pos2)
            distance = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2])
            if ( distance < minDistance ) and ( distance > 0.01 ) :     # don't return self
                minDistance = distance
                hero = object
    return hero
    
#  modified to get get minion nearest a position if hero is a length 3 tuple
def getNearestMinion(hero):
    if (type(hero) == TupleType) and ( len(hero) == 3 ):
        pos1 = hero
    else:
        pos1 = Get_ObjectPos(hero)
    minDistance = pow(10,6)
    minion = ''
    objects = js.Mission_GetDynamicObjects()
    for object in objects:
        objclass = js.Object_GetClass(object)
        if ((objclass & js.OC_MINION) or (objclass & js.OC_VILLIAN)) and Object_IsAlive(object):
            pos2 = Get_ObjectPos(object)
            vector = Vector_Sub(pos1, pos2)
            distance = sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2])
            if distance < minDistance:
                minDistance = distance
                minion = object
    if len(minion) == 0:
        print "GetNearestMinion(): There are no minions or villains left alive!"
    return minion

#--------------
# speaking
#--------------

def speakInternal(object, tag, fn = '', user = 0, isTemplate = 0, pan = 0.5, position = -1, telepathy = 2, forceNonMission = 0, interrupt = 0):
    print 'speakInternal'
    print forceNonMission
    if (csHelper.isTrivial()):
        sayDynamic(object, tag)
    else:
        if interrupt:
            AI_CancelAllSpeech(CHANNEL_SCRIPT)
        AI_SayPortraitSpeech(object, tag, fn, user, isTemplate, pan, position, telepathy, -1, forceNonMission)
    
def speak(object, tag, position = -1, forceNonMission = 0, interrupt = 0):
    print forceNonMission
    if js.Object_Exists(object):
        if js.Object_IsAlive(object):
            speakInternal(object, tag, fn = 'onStepCB', position = position, forceNonMission = forceNonMission, interrupt = interrupt)
        else:
            null()
    elif isHeroTemplate(object):
        # if not a minion template, try to find a hero with the same name
        hero = findHero(object)
        if hero != '':
            if Object_IsAlive(hero):
                speakInternal(hero, tag, fn = 'onStepCB')
            else:
                null()
        else:
            # otherwise, speak as template (i.e. via telepathy)
            speakInternal(object, tag, fn = 'onStepCB', isTemplate = 1, position = position, forceNonMission = forceNonMission, interrupt = interrupt)
    elif object == 'narrator' or object[-5:] == '_dark':
        speakInternal(object, tag, fn = 'onStepCB', isTemplate = 1, position = position, forceNonMission = forceNonMission, interrupt = interrupt)
    else:
        null()

def speakList(speakList, optional = 0):
    for speakInstance in speakList[:-1]:
        speaker = speakInstance[0]
        tag = speakInstance[1]
        if js.Object_Exists(speaker):
            object = speaker
        else:
            object = findHero(speaker)
        if object != '' and js.Object_IsAlive(object):
            speakInternal(object, tag, fn = 'onStepCB')
            return
    speaker = speakList[-1][0]
    tag = speakList[-1][1]
    if optional:
        optionalSpeak(speaker, tag)
    else:
        speak(speaker, tag)

def speakListNoCB(speakList):
    for speakInstance in speakList[:-1]:
        speaker = speakInstance[0]
        tag = speakInstance[1]
        if js.Object_Exists(speaker) and js.Object_IsAlive(speaker):
            speakInternal(speaker, tag)
            return
    speaker = speakList[-1][0]
    tag = speakList[-1][1]
    speakNoCB(speaker, tag)
    
def speakNoCB(object, tag):
    if js.Object_Exists(object):
        if js.Object_IsAlive(object):
            speakInternal(object, tag)
    elif isHeroTemplate(object):
        # if not a minion template, try to find a hero with the same name
        hero = findHero(object)
        if hero != '':
            if Object_IsAlive(hero):
                speakInternal(hero, tag)
        else:
            # otherwise, speak as template (i.e. via telepathy)
            speakInternal(object, tag, isTemplate = 1)
    elif object == 'narrator' or object[-5:] == '_dark':
        speakInternal(object, tag, isTemplate = 1)

def optionalSpeak(template, tag):
    if js.Object_Exists(template):
        object = template
    else:
        object = findHero(template)
    if object and js.Object_IsAlive(object):
        speakInternal(object, tag, fn = 'onStepCB')
        return object
    null()
    return ''

def optionalSpeakNoCB(template, tag):
    if js.Object_Exists(template):
        object = template
    else:
        object = findHero(template)
    if object and js.Object_IsAlive(object):
        speakInternal(object, tag)
        return object
    return ''

# template param is ignored!
def forceSpeak(object, tag, telepathy = 2, template = ''):
    if js.Object_Exists(object):
        speakInternal(object, tag, fn = 'onStepCB', telepathy = telepathy)
    elif isHeroTemplate(object):
        hero = findHero(object)
        if hero != '':
            speakInternal(hero, tag, fn = 'onStepCB', telepathy = telepathy)
        else:
            speakInternal(object, tag, fn = 'onStepCB', isTemplate = 1, telepathy = telepathy)
    elif object == 'narrator' or object[-5:] == '_dark' or isCharacterTemplate(object):
        speakInternal(object, tag, fn = 'onStepCB', isTemplate = 1, telepathy = telepathy)
    else:
        speakInternal("", tag, fn = 'onStepCB', isTemplate = 1)

def forceSpeakNoCB(object, tag, telepathy = 1):
    if js.Object_Exists(object):
        speakInternal(object, tag, isTemplate = 0, telepathy = telepathy)
    else:
        speakInternal(object, tag, isTemplate = 1, telepathy = telepathy)

def speakTo(speaker, listener, tag):
    turnTo(speaker, listener)
    speak(speaker, tag)
    
def speakToNoCB(speaker, listener, tag):
    turnTo(speaker, listener)
    speakNoCB(speaker, tag)

def sayDynamic(object, speechTag, portrait = 1):
    speaker = ''
    if js.Object_Exists(object):
        speaker = object
    elif isHeroTemplate(object):
        speaker = findHero(object)
    if speaker != '' and js.Object_IsAlive(speaker):
        if portrait:
            AI_SayPortraitSpeech(speaker, speechTag, 'onStepCB', 0, 0, 0.5, -1, 2, -1)
        else:
            AI_SayPortraitSpeech(speaker, speechTag, 'onStepCB', 0, 0, 0.0, -1, 2, 0)
    else:
        null()

def sayDynamicNoCB(object, speechTag, portrait = 0):
    speaker = ''
    if js.Object_Exists(object):
        speaker = object
    elif isHeroTemplate(object):
        speaker = findHero(object)
    if speaker != '' and js.Object_IsAlive(speaker):
        if portrait:
            AI_SayPortraitSpeech(speaker, speechTag, '', 0, 0, 0.5, -1, 2, -1)
        else:
            AI_SayPortraitSpeech(speaker, speechTag, '', 0, 0, 0.0, -1, 2, 0)

def testSpeak(test, object, tag):
    if test:
        speak(object, tag)
    else:
        null()

#--------------
# animation
#--------------

def optionalAnimate(template, animation):
    if js.Object_Exists(template):
        object = template
    else:
        object = findHero(template)
    if object and js.Object_IsAlive(object):
        AI_Animate(object, animation, 'onStepCB')
        return object
    null()
    return ''
        
def optionalAnimateNoCB(template, animation):
    if js.Object_Exists(template):
        object = template
    else:
        object = findHero(template)
    if object and js.Object_IsAlive(object):
        AI_Animate(object, animation, '')
        return object
    return ''
        
#--------------
# looking
#--------------

def look(thing, zoom = 0, heading = 0, time = 0, fn = ''):
    if js.Object_Exists(thing):
        lookAtObject(thing, zoom, heading, time, fn)
    else:
        lookAtMarker(thing, zoom, heading, time, fn)

def lookAtObject(object, zoom = 0, heading = 0, time = 0, fn = ''):
    if not js.Object_Exists(object):
        return
    
    if time == 0:
        js.Camera_LookAtObject(object, zoom, 0, heading, 0, js.CPM_SNAP, js.CA_MOVE, fn)
    else:
        js.Camera_LookAtObject(object, zoom, 0, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)

def lookAtMarker(marker, zoom = 0, heading = 0, time = 0, fn = ''):
    if time == 0:
        js.Camera_LookAtMarker(marker, zoom, 0, heading, time, js.CPM_SNAP, js.CA_MOVE, fn)
    else:
        js.Camera_LookAtMarker(marker, zoom, 0, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)
        
def lookAtPos(pos, zoom = 0, pitch = 0, heading = 0, time = 0, fn = ''):
    if time == 0:
        js.Camera_LookAtPos(pos, zoom, pitch, heading, 0, js.CPM_SNAP, js.CA_MOVE, fn)
    else:
        js.Camera_LookAtPos(pos, zoom, pitch, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)
    
def lookTo(thing, fn = '', time = 3):
    if js.Object_Exists(thing):
        lookToObject(thing, fn, time)
    else:
        lookToMarker(thing, fn, time)

def lookToObject(object, fn = '', time = 3):
    js.Camera_LookAtObject(object, 0, 0, 0, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)

def lookToMarker(marker, fn = '', time = 3):
    js.Camera_LookAtMarker(marker, 0, 0, 0, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)

def noLook():
    Camera_Release()

def findLastHero():
    heroes = getAllHeroes()
    for hero in heroes:
        if Object_IsAlive(hero):
            return hero
    return heroes[-1]

def lookAtLastHero():
    look(findLastHero())

#-----------------
# death/destruction
#-----------------

def kill(object, force = 1):
    if js.Object_Exists(object):
        if force:
            js.Object_SetAttr(object, 'minHealth', 0)
        js.Trigger_Damage(object, js.Object_GetAttr(object, 'health'))
    else:
        print "kill - specified object does not exist"

def killAll(template, force = 1):
    objects = js.Mission_GetDynamicObjects()
    found = []
    for object in objects:
        if (js.Object_GetTemplate(object) == template):
            kill(object)

def destroy(object):
    if js.Object_Exists(object):
        js.Object_Destroy(object)
    else:
        print "destroy - specified object does not exist"

def destroyAllByTemplate(template):
    objects = getAllStatic(template)
    for object in objects:
        js.Object_Destroy(object)

#-----------------
# movement
#-----------------

def move(character, marker, radius = 1):
    if Object_Exists(character):
        #js.AI_IssueMove(character, marker, '', 2, AIM_SPEED_RUN)
        js.Trigger_Move(character, 'move', Get_MarkerPos(marker), '', radius, 0)
    else:
        print "move() - no character " + character

def moveCB(character, marker, radius = 1):
    if Object_Exists(character):
        #js.AI_IssueMove(character, marker, 'onStepCB', 2, AIM_SPEED_RUN)
        js.Trigger_Move(character, 'move', Get_MarkerPos(marker), 'onStepCB', radius, 0)
    else:
        null()
        print "moveCB() - no character " + character

def walk(character, marker, radius = 1):
    if Object_Exists(character):
        #js.AI_IssueMove(character, marker, '', 2, AIM_SPEED_WALK)
        js.Trigger_Move(character, 'move', Get_MarkerPos(marker), '', radius, 1)
    else:
        print "move() - no character " + character

def walkCB(character, marker, radius = 1):
    if Object_Exists(character):
        #js.AI_IssueMove(character, marker, 'onStepCB', 2, AIM_SPEED_WALK)
        js.Trigger_Move(character, 'move', Get_MarkerPos(marker), 'onStepCB', radius, 1)
    else:
        null()
        print "moveCB() - no character " + character

# stumpy 2006-10-20: position may also be a tuple
def moveTo(character, position, command = 'move', fn = '', radius = 30, speed = 0):
    if isinstance(position,types.TupleType):
        js.Trigger_Move(character, command, position, fn, radius, speed)
    elif isinstance(position,types.StringType):
        if Object_Exists(position):
            js.Trigger_Move(character, command, js.Get_ObjectPos(position), fn, radius, speed)
        elif Marker_Exists(position):
            js.Trigger_Move(character, command, js.Get_MarkerPos(position), fn, radius, speed)
    else:
        print 'moveTo: ERROR - could not determine what to do with position = %s' % repr(position)
        raise RuntimeError

def turn(character, marker, fn = ''):
    if Object_Exists(character):
        js.AI_IssueMove(character, marker, fn, 0, 0, 0, 0, 'turn')
    else:
        print "turn() - no character " + character

# stumpy 2006-10-20: target may also be a tuple
def turnTo(character, target):
    if isinstance(target,types.TupleType):
        js.Trigger_Move(character, "turn", target)
    elif isinstance(target,types.StringType):
        if Object_Exists(target):
            js.Trigger_Move(character, "turn", Get_ObjectPos(target))
        elif Marker_Exists(target):
            js.Trigger_Move(character, "turn", Get_MarkerPos(target))
    else:
        print 'turnTo: ERROR - could not determine what to do with target = %s' % repr(target)
        raise RuntimeError
   
def teleport(character, marker):
    if Object_Exists(character):
        js.Object_MoveToMarker(character, marker)
    else:
        print "teleport() - no character " + character

#-------------------
# arrows
#-------------------
   
def addArrow(name, target, colour = js.ARROW_RED, delete = 0):
    print 'addArrow ' + name + ' ' + target + ' ' + `colour` + ' ' + `delete`
    if type(target) == TupleType:
        Mission_AddArrowPos(name, target, colour)
    elif type(target) == StringType:
        if js.Object_Exists(target):
            Mission_AddArrowObject(name, target, colour, delete)
        else:
            Mission_AddArrowPos(name, js.Get_MarkerPos(target), colour)
    else:
        print "addArrow: 'target' parameter invalid type, should be string or tuple."
        return

def replaceArrow(name, target, type = js.ARROW_RED):
    Mission_RemoveArrow(name)
    addArrow(name, target, type)

def removeArrow(name):
    Mission_RemoveArrow(name)

def arrowNextTarget(prefix, dead_target = ''):
    targets = getAllByPrefix(prefix)

    # Remove all arrows from this group    
    for target in targets:
        removeArrow(target + "_arrow")

    # Ignore a target that has just died
    if (dead_target != '') and (dead_target in targets):
        targets.remove(dead_target)

    # Sort target list in ascending order
    targets.sort()
    
    # Arrow the first living member of the group
    for target in targets:
        if Object_IsAlive(target):
            addArrow(target + "_arrow", target)
            return

#-------------------
# character states
#-------------------

def freeze(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_FROZEN, 10, js.PERMANENT_STATE)

def frozen(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_FROZEN:
        return 1
    else:
        return 0

def stun(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_STUNNED, 10, js.PERMANENT_STATE)

def stunned(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_STUNNED:
        return 1
    else:
        return 0

def static(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_STATIC, 10, js.PERMANENT_STATE)

def staticked(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_STATIC:
        return 1
    else:
        return 0

def setBlankState(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_BLANK, 10, js.PERMANENT_STATE)

def blanked(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_BLANK:
        return 1
    else:
        return 0

def enrage(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_ENRAGED, 10, js.PERMANENT_STATE)

def enraged(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_ENRAGED:
        return 1
    else:
        return 0

def panic(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_PANICKED, 10, js.PERMANENT_STATE)

def panicked(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_PANICKED:
        return 1
    else:
        return 0

def panicCivs(position, radius):
    civs = getAllCivilians()
    for civ in civs:
        distance = distance2D(Get_ObjectPos(civ), position)
        if (distance<=radius):
            panic(civ)
    
def hypnotize(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_HYPNOTISED, 10, js.PERMANENT_STATE)

def hypnotized(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_HYPNOTISED:
        return 1
    else:
        return 0

def energize(character):
    js.Object_SetSecondaryState(character, js.SCSTATE_ENERGIZED, 40, 0)
    
def setInterrogateState(character):
    js.Object_SetPrimaryState(character, js.PCSTATE_INTEROGATE, 0, js.PERMANENT_STATE)
    
def interrogatable(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_INTEROGATE:
        return 1
    else:
        return 0
    
def normalize(character):
    if Object_IsAlive(character):
        js.Object_SetPrimaryState(character, js.PCSTATE_NORMAL, 0)
        Object_SetSecondaryState(character, SCSTATE_IRRADIATED, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_BLIND, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_HEXED, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_ACID_BURNT, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_INVISIBLE, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_POWER_NULLIFICATION, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_POSSESSED, 0, REMOVE_STATE)
        Object_SetGenericState(character, OBJSTATE_ELECTRIFIED, 0, REMOVE_STATE)
        # clear some FFX states
        from ffx import FFX_ObjectGetAttr, FFX_ObjectSetAttr
        if FFX_ObjectGetAttr(character,'lifedrain'):
            FFX_ObjectSetAttr(character,'lifedrain',0)

def cure(character):
    if Object_IsAlive(character):
        js.Object_SetPrimaryState(character, js.PCSTATE_NORMAL, 0)
        Object_SetSecondaryState(character, SCSTATE_IRRADIATED, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_BLIND, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_HEXED, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_ACID_BURNT, 0, REMOVE_STATE)
        Object_SetSecondaryState(character, SCSTATE_POWER_NULLIFICATION, 0, REMOVE_STATE)
        Object_SetGenericState(character, OBJSTATE_ALTERED_STRUCTURE, 0, REMOVE_STATE)
        Object_SetGenericState(character, OBJSTATE_DENSITY_CHANGE, 0, REMOVE_STATE)

def heal(character):
    if Object_IsAlive(character):
        maxHealth = js.Object_GetAttr(character, "maxHealth")
        js.Object_SetAttr(character, "health", maxHealth)

def revive(character):
    if not Object_IsAlive(character):
        js.Object_SetPrimaryState(character, js.PCSTATE_NORMAL, 0)
    
def normal(character):
    if js.Object_GetPrimaryState(character) == PCSTATE_NORMAL:
        return 1
    else:
        return 0

#-------------------------
# event sink registrations
#-------------------------

def regMarkerEnter(object, marker, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_MARKER_ENTER, fn, object, marker, 0, user, persistent)

def regMarkerLeave(object, marker, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_MARKER_LEAVE, fn, object, marker, 0, user, persistent)

def regHeroMarkerEnter(marker, fn, user = 0, persistent = 0):
    regMarkerEnter('GAME_OBJ_HERO', marker, fn, user, persistent)
    
def regHeroMarkerLeave(marker, fn, user = 0, persistent = 0):
    regMarkerLeave('GAME_OBJ_HERO', marker, fn, user, persistent)
    
def regProximityEnter(object, focus, distance, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_OBJ_PROXIMITY_ENTER, fn, object, focus, distance, user, persistent)

def regProximityLeave(object, focus, distance, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_OBJ_PROXIMITY_LEAVE, fn, object, focus, distance, user, persistent)

def regPickup(object, fn, character = 'GAME_OBJ_HERO', user = 0):
    js.Event_RegisterSink(EVENT_PICKUP, fn, object, '', 0, 0, 0)
    
def regDeath(object, fn, user = 0):
    js.Event_RegisterSink(EVENT_OBJ_HEALTHCHANGED, fn, object, '', 0, user, 0)

def regDamage(object, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_OBJ_HEALTHCHANGED, fn, object, '', 1000, user, persistent)

def regDanger(object, fn, user = 0):
    js.Event_RegisterSink(EVENT_AI_DANGER, fn, object, '', 0, user, 0)

def regDetect(ai, character, fn, user = 0):
    js.Event_RegisterSink(EVENT_DETECT_ENEMY, fn, ai, character, 0, user, 0)

def regDetectHero(ai, fn, user = 0):
    js.Event_RegisterSink(EVENT_DETECT_ENEMY, fn, ai, "GAME_OBJ_HERO", 0, user, 0)
    
def regDeaths(template, fn, user = 0):
    js.Event_RegisterSink(EVENT_TEMPLATE_OBJ_DIED, fn, template, '', 0, user, 1)

def regAlmostDead(char, fn, health = 1, user = 0, string = '', persistent = 0):
    Object_SetAttr(char, 'minHealth', health)
    js.Event_RegisterSink(EVENT_OBJ_HEALTHCHANGED, fn, char, string, health+1, user, persistent)

def regCollapse(building, fn, user = 0):
    js.Event_RegisterSink(EVENT_BLD_SEGMENT_COLLAPSED, fn, building, '', 0, user, 0)

def regCharState(character, state, fn, persistent = 0):
    js.Event_RegisterSink(EVENT_CHARACTER_STATE_CHANGED, fn, character, '', 0, state, persistent)

def regPowerUse(character, power, fn, persistent = 0, user = 0):
    js.Event_RegisterPowerUse(character, power, fn, persistent, user)

def regTimer(fn, time, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_TIMER, fn, '', '', time, user, persistent)

def regShieldDamage(character, fn, user = 0, persistent = 0):
    js.Event_RegisterSink(EVENT_OBJ_SHIELD_DAMAGED, fn, character, '', 1000, user, persistent)

def regDialog(dialog, fn):
	Event_RegisterSink(js.EVENT_DLG_ONSHOW, fn, '', dialog)
           
#-----------
# group spawn
#-----------

def spawnGetMarker(prefix, number):
    return prefix + `number`

def spawnGetName(prefix, number, wave=0):
    if wave==0:
        return '_spwn_' + prefix + `number`
    else:
        return '_spwn_' + prefix + '_' + `wave` + '_' + `number`

def spawnCreate(prefix, template, number, wave=0):
    name = spawnGetName(prefix, number, wave)
    marker = spawnGetMarker(prefix, number)
    Object_Spawn(name, template, marker)
    
def spawnGroup(prefix, template, total, wave=0):
    for number in range(0, total):
        spawnCreate(prefix, template, number, wave)

def spawnAttack(prefix, target):
    minions = getAllByPrefix(prefix)
    for minion in minions:
        js.AI_SetDestroyObject(minion, target)

def spawn(name, template, position):
    if type(position) == TupleType:
        Object_SpawnAt(template, position, name)
    elif type(position) == StringType:
        if js.Object_Exists(position):
            Object_SpawnAt(template, Get_ObjectPos(position), name)
        else:
            Object_Spawn(name, template, position)
    else:
        print "spawn: 'position' parameter invalid type, should be string or tuple."
        
#-----------
# powerups
#-----------

def UsePowerup(object, hero):
    print 'UsePowerup: called for object=%s hero=%s' % (object,hero)
    cure(hero)
    kill(object)
    Sound_PlaySound('hero_powerup')
    if hasattr(ff, 'OnPowerup'):    
        ff.OnPowerup(object, hero)
    
def UseHPPowerup(object, hero):
    print 'UseHPPowerup: called for object=%s hero=%s' % (object,hero)
    objHealth = js.Object_GetAttr(object, "health")
    if objHealth > 0:
        maxHealth = js.Object_GetAttr(hero, "maxHealth")
        js.Object_SetAttr(hero, "health", maxHealth)
        Mission_StatusText("@POWERUP_MSG_HP", ("@"+Object_GetTemplate(hero),)) 
        UsePowerup(object, hero)
    
def UsePPPowerup(object, hero):
    print 'UsePPPowerup: called for object=%s hero=%s' % (object,hero)
    objHealth = js.Object_GetAttr(object, "health")
    if objHealth > 0:
        js.Mission_AddBonus("DLG_DEBRIEF_BONUS_POWERUP", 500)
        Mission_StatusText("@POWERUP_MSG_PP") 
        UsePowerup(object, hero)

def UseCPPowerup(object, hero):
    print 'UseCPPowerup: called for object=%s hero=%s' % (object,hero)
    objHealth = js.Object_GetAttr(object, "health")
    if objHealth > 0:
        Campaign_AddCP(Object_GetTemplate(hero), 325)
        Mission_StatusText("@POWERUP_MSG_CP", ("@"+Object_GetTemplate(hero),)) 
        UsePowerup(object, hero)
    
def UseEPPowerup(object, hero):
    print 'UseEPPowerup: called for object=%s hero=%s' % (object,hero)
    objHealth = js.Object_GetAttr(object, "health")
    if objHealth > 0:
        maxEnergy = js.Object_GetAttr(hero, "maxEnergyPoints")
        js.Object_SetAttr(hero, "energyPoints", maxEnergy)
        energize(hero)
        Mission_StatusText("@POWERUP_MSG_EP", ("@"+Object_GetTemplate(hero),)) 
        UsePowerup(object, hero)

#-----------------
# portrait speech
#-----------------

def talk(character, number, variant = 1):
    speak(character, getTalkTag(character, number, variant))
    
def talkNoCB(character, number, variant = 1):
    speakNoCB(character, getTalkTag(character, number, variant))

def talkCS(character, number, variant = 1):
    statement = "talk('" + character + "', " + `number` + ", " + `variant` + ")"
    cs = [(statement,),("blank",),]
    play(cs, "talkCS")
    
def brief(character, number, variant = 1):
    speak(character, getBriefTag(character, number, variant))

def briefNoCB(character, number, variant = 1):
    speak(character, getBriefTag(character, number, variant))

def getTalkTag(character, number, variant = 1):
    return 'MISSPCH_' + getTag(character, number, variant)

def getBriefTag(character, number, variant = 1):
    return 'BRFSPCH_' + getTag(character, number, variant)

def getChatTag(charID, number, variant = 1):
    return 'MISSPCH_' + getTagFixed(charID, number, variant)
    
def getCharTag(character, variant = 1):
    if not js.Object_Exists(character):
        return Template_GetSpeechID(character)
    else:
        return Object_GetSpeechID(character)

def getTagFixed(charID, number, variant = 1):
    if (number<=0) or (number>99):
        print "getTagFixed - bad speech number: " + `number`
        return
    if (number<10):
        numberString = '0'+`number`
    else:
        numberString = `number`
    mission = getSpeechMission()
    return `mission` + '_' + charID + '_' + numberString
    
def getTag(character, number, variant = 1):
    return getTagFixed(getCharTag(character, variant), number, variant)

def speakCS(speaker, tag, cs_name = 'cutscene', interrupt = 0, priority = 1):
    if interrupt:
        step = "speak('" + speaker + "','" + tag + "', interrupt = 1)"
    else:
        step = "speak('" + speaker + "','" + tag + "')"
    print step
    play([(step,),("blank",),], cs_name, priority, 0, interrupt)
    
#----------------
# trainers
#----------------

def UseTrainer(object, hero):
    train_number = ParseNumber(object)
    tag = getTalkTag('narrator', train_number)
    trainer_name = "_trainer_" + `train_number` + "_cs"
    speakCS('narrator', tag, cs_name = trainer_name, interrupt = 1, priority = -1)
    removeAttention(object)
    if hasattr(ff, 'OnTrainer'):    
        ff.OnTrainer(object, hero, train_number)

def addTrainerAttention(numbers = []):
    if (numbers == []):
        trainers = getAll('trainer')
    else:
        trainers = []
        for number in numbers:
            trainers.append('trainer_' + `number`)
    for trainer in trainers:
        addAttention(trainer)

#----------------
# mission stuff
#----------------

# this has become the unofficial residence of all stuff that has to happen for each mission
def setMission(number, customDeaths = 0, chatter = 1):
    Mission_SetAttr('_cs_mission_number', number)
    if not customDeaths:
        Event_RegisterSink(js.EVENT_MISSIONLOST, "onHeroesDead", '')
    if chatter:
        setupChatter()

def getMission():
    return int(Mission_GetAttr('_cs_mission_number'))

# text DB uses different numbering after mission 14... sigh
def getSpeechMission():
    mission = int(Mission_GetAttr('_cs_mission_number'))
    if mission>14:
        return mission-1
    return mission

def getMissionDigits(mission):
    if (mission<10):
        return '0'+`mission`
    return `mission`

#----------------
# on heroes dead
#----------------

# default handler for all heroes dying
def onHeroesDead(event):
    deadHero = event.object
    # check for heroic deeds
    heroes = getAllHeroes()
    hasHeroDeed = 0
    for hero in heroes:
        if Object_GetAttr(hero, 'heroPoints')>0:
            hasHeroDeed = 1
            break
    if hasHeroDeed:
        # re-register end mission callback
        Event_RegisterSink(js.EVENT_MISSIONLOST, "onHeroesDead", '')

        if hasattr(ff, 'OnAllDead'):    
            ff.OnAllDead(deadHero)
        else:
            js.Event_RegisterSink(js.EVENT_TIMER, 'endMission', deadHero, '', 15, 0)
    else:
        js.Event_RegisterSink(js.EVENT_TIMER, 'endMission', deadHero, '', 0.1, 0)

def endMission(event):
    # check to see if anyone has been revived
    heroes = getAllHeroes()
    isAlive = 0
    for hero in heroes:
        if Object_IsAlive(hero):
            isAlive = 1
            break
    if not isAlive:
        lookcmd = "Camera_LookAtObject('" + event.object + "', -100, 0, 0, 3, CPM_SCROLLTO, CA_MOVE)"
        heroes_dead_cs[0][4] = lookcmd
        play(heroes_dead_cs, "heroes_dead_cs", 1000)
        
def playDefeatMusic():
    Sound_PlayMusic('music_defeat', 1)
    
heroes_dead_cs = [
    [
        "Sound_PlayMusic('music_defeat', 1)",
        "see()",
        "peace()",
        "startCS()",
        "blank",
        "wait(10)",
    ],
    [
        "endCS()",
        "playTransition()",
    ],
    [
        "Mission_Lose()",
    ],
]

#----------------
# transitions
#----------------

def playTransition():
    Mission_Play2DCutscene("transition_mission_nif", "transition_mission_wav", "onStepCB")

def playTransitionNoCB():
    Mission_Play2DCutscene("transition_mission_nif", "transition_mission_wav", "")    

#-----------------
# attention marker
#-----------------

def addAttention(object):
    if object == '':
        print "WARNING: addAttentionMarker called with null object name"
        return
    handle = Object_GetAttr(object, '_attention_handle')
    if handle != -1:
        return
    handle = Object_PlayEffect(object, 'effect_attention_marker', '', FX_TRACK_OBJECT_POS | FX_LOOP, 0, 'top')
    Object_SetAttr(object, '_attention_handle', handle)

def removeAttention(object):
    if not Object_Exists(object):
        return
    handle = Object_GetAttr(object, '_attention_handle')
    if handle == -1:
        return
    Object_StopEffect(object, handle)
    Object_SetAttr(object, '_attention_handle', -1)

def hasAttention(object):
    handle = Object_GetAttr(object, '_attention_handle')
    if handle == -1:
        return 0
    return 1

#----------------
# talking
#----------------

speechIDs = [{
    'alchemiss': 'AL',
    'alien_sergeant': 'AW',
    'alien_warrior': 'AW',
    'ant_soldier': 'SA',
    'ant_worker': 'WA',
    'bacchite_eve': 'EV',
    'bacchite_green': 'GB',
    'bacchite_red': 'RB',
    'bacchite_silver': 'SB',
    'bullet': 'BL',
    'civilian_female': 'F1',
    'civilian_male': 'M1',
    'cop': 'P1',
    'cop_crooked': 'P1',
    'darkman_blue': 'BD',
    'darkman_purple': 'PD',
    'dejavu': 'DV',
    'dejavu_clone': 'DV',
    'dominion': 'LD',
    'duplicate_cop': 'DC',
    'duplicate_fem': 'DF',
    'duplicate_male': 'DM',
    'el_diablo': 'ED',
    'eve': 'EV',
    'eve_neutral': 'EV',
    'generic_male': 'GM',
    'generic_female': 'GF',
    'ice_queen': 'IQ',
    'ice_trooper': 'CW',
    'law': 'LW',
    'liberty_lad': 'LL',
    'mafioso': 'MF',
    'mafioso_gun': 'MF',
    'man_bot': 'MB',
    'manbull': 'ML',
    'mech_man': 'MC',
    'mech_man_flame': 'FM',
    'mentor': 'ME',
    'mentor_neutral': 'ME',
    'microwave': 'MW',
    'microwave_evil': 'MW',
    'minute_man': 'MM',
    'mm_dupe': 'MM',
    'mr_mechanical': 'MR',
    'narrator': 'NR',
    'nick_craft': 'NC',
    'nuclear_winter': 'NW',
    'oc_thug_bat': 'TB',
    'oc_thug_gun': 'TG',
    'oconnor': 'OC',
    'order': 'OR',
    'pan': 'PN',
    'pin_stripe': 'PS',
    'praetor': 'PR',
    'raptor': 'RP',
    'sea_urchin': 'SU',
    'su_thug_bat': 'TB',
    'su_thug_gun': 'TG',
    'su_thug_grenade': 'TB',
    'sukhov': 'NW',
    'the_ant': 'AN',
    'the_ant_evil': 'AN',
    'thug_with_bat': 'TB',
    'thug_with_gun': 'TG',
    'time_master': 'TM',
    'un_delegate': 'M1',
    'male_chatter': 'MC',
    'female_chatter': 'FC',
    },
    {
    'civilian_female': 'F2',
    'civilian_male': 'M2',
    'thug_with_bat': 'T2',
    'thug_with_gun': 'G2',
    }
]

def detectTalkCallback(event):
    number = int(event.user)
    frac = event.user - int(event.user)
    variant = int(frac * 10)
    talkCS(event.object, number, variant)

def regDetectTalk(ai, character, talkNumber, variant = 1):
	regDetect(ai, character, 'detectTalkCallback', user = talkNumber + variant * 0.1)

def addTalk(character, number, variant = 1, radius = 10, hero = ''):
    existingNumber = Object_GetAttr(character, '_talk_number')
    if existingNumber == -1:
        addCustomCommand("CUSTOM_TALKTO", character, 'customTalk', radius, hero)
    elif existingNumber != number:
        addAttention(character)
    Object_SetAttr(character, '_talk_number', number)
    Object_SetAttr(character, '_talk_variant', variant)

def customTalk(object, hero):
    turnTo(hero, object)
    number = int(Object_GetAttr(object, '_talk_number'))
    variant = int(Object_GetAttr(object, '_talk_variant'))
    talk(object, number, variant)
    removeAttention(object)

#-------------------
# interrogations
#-------------------

Global_RegisterAttr('_attention_handle', -1)
Global_RegisterAttr('_talk_number', -1)
Global_RegisterAttr('_talk_variant', -1)

def addInterrogation(char, fn, oneshot = 0, name = "CUSTOM_INTERROGATE", hero = ''):
    dynamicSpeakCS(char, 'INTEROGATEME')
    addCustomCommand(name, char, fn, oneshot, hero = hero)

def interrogateCharacter(event):
    char = event.object
    setInterrogateState(char)
    
def makeCustomCS(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0, attentionMarker = 1):
    addCustomCommand(name, char, 'interrogateCS', oneshot, hero = hero, attentionMarker = attentionMarker)

def makeSafeCustomCS(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0, attentionMarker = 1):
    addCustomCommand(name, char, 'safeInterrogateCS', oneshot, hero = hero, attentionMarker = attentionMarker)
    
def makeCriticalCustomCS(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0):
    regAlmostDead(char, 'interrogateCharacter', user = oneshot, string = hero)
    makeCustomCS(char, oneshot, name)
    
def makeSafeCriticalCustomCS(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0):
    regAlmostDead(char, 'interrogateCharacter', user = oneshot, string = hero)
    makeSafeCustomCS(char, oneshot, name)
    
def makeCustomFn(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0, attentionMarker = 1):
    addCustomCommand(name, char, 'interrogateFn', oneshot, hero = hero, attentionMarker = attentionMarker)
    
def makeSafeCustomFn(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0, attentionMarker = 1):
    addCustomCommand(name, char, 'safeInterrogateFn', oneshot, hero = hero, attentionMarker = attentionMarker)
    
def makeCriticalCustomFn(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0):
    regAlmostDead(char, 'interrogateCharacter', user = oneshot, string = hero)
    makeCustomFn(char, oneshot, name)
    
def makeSafeCriticalCustomFn(char, oneshot = 0, name = "CUSTOM_TALKTO", hero = '', radius = 0):
    regAlmostDead(char, 'interrogateCharacter', user = oneshot, string = hero)
    makeSafeCustomFn(char, oneshot, name)
    
def makeInterrogationCS(char, oneshot = 0, hero = 'none', radius = 0):
    regAlmostDead(char, 'setupInterrogationCS', user = oneshot, string = hero)

def makeInterrogationFn(char, oneshot = 0, hero = 'none', radius = 0):
    regAlmostDead(char, 'setupInterrogationFn', user = oneshot, string = hero)

def interrogateMe(event):
    char = event.object
    dynamicSpeakCS(char, 'INTEROGATEME')
    wait(7, 'setupInterrogateMe', string = char)
    
def setupInterrogateMe(event):
    char = event.string
    regAlmostDead(char, 'interrogateMe', persistent = 0)
    
def setupInterrogationCS(event):
    char = event.object
    hero = event.string
    oneshot = event.user
    setInterrogateState(char)
    if hero == 'none' or hero == char:
        addInterrogation(char, 'safeInterrogateCS', oneshot)
    else:
        addInterrogation(char, 'safeInterrogateCS', oneshot, hero = hero)
    # re-trigger interrogation speech if hurt again
    wait(5, 'setupInterrogateMe', string = char)
    if hasattr(ff, 'OnSetupInterrogation'):    
        ff.OnSetupInterrogation(char, hero)
        
def setupInterrogationFn(event):
    char = event.object
    hero = event.string
    oneshot = event.user
    setInterrogateState(char)
    if hero == 'none' or hero == char:
        addInterrogation(char, 'safeInterrogateFn', oneshot)
    else:
        addInterrogation(char, 'safeInterrogateFn', oneshot, hero = hero)
    # re-trigger interrogation speech if hurt again
    wait(5, 'setupInterrogateMe', string = char)
    if hasattr(ff, 'OnSetupInterrogation'):    
        ff.OnSetupInterrogation(char, hero)

def checkInterrogation(object, char):
    baddies = getNearBaddies(object, 350)
    removebaddies = []
    for baddy in baddies:
        if interrogatable(baddy) or not AI_IsEnabled(baddy):
            removebaddies.append(baddy)
    for removebaddy in removebaddies:
        baddies.remove(removebaddy)
    if object in baddies:
        baddies.remove(object)
    if len(baddies):
        if hasattr(ff, 'OnCantInterrogate'):
            speak = ff.OnCantInterrogate(object, char)
        else:
            speak = 1
        if speak:
            dynamicSpeakCS(char, 'CANTINTEROGATE')
        arrowName = '_' + object + '_int_arrow'
        addArrow(arrowName, baddies[0], delete = 1)
        return 0
    return 1

def dynamicSpeakCS(char, suffix):
    tag = 'SPEECH_' + getCharTag(char) + '_' + suffix + '_01'
    # may be some need to do something differnt for localized version here?
    speakCS(char, tag)
    
def dynamicSpeak(char, suffix):
    tag = 'SPEECH_' + getCharTag(char) + '_' + suffix + '_01'
    if Language != 'english':
        sayDynamicNoCB(char, tag, portrait = 0)
    else:
        sayDynamicNoCB(char, tag, portrait = 1)

def safeInterrogateCS(object, char):
    if not checkInterrogation(object, char):
        return -1
    interrogateCS(object, char)

def interrogateCS(object, char):
    interrogate(object, char)
    cs = 'ff.interrogate_' + object + '_cs'
    cs_play = 'play(' + cs + ', "' + cs + '", 3)'
    exec cs_play
    
def safeInterrogateFn(object, char):
    if not checkInterrogation(object, char):
        return -1
    interrogateFn(object, char)
    
def interrogateFn(object, char):
    interrogate(object, char)
    fn_call = 'ff.interrogate_' + object + '_fn("' + char + '")'
    exec fn_call

def interrogate(char, hero):
    removeAttention(char)
    # sanity check arrow removal
    arrowName = '_' + char + '_int_arrow'
    removeArrow(arrowName)
    Object_SetAttr(char, 'minHealth', 0)
    turnTo(hero, char)
    if hasattr(ff, 'OnInterrogation'):    
        ff.OnInterrogation(char, hero)

#-------------
# debug cheats
#-------------

def god():
    heroes = getAllHeroes()
    print 'god: heroes = %s' % (heroes,)
    for hero in heroes:
        js.Object_SetAttr(hero, 'minHealth', 100)
        
def mortal():
    heroes = getAllHeroes()
    print 'mortal: heroes = %s' % (heroes,)
    for hero in heroes:
        js.Object_SetAttr(hero, 'minHealth', 0)
        
def killthemall():
    objects = js.Mission_GetDynamicObjects()
    for object in objects:
        objclass = js.Object_GetClass(object)
        if (objclass & js.OC_MINION) or (objclass & js.OC_VILLIAN):
            print 'killthemall: Killing object: <' + object + '> class: <' + js.Object_GetTemplate(object) + '>'
            kill(object)

def killTemplate(template):
    objects = js.Mission_GetDynamicObjects()
    for object in objects:
        if Object_GetTemplate(object) == template:
            print 'killTemplate: Removing object: <' + object + '> template: <' + template + '>'
            js.Object_Destroy(object)
            
def gimmepoints(level = 0):
	Campaign_AddPrestige(10000000)
	allheroes = ['alchemiss',
			  'black_jack',
			  'blackbird',
			  'bullet',
			  'el_diablo',
			  'eve',
			  'green_genie',
			  'ironox',
			  'law',
			  'liberty_lad',
			  'man_bot',
			  'man_o_war',
			  'mentor',
			  'microwave',
			  'minute_man',
			  'order',
			  'quetzalcoatl',
			  'sea_urchin',
			  'sky_king',
			  'supercollider',
			  'the_ant',
			  'the_bard',
			  'tombstone',
			  'tricolour']
	if (level == 0):
		for ahero in allheroes:
			Campaign_AddCP(ahero,10000)
	elif (level == 1):
		print "no extra points for Dream cheats"
	elif (level == 2):
		print "cuba cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(1,0))
	elif (level == 3):
		print "red cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(2,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(1,0))
		Campaign_AddCP('mentor',CalcCheatPoints(1,0))
		Campaign_AddCP('minute_man',CalcCheatPoints(0,1))
	elif (level == 4):
		print "inside cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(3,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(2,0))
		Campaign_AddCP('mentor',CalcCheatPoints(1,1))
		Campaign_AddCP('minute_man',CalcCheatPoints(1,1))
		Campaign_AddCP('tombstone',CalcCheatPoints(1,0))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(1,0))
	elif (level == 5):
		print "winter cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(4,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(3,0))
		Campaign_AddCP('mentor',CalcCheatPoints(1,2))
		Campaign_AddCP('minute_man',CalcCheatPoints(2,1))
		Campaign_AddCP('tombstone',CalcCheatPoints(2,0))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(2,0))
		Campaign_AddCP('bullet',CalcCheatPoints(1,0))
		Campaign_AddCP('green_genie',CalcCheatPoints(1,0))
	elif (level == 6):
		print "berlin/fortissimo cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(5,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(4,0))
		Campaign_AddCP('mentor',CalcCheatPoints(2,2))
		Campaign_AddCP('minute_man',CalcCheatPoints(2,2))
		Campaign_AddCP('tombstone',CalcCheatPoints(3,0))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(3,0))
		Campaign_AddCP('bullet',CalcCheatPoints(2,0))
		Campaign_AddCP('green_genie',CalcCheatPoints(2,0))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(1,0))
		Campaign_AddCP('the_ant',CalcCheatPoints(1,0))
	elif (level == 7):
		print "Red Sun/crypto cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(6,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(5,0))
		Campaign_AddCP('mentor',CalcCheatPoints(2,3))
		Campaign_AddCP('minute_man',CalcCheatPoints(3,2))
		Campaign_AddCP('tombstone',CalcCheatPoints(4,0))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(4,0))
		Campaign_AddCP('bullet',CalcCheatPoints(2,1))
		Campaign_AddCP('green_genie',CalcCheatPoints(3,0))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(2,0))
		Campaign_AddCP('the_ant',CalcCheatPoints(2,0))
		Campaign_AddCP('microwave',CalcCheatPoints(1,0))
		Campaign_AddCP('eve',CalcCheatPoints(1,0))
		Campaign_AddCP('supercollider',CalcCheatPoints(1,0))
		Campaign_AddCP('black_jack',CalcCheatPoints(1,0))
		Campaign_AddCP('the_bard',CalcCheatPoints(1,0))
	elif (level == 8):
		print "Euros/bullet cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(7,0))
		Campaign_AddCP('el_diablo',CalcCheatPoints(6,0))
		Campaign_AddCP('mentor',CalcCheatPoints(2,4))
		Campaign_AddCP('minute_man',CalcCheatPoints(4,2))
		Campaign_AddCP('tombstone',CalcCheatPoints(5,0))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(5,0))
		Campaign_AddCP('bullet',CalcCheatPoints(3,2))
		Campaign_AddCP('green_genie',CalcCheatPoints(4,0))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(3,0))
		Campaign_AddCP('the_ant',CalcCheatPoints(3,0))
		Campaign_AddCP('microwave',CalcCheatPoints(2,0))
		Campaign_AddCP('eve',CalcCheatPoints(2,0))
		Campaign_AddCP('supercollider',CalcCheatPoints(2,0))
		Campaign_AddCP('black_jack',CalcCheatPoints(2,0))
		Campaign_AddCP('tricolour',CalcCheatPoints(1,0))
		Campaign_AddCP('law',CalcCheatPoints(1,0))
		Campaign_AddCP('order',CalcCheatPoints(1,0))
		Campaign_AddCP('ironox',CalcCheatPoints(1,0))
		Campaign_AddCP('the_bard',CalcCheatPoints(2,0))
	elif (level == 9):
		print "sky king set cheats"
		Campaign_AddCP('alchemiss',CalcCheatPoints(7,1))
		Campaign_AddCP('el_diablo',CalcCheatPoints(6,1))
		Campaign_AddCP('mentor',CalcCheatPoints(2,5))
		Campaign_AddCP('minute_man',CalcCheatPoints(4,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(5,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(5,1))
		Campaign_AddCP('bullet',CalcCheatPoints(4,2))
		Campaign_AddCP('green_genie',CalcCheatPoints(4,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(3,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(3,1))
		Campaign_AddCP('microwave',CalcCheatPoints(2,1))
		Campaign_AddCP('eve',CalcCheatPoints(2,1))
		Campaign_AddCP('supercollider',CalcCheatPoints(2,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(3,0))
		Campaign_AddCP('tricolour',CalcCheatPoints(2,0))
		Campaign_AddCP('law',CalcCheatPoints(1,1))
		Campaign_AddCP('order',CalcCheatPoints(1,1))
		Campaign_AddCP('ironox',CalcCheatPoints(1,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(1,1))
		Campaign_AddCP('man_o_war',CalcCheatPoints(0,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(0,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(0,1))
		Campaign_AddCP('the_bard',CalcCheatPoints(2,1))
	elif (level == 10):
		print "blitz/entropy cheats"
		Campaign_AddCP('el_diablo',CalcCheatPoints(7,1))
		Campaign_AddCP('mentor',CalcCheatPoints(3,5))
		Campaign_AddCP('minute_man',CalcCheatPoints(5,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(6,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(6,1))
		Campaign_AddCP('bullet',CalcCheatPoints(4,3))
		Campaign_AddCP('green_genie',CalcCheatPoints(5,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(4,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(4,1))
		Campaign_AddCP('microwave',CalcCheatPoints(3,1))
		Campaign_AddCP('eve',CalcCheatPoints(3,1))
		Campaign_AddCP('supercollider',CalcCheatPoints(3,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(4,0))
		Campaign_AddCP('tricolour',CalcCheatPoints(3,0))
		Campaign_AddCP('law',CalcCheatPoints(2,1))
		Campaign_AddCP('order',CalcCheatPoints(2,1))
		Campaign_AddCP('ironox',CalcCheatPoints(2,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(2,0))
		Campaign_AddCP('man_o_war',CalcCheatPoints(1,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(1,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(1,1))
		Campaign_AddCP('the_bard',CalcCheatPoints(3,1))
	elif (level == 11):
		print "symphony cheats"
#		Campaign_AddCP('el_diablo',CalcCheatPoints(8,1))
		Campaign_AddCP('mentor',CalcCheatPoints(4,5))
		Campaign_AddCP('minute_man',CalcCheatPoints(6,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(7,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(7,1))
		Campaign_AddCP('bullet',CalcCheatPoints(4,4))
		Campaign_AddCP('green_genie',CalcCheatPoints(6,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(5,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(5,1))
		Campaign_AddCP('microwave',CalcCheatPoints(4,1))
		Campaign_AddCP('eve',CalcCheatPoints(4,1))
		Campaign_AddCP('supercollider',CalcCheatPoints(4,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(4,1))
		Campaign_AddCP('tricolour',CalcCheatPoints(3,1))
		Campaign_AddCP('law',CalcCheatPoints(3,1))
		Campaign_AddCP('order',CalcCheatPoints(3,1))
		Campaign_AddCP('ironox',CalcCheatPoints(3,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(2,1))
		Campaign_AddCP('man_o_war',CalcCheatPoints(2,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(2,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(2,1))
		Campaign_AddCP('the_bard',CalcCheatPoints(4,1))
	elif (level == 12):
		print "Elementals cheats"
#		Campaign_AddCP('el_diablo',CalcCheatPoints(9,1))
		Campaign_AddCP('mentor',CalcCheatPoints(5,5))
		Campaign_AddCP('minute_man',CalcCheatPoints(7,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(8,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(8,1))
		Campaign_AddCP('bullet',CalcCheatPoints(5,4))
		Campaign_AddCP('green_genie',CalcCheatPoints(7,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(6,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(5,2))
		Campaign_AddCP('microwave',CalcCheatPoints(5,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(5,1))
		Campaign_AddCP('tricolour',CalcCheatPoints(4,1))
		Campaign_AddCP('law',CalcCheatPoints(4,1))
		Campaign_AddCP('order',CalcCheatPoints(4,1))
		Campaign_AddCP('ironox',CalcCheatPoints(4,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(3,1))
		Campaign_AddCP('man_o_war',CalcCheatPoints(3,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(3,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(3,1))
		Campaign_AddCP('man_bot',CalcCheatPoints(1,0))
		Campaign_AddCP('the_bard',CalcCheatPoints(51))
	elif (level == 13):
		print "Disappear cheats"
#		Campaign_AddCP('el_diablo',CalcCheatPoints(10,1))
		Campaign_AddCP('minute_man',CalcCheatPoints(8,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(9,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(9,1))
		Campaign_AddCP('bullet',CalcCheatPoints(6,4))
		Campaign_AddCP('green_genie',CalcCheatPoints(8,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(7,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(6,2))
		Campaign_AddCP('microwave',CalcCheatPoints(6,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(6,1))
		Campaign_AddCP('tricolour',CalcCheatPoints(5,1))
		Campaign_AddCP('law',CalcCheatPoints(5,1))
		Campaign_AddCP('order',CalcCheatPoints(5,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(4,1))
		Campaign_AddCP('man_o_war',CalcCheatPoints(4,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(4,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(4,1))
		Campaign_AddCP('man_bot',CalcCheatPoints(2,0))
		Campaign_AddCP('the_bard',CalcCheatPoints(6,1))
	elif (level == 14):
		print "Man-Bot cheats"
#		Campaign_AddCP('el_diablo',CalcCheatPoints(11,1))
		Campaign_AddCP('minute_man',CalcCheatPoints(9,3))
		Campaign_AddCP('tombstone',CalcCheatPoints(10,1))
		Campaign_AddCP('liberty_lad',CalcCheatPoints(10,1))
		Campaign_AddCP('bullet',CalcCheatPoints(7,4))
		Campaign_AddCP('green_genie',CalcCheatPoints(9,1))
		Campaign_AddCP('quetzalcoatl',CalcCheatPoints(8,1))
		Campaign_AddCP('the_ant',CalcCheatPoints(7,2))
		Campaign_AddCP('microwave',CalcCheatPoints(7,1))
		Campaign_AddCP('black_jack',CalcCheatPoints(7,1))
		Campaign_AddCP('tricolour',CalcCheatPoints(6,1))
		Campaign_AddCP('law',CalcCheatPoints(6,1))
		Campaign_AddCP('order',CalcCheatPoints(6,1))
		Campaign_AddCP('sky_king',CalcCheatPoints(5,1))
		Campaign_AddCP('man_o_war',CalcCheatPoints(5,1))
		Campaign_AddCP('sea_urchin',CalcCheatPoints(5,1))
		Campaign_AddCP('blackbird',CalcCheatPoints(5,1))
		Campaign_AddCP('man_bot',CalcCheatPoints(3,0))
		Campaign_AddCP('the_bard',CalcCheatPoints(7,1))
def CalcCheatPoints(onteam,offteam):
	return ((onteam * 450) + (offteam * 275))

#----------------
# objectives
#----------------
            
def newObjective(type, number, name = ''):
    mission = getMission()
    tag = 'OBJ_'+ getMissionDigits(mission)
    if (type == 'primary'):
        if (name == ''):
            name = 'primary_' + `number`
        typeIndex = 0
        tag = tag + '_P'
    else:
        if (name == ''):
            name = 'secondary_' + `number`
        typeIndex = 1
        tag = tag + '_S'
    tag = tag + `number`
    prestige = getPrestige(type, mission, number)
    Objective_Add(name, tag, typeIndex, prestige)

def registerObjective(type, number, name = ''):
    mission = getMission()
    tag = 'OBJ_'+ getMissionDigits(mission)
    if (type == 'primary'):
        if (name == ''):
            name = 'primary_' + `number`
        typeIndex = 0
        tag = tag + '_P'
    else:
        if (name == ''):
            name = 'secondary_' + `number`
        typeIndex = 1
        tag = tag + '_S'
    tag = tag + `number`
    prestige = getPrestige(type, mission, number)
    Objective_Register(name, tag, typeIndex, prestige)

#----------------
# goals
#----------------

def addKillGoal(char, target, priority = ai.goal.PRI_MED, fn = '', user = 0, permanent = 0, replace = 0):
    if type(target) == StringType:
        if (target == "buildings"):
            goal = ai.goal.KillBuildings()
        else:
            goal = ai.goal.KillObjects(target)
    elif type(target) == TupleType:
        goal = ai.goal.KillObjectsTuple(target)
    else:
        print "addKillGoal: 'target' parameter invalid type, should be string or tuple."
        return
    addGoalToAI(char, goal, priority, fn, user, permanent, replace)

def addKillAllGoal(char, targetClass, priority = ai.goal.PRI_MED, fn = '', user = 0, permanent = 0, replace = 0):
    goal = ai.goal.KillClass(targetClass)
    addGoalToAI(char, goal, priority, fn, user, permanent, replace)

def addStayGoal(char, priority = ai.goal.PRI_ZERO+1):
    goal = ai.goal.Null()
    addGoalToAI(char, goal, priority, permanent = 1)

def addMoveGoal(char, dest, priority = ai.goal.PRI_MED, fn = '', user = 0, command = 'move', radius = 4 , speed = 0, permanent = 0, replace = 0):
    if type(dest) == TupleType:
        goal = ai.goal.Move(dest, command, radius, speed)
    elif type(dest) == StringType:
        if js.Object_Exists(dest):
            goal = ai.goal.Move(js.Get_ObjectPos(dest), command, radius, speed)
        else:
            goal = ai.goal.Move(js.Get_MarkerPos(dest), command, radius, speed)
    else:
        print "addMoveGoal: 'dest' parameter invalid type, should be string or tuple."
        return
    addGoalToAI(char, goal, priority, fn, user, permanent, replace)

def addKillBuildingsGoal(char, priority = ai.goal.PRI_MED, fn = '', user = 0, permanent = 0, replace = 0):
    goal = ai.goal.KillBuildings()
    addGoalToAI(char, goal, priority, fn, user, permanent, replace)

def clearGoals(char):
    aiObj = ai.findAI(char)
    aiObj.clearGoals()

def makePeaceful(char):
    aiObj = ai.findAI(char)
    aiObj.removeNaturalKillGoal()
    
#----------------
# patrols
#----------------
    
def startPatrol(name, points, priority = ai.goal.PRI_ZERO-1, run = 0):
    Object_SetAttr(name, "_patrolpoints", points)
    if run:
        moveRun(name, 1, priority)
    else:
        movePatrol(name, 1, priority)

def movePatrol(name, point, priority = ai.goal.PRI_ZERO-1):
    Object_SetAttr(name, "_patrolpoint", point)
    points  = int(Object_GetAttr(name, "_patrolpoints"))
    patrolPoint = name + "_patrol_" + `point`
#    print "Patrol: " + name + " is moving towards point " + str(point) + " of " + str(points)
    addMoveGoal(name, patrolPoint, priority, 'OnPatrol', priority, radius = 1, speed = 1, replace = 1)

def abortPatrol(name):
    clearGoals(name)
    Object_SetAttr(name, "_patrolpoints", -1)
    
def OnPatrol(event):
    name = event.object
    point = int(Object_GetAttr(name, "_patrolpoint"))
    points  = int(Object_GetAttr(name, "_patrolpoints"))
    if (points<0):
        # this is set when we decide to cancel the patrol
        return
#    print "Patrol: " + name + " arrived at point " + str(point) + " of " + str(points)
    nextPoint = point + 1
    if (nextPoint>points):
        nextPoint = 1
    movePatrol(name, nextPoint, event.user)

def moveRun(name, point, priority = ai.goal.PRI_ZERO-1):
    Object_SetAttr(name, "_patrolpoint", point)
    points  = int(Object_GetAttr(name, "_patrolpoints"))
    patrolPoint = name + "_patrol_" + `point`
#    print "Patrol: " + name + " is moving towards point " + str(point) + " of " + str(points)
    addMoveGoal(name, patrolPoint, priority, 'OnRun', priority, radius = 1, speed = 0, replace = 1)

def OnRun(event):
    name = event.object
    point = int(Object_GetAttr(name, "_patrolpoint"))
    points  = int(Object_GetAttr(name, "_patrolpoints"))
    if (points<0):
        # this is set when we decide to cancel the patrol
        return
#    print "Patrol: " + name + " arrived at point " + str(point) + " of " + str(points)
    nextPoint = point + 1
    if (nextPoint>points):
        nextPoint = 1
    moveRun(name, nextPoint, event.user)

#----------------
# civ chatter
#----------------

def setChatterNumber(number):
    Mission_SetAttr('_chatter_number', number)

def getMaleChatterNumber():
    return _maleChatterNumber[getMission()-1]

def getFemaleChatterNumber():
    return _femaleChatterNumber[getMission()-1]

_femaleChatterNumber = (
    # NEWBEGIN
    4,
    # TIPOFF
    4,
    # THINICE
    4,
    # NUCLEAR
    0,
    # NICKTIME
    4,
    # SKIP
    0,
    # PANIC
    3,
    # VANISH
    4,
    # DARKNESS
    0,
    # SHADOWS
    0,
    # TROUBLE
    3,
    # WANTED
    3,
    # DEJAWHO
    3,
    # SEEING
    4,
    # SKIP
    0,
    # ROBOTS
    4,
    # DESTRUCT
    0,
    # MECHANIC
    4,
    # UNITED
    3,
    # RETURN
    4,
    # FORBID
    0,
    # PANDEMON
    0,
    # DAWN
    4,
    # END
    0,
)

_maleChatterNumber = (
    # NEWBEGIN
    4,
    # TIPOFF
    3,
    # THINICE
    3,
    # NUCLEAR
    0,
    # NICKTIME
    4,
    # SKIP
    0,
    # PANIC
    4,
    # VANISH
    4,
    # DARKNESS
    0,
    # SHADOWS
    0,
    # TROUBLE
    3,
    # WANTED
    4,
    # DEJAWHO
    3,
    # SEEING
    3,
    # SKIP
    0,
    # ROBOTS
    4,
    # DESTRUCT
    0,
    # MECHANIC
    4,
    # UNITED
    4,
    # RETURN
    3,
    # FORBID
    0,
    # PANDEMON
    0,
    # DAWN
    3,
    # END
    0,
)

def setupCivChatter(civ):
    number = ParseNumber(civ)
    if (number>=0):
        maleName = "civ_"+`number`
        femaleName = "fciv_"+`number`
        if (civ == maleName) or (civ == femaleName):
            Mission_CustomAction('CUSTOM_TALKTO', '', civ, 'chatCB', 2)
                    
def setupChatter():
    civs = getAllCivilians()
    if getMaleChatterNumber()>0:
        for civ in civs:
            if isMaleCivilian(civ):
                setupCivChatter(civ)
    if getFemaleChatterNumber()>0:
        for civ in civs:
            if isFemaleCivilian(civ):
                setupCivChatter(civ)

def chatCB(civ, hero):
    print hero + ' chats to ' + civ
    if (isMaleCivilian(civ)):
        number = (ParseNumber(civ)-1) % getMaleChatterNumber() + 1
        tag = getChatTag('MK', number)
    else: # isFemaleCivilian(civ)
        number = (ParseNumber(civ)-1) % getFemaleChatterNumber() + 1
        tag = getChatTag('FC', number)
    if Language != 'english':
        sayDynamicNoCB(civ, tag, portrait = 0)
    else:
        sayDynamicNoCB(civ, tag, portrait = 1)

#----------------
# widgets
#----------------

def flashWidget(widget):
    UI_HideWidget(widget)
    js.Event_RegisterSink(js.EVENT_TIMER, 'flashOnWidget', '', widget, 0.05, 6)

def flashOnWidget(event):
    widget = event.string
    count = event.user-1
    UI_ShowWidget(widget)
    if (count>0):
        js.Event_RegisterSink(js.EVENT_TIMER, 'flashOffWidget', '', widget, 0.1, count)
    
def flashOffWidget(event):
    widget = event.string
    count = event.user
    UI_HideWidget(widget)
    js.Event_RegisterSink(js.EVENT_TIMER, 'flashOnWidget', '', widget, 0.05, count)

#----------------
# interludes
#----------------

def setupInterlude(marker, actors = (), priority = 1):
    Mission_SetAttr(getInterruptStatusName(marker), 0)
    regMarkerEnter('GAME_OBJ_HERO', marker, 'playInterlude', user = priority)
    for actor in actors:
        regDetect(actor, 'GAME_OBJ_HERO', 'detectInterlude')
        regDeath(actor, 'detectDeath')

def getInterruptStatusName(markerName):
    return "_" + markerName + "_interrupt"

def getMainInterludeName(markerName):
    return markerName + "_cs"

def getSetupInterludeName(markerName):
    return markerName + "_setup_cs"

def detectDeath(event):
    markerName = ParsePrefix(event.object)
    interruptInterlude(markerName)
    
def detectInterlude(event):
    markerName = ParsePrefix(event.object)
    interruptInterlude(markerName)

def interruptInterlude(markerName):
    currentCS = csHelper.getName()
    if currentCS == getMainInterludeName(markerName):
        csHelper.end()
    else:
        Mission_SetAttr(getInterruptStatusName(markerName), 1)
        
def playInterlude(event):
    markerName = event.string
    priority = event.user
    if (not Mission_GetAttr(getInterruptStatusName(markerName))):
        csName = 'ff.'+ getMainInterludeName(markerName)
        cs_play = 'play(' + csName + ', "' + getMainInterludeName(markerName) + '", priority)'
        exec cs_play
    csName = 'ff.'+ getSetupInterludeName(markerName)
    cs_play = 'play(' + csName + ', "' + getSetupInterludeName(markerName) + '", priority)'
    exec cs_play

#-------------
# cars
#-------------

def CarOn(car):
    Object_Drive(car, 1)

def CarsOn():
    Object_Drive('', 1)

def CarOff(car):
    Object_Drive(car, 0)
    
def CarsOff():
    Object_Drive('', 0)
 
#---------------
# miscellaneous
#---------------

def addCustomCommand(commandName, object, funcName, oneshot = 0, radius = 4, hero = '', attentionMarker = 1):
    Mission_CustomAction(commandName, hero, object, funcName, radius, oneshot)
    if attentionMarker:
        addAttention(object)

def isDigits(test):
    for l in test:
        if not l in string.digits:
            return 0
    return 1

### gets a number out of an object name where object name is <blah>_<number>
##def ParseNumber(object):      
##    # parse name - gross but no real alternative
##    # train_number = int(object[object.rfind("_")+1:])
##    i = len(object)-1
##    while not (object[i] == '_') and (i>=0):
##        i = i-1
##    if ((i<0) or (i == len(object)-1)) or not isDigits(object[i+1:]):
##        print 'cshelper.ParseNumber - bad name: ' + object
##        return -1
##    return int(object[i+1:])

# gets a number at the end of an object name where object name is blah<number>
def ParseNumber(object):      
    i = len(object)-1
    hasnum = 0
    while ( object[i] in '1234567890' ) and ( i >= 0 ):
        hasnum = 1
        i = i-1
    if ( (i<0) or not hasnum ):
        print 'cshelper.ParseNumber - bad name: ' + object
        return -1
    return int(object[i+1:])

def ParsePrefix(object):      
    # parse name - gets prefix before _
    i = 0
    while i<len(object) and object[i] != '_':
        i = i+1
    if (i>=len(object)):
        print 'cshelper.ParseNumber - bad name: ' + object
        return ''
    return object[0:i]

def addAttr(attr, inc):
    js.Mission_SetAttr(attr, js.Mission_TestAttr(attr) + inc)

def see():
    js.Mission_SetDetectionEnabled(0)

def blind():
    js.Mission_SetDetectionEnabled(1)

def zeroDetect(char):
    Object_SetAttr(char, 'detectradius', 0)
    
def peace():
    js.AI_SetAllEnabled(0)
    
def war():
    js.AI_SetAllEnabled(1)

def enableHeroes():
    for i in range(1,5):
        hero = 'hero_' + `i`
        if Object_Exists(hero):
            enable(hero)
        
def enable(char):
    js.AI_SetEnabled(char, 1)

def disable(char):
    js.AI_SetEnabled(char, 0)

def alive(char):
    return js.Object_IsAlive(char)

def health(char):
    if Object_Exists(char):
        return Object_GetAttr(char, 'health')
    else:
        return 0

def energy():
    RPG_EP_RECHARGE=10000
    
def sun(rot, tilt, bright, ambient, color):
    js.Mission_SetSunlight(rot, tilt, bright, ambient, color)
    
def RPGOn():
    RPG_SetActive(1)

def RPGOff():
    RPG_SetActive(0)

def findHero(heroTemplate):
    for i in range(1,5):
        name = 'hero_'+`i`
        if js.Object_Exists(name) and (js.Object_GetTemplate(name) == heroTemplate):
            return name
    return ''

def wait(time, fn = 'onStepCB', user = 0, string = ''):
    js.Event_RegisterSink(js.EVENT_TIMER, fn, '', string, time, user)

def explode(object, power, intensity = 0):
    js.Trigger_Explosion(object, power, intensity)

#-----------
# SkXCamera2
#-----------
def fSkXCamera2 (tevent = None):
	try: test = skXCamera2 or skX2
	except:
		import skXCamera2
		ff.skXCamera2 = skXCamera2
		skXCamera2.fSkXCameraInitialise (0)
	regTimer ("fSkXCamera2", 1)#this re-initialises skXCamera2 after a saved-game is loaded

#rewrite setMission to include skX2 and skXCamera2 triggers
def setMission (number, customDeaths = 0, chatter = 1):
    Mission_SetAttr('_cs_mission_number', number)
    if not customDeaths:
        Event_RegisterSink(js.EVENT_MISSIONLOST, "onHeroesDead", '')
    if chatter:
        setupChatter()
    #skXCamera
    regTimer ("fSkXCamera2", 1)

#################################################################
# miscellaneous 2, stumpy added 
#################################################################

################################
# state info functions
################################

PCSTATE_EXILE = 3 
PCSTATES = {PCSTATE_FROZEN:'PCSTATE_FROZEN',
            PCSTATE_STUNNED:'PCSTATE_STUNNED',
            PCSTATE_STATIC:'PCSTATE_STATIC',
            PCSTATE_BLANK:'PCSTATE_BLANK',
            PCSTATE_ENRAGED:'PCSTATE_ENRAGED',
            PCSTATE_PANICKED:'PCSTATE_PANICKED',
            PCSTATE_HYPNOTISED:'PCSTATE_HYPNOTISED',
            PCSTATE_NORMAL:'PCSTATE_NORMAL',
            PCSTATE_INTEROGATE:'PCSTATE_INTEROGATE',
            PCSTATE_EXILE:'PCSTATE_EXILE'}
if ( 'datfiles' in sys.modules.keys() ) and ( datfiles.ffver() == 2 ):
    PCSTATES[PCSTATE_DIMENSION_SHIFT] = 'PCSTATE_DIMENSION_SHIFT'
    PCSTATES[PCSTATE_INTEROGATE] = 'PCSTATE_TOAD'
    PCSTATES[PCSTATE_OBJECT] = 'PCSTATE_OBJECT'
    PCSTATES[PCSTATE_TOAD] = 'PCSTATE_TOAD'
    PCSTATES[PCSTATE_DOPPELGANGER] = 'PCSTATE_DOPPELGANGER'
    PCSTATES[PCSTATE_POLYMORPHED] = 'PCSTATE_POLYMORPHED'

SCSTATES = {SCSTATE_IRRADIATED:'SCSTATE_IRRADIATED', \
            SCSTATE_BLIND:'SCSTATE_BLIND', \
            SCSTATE_HEXED:'SCSTATE_HEXED', \
            SCSTATE_ACID_BURNT:'SCSTATE_ACID_BURNT', \
            SCSTATE_INVISIBLE:'SCSTATE_INVISIBLE', \
            SCSTATE_POWER_NULLIFICATION:'SCSTATE_POWER_NULLIFICATION', \
            SCSTATE_ENERGIZED:'SCSTATE_ENERGIZED', \
            SCSTATE_EMPATHY:'SCSTATE_EMPATHY', \
            SCSTATE_AREA_ENERGIZED:'SCSTATE_AREA_ENERGIZED', \
            SCSTATE_DISPLACE_IMAGE:'SCSTATE_DISPLACE_IMAGE'}
if datfiles.ffver() == 2:
    SCSTATES[SCSTATE_SPEED_INCREASED] = 'SCSTATE_SPEED_INCREASED'
    SCSTATES[SCSTATE_SPEED_DECREASED] = 'SCSTATE_SPEED_DECREASED'
    SCSTATES[SCSTATE_WEAKENED] = 'SCSTATE_WEAKENED'
    SCSTATES[SCSTATE_BUOYANT] = 'SCSTATE_BUOYANT'
    SCSTATES[SCSTATE_ZOMBIE] = 'SCSTATE_ZOMBIE'
    SCSTATES[SCSTATE_DEEP_SENSORED] = 'SCSTATE_DEEP_SENSORED'
    SCSTATES[SCSTATE_POSSESSED] = 'SCSTATE_POSSESSED'  # = 134217728

OBJSTATES = {OBJSTATE_ELECTRIFIED:'OBJSTATE_ELECTRIFIED', \
            OBJSTATE_MOLECULAR_EXCITE:'OBJSTATE_MOLECULAR_EXCITE', \
            OBJSTATE_ALTERED_STRUCTURE:'OBJSTATE_ALTERED_STRUCTURE', \
            OBJSTATE_ATTACKBONUS:'OBJSTATE_ATTACKBONUS', \
            OBJSTATE_DENSITY_CHANGE:'OBJSTATE_DENSITY_CHANGE'}

OC_GENERIC = 0x400000       # via brute force testing
OC_POWER_PROXY = 0x800000   # via brute force testing

OBJCLASSES = {OC_CONTROLLABLE:'OC_CONTROLLABLE',
              OC_VILLIAN:'OC_VILLIAN',
              OC_GENERIC:'OC_GENERIC',
              OC_BUILDING:'OC_BUILDING',
              OC_POWER_PROXY:'OC_POWER_PROXY',
              OC_CIVILIAN:'OC_CIVILIAN',
              OC_MINION:'OC_MINION',
              OC_POLICE:'OC_POLICE',
              OC_VEHICLE:'OC_VEHICLE',
              OC_POWERUP:'OC_POWERUP',
              OC_TERRAIN:'OC_TERRAIN',
              OC_CHARACTER:'OC_CHARACTER'}

def Object_GetAllPrimaryStates(obj):
    objstate = Object_GetPrimaryState(obj)
    for stateval in PCSTATES.keys():
        if ( stateval == objstate ) :
            return PCSTATES[stateval]

def Object_GetAllSecondaryStates(obj):
    statenames = []
    objstates = Object_GetSecondaryStates(obj)
    for stateval in SCSTATES.keys():
        if ( stateval & objstates ) :
            statenames.append(SCSTATES[stateval])
    return statenames

# Setup a custom command that allows a character to click on an object and have
# the object's name be printed to the console.
def Object_Identify(char,oneShot=1):
    Object_SetAttr(char,'Identify_IsOneShot',oneShot)
    Mission_CustomAction('identify', char, '', 'Object_EchoID', 2000, oneShot)

IDName = '' # note that this still is cshelper.IDName from the console
def Object_EchoID(object,char):
    global IDName
    classes = filter(lambda x, o=object: Object_GetClass(o)&x, OBJCLASSES.keys())
    classnames = map(lambda x: OBJCLASSES[x],classes)
    print '%s has clicked on %s, which has template %s, class=%s.' \
          % (char,repr(object),Object_GetTemplate(object),classnames)
    if ( Object_GetAttr(char,'Identify_IsOneShot') == 1 ):
        Mission_RemoveCustomAction('identify', char, '')
    IDName = object

def ShowTeam():
    print getAllHeroes()
    print map(Object_GetTemplate,getAllHeroes())

def getAllPowerups():
    return filter(lambda obj: Object_GetClass(obj)&OC_POWERUP, Mission_GetMostObjects())
    
################################
# testing functions
################################

def GenericCallback(event):
    fields = dir(event)
    for field in fields:
        print 'GenericCallback: event.%s = %s' % (field,repr(eval('event.%s'%field)))

# show the current XP and XP needed to level up for recruited
# characters.  If char=='', then show for all characters, otherwise
# just for char or for each character in char if char is a list of
# character names
def ShowXP(char='', ForceRead=0):
    '''Show current XP and XP to level up. For empty string, show for all recruits; for string show for that char; for list of strings show for those chars.'''
    sgd = datfiles.Campaign_ReadCharactersFromSavedGame(ForceRead=ForceRead)
    if char == '':
        chars = filter(Campaign_IsRecruited,sgd.keys())
    elif isinstance(char,types.StringType):
        chars = filter(Campaign_IsRecruited,[char])
    elif isinstance(char,types.ListType) or isinstance(char,types.TupleType):
        chars = []
        for c in char:
            if c in sgd.keys():
                chars.append(c)
            elif Object_GetTemplate(c) in sgd.keys(): 
                chars.append(Object_GetTemplate(c))
    else:
        print 'ShowXP: improper argument (char=%s).' % repr(char)
    chars = map(lambda c,sgd=sgd: (sgd[c]['charName'],c),chars)
    chars.sort()
    CPPL = int(RPG_CP_PER_LEVEL)
    print '%30s:  %5s  %5s  %3s/%d' % ('character', 'CP', 'XP', 'XP needed', CPPL)
    for c in chars:
        XP = sgd[c[1]]['XP']
        CP = sgd[c[1]]['CP']
        print '%30s:  %5d  %5d  %3d/%d' % \
              (c[0], CP, XP, (XP/CPPL+1)*CPPL-XP,CPPL)

def land(c):
    Trigger_Power(c,c,'land_from_jump','')
    
################################
# random helper functions
################################

# Restore the materials for the given object to the one specified in
# the object's FFEdit template. If object=='AllObjects', this applies
# to multiple objects as specified by the CharactersOnly flag,
# otherwise only to the object given. If CharactersOnly==1, acts only
# on dynamic character objects, otherwise on all game objects.
# A handy sink to catch new spawns is
# regCharacterSpawn('Object_ResetToTemplateMaterial', persistent = 1)
def Object_ResetToTemplateMaterial(object='AllObjects',CharactersOnly=1):
    '''Reset character objects to their FFEdit template materials'''
    import chardata
    if 'object' in dir(object): # this was a sink callback, assume object name in the 'object' field.
        object = object.object
    if object != 'AllObjects':
        if Object_Exists(object):
            template = Object_GetTemplate(object)
            if template[:7] == 'custom_': # this is a HERO file character or special internal template
                Object_SetAttr(object,'material',chardata.GetCharacterData(object)['material'])
            else:
                Object_SetAttr(object,'material',datfiles.Campaign_ReadObjects()[template]['material'])
    else:
        if CharactersOnly:
            for object in filter(lambda o: Object_GetClass(o)&OC_CHARACTER,Mission_GetDynamicObjects()):
                Object_SetAttr(object,'material',chardata.GetCharacterData(object)['material'])
        else:
            temps = datfiles.Campaign_ReadObjects()
            for object in Mission_GetObjects()[1:]: # skip the map object
                template = Object_GetTemplate(object)
                if template[:17] == 'custom_': # this is a HERO file character or special internal template
                    Object_SetAttr(object,'material',chardata.GetCharacterData(object)['material'])
                else:
                    Object_SetAttr(object,'material',temps[template]['material'])

