#! /usr/bin/env python
# datfiles.py

# Some attempts to extract data from the Freedom Force DAT files.
#
# For documentation, search the Scripting forum on Freedom Reborn for
#     +Documentation +datfiles +chardata
# and subject "datfiles and chardata: documentation...".
#
# (c) stumpy 2004, 2005, 2006

import sys
import os
import os.path
import types
import struct
import string
import time

if __name__ == '__main__':  # outside game
    try:
        import psyco    # very cool when available
        psyco.profile()
    except ImportError:
        pass

DATFILES_VER = 0.256
def datfilesVersion():
    return DATFILES_VER

print 'loading datfiles version %f' % datfilesVersion()

# Have we reloaded since the last time this function asked?
def CheckReload(funcname):
    global _FuncCheckList
    if not '_FuncCheckList' in globals().keys():
        _FuncCheckList = []
    if funcname in _FuncCheckList:
        return 0    # false
    else:
        _FuncCheckList.append(funcname)
        return 1    # True

#################################
### DAT Constants ###############
#################################

# Marker Types.  The Type field that shows up in FFEdit's Marker
# Details in the Mission tab.  (The constant names are unofficial.)
# Don't know what 3 was to be.
MT_GENERIC = 0
MT_CHARACTER = 1
MT_LIGHT = 2
MT_SOUND = 4
MT_POSITIONAL = 5
MT_ROAD_NODE = 6
MT_CIV_NODE = 7
MT_TRAFFIC = 8
# lookup constant name via constant value
MT_MARKER_TYPES = {MT_GENERIC : 'MT_GENERIC',
                   MT_CHARACTER : 'MT_CHARACTER',
                   MT_LIGHT : 'MT_LIGHT',
                   MT_SOUND : 'MT_SOUND',
                   MT_POSITIONAL : 'MT_POSITIONAL',
                   MT_ROAD_NODE : 'MT_ROAD_NODE',
                   MT_CIV_NODE : 'MT_CIV_NODE',
                   MT_TRAFFIC : 'MT_TRAFFIC'}

# Power Types. The Attack or Defence Type field that shows up in FFEdit's
# Power Details and Defence Details in the Powers and Defences tabs,
# respecitively.  (The constant names are unofficial.)
PT_MELEE = 1
PT_RANGED = 2
PT_AREA = 3
PT_DIRECT = 4
PT_ACTIVE_DEFENCE = 5
PT_PASSIVE_DEFENCE = 6
PT_SPECIAL = 7
# lookup constant name via constant value
PT_POWER_TYPES = {PT_MELEE : 'PT_MELEE',
                  PT_RANGED : 'PT_RANGED',
                  PT_AREA : 'PT_AREA',
                  PT_DIRECT : 'PT_DIRECT',
                  PT_ACTIVE_DEFENCE : 'PT_ACTIVE_DEFENCE',
                  PT_PASSIVE_DEFENCE : 'PT_PASSIVE_DEFENCE',
                  PT_SPECIAL : 'PT_SPECIAL'}

# attack sub-type for ranged attacks
PT_ATTACK_SUBTYPE_NONE = 0
PT_ATTACK_SUBTYPE_BEAM = 1
PT_ATTACK_SUBTYPE_PROJECTILE = 2
PT_ATTACK_SUBTYPE_EXPLOSIVE = 3
PT_ATTACK_SUBTYPE_CONE = 4
PT_ATTACK_SUBTYPES = {PT_ATTACK_SUBTYPE_NONE : 'PT_ATTACK_SUBTYPE_NONE',
                      PT_ATTACK_SUBTYPE_BEAM : 'PT_ATTACK_SUBTYPE_BEAM',
                      PT_ATTACK_SUBTYPE_PROJECTILE : 'PT_ATTACK_SUBTYPE_PROJECTILE',
                      PT_ATTACK_SUBTYPE_EXPLOSIVE : 'PT_ATTACK_SUBTYPE_EXPLOSIVE',
                      PT_ATTACK_SUBTYPE_CONE : 'PT_ATTACK_SUBTYPE_CONE',
                      }

# attack damage types
PT_DAMAGE_ACID = 6
PT_DAMAGE_ACID_BURN = 18
PT_DAMAGE_BLIND = 17
PT_DAMAGE_COLD = 7
PT_DAMAGE_CRUSH = 0
PT_DAMAGE_ELECTRICAL = 5
PT_DAMAGE_ENERGY = 2
PT_DAMAGE_FIRE = 3
PT_DAMAGE_FREEZE = 11
PT_DAMAGE_HEX = 19
PT_DAMAGE_HYPNOSIS = 14
PT_DAMAGE_IRRADIATE = 16
#PT_DAMAGE_MENTAL = 
PT_DAMAGE_MENTAL_BLANK = 15
#PT_DAMAGE_MYSTICAL =
PT_DAMAGE_OBJECTIFY = 23
PT_DAMAGE_PANIC = 13
PT_DAMAGE_PIERCE = 1
#PT_DAMAGE_PIERCING =
PT_DAMAGE_POLYMORPH = 25
PT_DAMAGE_RADIATION = 4
PT_DAMAGE_RAGE = 12
PT_DAMAGE_SPECIAL = 9
PT_DAMAGE_SPEED_DECREASE = 21
PT_DAMAGE_SPEED_INCREASE = 20 
PT_DAMAGE_STASIS = 10
PT_DAMAGE_TOADIFY = 24
PT_DAMAGE_WATER = 8
PT_DAMAGE_WEAKEN = 22
PT_DAMAGE_TYPES = {PT_DAMAGE_ACID : 'PT_DAMAGE_ACID',
                   PT_DAMAGE_ACID_BURN : 'PT_DAMAGE_ACID_BURN',
                   PT_DAMAGE_BLIND : 'PT_DAMAGE_BLIND',
                   PT_DAMAGE_COLD : 'PT_DAMAGE_COLD',
                   PT_DAMAGE_CRUSH : 'PT_DAMAGE_CRUSH',
                   PT_DAMAGE_ELECTRICAL : 'PT_DAMAGE_ELECTRICAL',
                   PT_DAMAGE_ENERGY : 'PT_DAMAGE_ENERGY',
                   PT_DAMAGE_FIRE : 'PT_DAMAGE_FIRE',
                   PT_DAMAGE_FREEZE : 'PT_DAMAGE_FREEZE',
                   PT_DAMAGE_HEX : 'PT_DAMAGE_HEX',
                   PT_DAMAGE_HYPNOSIS : 'PT_DAMAGE_HYPNOSIS',
                   PT_DAMAGE_IRRADIATE : 'PT_DAMAGE_IRRADIATE',
                   #PT_DAMAGE_MENTAL : 'PT_DAMAGE_MENTAL',
                   PT_DAMAGE_MENTAL_BLANK : 'PT_DAMAGE_MENTAL_BLANK',
                   #PT_DAMAGE_MYSTICAL : 'PT_DAMAGE_MYSTICAL',
                   PT_DAMAGE_OBJECTIFY : 'PT_DAMAGE_OBJECTIFY',
                   PT_DAMAGE_PANIC : 'PT_DAMAGE_PANIC',
                   PT_DAMAGE_PIERCE : 'PT_DAMAGE_PIERCE',
                   #PT_DAMAGE_PIERCING : 'PT_DAMAGE_PIERCING',
                   PT_DAMAGE_POLYMORPH : 'PT_DAMAGE_POLYMORPH',
                   PT_DAMAGE_RADIATION : 'PT_DAMAGE_RADIATION',
                   PT_DAMAGE_RAGE : 'PT_DAMAGE_RAGE',
                   PT_DAMAGE_SPECIAL : 'PT_DAMAGE_SPECIAL',
                   PT_DAMAGE_SPEED_DECREASE : 'PT_DAMAGE_SPEED_DECREASE',
                   PT_DAMAGE_SPEED_INCREASE : 'PT_DAMAGE_SPEED_INCREASE',
                   PT_DAMAGE_STASIS : 'PT_DAMAGE_STASIS',
                   PT_DAMAGE_TOADIFY : 'PT_DAMAGE_TOADIFY',
                   PT_DAMAGE_WATER : 'PT_DAMAGE_WATER',
                   PT_DAMAGE_WEAKEN : 'PT_DAMAGE_WEAKEN'}

# special attack types
PT_SPECIAL_AFTERBURNER = 52             # bought to one level only
PT_SPECIAL_ALTERATION = 1
PT_SPECIAL_ALTER_GENETIC_STRUCTURE = 24
PT_SPECIAL_ANIMATE_DEAD = 43
PT_SPECIAL_ANIMATE_DEAD_MINIONS = 46
PT_SPECIAL_BEZERKER = 22                # bought to one level only
PT_SPECIAL_BUOYANCY = 45
PT_SPECIAL_BURROW = 8                   # bought to one level only
PT_SPECIAL_CEREBRAL_NORMALISATION = 23  # bought to one level only
PT_SPECIAL_CLOAK = 10                   # bought to one level only
PT_SPECIAL_CLONE_SELF = 15              # bought to one level only
PT_SPECIAL_CLONE_TARGET = 14
PT_SPECIAL_DEEP_SENSOR = 47
PT_SPECIAL_DENSITY_MAXIMIZE = 32
PT_SPECIAL_DIMENSION_SHIFT = 5
PT_SPECIAL_DISPLACE_IMAGE = 36          # bought to one level only
PT_SPECIAL_DISRUPT_OBJECT = 37  # = 39 in parts of ffvt3r.exe
PT_SPECIAL_DOPPELGANGER = 51
PT_SPECIAL_ELECTRIFY_SELF = 26          # bought to one level only
PT_SPECIAL_EMPATHY = 33                 # bought to one level only
PT_SPECIAL_ENERGIZE = 30        # energy surge + chance of stun, as per Bullet
PT_SPECIAL_ENERGIZE_AREA = 35   # energy boost + unstable energy, as per Man-Bot
PT_SPECIAL_EP_DRAW = 3
PT_SPECIAL_EP_TRANSFER = 2
PT_SPECIAL_HP_BALANCE = 20
PT_SPECIAL_IGNITE = 12
PT_SPECIAL_IMAGE_PROJECTOR = 40         # bought to one level only
#PT_SPECIAL_KINETIC_DAMPING = 
PT_SPECIAL_LAWORDER = 17                # bought to one level only
#PT_SPECIAL_LIFE_FORCE = 
PT_SPECIAL_LIFE_GIFT = 18
PT_SPECIAL_LIVING_FIRES = 48            # bought to one level only
PT_SPECIAL_MOLECULAR_EXCITE = 6
PT_SPECIAL_MORPH = 16
PT_SPECIAL_NONE = 0
PT_SPECIAL_NORMALISE = 21               # bought to one level only
PT_SPECIAL_PATRIOT_CHARGE = 25          # bought to one level only
PT_SPECIAL_POSSESSION = 41
PT_SPECIAL_POWER_NULLIFICATION = 34
PT_SPECIAL_RALLY = 19
PT_SPECIAL_REMOVE_DEFENCE = 44          # bought to one level only
PT_SPECIAL_SACRIFICE = 42               # bought to one level only
PT_SPECIAL_SONIC_BOOM = 50              # bought to one level only
PT_SPECIAL_SPEEDING_BULLET = 28         # bought to one level only
PT_SPECIAL_SPRINT = 27                  # bought to one level only
PT_SPECIAL_TELEPORT_SELF = 31           # bought to one level only
PT_SPECIAL_TELEPORT_SWAP = 49           # bought to one level only
PT_SPECIAL_TUMBLE = 7                   # bought to one level only
PT_SPECIAL_TUNNEL = 39  # = 09 in parts of ffvt3r.exe   # bought to one level only
PT_SPECIAL_UNCLOAK = 11                 # bought to one level only
PT_SPECIAL_TYPES = {PT_SPECIAL_AFTERBURNER : 'PT_SPECIAL_AFTERBURNER',
                    PT_SPECIAL_ALTERATION : 'PT_SPECIAL_ALTERATION',
                    PT_SPECIAL_ALTER_GENETIC_STRUCTURE : 'PT_SPECIAL_ALTER_GENETIC_STRUCTURE',
                    PT_SPECIAL_ANIMATE_DEAD : 'PT_SPECIAL_ANIMATE_DEAD',
                    PT_SPECIAL_ANIMATE_DEAD_MINIONS : 'PT_SPECIAL_ANIMATE_DEAD_MINIONS',
                    PT_SPECIAL_BEZERKER : 'PT_SPECIAL_BEZERKER',
                    PT_SPECIAL_BUOYANCY : 'PT_SPECIAL_BUOYANCY',
                    PT_SPECIAL_BURROW : 'PT_SPECIAL_BURROW',
                    PT_SPECIAL_CEREBRAL_NORMALISATION : 'PT_SPECIAL_CEREBRAL_NORMALISATION',
                    PT_SPECIAL_CLOAK : 'PT_SPECIAL_CLOAK',
                    PT_SPECIAL_CLONE_SELF : 'PT_SPECIAL_CLONE_SELF',
                    PT_SPECIAL_CLONE_TARGET : 'PT_SPECIAL_CLONE_TARGET',
                    PT_SPECIAL_DEEP_SENSOR : 'PT_SPECIAL_DEEP_SENSOR',
                    PT_SPECIAL_DENSITY_MAXIMIZE : 'PT_SPECIAL_DENSITY_MAXIMIZE',
                    PT_SPECIAL_DIMENSION_SHIFT : 'PT_SPECIAL_DIMENSION_SHIFT',
                    PT_SPECIAL_DISPLACE_IMAGE : 'PT_SPECIAL_DISPLACE_IMAGE',
                    PT_SPECIAL_DISRUPT_OBJECT : 'PT_SPECIAL_DISRUPT_OBJECT',
                    PT_SPECIAL_DOPPELGANGER : 'PT_SPECIAL_DOPPELGANGER',
                    PT_SPECIAL_ELECTRIFY_SELF : 'PT_SPECIAL_ELECTRIFY_SELF',
                    PT_SPECIAL_EMPATHY : 'PT_SPECIAL_EMPATHY',
                    PT_SPECIAL_ENERGIZE : 'PT_SPECIAL_ENERGIZE',
                    PT_SPECIAL_ENERGIZE_AREA : 'PT_SPECIAL_ENERGIZE_AREA',
                    PT_SPECIAL_EP_DRAW : 'PT_SPECIAL_EP_DRAW',
                    PT_SPECIAL_EP_TRANSFER : 'PT_SPECIAL_EP_TRANSFER',
                    PT_SPECIAL_HP_BALANCE : 'PT_SPECIAL_HP_BALANCE',
                    PT_SPECIAL_IGNITE : 'PT_SPECIAL_IGNITE',
                    PT_SPECIAL_IMAGE_PROJECTOR : 'PT_SPECIAL_IMAGE_PROJECTOR',
                    #PT_SPECIAL_KINETIC_DAMPING : '#PT_SPECIAL_KINETIC_DAMPING',
                    PT_SPECIAL_LAWORDER : 'PT_SPECIAL_LAWORDER',
                    #PT_SPECIAL_LIFE_FORCE : '#PT_SPECIAL_LIFE_FORCE',
                    PT_SPECIAL_LIFE_GIFT : 'PT_SPECIAL_LIFE_GIFT',
                    PT_SPECIAL_LIVING_FIRES : 'PT_SPECIAL_LIVING_FIRES',
                    PT_SPECIAL_MOLECULAR_EXCITE : 'PT_SPECIAL_MOLECULAR_EXCITE',
                    PT_SPECIAL_MORPH : 'PT_SPECIAL_MORPH',
                    PT_SPECIAL_NONE : 'PT_SPECIAL_NONE',
                    PT_SPECIAL_NORMALISE : 'PT_SPECIAL_NORMALISE',
                    PT_SPECIAL_PATRIOT_CHARGE : 'PT_SPECIAL_PATRIOT_CHARGE',
                    PT_SPECIAL_POSSESSION : 'PT_SPECIAL_POSSESSION',
                    PT_SPECIAL_POWER_NULLIFICATION : 'PT_SPECIAL_POWER_NULLIFICATION',
                    PT_SPECIAL_RALLY : 'PT_SPECIAL_RALLY',
                    PT_SPECIAL_REMOVE_DEFENCE : 'PT_SPECIAL_REMOVE_DEFENCE',
                    PT_SPECIAL_SACRIFICE : 'PT_SPECIAL_SACRIFICE',
                    PT_SPECIAL_SONIC_BOOM : 'PT_SPECIAL_SONIC_BOOM',
                    PT_SPECIAL_SPEEDING_BULLET : 'PT_SPECIAL_SPEEDING_BULLET',
                    PT_SPECIAL_SPRINT : 'PT_SPECIAL_SPRINT',
                    PT_SPECIAL_TELEPORT_SELF : 'PT_SPECIAL_TELEPORT_SELF',
                    PT_SPECIAL_TELEPORT_SWAP : 'PT_SPECIAL_TELEPORT_SWAP',
                    PT_SPECIAL_TUMBLE : 'PT_SPECIAL_TUMBLE',
                    PT_SPECIAL_TUNNEL : 'PT_SPECIAL_TUNNEL',
                    PT_SPECIAL_UNCLOAK : 'PT_SPECIAL_UNCLOAK'}

# attack flags
PT_ATTACK_CHAINED = 32
PT_ATTACK_FLIGHT_SPAWN = 16
PT_ATTACK_GRENADE = 2
PT_ATTACK_GROUND_TARG_ONLY = (1 << 8)
PT_ATTACK_ENEMIES_ONLY = (32 << 8)
PT_ATTACK_AIRBORN_TARG_ONLY = (16 << 8)
PT_ATTACK_KNOCKBACK_SELF = (64 << 8)
PT_ATTACK_SACRIFICE = (128 << 8)
PT_ATTACK_HOMING = 8
PT_ATTACK_IMPACT_SPAWN = 1
PT_ATTACK_PENETRATING = 64
PT_ATTACK_PROXIMITY = (2 << 8)
PT_ATTACK_UP_KNOCKBACK = 128
PT_ATTACK_TURRET = (4 << 8)
PT_ATTACK_MELEE_AREA = 4
PT_ATTACK_MELEE_DINOSAUR = (8 << 8)
PT_ATTACK_FLAGS = {PT_ATTACK_CHAINED : 'PT_ATTACK_CHAINED',
                   PT_ATTACK_FLIGHT_SPAWN : 'PT_ATTACK_FLIGHT_SPAWN',
                   PT_ATTACK_GRENADE : 'PT_ATTACK_GRENADE',
                   PT_ATTACK_GROUND_TARG_ONLY : 'PT_ATTACK_GROUND_TARG_ONLY',
                   PT_ATTACK_ENEMIES_ONLY : 'PT_ATTACK_ENEMIES_ONLY',
                   PT_ATTACK_AIRBORN_TARG_ONLY : 'PT_ATTACK_AIRBORN_TARG_ONLY',
                   PT_ATTACK_KNOCKBACK_SELF : 'PT_ATTACK_KNOCKBACK_SELF',
                   PT_ATTACK_SACRIFICE : 'PT_ATTACK_SACRIFICE',
                   PT_ATTACK_HOMING : 'PT_ATTACK_HOMING',
                   PT_ATTACK_IMPACT_SPAWN : 'PT_ATTACK_IMPACT_SPAWN',
                   PT_ATTACK_PENETRATING : 'PT_ATTACK_PENETRATING',
                   PT_ATTACK_PROXIMITY : 'PT_ATTACK_PROXIMITY',
                   PT_ATTACK_UP_KNOCKBACK : 'PT_ATTACK_UP_KNOCKBACK',
                   PT_ATTACK_TURRET : 'PT_ATTACK_TURRET',
                   PT_ATTACK_MELEE_AREA : 'PT_ATTACK_MELEE_AREA',
                   PT_ATTACK_MELEE_DINOSAUR : 'PT_ATTACK_MELEE_DINOSAUR'}

# For Defences, the damage types blocked.  These are bitwise OR-able.
# Choice to make the byte containing water, mental and mystical the
# high-order bit is arbitrary.
PT_DAMAGE_BLOCKED_CRUSH = 1
PT_DAMAGE_BLOCKED_PIERCE = 2
PT_DAMAGE_BLOCKED_ENERGY = 4
PT_DAMAGE_BLOCKED_FIRE = 8
PT_DAMAGE_BLOCKED_RADIATION = 16
PT_DAMAGE_BLOCKED_ELECTRICAL = 32
PT_DAMAGE_BLOCKED_ACID = 64
PT_DAMAGE_BLOCKED_COLD = 128
PT_DAMAGE_BLOCKED_WATER = 256
PT_DAMAGE_BLOCKED_MENTAL = 512
PT_DAMAGE_BLOCKED_MYSTICAL = 1024
# lookup constant name via constant value
PT_DAMAGES_BLOCKED = {PT_DAMAGE_BLOCKED_CRUSH : 'PT_DAMAGE_BLOCKED_CRUSH',
                      PT_DAMAGE_BLOCKED_PIERCE : 'PT_DAMAGE_BLOCKED_PIERCE',
                      PT_DAMAGE_BLOCKED_ENERGY : 'PT_DAMAGE_BLOCKED_ENERGY',
                      PT_DAMAGE_BLOCKED_FIRE : 'PT_DAMAGE_BLOCKED_FIRE',
                      PT_DAMAGE_BLOCKED_RADIATION : 'PT_DAMAGE_BLOCKED_RADIATION',
                      PT_DAMAGE_BLOCKED_ELECTRICAL : 'PT_DAMAGE_BLOCKED_ELECTRICAL',
                      PT_DAMAGE_BLOCKED_ACID : 'PT_DAMAGE_BLOCKED_ACID',
                      PT_DAMAGE_BLOCKED_COLD : 'PT_DAMAGE_BLOCKED_COLD',
                      PT_DAMAGE_BLOCKED_WATER : 'PT_DAMAGE_BLOCKED_WATER',
                      PT_DAMAGE_BLOCKED_MENTAL : 'PT_DAMAGE_BLOCKED_MENTAL',
                      PT_DAMAGE_BLOCKED_MYSTICAL : 'PT_DAMAGE_BLOCKED_MYSTICAL'}

# For Defences, the attack modes blocked.  E.g. PT_RANGED_BLOCKED
PT_MELEE_BLOCKED = 2
PT_RANGED_BLOCKED = 4
PT_AREA_BLOCKED = 8
PT_DIRECT_BLOCKED = 16
PT_ACTIVE_DEFENCE_BLOCKED = 32
PT_PASSIVE_DEFENCE_BLOCKED = 64
PT_SPECIAL_BLOCKED = 128
# lookup constant name via constant value
PT_MODES_BLOCKED = {PT_MELEE_BLOCKED : 'PT_MELEE_BLOCKED',
                    PT_RANGED_BLOCKED : 'PT_RANGED_BLOCKED',
                    PT_AREA_BLOCKED : 'PT_AREA_BLOCKED',
                    PT_DIRECT_BLOCKED : 'PT_DIRECT_BLOCKED',
                    PT_ACTIVE_DEFENCE_BLOCKED : 'PT_ACTIVE_DEFENCE_BLOCKED',
                    PT_PASSIVE_DEFENCE_BLOCKED : 'PT_PASSIVE_DEFENCE_BLOCKED',
                    PT_SPECIAL_BLOCKED : 'PT_SPECIAL_BLOCKED'}

# For Defences, the type of blocking.
PT_BLOCK_TYPE_NORMAL = 1
PT_BLOCK_TYPE_DEFLECT = 2
PT_BLOCK_TYPE_REFLECT = 3
PT_BLOCK_TYPE_REDIRECT = 4
PT_BLOCK_TYPE_ABSORB = 5
PT_BLOCK_TYPE_PASSTHROUGH = 6
# lookup constant name via constant value
PT_BLOCK_TYPES = {PT_BLOCK_TYPE_NORMAL : 'PT_BLOCK_TYPE_NORMAL',
                  PT_BLOCK_TYPE_DEFLECT : 'PT_BLOCK_TYPE_DEFLECT',
                  PT_BLOCK_TYPE_REFLECT : 'PT_BLOCK_TYPE_REFLECT',
                  PT_BLOCK_TYPE_REDIRECT : 'PT_BLOCK_TYPE_REDIRECT',
                  PT_BLOCK_TYPE_ABSORB : 'PT_BLOCK_TYPE_ABSORB',
                  PT_BLOCK_TYPE_PASSTHROUGH : 'PT_BLOCK_TYPE_PASSTHROUGH'}

# For Passive Defences, the frequency of success. NB: These 'percents' are really fractions.
PT_BLOCK_SUCCESS_RARE = 0
PT_BLOCK_SUCCESS_OCCASIONAL = 1
PT_BLOCK_SUCCESS_FREQUENT = 2
PT_BLOCK_SUCCESS_VERY_FREQUENT = 3
PT_BLOCK_SUCCESS_ALWAYS = 4
# lookup constant name via constant value
PT_BLOCK_SUCCESSES = {PT_BLOCK_SUCCESS_RARE : 'PT_BLOCK_SUCCESS_RARE',
                      PT_BLOCK_SUCCESS_OCCASIONAL : 'PT_BLOCK_SUCCESS_OCCASIONAL',
                      PT_BLOCK_SUCCESS_FREQUENT : 'PT_BLOCK_SUCCESS_FREQUENT',
                      PT_BLOCK_SUCCESS_VERY_FREQUENT : 'PT_BLOCK_SUCCESS_VERY_FREQUENT',
                      PT_BLOCK_SUCCESS_ALWAYS : 'PT_BLOCK_SUCCESS_ALWAYS'}
PT_BLOCK_SUCCESS_MIN_PCT_RARE = 0.10
PT_BLOCK_SUCCESS_MIN_PCT_OCCASIONAL = 0.25
PT_BLOCK_SUCCESS_MIN_PCT_FREQUENT = 0.50
PT_BLOCK_SUCCESS_MIN_PCT_VERY_FREQUENT = 0.75
PT_BLOCK_SUCCESS_MIN_PCT_ALWAYS = 0.85
# lookup percent values via frequency constant values
PT_BLOCK_SUCCESS_MIN_PCTS = {PT_BLOCK_SUCCESS_RARE : PT_BLOCK_SUCCESS_MIN_PCT_RARE,
                             PT_BLOCK_SUCCESS_OCCASIONAL : PT_BLOCK_SUCCESS_MIN_PCT_OCCASIONAL,
                             PT_BLOCK_SUCCESS_FREQUENT : PT_BLOCK_SUCCESS_MIN_PCT_FREQUENT,
                             PT_BLOCK_SUCCESS_VERY_FREQUENT : PT_BLOCK_SUCCESS_MIN_PCT_VERY_FREQUENT,
                             PT_BLOCK_SUCCESS_ALWAYS : PT_BLOCK_SUCCESS_MIN_PCT_ALWAYS}

# defence flags (I wonder what 2 was?)
PT_DEFENCE_FLAG_INACTIVE = 1
PT_DEFENCE_FLAG_REMOTE = 2
PT_DEFENCE_FLAG_MOVE = 4
PT_DEFENCE_FLAG_REGENERATE = 8
PT_DEFENCE_FLAG_INFINITE = 16
PT_DEFENCE_FLAGS = {PT_DEFENCE_FLAG_INACTIVE : 'PT_DEFENCE_FLAG_INACTIVE',
                    PT_DEFENCE_FLAG_REMOTE : 'PT_DEFENCE_FLAG_REMOTE',
                    PT_DEFENCE_FLAG_MOVE : 'PT_DEFENCE_FLAG_MOVE',
                    PT_DEFENCE_FLAG_REGENERATE : 'PT_DEFENCE_FLAG_REGENERATE',
                    PT_DEFENCE_FLAG_INFINITE : 'PT_DEFENCE_FLAG_INFINITE'}

# Other power string names
PT_DAMAGE_MAGNITUDE_NAMES = ['none','very low','low','medium','high','very high','extreme']
PT_EP_COST_NAMES = ['none','trace','minor','very low','low','medium','high','very high','extreme']
PT_SPEED_NAMES = ['very slow','slow','normal','fast','very fast']
PT_STUN_NAMES = ['none','low','medium','high','very high','extreme']
PT_KB_NAMES = ['none','low','medium','high','extreme']
PT_RANGE_NAMES = ['short','medium','long','infinite']
PT_RANGE_NAMES.extend(map(str,range(4,100)))
PT_ACCURACY_NAMES = ['very low','low','medium','high','very high']
PT_RADIUS_NAMES = ['none','small','medium','large','extreme']

#################################
### Reading DAT Files ###########
#################################

# Read and return all the markers (objects, positionals, sounds, etc.) in
# a mission.dat file.  It returns these in a dictionary whose keys are
# the marker names and whose values are either the template names (for
# objects) or the MT_ constants defined above for all other marker types.
# By default, the mission dat file is the one in the temp directory (e.g.
# Data/temp) of the current mod, which should be the current mission.
# By default, if there has been a previous read, we return the dictionary
# defined by that read.  If ForceRead is not 0, we re-read the dat file,
# even if there has been a previous read.
# NB: This will NOT work if the dat files don't exist.  E.g., the main
# campaign keeps them compressed in the .ff files unless one explicitely
# unzips them.
def Mission_ReadStartingMarkers(FileName='',ForceRead=0,verbose=0):
    global _StartingMarkers
##    tstart = time.clock()
    if FileName=='': # try to guess the current temp directory
        FileName = GetMissionDatFile()
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ('_StartingMarkers' in globals().keys() ) and ( FileName in _StartingMarkers.keys() ):
##        print 'Mission_ReadStartingMarkers: time = %3.5f' % (time.clock()-tstart)
        return _StartingMarkers[FileName].copy()
    if not ( '_StartingMarkers' in globals().keys() ):      # entirely new dictionary
        _StartingMarkers = {}
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    ITEMSUBSIZE1 = 764
    f = open(FileName,'rb')
    dat = f.read()
    f.close()
    _StartingMarkers[FileName] = {}
    datlen = len(dat)
    pos = FIRSTITEMPOS
    while pos < datlen:
        if ( dat[pos] > '\x08') and ( pos == GetDatWordPos(dat,pos) ) :
            words = GetDatWords(dat,pos,3)
            itemName = words[0]
            lenitem = len(itemName)
            if lenitem < 30:
                itemType = ord(dat[pos+lenitem+1])
                if ( itemType == 0 ) and ( itemName == words[1] ):   # an object
                    itemTemplate = words[2]
                    _StartingMarkers[FileName][itemName] = itemTemplate
                    pos2 = pos + 5 + lenitem + ITEMSUBSIZE1
                elif itemType == 1: # a char
                    templatepos = GetDatWordPos(dat,pos,2)+30
                    itemTemplate = GetDatWords(dat,templatepos,1)[0]
                    _StartingMarkers[FileName][itemName] = itemTemplate
                    pos2 = pos + 5 + lenitem + ITEMSUBSIZE1
                elif ( itemType <= 8 ) and ( itemType > 1 ):
                    _StartingMarkers[FileName][itemName] = itemType
                    pos2 = pos + 5 + lenitem + ITEMSUBSIZE1
                else:
                    if itemName in ('_textureDir', '_missionName', '_missionExtents','_layoutFile'):
                        if itemName == '_missionExtents':
                            _StartingMarkers[FileName][itemName] = struct.unpack('6f',dat[pos+lenitem+1:pos+lenitem+25])
                            pos2 = pos + 5 + lenitem + ITEMSUBSIZE1
                            ##print 'caught _missionExtents'
                        else:
                            _StartingMarkers[FileName][itemName] = words[1]
                            ##print 'caught %s' % itemName
                            pos2 = pos + lenitem + len(words[1]) + 2
                    else:
                        ##print 'unrecognized extra data type at %6d\t%30s\t%d\t%s' % (pos,itemName,itemType,words)
                        pos2 = pos + 5 + lenitem + ITEMSUBSIZE1
                if verbose: print '%6d\t%30s\t%d\t%s' % (pos,itemName,itemType,words)
                if ( pos2 < datlen ) and ( dat[pos2] > '\x08' ) and ( pos2 == GetDatWordPos(dat,pos2) ):
                    words = GetDatWords(dat,pos2,3)
                    itemName = words[0]
                    lenitem = len(itemName)
                    if itemName in ('_textureDir', '_missionName', '_missionExtents','_layoutFile'):
                        if itemName == '_missionExtents':
                            _StartingMarkers[FileName][itemName] = struct.unpack('6f',dat[pos2+lenitem+1:pos2+lenitem+25])
                        else:
                            _StartingMarkers[FileName][itemName] = words[1]
                    else:
                        ##print 'not a recognized extra data type at %6d\t%30s\t%d\t%s' % (pos2,itemName,itemType,words)
                        if ( lenitem < 30 ) and ( dat[pos2+lenitem] == '\x00' ):
                            itemType = ord(dat[pos2+lenitem+1])
                            if ( itemType == 0 ) and ( itemName == words[1] ):   # an object
                                itemTemplate = words[2]
                                _StartingMarkers[FileName][itemName] = itemTemplate
                                ##print 'caught an object at %d' % pos2
                            elif itemType == 1: # a char
                                templatepos = GetDatWordPos(dat,pos2,2)+30
                                itemTemplate = GetDatWords(dat,templatepos,1)[0]
                                _StartingMarkers[FileName][itemName] = itemTemplate
                                ##print 'caught a char <%s> at %d' % (itemTemplate,pos2)
                            elif ( itemType <= 8 ) and ( itemType > 1 ):
                                _StartingMarkers[FileName][itemName] = itemType
                                ##print 'caught a node at %d' % pos2
                            else:
                                ##print 'no good interp of data found'
                                pass
                    if verbose: print '%6d\t%30s\t%d\t%s' % (pos2,itemName,itemType,words)
        pos = pos + ITEMSIZE
##    print 'Mission_ReadStartingMarkers: time = %3.5f' % (time.clock()-tstart)
    return _StartingMarkers[FileName].copy()

def Mission_GetStartingObjects(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    mkeys = markers.keys()
    for nonobj in ('_textureDir', '_missionName', '_missionExtents','_layoutFile'):
        if nonobj in mkeys:
            mkeys.remove(nonobj)
    return filter(lambda k,markers=markers: isinstance(markers[k],types.StringType),mkeys)

def Mission_GetStartingLights(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_LIGHT,markers.keys())

def Mission_GetStartingSounds(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_SOUND,markers.keys())

def Mission_GetStartingPositionals(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_POSITIONAL,markers.keys())

def Mission_GetStartingRoadNodes(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_ROAD_NODE,markers.keys())

def Mission_GetStartingCivNodes(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_CIV_NODE,markers.keys())

def Mission_GetStartingTraffics(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    return filter(lambda k,markers=markers: markers[k]==MT_TRAFFIC,markers.keys())

def Mission_GetMissionName(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    if '_missionName' in markers.keys():
        return markers['_missionName']
    return ''

def Mission_GetTextureDir(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    if '_textureDir' in markers.keys():
        return markers['_textureDir']
    return ''

def Mission_GetLayoutFile(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    if '_layoutFile' in markers.keys():
        return markers['_layoutFile']
    return ''

def Mission_GetMapExtents(MissionFileName=''):
    markers = Mission_ReadStartingMarkers(MissionFileName)
    if '_missionExtents' in markers.keys():
        return markers['_missionExtents']
    return ''

# Read a characters.dat file and return a dictionary with the characters' data.
# The dictionary keys are the character names and each value is another dictionary of
# field names and values where the fields are 'CSBase','attribs','tier_a','tier_b',
# 'tier_a_start','tier_b_start','attrib_start', and 'camp_only'.
def Campaign_ReadCharacters(FileName='',ForceRead=0,verbose=0):
    charfields = ['charName', 'isCustom', 'CSBase', 'strength', 'speed', 'endurance',
                  'energy', 'agility', 'characterAttributes', 'VID', 'AI', 'tier_a',
                  'tier_b', 'powers', 'tier_a_start', 'tier_b_start', 'attrib_start',
                  'camp_only', 'movementRadius', 'alterEgo']
    global _CharacterData
##    tstart = time.clock()
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('characters.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_CharacterData' in globals().keys() ) and ( FileName in _CharacterData.keys() ):
##        print 'Campaign_ReadCharacters: time = %3.5f' % (time.clock()-tstart)
        return _CharacterData[FileName].copy()
    if not ( '_CharacterData' in globals().keys() ):      # entirely new dictionary
        _CharacterData = {}
    _CharacterData[FileName] = {}
    try:
        f = open(FileName,'rb')
        alldat = f.read()
        f.close()
    except:
        print 'Campaign_ReadCharacters: unable to read data from %s' % FileName
        return _CharacterData[FileName]
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    headermark = 'HDR4'
    datlen = len(alldat)
    pos = FIRSTITEMPOS
    while pos < datlen:
        if ord(alldat[pos]) > 8:   # we have a character name (potentially)
            dat = alldat[pos:pos+ITEMSIZE]
            words = GetDatWords(dat,0,2)
            if words[1][:len(headermark)] == headermark:
                charName = words[0]
                offset = GetDatWordPos(dat,0,5)
                CSBase = string.split(dat[offset+30:],'\x00',1)[0]
                strength = ord(dat[offset+285])
                speed = ord(dat[offset+286])
                endurance = ord(dat[offset+287])
                if dat[offset+288]!='\x00': print 'for %s, %d+288 = %d (%x)' % (charName,offset,ord(dat[offset+288]),ord(dat[offset+288]))
                energy = ord(dat[offset+289])
                agility = ord(dat[offset+290])
                attribs = []
                count = 0
                while ( dat[offset+291+30*count] >= '\x20' ) and ( count < 5 ):
                    attribs.append(string.split(dat[offset+291+30*count:],'\x00',1)[0])
                    count = count + 1
                powers = {}
                tier_a = []
                count = 0
                while ( dat[offset+441+30*count] >= '\x20' ) and ( count < 5 ):
                    power = string.split(dat[offset+441+30*count:],'\x00',1)[0]
                    tier_a.append(power)
                    powers[power] = {}
                    count = count + 1
                tier_b = []
                count = 0
                while ( dat[offset+591+30*count] >= '\x20' ) and ( count < 5 ):
                    power = string.split(dat[offset+591+30*count:],'\x00',1)[0]
                    tier_b.append(power)
                    powers[power] = {}
                    count = count + 1
                if dat[offset+751] != '\x00':
                    VID = string.split(dat[offset+751:],'\x00',1)[0]
                else:
                    VID = ''
                AI = string.split(dat[offset+754:],'\x00',1)[0]
                tier_a_start = ord(dat[offset+786])
                tier_b_start = ord(dat[offset+787])
                attrib_start = ord(dat[offset+790])
                camp_only = ord(dat[offset+791])
                movementRadius = struct.unpack('f',dat[offset+797:offset+797+4])[0]
                alterEgo = string.split(dat[offset + 801:],'\x00',1)[0]
                _CharacterData[FileName][charName] = {'charName':charName,
                                                      'isCustom':0,
                                                      'CSBase':CSBase,
                                                      'strength':strength,
                                                      'speed':speed,
                                                      'endurance':endurance,
                                                      'energy':energy,
                                                      'agility':agility,
                                                      'characterAttributes':attribs,
                                                      'VID':VID,
                                                      'AI':AI,
                                                      'tier_a':tier_a,
                                                      'tier_b':tier_b,
                                                      'powers':powers,
                                                      'tier_a_start':tier_a_start,
                                                      'tier_b_start':tier_b_start,
                                                      'attrib_start':attrib_start,
                                                      'camp_only':camp_only,
                                                      'movementRadius':movementRadius,
                                                      'alterEgo':alterEgo}
                if verbose:
                    print '\nCampaign_ReadCharacters: reading <%s>' % charName
                    for field in charfields:
                        print '%15s : %s' % (field,_CharacterData[FileName][charName][field])
        pos = pos + ITEMSIZE
##    print 'Campaign_ReadCharacters: time = %3.5f' % (time.clock()-tstart)
    return _CharacterData[FileName].copy()

# Given a character template name and a list of attribute names, read
# the characters.dat file and replace that character's attributes with
# those in the list and rewrite the DAT file.  Optional attribStart
# specifies the number of starting attributes.  Optional FileName
# specifies the full path the of the characters.dat file.  Optional
# updatePreviousRead tells whether to update any previous read of this
# characters.dat file with the new attribute and attribStart
# information  That is, if updatePreviousRead is true (the default),
# then there is no need to call Campaign_ReadCharacters for that file
# (if it has previously been read) for that function to get the
# updated data.

# NB: Be carefull with this.  1) It does not make any backups of the
# DAT file.  2) It does refresh the datfiles data for that character,
# but it doesn't update anything that doesn't call
# Campaign_ReadCharacters().  3) It doesn't check that the given
# attribute list contains valid attributes.
def Campaign_RewriteCharacterAttributes(char,attribList,attribStart=5,FileName='', updatePreviousRead=1):
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('characters.dat')
    FileName = os.path.abspath(FileName)
    try:
        f = open(FileName,'rb')
        alldat = f.read()
        f.close()
    except:
        print 'Campaign_RewriteCharacterAttributes: unable to read data from %s' % FileName
        return
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    headermark = 'HDR4'
    datlen = len(alldat)
    pos = FIRSTITEMPOS
    changed = 0
    while pos < datlen:
        if ord(alldat[pos]) > 8:   # we have a character name (potentially)
            dat = alldat[pos:pos+ITEMSIZE]
            words = GetDatWords(dat,0,2)
            if words[1][:len(headermark)] == headermark:
                charName = words[0]
                if char == charName:
                    origdat = dat
                    offset = GetDatWordPos(dat,0,5)
                    attribstr = ''
                    for attrib in attribList[:5]:
                        attribstr = attribstr + attrib[:30] + '\x00'*(30-len(attrib[:30]))
                    attribstr = attribstr + '\x00'*(150-len(attribstr))
                    dat = dat[:offset+291]+attribstr+dat[offset+441:]
                    dat = dat[:offset+790]+chr(attribStart)+dat[offset+791:]
                    if not origdat==dat:
                        alldat = alldat[:pos] + dat + alldat[pos+ITEMSIZE:]
                        changed = 1
                    break
        pos = pos + ITEMSIZE
    # now rewrite the DAT file
    if changed:
        try:
            f = open(FileName,'wb')
            f.write('%s' % alldat)
            f.close()
            global _CharacterData
            if updatePreviousRead and ( '_CharacterData' in globals().keys() ) and ( FileName in _CharacterData.keys() ):
                _CharacterData[FileName][char]['characterAttributes'] = attribList
                _CharacterData[FileName][char]['attrib_start'] = attribStart        
                # uodate _ModCharData for calls to Campaign_GetCharacterData()
                ModPath = os.path.dirname(FileName)
                global _ModCharData
                if ( '_ModCharData' in globals().keys() ) and ( ModPath in _ModCharData.keys() ):
                    _ModCharData[ModPath][char]['characterAttributes'] = attribList
                    _ModCharData[ModPath][char]['activeAttributes'] = \
                            min(_ModCharData[ModPath][char]['attrib_start'],len(_ModCharData[ModPath][char]['characterAttributes']))
        except:
            print 'Campaign_RewriteCharacterAttributes: unable to write changed data to %s' % FileName
            return
    else:
        print 'Campaign_RewriteCharacterAttributes: no change made to %s' % FileName

# Given a character template name and a list of power names, read the
# characters.dat file and replace that character's attributes with
# those in the list and rewrite the DAT file.  Note that the powerList
# must have the power names for both tier A and tier B and any empty
# powers in tier A must be empty strings.  Optional TierAStart and
# TierBStart specify the number of starting powers in each tier.
# Optional FileName specifies the full path the of the characters.dat
# file.  Optional updatePreviousRead tells whether to update any
# previous read of this characters.dat file with the new attribute and
# attribStart information  That is, if updatePreviousRead is true (the
# default), then there is no need to call Campaign_ReadCharacters for
# that file (if it has previously been read) for that function to get
# the updated data.

# NB: Be carefull with this.  1) It does not make any backups of the
# DAT file.  2) It does refresh the datfiles data for that character,
# but it doesn't update anything that doesn't call
# Campaign_ReadCharacters().  3) It doesn't check that the given
# attribute list contains valid attributes.
def Campaign_RewriteCharacterPowers(char,powersList,tierAStart=5,tierBStart=5,FileName='', updatePreviousRead=1):
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('characters.dat')
    FileName = os.path.abspath(FileName)
    try:
        f = open(FileName,'rb')
        alldat = f.read()
        f.close()
    except:
        print 'Campaign_RewriteCharacterPowers: unable to read data from %s' % FileName
        return
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    headermark = 'HDR4'
    datlen = len(alldat)
    pos = FIRSTITEMPOS
    changed = 0
    while pos < datlen:
        if ord(alldat[pos]) > 8:   # we have a character name (potentially)
            dat = alldat[pos:pos+ITEMSIZE]
            words = GetDatWords(dat,0,2)
            if words[1][:len(headermark)] == headermark:
                charName = words[0]
                if char == charName:
                    origdat = dat
                    offset = GetDatWordPos(dat,0,5)
                    powerstr = ''
                    for power in powersList[:10]:
                        powerstr = powerstr + power[:30] + '\x00'*(30-len(power[:30]))
                    powerstr = powerstr + '\x00'*(300-len(powerstr))
                    dat = dat[:offset+441]+powerstr+dat[offset+741:]
                    dat = dat[:offset+786]+chr(tierAStart)+chr(tierBStart)+dat[offset+788:]
                    if not origdat==dat:
                        alldat = alldat[:pos] + dat + alldat[pos+ITEMSIZE:]
                        changed = 1
                    break
        pos = pos + ITEMSIZE
    # now rewrite the DAT file
    if changed:
        try:
            f = open(FileName,'wb')
            f.write('%s' % alldat)
            f.close()
            global _CharacterData
            if updatePreviousRead and ( '_CharacterData' in globals().keys() ) and ( FileName in _CharacterData.keys() ):
                _CharacterData[FileName][char]['tier_a'] = filter(lambda p:p,powersList[:5])
                _CharacterData[FileName][char]['tier_b'] = filter(lambda p:p,powersList[5:])
                _CharacterData[FileName][char]['tier_a_start'] = tierAStart
                _CharacterData[FileName][char]['tier_b_start'] = tierBStart
                _CharacterData[FileName][char]['powers'] = {}
                for power in filter(lambda p:p,powersList):
                    _CharacterData[FileName][char]['powers'][power] = {}
                # uodate _ModCharData for calls to Campaign_GetCharacterData()
                ModPath = os.path.dirname(FileName)
                global _ModCharData
                if ( '_ModCharData' in globals().keys() ) and ( ModPath in _ModCharData.keys() ):
                    _ModCharData[ModPath][char]['tier_a'] = filter(lambda p:p,powersList[:5])
                    _ModCharData[ModPath][char]['tier_b'] = filter(lambda p:p,powersList[5:])
                    _ModCharData[ModPath][char]['tier_a_start'] = tierAStart
                    _ModCharData[ModPath][char]['tier_b_start'] = tierBStart
                    _ModCharData[ModPath][char]['powerLevels'] = {}
                    for i in range(len(_ModCharData[ModPath][char]['tier_a'])):
                        if i < _ModCharData[ModPath][char]['tier_a_start'] - 1:
                            _ModCharData[ModPath][char]['powerLevels'][_ModCharData[ModPath][char]['tier_a'][i]] = 3
                        else:
                            _ModCharData[ModPath][char]['powerLevels'][_ModCharData[ModPath][char]['tier_a'][i]] = 1
                    for i in range(len(_ModCharData[ModPath][char]['tier_b'])):
                        if i < _ModCharData[ModPath][char]['tier_b_start'] - 1:
                            _ModCharData[ModPath][char]['powerLevels'][_ModCharData[ModPath][char]['tier_b'][i]] = 3
                        else:
                            _ModCharData[ModPath][char]['powerLevels'][_ModCharData[ModPath][char]['tier_b'][i]] = 1
                    _ModCharData[ModPath][char]['powers'] = {}
                    powerdata = Campaign_ReadPowers(os.path.join(ModPath,'powers.dat'))
                    for power in filter(lambda p:p,powersList):
                        _ModCharData[ModPath][char]['powers'][power] = {}
                    for p in _ModCharData[ModPath][char]['powers'].keys():
                        if p in powerdata.keys():
                            _ModCharData[ModPath][char]['powers'][p] = powerdata[p]
                        else:
                            if verbose:
                                print 'Campaign_RewriteCharacterPowers: character <%s> has power <%s> with no corresponding powers.dat entry.' % (char,p)
                            pass

        except:
            print 'Campaign_RewriteCharacterPowers: unable to write changed data to %s' % FileName
            return
    else:
        print 'Campaign_RewriteCharacterPowers: no change made to %s' % FileName

# Read a object.dat file and return a dictionary with the objects' data
def Campaign_ReadObjects(FileName='',ForceRead=0,verbose=0):
    global _ObjectData
##    tstart = time.clock()
    if FileName=='':                                    # try to guess the current DAT file
        FileName = GetDatFile('objects.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_ObjectData' in globals().keys() ) and ( FileName in _ObjectData.keys() ):
##        print 'Campaign_ReadObjects: time = %3.5f' % (time.clock()-tstart)
        return _ObjectData[FileName].copy()
    if not ( '_ObjectData' in globals().keys() ):      # entirely new dictionary
        _ObjectData = {}
    _ObjectData[FileName] = {}
    try:
        f = open(FileName,'rb')
        alldat = f.read()
        f.close()
    except:
        print 'Campaign_ReadObjects: unable to read data from %s' % FileName
        return _ObjectData[FileName]
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    datlen = len(alldat)
    pos = 0
    while pos < datlen:
        gotobj = 0
        if verbose:  print 'Campaign_ReadObjects: looking at pos %d' % (pos,)
        if ord(alldat[pos]) > 8:                        # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = string.split(dat,'\x00',2)
            if verbose:  print 'Campaign_ReadObjects: words=%s' % (words,)
            if words[0] == words[1]:                    # check that it's a legit name
                gotobj = 1
                offset = GetDatWordPos(dat,0,2)         # first attrib
                objName = words[0]
                _ObjectData[FileName][objName] = {}
                _ObjectData[FileName][objName]['templateName'] = objName
                _ObjectData[FileName][objName]['NIF'] = string.split(dat[offset+30:],'\x00',1)[0]
                _ObjectData[FileName][objName]['class'] = string.split(dat[offset+285:],'\x00',1)[0]
                offset = offset + 316
                for count in range(15):
                    if dat[offset+count*36] != '\x00':
                        attrib = string.split(dat[offset+36*count:],'\x00',1)[0]
                        val = struct.unpack('f',dat[offset+36*count+32:offset+36*count+36])[0]
                        if _ObjectData[FileName][objName].has_key(attrib):
                            print 'Campaign_ReadObjects() warning: object template <%s> has more than one entry for attribute <%s>, values %f and %f' % (objName,attrib,_ObjectData[FileName][objName][attrib],val)
                        _ObjectData[FileName][objName][attrib] = val
                if dat[offset+570] != '\x00' :  _ObjectData[FileName][objName]['skin'] = string.split(dat[offset+570:],'\x00',1)[0]
                if dat[offset+855] != '\x00' :  _ObjectData[FileName][objName]['explosion'] = string.split(dat[offset+855:],'\x00',1)[0]
                if dat[offset+1140] != '\x00' :  _ObjectData[FileName][objName]['arg2'] = string.split(dat[offset+1140:],'\x00',1)[0]
                if dat[offset+1425] != '\x00' :  _ObjectData[FileName][objName]['arg3'] = string.split(dat[offset+1425:],'\x00',1)[0]
                if dat[offset+1710] != '\x00' :  _ObjectData[FileName][objName]['arg4'] = string.split(dat[offset+1710:],'\x00',1)[0]
                if verbose:
                    print '\nCampaign_ReadObjects: reading <%s>' % objName
                    for key in _ObjectData[FileName][objName].keys():
                        print '%36s = %s' % (key,str(_ObjectData[FileName][objName][key]))
        if gotobj:
            pos = pos + ITEMSIZE
        else:
            pos = pos + SKIPSIZE
##    print 'Campaign_ReadObjects: time = %3.5f' % (time.clock()-tstart)
    return _ObjectData[FileName].copy()

# Read a power.dat file and return a dictionary with the powers' data
def Campaign_ReadPowers(FileName='',ForceRead=0,verbose=0):
    global _PowerData
##    tstart = time.clock()
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('powers.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_PowerData' in globals().keys() ) and ( FileName in _PowerData.keys() ):
##        tend = time.clock()
##        print 'Campaign_ReadPowers: (cached) time = %3.4f' % (time.clock()-tstart)
        return _PowerData[FileName].copy()
    if not ( '_PowerData' in globals().keys() ):      # entirely new dictionary
        _PowerData = {}
    _PowerData[FileName] = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Campaign_ReadPowers: unable to read data from %s' % FileName
        return _PowerData[FileName]
    FIRSTITEMPOS = 2048
    positions = GetPowerRecordPositions(dat,FIRSTITEMPOS)
    for pos in positions:
        npr = dat[pos:pos+184]
        pname = string.split(npr,'\x00',1)[0]
        _PowerData[FileName][pname] = ParsePowerBlock(npr)
        if verbose: print 'Campaign_ReadPowers: <%s> pos = %d' % (pname,pos)
##    print 'Campaign_ReadPowers: (read) time = %3.4f' % (time.clock()-tstart)
    return _PowerData[FileName].copy()

# Change powers.dat for a particular power.  This might be useful
# because FF2 allows re-reading powers.dat during a mission.  THIS MAY
# BE RISKY!!!
#
# We give a powers.dat file name or empty string if we are to guess
# the current mod's powers.dat.  We also give list of power
# dictionaries to overwrite in that DAT file. These dictiomaries are
# of the sort returned by Campaign_ReadPowers().
#
# This function Also updates the _PowerData dictionary, so that it is
# unnecessary to force a re-read of powers.dat to update the local
# dictionary.  It may, however, be necessary to call
# Campaign_ReadPowers() again to get a fresh copy of the dictionery
# outside this module, but at least that will require no file i/o.

# list of functions that want to be informed when powers are re-written.
if 'POWER_REWRITE_CALLBACKS' not in globals().keys():
    POWER_REWRITE_CALLBACKS = {}
    
# have powers been rewritten by Campaign_RewritePowers()?
def Campaign_PowersRewritten(FileName=''):
    if FileName=='': # try to guess the current temp directory
        FileName = GetDatFile('powers.dat')
    FileName = os.path.abspath(FileName)
    BackupFileName = '%s.bak' % (FileName,)
    return os.path.isfile(BackupFileName)

# Delete the current (rewritten) powers.dat and rename the .bak
# version in its place.  ForceReread (==1 by default) forces a re-read
# of the reverted power DAT to keep things up to date.  Also clears the
# REWRITES attrib.
def Campaign_RevertPowers(FileName='',ForceReread=1):
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('powers.dat')
    FileName = os.path.abspath(FileName)
    BackupFileName = '%s.bak' % (FileName,)
    if os.path.isfile(BackupFileName):
        try:
            if os.path.isfile(FileName):
                os.remove(FileName)
            os.rename(FileName,BackupFileName)
            Campaign_ReadPowers(FileName,ForceRead=ForceReread)
        except OSError:
            print 'Campaign_RevertPowers: Unable to revert to backup of file <%s>, possible Read-Only problem.  Quitting' % (FileName,)
            raise
        except:
            print 'Campaign_RevertPowers: Unable to revert to backup of file <%s>.  Quitting' % (FileName,)
            raise
    else:
        print 'Campaign_RevertPowers: no backup file <%s> found.' % BackupFileName
    
def Campaign_RewritePowers(FileName='',ChangePowers=[],ForceRead=0):
    global _PowerData
    ChangedPowers = []
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('powers.dat')
    FileName = os.path.abspath(FileName)
    if not ( '_PowerData' in globals().keys() ):      # entirely new dictionary
        _PowerData = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Campaign_RewritePowers: unable to read data from %s' % FileName
        return _PowerData[FileName]
    # some healf-hearted attempt at safety
    if not Campaign_PowersRewritten(FileName):
        try:
            i = 1
            BackupFileName = '%s.bak' % (FileName,)
            while os.path.isfile(BackupFileName):
                i = i + 1   # first numbered .bak will be '.bak2'
                BackupFileName = '%s.bak%d' % (FileName,i)
            os.rename(FileName,BackupFileName)
        except OSError:
            print 'Campaign_RewritePowers: Unable to create backup of file <%s>, possible Read-Only problem.  Quitting' % (FileName,)
            raise
        except:
            print 'Campaign_RewritePowers: Unable to create backup of file <%s>.  Quitting' % (FileName,)
            raise
        
    FIRSTITEMPOS = 2048
    if ( ForceRead ) or ( FileName not in _PowerData.keys() ):
        _PowerData[FileName] = {}
        positions = GetPowerRecordPositions(dat,FIRSTITEMPOS)
        ppositions = {}
        for pos in positions:
            npr = dat[pos:pos+184]
            pname = string.split(npr,'\x00',1)[0]
            ppositions[pname] = pos
            _PowerData[FileName][pname] = ParsePowerBlock(npr)
    else:
        positions = GetPowerRecordPositions(dat,FIRSTITEMPOS)
        ppositions = {}
        for pos in positions:
            npr = dat[pos:pos+184]
            pname = string.split(npr,'\x00',1)[0]
            ppositions[pname] = pos
    # We need to do this in reverse order of where the blocks start in the data.  Or, we
    # need to ge the positions anew after each pwoer is done.  Trying the first way...
    CPs = []
    for power in ChangePowers:
        pname = power['PowerName']
        blockpos = ppositions[pname]            # where in the DAT the 184 byte block starts
        startpos = blockpos - len(pname) - 1
        endpos = blockpos + 184
        newblock = pname + '\x00' + CreateDATBlock(power,ptemp,0)
##        newblock = pname + '\x00' + CreateDATBlock(power,ptemp,1)
        CPs.append((startpos,endpos,newblock))
        _PowerData[FileName][pname] = power
        ChangedPowers.append(pname)
    CPs.sort()
    CPs.reverse()
    for cp in CPs:
        dat = dat[:cp[0]] + cp[2] + dat[cp[1]:]
    f = open(FileName,'wb')
    f.write('%s' % dat)
    f.close()
    if inGame():
        HandlePowerChangeCallbacks(ChangedPowers)

# Make a pseudo-event register sink for power changes to make
# scripting them more familiar
REWRITE_EVENT = 65  # I'm just stealing this number.  I hope it's not used.
class PseudoEvent:
    def __init__(self, eventType=REWRITE_EVENT, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal

# Add a sink so that a script function is called after a power in
# powers.dat is changed using Campaign_RewritePowers().
def regPowerRewrite(powername,callback,persistent=0,string='',float=0,user=0):
    import missionobjvar
    global POWER_REWRITE_CALLBACKS
    if CheckReload('regPowerRewrite') and inGame() and missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS'):
        POWER_REWRITE_CALLBACKS = missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS')
    if powername in POWER_REWRITE_CALLBACKS.keys():
        POWER_REWRITE_CALLBACKS[powername].append((callback,string,persistent,float,user))
    else:
        POWER_REWRITE_CALLBACKS[powername] = [(callback,string,persistent,float,user)]
    if inGame():
        missionobjvar.Mission_SetVar('_datfiles_POWER_REWRITE_CALLBACKS',POWER_REWRITE_CALLBACKS)

        
# Cancel all regPowerRewrite() sinks to a particular callback function
def cancelRewriteCallbackSinks(callback):
    import missionobjvar
    global POWER_REWRITE_CALLBACKS
    if CheckReload('regPowerRewrite') and inGame() and missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS'):
        POWER_REWRITE_CALLBACKS = missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS')
    for pname in POWER_REWRITE_CALLBACKS.keys():
        callbacks = POWER_REWRITE_CALLBACKS[pname]
        cbremove = []
        for cb in callbacks:
            if cb[0] == callback:
                cbremove.append(cb)
        for cb in cbremove:
            callbacks.remove(cb)
        if callbacks:   # some were persistent
            POWER_REWRITE_CALLBACKS[pname] = callbacks
        else:
            del POWER_REWRITE_CALLBACKS[pname]
    if inGame():
        missionobjvar.Mission_SetVar('_datfiles_POWER_REWRITE_CALLBACKS',POWER_REWRITE_CALLBACKS)
    
# Cancel all regPowerRewrite() sinks for a particular power change
def cancelRewritePowerSinks(powername):
    import missionobjvar
    global POWER_REWRITE_CALLBACKS
    if CheckReload('regPowerRewrite') and inGame() and missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS'):
        POWER_REWRITE_CALLBACKS = missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS')
    if powername in POWER_REWRITE_CALLBACKS.keys():
        del POWER_REWRITE_CALLBACKS[powername]
    if inGame():
        missionobjvar.Mission_SetVar('_datfiles_POWER_REWRITE_CALLBACKS',POWER_REWRITE_CALLBACKS)

# Cancel all regPowerRewrite() sinks to a particular callback function
def cancelAllRewriteSinks():
    import missionobjvar
    global POWER_REWRITE_CALLBACKS
    POWER_REWRITE_CALLBACKS = {}
    if inGame():
        missionobjvar.Mission_SetVar('_datfiles_POWER_REWRITE_CALLBACKS',POWER_REWRITE_CALLBACKS)
    
# Execute the callback functions
def HandlePowerChangeCallbacks(powerslist):
    import missionobjvar
    global POWER_REWRITE_CALLBACKS
    if CheckReload('regPowerRewrite') and inGame() and missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS',{}):
        POWER_REWRITE_CALLBACKS = missionobjvar.Mission_GetVar('_datfiles_POWER_REWRITE_CALLBACKS')
    matchall = ( '' in POWER_REWRITE_CALLBACKS.keys() )
    for pname in powerslist:
        if ( pname in POWER_REWRITE_CALLBACKS.keys() ):
            callbacks = POWER_REWRITE_CALLBACKS[pname]
            cbremove = []
            for cb in callbacks:
                event = PseudoEvent(REWRITE_EVENT,objectStr=pname, stringStr=cb[1], floatVal=cb[3], dataVal=cb[2], userVal=cb[4])
                fn = cb[0]
                eventeval(fn,event)
                if event.data == 0:     # remove it from callback list for that this power
                    cbremove.append(cb)
            for cb in cbremove:
                callbacks.remove(cb)
            if callbacks:   # some were persistent
                POWER_REWRITE_CALLBACKS[pname] = callbacks
            else:
                del POWER_REWRITE_CALLBACKS[pname]
        if ( matchall ):
            callbacks = POWER_REWRITE_CALLBACKS['']
            cbremove = []
            for cb in callbacks:
                event = PseudoEvent(REWRITE_EVENT,objectStr=pname, stringStr=cb[1], floatVal=cb[3], dataVal=cb[2], userVal=cb[4])
                fn = cb[0]
                eventeval(fn,event)
                if event.data == 0:     # remove it from callback list for that this power
                    cbremove.append(cb)
            for cb in cbremove:
                callbacks.remove(cb)
            if callbacks:   # some were persistent
                POWER_REWRITE_CALLBACKS[''] = callbacks
            else:
                del POWER_REWRITE_CALLBACKS['']
    if inGame():
        missionobjvar.Mission_SetVar('_datfiles_POWER_REWRITE_CALLBACKS',POWER_REWRITE_CALLBACKS)
                
# Read a fx.dat file and return a dictionary with the FXs' data
def Campaign_ReadFX(FileName='',ForceRead=0,verbose=0):
    global _FXData
    tstart = time.clock()
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('fx.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_FXData' in globals().keys() ) and ( FileName in _FXData.keys() ):
        ##print 'Campaign_ReadFX: time = %3.4f' % (time.clock()-tstart)
        return _FXData[FileName].copy()
    if not ( '_FXData' in globals().keys() ):      # entirely new dictionary
        _FXData = {}
    _FXData[FileName] = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Campaign_ReadFX: unable to read data from %s' % FileName
        return _FXData[FileName]
    ## decide which game version we are using
    if inGame():
        gameVer = ffver()
    else: # use routine above
        gameVer = FindFXDATVersion(dat)
    FIRSTITEMPOS = 2048
    ITEMSIZE = 1024
    datlen = len(dat)
    if gameVer == 1:  
        pos = FIRSTITEMPOS
        npr, foundat = GetNextFXRecord(dat,pos)
        words = GetDatWords(npr,0,3,'\x1f')
        fxname = words[0]
        offset = GetDatWordPos(npr,0,2,'\x1f')
        while ( npr ):### and len(_FXData[FileName])<10:
            if verbose: print 'Campaign_ReadFX: (FF1 format) <%30s>           (%d,%d,%d)' % (fxname,pos,offset,len(npr))
            _FXData[FileName][fxname] = {}
            _FXData[FileName][fxname]['FXName'] = fxname
            pType = ord(npr[offset+468])
            _FXData[FileName][fxname]['PowerType'] = pType
            if pType == PT_RANGED:
                SubType = ord(npr[offset+472])
                _FXData[FileName][fxname]['SubType'] = SubType
            _FXData[FileName][fxname]['NIF'] = words[2]
            if npr[offset+375] != '\x00' :  _FXData[FileName][fxname]['NodeStart'] = GetDatWords(npr,offset+375,1,'\x1f')[0]
            if npr[offset+405] != '\x00' :  _FXData[FileName][fxname]['NodeCore'] = GetDatWords(npr,offset+405,1,'\x1f')[0]
            if npr[offset+435] != '\x00' :  _FXData[FileName][fxname]['NodeEnd'] = GetDatWords(npr,offset+435,1,'\x1f')[0]
            _FXData[FileName][fxname]['NodesTrackRot'] = ord(npr[offset+476])
            #now set up next read
            pos = foundat + len(npr)
            npr, foundat = GetNextFXRecord(dat,pos)
            if npr:
                words = GetDatWords(npr,0,3,'\x1f')
                fxname = words[0]
                offset = GetDatWordPos(npr,0,2,'\x1f')
    elif gameVer == 2:   # FFv2 format
        hdr = '\x00HDR4\x06\x00\x00\x00sFXDef\x01'
        hpos = find(dat,hdr,FIRSTITEMPOS)
        hposnext = hpos
        clen = struct.unpack('h',dat[hpos+19:hpos+21])[0]
        npr = dat[hpos+23:hpos+23+clen]
        while ( hpos >= 0 ):
            words = GetDatWords(npr,0,2)
            fxname = words[0]
            if verbose: print 'Campaign_ReadFX: (FF2 format) <%30s>           (%d,%d,%d)' % (fxname,hpos,hposnext,len(npr))
            _FXData[FileName][fxname] = {}
            _FXData[FileName][fxname]['FXName'] = fxname
            _FXData[FileName][fxname]['NIF'] = words[1]
            pType = ord(npr[375])
            _FXData[FileName][fxname]['PowerType'] = pType
            if pType == PT_RANGED:
                _FXData[FileName][fxname]['SubType'] = ord(npr[379])
            if npr[285] != '\x00' :  _FXData[FileName][fxname]['NodeStart'] = string.split(npr[285:],'\x00',1)[0]
            if npr[315] != '\x00' :  _FXData[FileName][fxname]['NodeCore'] = string.split(npr[315:],'\x00',1)[0]
            if npr[345] != '\x00' :  _FXData[FileName][fxname]['NodeEnd'] = string.split(npr[345:],'\x00',1)[0]
            if npr[387] != '\x00' :  _FXData[FileName][fxname]['NodeLaunch'] = string.split(npr[387:],'\x00',1)[0]
            _FXData[FileName][fxname]['NodesTrackRot'] = ord(npr[383])
            #now set up next read
            hpos = string.find(dat,hdr,hpos+22+clen)
            if ( hpos >= 0 ):
                clen = struct.unpack('h',dat[hpos+19:hpos+21])[0]
                npr = dat[hpos+23:hpos+23+clen]
    ##print 'Campaign_ReadFX: time = %3.4f' % (time.clock()-tstart)
    return _FXData[FileName].copy()

# Read a attributes.dat file and return a dictionary with the attributes' data,
# which is just prestige cost.
def Campaign_ReadAttributes(FileName='',ForceRead=0,verbose=0):
    global _AttributeData
    ##tstart = time.clock()
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('attributes.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_AttributeData' in globals().keys() ) and ( FileName in _AttributeData.keys() ):
        return _AttributeData[FileName].copy()
    if not ( '_AttributeData' in globals().keys() ):      # entirely new dictionary
        _AttributeData = {}
    _AttributeData[FileName] = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Campaign_ReadAttributes: unable to read data from %s' % FileName
        return _AttributeData[FileName]
    FIRSTITEMPOS = 2048
    strings = string.split(dat[FIRSTITEMPOS:],'\x00')
    attribs = []
    attriballowable = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!#$%&()*+,-./:;<=>?@[\\]^_`{|}~ '
    for i in range(len(strings)-1):
        if ( strings[i] ) and ( strings[i+1] == strings[i][-len(strings[i+1]):] ) and \
           ( not filter(lambda c,allow=attriballowable: c not in allow, strings[i+1]) ):
            attribs.append(strings[i+1])
            i = i + 1
    pos = FIRSTITEMPOS
    for attrib in attribs:
        pos = string.find(dat,attrib+'\x00'+attrib,pos)
        valpos = pos + len(attrib) + 33
        if not string.find(dat[pos+1+2*len(attrib):valpos-1],'\xFF\xFF\xFF\xFF') > -1: # no end of field marker between name and value
            _AttributeData[FileName][attrib] = struct.unpack('h',dat[valpos:valpos+2])[0]
            if verbose: print '%30s:  pos = %6d  valpos = %6d  val = %d' % (attrib,pos,valpos,_AttributeData[FileName][attrib])
        else:
            if verbose: print '%30s:  pos = %6d  valpos = %6d  likely bogus entry: skipping' % (attrib,pos,valpos)

        pos = valpos + 35
    ##print 'Campaign_ReadAttributes: time = %3.4f' % (time.clock()-tstart)
    return _AttributeData[FileName].copy()

# Read a keyframes file and return a dictionary with the time info for
# each animation in the file.
def Keyframe_ReadAnimationTimes(FileName='',ForceRead=0,verbose=0):
    global _AnimationData
    ##tstart = time.clock()
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_AnimationData' in globals().keys() ) and ( FileName in _AnimationData.keys() ):
        ##print 'Keyframe_ReadAnimationTimes: cached time = %3.4f' % (time.clock()-tstart)
        return _AnimationData[FileName].copy()
    if not ( '_AnimationData' in globals().keys() ):      # entirely new dictionary
        _AnimationData = {}
    _AnimationData[FileName] = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Keyframe_ReadAnimationTimes: unable to read data from %s' % FileName
        return _AnimationData[FileName]
    datlen = len(dat)
    if dat[:22] == 'NetImmerse File Format':    # FF keyframe file
        huntstr = 'NiSequenceStreamHelper'
        pos = string.find(dat,huntstr)
        while (pos<datlen) and (pos!=-1) :
            namepos = pos + 26
            namelen = ord(dat[namepos - 4])
            name = dat[namepos:namepos+namelen]
            if verbose:  print '%6d %s %d' % (pos,name,namepos)
            nextpos = string.find(dat,huntstr,pos+1)
            startp = string.find(dat,'\x05\x00\x00\x00start',pos)
            startval = struct.unpack('f',dat[startp-4:startp])[0]
            nodes = [('start',startval)]
            nextp = startp+17
            nextval = struct.unpack('f',dat[startp+9:startp+13])[0]
            nextstrlen = ord(dat[startp+13])
            nextstr = dat[nextp:nextp+nextstrlen]
            while ( ( nextp < nextpos ) or ( nextpos == -1 ) ) and ( nextstr != 'end' ):
                ##print '  nextp = %d  <%s>' % (nextp,nextstr)
                nodes.append((nextstr,nextval))
                strlen = nextstrlen
                nextval = struct.unpack('f',dat[nextp+strlen:nextp+strlen+4])[0]
                nextstrlen = ord(dat[nextp+strlen+4])
                nextp = nextp+strlen+8
                nextstr = dat[nextp:nextp+nextstrlen]
            ##print '  nextp = %d  <%s>' % (nextp,nextstr)
            nodes.append((nextstr,nextval))
            if verbose:  print nodes
            _AnimationData[FileName][name]=nodes
            pos = nextpos
    else:   # This is an FFvT3R (GameBryo) file, orginally or converted
        huntstr = 'Root_NoteTrack'      # FFvT3R keyframe file
        FF1converted = 0
        pos = string.find(dat,huntstr)
        if pos == -1:    # FF keyframe file
            for huntstart in ['\x2D','\x29','\x2F','\x1F']:
                huntstr = huntstart + '\x00\x00\x00\x05\x00\x00\x00'+'Bip01'
                FF1converted = 1
                pos = string.find(dat,huntstr)
                if pos != -1:
                    break
        if verbose:  print 'pos=%d  FF1converted=%d' % (pos,FF1converted)
        while (pos<datlen) and (pos!=-1) :
            namepos = pos - 1
            if FF1converted:  namepos = namepos - 4
            while dat[namepos]=='\x00':  namepos = namepos - 1      # skip zeroes
            while dat[namepos-1]!='\x00':  namepos = namepos - 1    # go to start of anim name
            namelen = ord(dat[namepos - 4])
            name = dat[namepos:namepos+namelen]
            if verbose:  print '%6d %s %d' % (pos,name,namepos)
            nextpos = string.find(dat,'\xff\xff\xff\xff',pos+1)
            nextpos = string.find(dat,huntstr,nextpos+1)
            startp = string.find(dat,'\x05\x00\x00\x00start',pos)
            startval = struct.unpack('f',dat[startp-4:startp])[0]
            nodes = [('start',startval)]
            nextval = struct.unpack('f',dat[startp+9:startp+13])[0]
            nextp = startp+17
            nextstrlen = ord(dat[startp+13])
            nextstr = dat[nextp:nextp+nextstrlen]
            while ( ( nextp < nextpos ) or ( nextpos == -1 ) ) and ( nextstr != 'end' ):
                ##print '  nextp = %d  <%s>' % (nextp,nextstr)
                nodes.append((nextstr,nextval))
                strlen = nextstrlen
                nextval = struct.unpack('f',dat[nextp+strlen:nextp+strlen+4])[0]
                nextstrlen = ord(dat[nextp+strlen+4])
                nextp = nextp+strlen+8
                nextstr = dat[nextp:nextp+nextstrlen]
            ##print '  nextp = %d  <%s>' % (nextp,nextstr)
            nodes.append((nextstr,nextval))
            if verbose:  print nodes
            _AnimationData[FileName][name]=nodes
            pos = nextpos
    ##print 'Keyframe_ReadAnimationTimes: time = %3.4f' % (time.clock()-tstart)
    return _AnimationData[FileName].copy()

# Utility to display the KF names and times
def ShowKFTimes(KFdata,attackonly=1):
    anims = KFdata.keys()
    if attackonly:
        attacknames = ['melee','ranged','area','direct','special','active']
        tmp = []
        for aname in attacknames:
            tmp.extend(filter(lambda name,aname=aname: string.find(name,aname)>=0,anims))
        anims = tmp
    anims.sort()
    for anim in anims:
        print '%s:' % anim
        animstr = '  '
        for node in KFdata[anim]:
            ##print node
            animstr = animstr + ('%s: %1.2f\t' % (node[0],node[1]))
        print animstr

# Read a campdef.dat file and return a dictionary with the data for each
# mission, which is the list of mission names, and for each mission whether
# the mission is a submission, the number of allowable characters, the
# list of required characters, and the list of disallowed characters.
def Campaign_ReadCampaignDefinitions(FileName='',ForceRead=0,verbose=0):
    global _CampdefData
    if FileName=='': # try to guess the current DAT file
        FileName = GetDatFile('campdef.dat')
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_CampdefData' in globals().keys() ) and ( FileName in _CampdefData.keys() ):
        return _CampdefData[FileName].copy()
    if not ( '_CampdefData' in globals().keys() ):      # entirely new dictionary
        _CampdefData = {}
    _CampdefData[FileName] = {}
    try:
        f = open(FileName,'rb')
        dat = f.read()
        f.close()
    except:
        print 'Campaign_ReadCampaignDefinitions: unable to read data from %s' % FileName
        return _CampdefData[FileName]
    datlen = len(dat)
    _CampdefData[FileName]['totalMissions'] = ord(dat[4389])
    pos = 4393
    missions = []
    while ( len(missions) < _CampdefData[FileName]['totalMissions'] ):
        missionName = string.split(dat[pos:],'\x00',1)[0]
        #print 'Campaign_ReadCampaignDefinitions: pos=%5d len(missions)=%d read missionName=<%s>' % (pos,len(missions),missionName)
        if missionName: # check shouldn't be needed
            missions.append(missionName)
            _CampdefData[FileName][missionName] = {}
            _CampdefData[FileName][missionName]['missionName'] = missionName
            _CampdefData[FileName][missionName]['isSubmission'] = ord(dat[pos+30])
            num = ord(dat[pos+36])
            _CampdefData[FileName][missionName]['numAllowed'] = num
            required = []
            for i in range(num):
                char = string.split(dat[pos+40+i*30:],'\x00',1)[0]
                if char:
                    required.append(char)
            _CampdefData[FileName][missionName]['required'] = required
            disallowed = []
            for i in range(10):
                char = string.split(dat[pos+160+i*30:],'\x00',1)[0]
                if char:
                    disallowed.append(char)
            _CampdefData[FileName][missionName]['disallowed'] = disallowed
        pos = pos + 460
    _CampdefData[FileName]['missions'] = missions
    return _CampdefData[FileName].copy()

# This function is likely the most useful of the suite for mods and
# scripted attributes, since it reads the most recent version of the
# characters on a mission.  (Too bad there can be no equivalent for
# Danger Rooms.)  But there is an issue with caching: Since the quick
# save and autosave games will have the same filenames each time, it
# is possible that the player will save during the previous mission
# and not save again until the middle of this one.  If a scripted
# attribute initializes at the biggining of the mission, this function
# will read the old quicksave and (with caching on) never update with
# the newer data.  I am willing to live with this; the moral is: save
# after each buyup.  Also, the problem doesn't happen with FF2 because
# the autosaves mean that the first in-mission quicksave will be
# newer.
def Campaign_ReadCharactersFromSavedGame(FileName='',ModPath='',ForceRead=0,fill=1,verbose=0):
##    tstart = time.clock()
    global _SavedCharData
    if FileName=='': # try to guess the current saved game DAT file
        FileName = GetMostRecentSavedGameFile()
    if FileName == '':      # no saved game DAT found
        return {}
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_SavedCharData' in globals().keys() ) and ( FileName in _SavedCharData.keys() ):
##        print 'Campaign_ReadCharactersFromSavedGame: time = %3.4f' % (time.clock()-tstart)
        return _SavedCharData[FileName].copy()
    if not ( '_SavedCharData' in globals().keys() ):      # entirely new dictionary
        _SavedCharData = {}
    f = open(FileName,'rb')
    alldat = f.read()
    f.close()
    _SavedCharData[FileName] = {}
    if ( ModPath == '' ) and fill:
        ModPath = GetModPath(ForceRead)
    ghdr = '_GameHeader'                                        # Game Header
    chdr = 'HDR4'+'\x0A'+'\x00'+'\x00'+'\x00'+'FFCampaign'      # Campaign header
    hdr = 'HDR4'+'\x05'+'\x00'+'\x00'+'\x00'+'sCDef'            # CDef header
    pos = find(alldat,ghdr)
    SaveNameLen = ord(GetDatWords(alldat,pos,3)[2])
    pos = GetDatWordPos(alldat,pos,4)
    saveName = alldat[pos:pos+SaveNameLen]
    _SavedCharData[FileName]['saveName'] = saveName
    MissionDATFileLen = ord(alldat[pos+SaveNameLen])
    missionDATFile = alldat[pos+SaveNameLen+4:pos+SaveNameLen+4+MissionDATFileLen]
    ##print 'pos=%d SaveNameLen=%d MissionDATFileLen=%d' % (pos,SaveNameLen,MissionDATFileLen)
    _SavedCharData[FileName]['missionDATFile'] = missionDATFile
    pos = find(alldat,chdr)
    clen = struct.unpack('l',alldat[pos+22:pos+26])[0]
    _SavedCharData[FileName]['PP'] = struct.unpack('i',alldat[pos+90:pos+94])[0]
    ##print 'clen = %d' % clen
    dat = alldat[pos:pos+22+clen]
    pos = find(dat,hdr)
    if fill:
        objdata = Campaign_ReadObjects(os.path.join(ModPath,'objects.dat'), ForceRead)
        powerdata = Campaign_ReadPowers(os.path.join(ModPath,'powers.dat'), ForceRead)
    while pos >= 0:
        block = dat[pos:pos+1100]
        CDeflen = struct.unpack('h',block[17:19])[0]
        if ( CDeflen == 801 ) or ( CDeflen == 1057 ):
            charName, charData = ReadSavedTemplateCharacter(block[21:],0,CDefLen=CDeflen,verbose=verbose)
            if fill:
                if charName in objdata.keys():
                    if 'objectAttributes' not in charData.keys():
                        charData['objectAttributes'] = []
                    for k in objdata[charName].keys():
                        charData['objectAttributes'].append(k)
                        charData[k] = objdata[charName][k]
                else:
                    print 'Campaign_ReadCharactersFromSavedGame: character <%s> has no corresponding object template.' % charName
                for p in charData['powers'].keys():
                    if p in powerdata.keys():
                        charData['powers'][p] = powerdata[p]
                    else:
                        print 'Campaign_ReadCharactersFromSavedGame: character <%s> has power <%s> with no corresponding powers.dat entry.' % (charName,p)
        elif ( CDeflen == 4955 ) or ( CDeflen == 5211 ) :
            block = dat[pos:pos+5400]
            charName, charData = ReadSavedCustomCharacter(block[21:],0,CDefLen=CDeflen,verbose=verbose)
        else:
            print 'Campaign_ReadCharactersFromSavedGame: pos=%d unrecognized CDeflen=%d' % (pos,CDeflen)
            block = dat[pos:pos+CDeflen+32]
            charName, charData = ReadSavedTemplateCharacter(block[21:],0,CDefLen=CDeflen,verbose=verbose)
        _SavedCharData[FileName][charName] = charData
        pb = find(block,hdr,21+CDeflen)
        if pb > 0:
            pos = pos + pb
        else:
            pos = find(dat,hdr,pos+21+CDeflen)
##    print 'Campaign_ReadCharactersFromSavedGame: time = %3.4f' % (time.clock()-tstart)
    return _SavedCharData[FileName].copy()

# Read character data from a hero file.  Note that the function will try to
# guess the path where the HERO file exists for the current mod, but the actual
# file name *must* be given (unlike in many of these functions).
def Campaign_ReadHeroFile(FileName,ForceRead=0,verbose=0):
##    tstart = time.clock()
    global _HeroFileData
    if not os.path.isfile(FileName): # maybe we weren't given the full path
        FileName = os.path.join(GetHeroFilePath(),FileName)
    if not os.path.isfile(FileName):
        print 'Campaign_ReadHeroFile: file <%s> was not found' % (FileName,)
        return {}
    FileName = os.path.abspath(FileName)
    if ( not ForceRead ) and ( '_HeroFileData' in globals().keys() ) and ( FileName in _HeroFileData.keys() ):
##        print 'Campaign_ReadHeroFile: time = %3.4f' % (time.clock()-tstart)
        return _HeroFileData[FileName].copy()
    if not ( '_HeroFileData' in globals().keys() ):      # entirely new dictionary
        _HeroFileData = {}
    f = open(FileName,'rb')
    dat = f.read()
    f.close()
    _HeroFileData[FileName] = {}
    CDeflen = struct.unpack('h',dat[17:19])[0]
    customName, customData = ReadSavedCustomCharacter(dat,21,CDefLen=CDeflen,verbose=verbose)
    _HeroFileData[FileName] = customData
##    print 'Campaign_ReadHeroFile: time = %3.4f' % (time.clock()-tstart)
    return _HeroFileData[FileName].copy()

# Generate a list of locations where character data can be found given
# a 'complex' template attribute value.  This is for use in Danger
# Room missions where there is no saved game data and template name
# information is unreliable. The indices are the char's 'complex'
# object attribute and the values will be the full hero file name
# for customs and the template names for built-ins.
# 
# This should be run once for each mod path.  Maybe another wrapper
# function should be writted to find all mod paths at once, but I
# won't do that here in case this needs to be run during a mission.
def Campaign_MakeHeroIndex(ModPath='', indexFileName='', ForceRead=0, verbose=0):
    if ModPath == '':
        ModPath = GetModPath(ForceCheck=ForceRead)
    if indexFileName == '':                     # no filename given
        indexFileName = os.path.join(ModPath,'HeroIndexData.txt')
    if os.path.dirname(indexFileName) == '':    # filename given without directory
        indexFileName = os.path.join(ModPath,indexFileName)
    print 'Campaign_MakeHeroIndex: Making index <%s>' % (indexFileName,)
    if inGame():  # we are in the game
        import ff
    hlist = {}
    # First do hero files
    HeroPath = GetHeroFilePath(ModPath=ModPath,ForceCheck=ForceRead)
    if os.path.isdir(HeroPath):
        HeroFiles = map(lambda f, HeroPath=HeroPath: os.path.join(HeroPath,f), os.listdir(HeroPath))
    else:       # Some mods may not have this directory set up yet?
        HeroFiles = []
    HeroFiles = filter(lambda f: f[-5:] == '.hero', HeroFiles)
    for h in HeroFiles:
        if verbose: print 'Campaign_MakeHeroIndex: reading custom hero %s' % repr(h)
        hero = Campaign_ReadHeroFile(h,ForceRead=ForceRead, verbose=verbose)
        if verbose: print 'Campaign_MakeHeroIndex: custom hero %s has complex set to %s' % (hero.get('charName'),hero.get('complex'))
        if 'complex' not in hero['objectAttributes']:
            print 'Campaign_MakeHeroIndex: Re-run FFX Control Center; complex not set for %s.' % h
            if inGame():  # we are in the game
                ff.Mission_StatusText('new hero file: %s. quit and run FFX CONTROL CENTRE' % h)
            pass
        else:
            complex = int(hero['complex'])
            if complex in hlist.keys():
                print 'Campaign_MakeHeroIndex: duplicate complex numbers found in <%s> and <%s>,  \
                Re-run FFX Control Center' % (h,hlist[complex])
                if inGame():  # we are in the game
                    ff.Mission_StatusText('character found with duplicate "complex" value. quit and run FFX CONTROL CENTRE')
                pass
            else:
                hlist[complex] = str(h)
    # Now do campaign files
    CampaignChars = Campaign_ReadCharacterData(ModPath, ForceRead=ForceRead)
    for c in CampaignChars.keys():
        if verbose: print 'Campaign_MakeHeroIndex: looking at campaign character %s' % c
        if c == 'narrator': continue
        if not CampaignChars[c].has_key('class'):
            print 'Campaign_MakeHeroIndex warning: <%s> likely missing entry in objects.dat; skipping.' % (c,)
            continue
        hero = CampaignChars[c]
        if 'objectAttributes' in hero.keys():
            if ( 'complex' not in hero['objectAttributes'] ):
                if hero.get('class') in ('GAME_OBJ_TANK','GAME_OBJ_CIVILIAN'):
                    print 'Campaign_MakeHeroIndex: Complex not set for %s. Run datfiles.Campaign_SetMissingComplexAttribs()' % c
                    if inGame():  # we are in the game
                        ff.Mission_StatusText('new civilian/tank: %s. run datfiles.Campaign_SetMissingComplexAttribs()' % c)
                    pass
                else:
                    print 'Campaign_MakeHeroIndex: Re-run FFX Control Center; complex not set for %s.' % c
                    if inGame():  # we are in the game
                        ff.Mission_StatusText('new character: %s. quit and run FFX CONTROL CENTRE' % c)
                    pass
            else:
                complex = int(hero['complex'])
                if complex in hlist.keys():
                    print 'Campaign_MakeHeroIndex: duplicate complex numbers found in <%s> and <%s>,  \
                    Re-run FFX Control Center' % (hlist[complex],c)
                    if inGame():  # we are in the game
                        ff.Mission_StatusText('character found with duplicate "complex" value. quit and run FFX CONTROL CENTRE')
                    pass
                else:
                    hlist[complex] = c
        else:
            print 'Campaign_MakeHeroIndex: %s has no object attributes and likely no object template, skipping.' % c
    
    f = open(indexFileName,'w')
    f.write('# Complex index file for ModPath=%s\n' % ModPath)
    f.write('# (Generated by datfiles.Campaign_MakeHeroIndex on %s)\n' % time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime(time.time())))
    hlkeys = hlist.keys()
    hlkeys.sort()
    f.write('{\n')
    for key in hlkeys:
        f.write('%d: %s,\n' % (key,repr(hlist[key])))
    f.write('}\n')
    f.close()

# Return a completely filled-out character dictionary for all
# built-ins in a mod.  The ModPath can be left blank when run from
# within the game. argument is the character's template name.  
#
# This calls all the character-, object- and power-reading functions
# to return a dictionary of all characters with both the data from
# their FFEdit Characters pages, Objects pages, and their powers
# filled out. This effectively gives all the data on the built-in
# characters as there is for a custom character.
def Campaign_ReadCharacterData(ModPath='', ForceRead=0, verbose=0):
    global _ModCharData
    tstart = time.clock()
    if not ModPath:
        ModPath = GetModPath()
    ModPath = os.path.abspath(ModPath)
    if ( not ForceRead ) and ( '_ModCharData' in globals().keys() ) and ( ModPath in _ModCharData.keys() ):
##        print 'Campaign_ReadCharacterData: time = %3.5f' % (time.clock()-tstart)
        return _ModCharData[ModPath].copy()
    if not ( '_ModCharData' in globals().keys() ):      # entirely new dictionary
        _ModCharData = {}
    _ModCharData[ModPath] = {}

    _ModCharData[ModPath] = Campaign_ReadCharacters(os.path.join(ModPath,'characters.dat'), ForceRead=ForceRead, verbose=verbose)
    objdata = Campaign_ReadObjects(os.path.join(ModPath,'objects.dat'), ForceRead=ForceRead, verbose=verbose)
    powerdata = Campaign_ReadPowers(os.path.join(ModPath,'powers.dat'), ForceRead=ForceRead, verbose=verbose)
    for c in _ModCharData[ModPath].keys():
        # set up the starting attributes and power levels
        if 'attrib_start' in _ModCharData[ModPath][c].keys():     # this is a char
            _ModCharData[ModPath][c]['activeAttributes'] = \
                    min(_ModCharData[ModPath][c]['attrib_start'],len(_ModCharData[ModPath][c]['characterAttributes']))
            _ModCharData[ModPath][c]['powerLevels'] = {}
            for i in range(len(_ModCharData[ModPath][c]['tier_a'])):
                if i < _ModCharData[ModPath][c]['tier_a_start'] - 1:
                    _ModCharData[ModPath][c]['powerLevels'][_ModCharData[ModPath][c]['tier_a'][i]] = 3
                else:
                    _ModCharData[ModPath][c]['powerLevels'][_ModCharData[ModPath][c]['tier_a'][i]] = 1
            for i in range(len(_ModCharData[ModPath][c]['tier_b'])):
                if i < _ModCharData[ModPath][c]['tier_b_start'] - 1:
                    _ModCharData[ModPath][c]['powerLevels'][_ModCharData[ModPath][c]['tier_b'][i]] = 3
                else:
                    _ModCharData[ModPath][c]['powerLevels'][_ModCharData[ModPath][c]['tier_b'][i]] = 1
        # add the object attributes
        if 'objectAttributes' not in _ModCharData[ModPath][c].keys():
            _ModCharData[ModPath][c]['objectAttributes'] = []
        if c in objdata.keys():
            for k in objdata[c].keys():
                _ModCharData[ModPath][c]['objectAttributes'].append(k)
                _ModCharData[ModPath][c][k] = objdata[c][k]
        else:
            if verbose:
                print 'Campaign_ReadCharacterData: character <%s> has no corresponding object template.' % c
            pass
        # fill in the power data
        for p in _ModCharData[ModPath][c]['powers'].keys():
            if p in powerdata.keys():
                _ModCharData[ModPath][c]['powers'][p] = powerdata[p]
            else:
                if verbose:
                    print 'Campaign_ReadCharacterData: character <%s> has power <%s> with no corresponding powers.dat entry.' % (c,p)
                pass
##    print 'Campaign_ReadCharacterData: time = %3.5f' % (time.clock()-tstart)
    return _ModCharData[ModPath]
                                       
def ParseLangDATFile(FileName):
    f = open(FileName, 'rb')
    dat = f.read()
    f.close()
    pairs = []
    pos = 2048
    while pos<len(dat):
        if ( dat[pos] >= ' ' ) and ( dat[pos] <= '~' ):
            nullpos = string.find(dat[pos:pos+1024],'\x00\x00')
            if nullpos>-1:
                strings = string.split(dat[pos:pos+nullpos],'\x00')
            else:
                strings = string.split(dat[pos:pos+1024],'\x00')
            somepairs = map(lambda x,strings=strings: (strings[2*x],strings[2*x+1]),range(len(strings)/2))
            somepairs = filter(lambda x: ( len(x[0]) > 3 ) and ( x[1] ) \
                               and ( x[0][-3] == '_' ) and ( x[0][-1] in '0123465789' ) and ( x[0][-2] in '0123465789' ) \
                               ##and ( x[0][0] >= ' ' ) and ( x[0][0] <= '~' ) and ( x[1][0] >= ' ' ) and ( x[1][0] <= '~' )\
                               ,somepairs)
            pairs.extend(somepairs)
        pos = pos + 1024
    return pairs

def Campaign_ReadLangDAT(FileName='',ForceRead=0,verbose=0):
    global _LangData
    if FileName=='': # try to guess the current DAT file
        FileName = os.path.join(GetLangPath(),'strings.dat')
    if ( not ForceRead ) and ( '_LangData' in globals().keys() ) and ( FileName in _LangData.keys() ):
        return _LangData[FileName].copy()
    if not ( '_LangData' in globals().keys() ):      # entirely new dictionary
        _LangData = {}
    _LangData[FileName] = {}
    try:
        pairs = ParseLangDATFile(FileName)
    except:
        print 'Campaign_ReadLangDAT: unable to read data from %s' % FileName
        return _LangData[FileName]
    try:
        _LangData[FileName] = dict(pairs)
    except NameError:
        for p in pairs:
            _LangData[FileName][p[0]] = p[1]
    return _LangData[FileName].copy()

def Campaign_FindDisplayString(tag, FileName='', default=None):
    return Campaign_ReadLangDAT(FileName,ForceRead=0,verbose=0).get(string.lower(tag)) or default

#################################
### Parsing Functions ###########
#################################

def ReadSavedTemplateCharacter(dat,startpos,CDefLen=0,verbose=0):
    if CDefLen == 0:
        if ffver() == 1:
            CDefLen = 801
        else:
            CDefLen = 1057
    dict = {}
    words = GetDatWords(dat,startpos,2,'\x1f')
    templateName = words[0]
    dict['charName'] = templateName
    dict['isCustom'] = 0
    if verbose:  print 'ReadSavedTemplateCharacter: templateName = <%s>' % templateName
    # get cut-scene base
    CSBase = ''
    if GetDatWordPos(dat,startpos,2,'\x1f') == startpos + 30:
        CSBase = words[1]
    dict['CSBase'] = CSBase
    if verbose:  print 'ReadSavedTemplateCharacter: CSBase = <%s>' % dict['CSBase']
    # read strength, speed, etc.
    dict['strength'] = ord(dat[startpos+285])
    dict['speed'] = ord(dat[startpos+286])
    dict['endurance'] = ord(dat[startpos+287])
    dict['energy'] = ord(dat[startpos+289])
    dict['agility'] = ord(dat[startpos+290])
    if verbose:
        print 'ReadSavedTemplateCharacter: strength = %s, speed = %s, endurance = %s, energy = %s, agility = %s' % \
              (dict['strength'], dict['speed'], dict['endurance'], dict['energy'], dict['agility'])
    # get character attributes
    pos = startpos + 291
    characterAttributes = []
    for i in range(5):
        attrib = string.split(dat[pos:],'\x00',1)[0]
        if attrib != '':  characterAttributes.append(attrib)
        pos = pos + 30
    dict['characterAttributes'] = characterAttributes
    if verbose:  print 'ReadSavedTemplateCharacter: characterAttributes = <%s>' % dict['characterAttributes']
    # get list of template powers and power levels
    pos = startpos + 441
    powers = []
    powerLevels =  map(lambda c: ord(c), dat[startpos+741:startpos+741+10])
    dict['powerLevels'] = {}
    dict['powers'] = {}
    for i in range(5):
        pname = string.split(dat[pos:],'\x00',1)[0]
        if pname != '':
            powers.append(pname)
            dict['powerLevels'][pname] = powerLevels[i]
            dict['powers'][pname] = {}      # to be filled in from powers.dat info
        pos = pos + 30
    dict['tier_a'] = powers
    if verbose:  print 'ReadSavedTemplateCharacter: tier_a powers and levels %s' % (map(lambda p: [p,dict['powerLevels'][p]],dict['tier_a']),)
    powers = []
    for i in range(5):
        pname = string.split(dat[pos:],'\x00',1)[0]
        if pname != '':
            powers.append(pname)
            dict['powerLevels'][pname] = powerLevels[i+5]
            dict['powers'][pname] = {}      # to be filled in from powers.dat info
        pos = pos + 30
    dict['tier_b'] = powers
    if verbose:  print 'ReadSavedTemplateCharacter: tier_b powers and levels %s' % (map(lambda p: [p,dict['powerLevels'][p]],dict['tier_b']),)
    # get voice ID
    if dat[startpos+751] != '\x00':     dict['VID'] = dat[startpos+751:startpos+751+2]
    else:   dict['VID'] = ''
    if verbose:  print 'ReadSavedTemplateCharacter: VID = <%s>' % dict['VID']
    # get AI
    dict['AI'] = string.split(dat[startpos+754:],'\x00',1)[0]
    if verbose:  print 'ReadSavedTemplateCharacter: AI = <%s>' % dict['AI']
    # skipping starting tier powers and attributes, not valid for custom templates
    # get Campaign-Only flag
    dict['camp_only'] = ord(dat[startpos+791])
    # get movement radius
    dict['movementRadius'] = struct.unpack('f',dat[startpos+797:startpos+797+4])[0]
    if verbose:  print 'ReadSavedTemplateCharacter: movementRadius = <%s>' % dict['movementRadius']
    if CDefLen != 801:
        dict['alterEgo'] = string.split(dat[startpos + 801:],'\x00',1)[0]
    else:
        dict['alterEgo'] = ''
    if verbose:  print 'ReadSavedTemplateCharacter: alterEgo = <%s>' % dict['alterEgo']
    # get character's remaining CP
    dict['CP'] = struct.unpack('h',dat[startpos+784:startpos+784+2])[0]
    if verbose:  print 'ReadSavedTemplateCharacter: CP = <%s>' % dict['CP']
    # get character's XP
    dict['XP'] = struct.unpack('h',dat[startpos+788:startpos+788+2])[0]
    if verbose:  print 'ReadSavedTemplateCharacter: XP = <%s>' % dict['XP']
    # get character's number of active attributes
    dict['activeAttributes'] = ord(dat[startpos+790])
    if verbose:  print 'ReadSavedTemplateCharacter: activeAttributes = <%s>' % dict['activeAttributes']
    return templateName, dict
    
def ReadSavedCustomCharacter(dat,startpos,CDefLen=0,verbose=0):
    if CDefLen == 0:
        if ffver() == 1:
            CDefLen = 801
        else:
            CDefLen = 1057
    dict = {}
    if dat[startpos] != '\x00': templateName = GetDatWords(dat,startpos,1,'\x1f')[0]
    else:   templateName = ''
    if verbose:  print 'ReadSavedCustomCharacter: templateName = <%s>' % templateName
    dict['isCustom'] = 1
    # get cut-scene base
    CSBase = ''
    if dat[startpos+30] != '\x00':  CSBase = GetDatWords(dat,startpos+30,1,'\x1f')[0]
    else:   CSBase = ''
    dict['CSBase'] = CSBase
    if verbose:  print 'ReadSavedCustomCharacter: CSBase = <%s>' % dict['CSBase']
    # read strength, speed, etc.
    dict['strength'] = ord(dat[startpos+285])
    dict['speed'] = ord(dat[startpos+286])
    dict['endurance'] = ord(dat[startpos+287])
    dict['energy'] = ord(dat[startpos+289])
    dict['agility'] = ord(dat[startpos+290])
    if verbose:
        print 'ReadSavedCustomCharacter: strength = %s, speed = %s, endurance = %s, energy = %s, agility = %s' % \
              (dict['strength'], dict['speed'], dict['endurance'], dict['energy'], dict['agility'])
    # get character attributes
    pos = startpos + 291
    characterAttributes = []
    for i in range(5):
        attrib = string.split(dat[pos:],'\x00',1)[0]
        if attrib != '':  characterAttributes.append(attrib)
        pos = pos + 30
    dict['characterAttributes'] = characterAttributes
    if verbose:  print 'ReadSavedCustomCharacter: characterAttributes = <%s>' % dict['characterAttributes']
    # get voice ID
    if dat[startpos+751] != '\x00':     dict['VID'] = dat[startpos+751:startpos+751+2]
    else:   dict['VID'] = ''
    if verbose:  print 'ReadSavedCustomCharacter: VID = <%s>' % dict['VID']
    # get AI
    dict['AI'] = GetDatWords(dat,startpos+754,1,'\x1f')[0]
    if verbose:  print 'ReadSavedCustomCharacter: AI = <%s>' % dict['AI']
    # skipping starting tier powers and attributes, not valid for custom templates
    # get character's remaining CP
    dict['CP'] = struct.unpack('h',dat[startpos+784:startpos+784+2])[0]
    if verbose:  print 'ReadSavedCustomCharacter: CP = <%s>' % dict['CP']
    # get character's XP
    dict['XP'] = struct.unpack('h',dat[startpos+788:startpos+788+2])[0]
    if verbose:  print 'ReadSavedCustomCharacter: XP = <%s>' % dict['XP']
    # get character's number of active attributes
    dict['activeAttributes'] = ord(dat[startpos+790])
    if verbose:  print 'ReadSavedCustomCharacter: activeAttributes = <%s>' % dict['activeAttributes']
    # get Campaign-Only flag
    dict['camp_only'] = ord(dat[startpos+791])
    # get movement radius     # This doesn't work for hero files
    dict['movementRadius'] = struct.unpack('f',dat[startpos+4951:startpos+4951+4])[0]
    if verbose:  print 'ReadSavedCustomCharacter: movementRadius = <%s>' % dict['movementRadius']
    #### I am assuming this until there is a patch for the Law/Order transform in customs
    ##dict['alterEgo'] = string.split(dat[startpos + 801:],'\x00',1)[0]
    ##if verbose:  print 'ReadSavedCustomCharacter: alterEgo = <%s>' % dict['alterEgo']
    # get dictionary filled with power data and power levels
    powerLevels =  map(lambda c: ord(c), dat[startpos+741:startpos+741+10])
    pos = startpos + 793
    dict['powers'] = {}
    dict['powerLevels'] = {}
    dict['powerPositions'] = {}     # beta code for power ID number tests
    for i in range(10):
        pname = string.split(dat[pos:],'\x00',1)[0]
        if pname != '':
            dict['powerLevels'][pname] = powerLevels[i]
            dict['powerPositions'][pname] = (pos-793)/184
            dict['powers'][pname] = ParsePowerBlock(dat[pos:pos+184])
            if verbose:  
                print '<%s> at level %d' % (pname,dict['powerLevels'][pname])
                ShowPower(dict['powers'][pname])
        pos = pos + 184
    # get base template (nif) name  (Don't know exactly what this is.)
    dict['BaseTemplateName'] = string.split(dat[startpos+2633:],'\x00',1)[0]
    if verbose:  print 'ReadSavedCustomCharacter: BaseTemplateName = <%s>' % dict['BaseTemplateName']
    # get NIF location
    dict['NIF'] = string.split(dat[startpos+2663:],'\x00',1)[0]
    if verbose:  print 'ReadSavedCustomCharacter: NIF = <%s>' % dict['NIF']
    # get object class
    dict['class'] = string.split(dat[startpos+2918:],'\x00',1)[0]
    if verbose:  print 'ReadSavedCustomCharacter: class = <%s>' % dict['class']
    # get template attributes
    pos = startpos + 2949
    dict['objectAttributes'] = []
    for i in range(15):
        attrib = string.split(dat[pos:],'\x00',1)[0]
        if attrib != '':
            dict['objectAttributes'].append(attrib)
            dict[attrib]= struct.unpack('f',dat[pos+36-4:pos+36])[0]
        pos = pos + 36
    if verbose:  print 'ReadSavedCustomCharacter: objectAttributes = <%s>' % dict['objectAttributes']
    # get template attributes
    dict['skin'] = string.split(dat[startpos+3519:],'\x00',1)[0]
    if verbose:  print 'ReadSavedCustomCharacter: skin = <%s>' % dict['skin']
    # get in-game character name
    dict['name'] = string.split(dat[startpos+4917:],'\x00',1)[0]
    if verbose:  print 'ReadSavedCustomCharacter: name = <%s>' % dict['name']
    dict['charName'] = dict['name']
    if templateName == '':  # this is a hero file
        templateName = 'custom_template_' + dict['name']
##    else:
##        dict['charName'] = templateName
    return templateName, dict

# Return a list of the starting positions of each power in a powers.dat DAT that
# starts at startpos or later in the DAT.
def GetPowerRecordPositions(dat,startpos):
    positions = []
    # get the first matching pair
    cword = ''
    cwordpos = startpos
    lencword = len(cword)
    nword = GetDatWords(dat,startpos,1,'\x1f')[0]
    lennword = len(nword)
    nwordpos = GetDatWordPos(dat,startpos,1,'\x1f')
    while ( nword ):
        while ( nword ) and not ( ( nword == cword[-lennword:] ) and ( dat[cwordpos+lencword] == '\x00') and ( nwordpos - cwordpos == lencword + 1 ) ):
            cword = nword
            cwordpos = nwordpos
            lencword = lennword
            nword = GetDatWords(dat,cwordpos+len(cword),1,'\x1f')[0]
            lennword = len(nword)
            nwordpos = GetDatWordPos(dat,cwordpos+len(cword),1,'\x1f')
            ##print 'cword = <%s> @ %d  nword = <%s> @ %d' % (cword,cwordpos,nword,nwordpos)
        # now check that and return the chunk of dat
        if ( nword ):
            positions.append(nwordpos)
            cword = nword
            cwordpos = nwordpos
            lencword = lennword
            nword = GetDatWords(dat,cwordpos+184,1,'\x1f')[0]
            lennword = len(nword)
            nwordpos = GetDatWordPos(dat,cwordpos+184,1,'\x1f')
            ##print 'cword = <%s> @ %d  nword = <%s> @ %d' % (cword,cwordpos,nword,nwordpos)
        else:
            if nwordpos + 183 == len(dat):
                positions.append(nwordpos)
    ##print 'GetPowerRecordPositions: positions = %s' % positions
    return positions

# Given a 184 byte block of a dat (powers.dat or custom hero dat), return a dictionary with power data.
def ParsePowerBlock(dat):
    pdata = {}
    pdata['PowerName'] = string.split(dat,'\x00',1)[0]
    pType = ord(dat[132])
    pdata['PowerType'] = pType
    if ( pType == PT_PASSIVE_DEFENCE ) or ( pType == PT_ACTIVE_DEFENCE ):
        pdata['BlockType'] = ord(dat[182])
        DamageTypeBlocked = ( ord(dat[177]) << 8 ) + ord(dat[176])
        pdata['DamageTypesBlocked'] = DamageTypeBlocked
        pdata['AttackModesBlocked'] = ord(dat[172])
        pdata['DefenceFlags'] = ord(dat[156])
        if pType == PT_PASSIVE_DEFENCE:
            pdata['Success'] = ord(dat[180])
        if pType == PT_ACTIVE_DEFENCE:
            pdata['EPCost'] = ord(dat[121])
            pdata['Duration'] = ord(dat[181])
        pdata['animation'] = string.split(dat[60:],'\x00',1)[0]
        pdata['FX'] = string.split(dat[90:],'\x00',1)[0]
    else:   # it's an attack type
        pdata['EPCost'] = ord(dat[121])
        pdata['animation'] = string.split(dat[60:],'\x00',1)[0]
        pdata['FX'] = string.split(dat[90:],'\x00',1)[0]
        pdata['SubType'] = ord(dat[136])
        pdata['Magnitude'] = ord(dat[140])
        pdata['DamageType'] = ord(dat[144])
        pdata['Speed'] = ord(dat[122])
        pdata['Stun'] = ord(dat[124])
        pdata['Knockback'] = ord(dat[125])
        pdata['RangeMin'] = ord(dat[126])
        pdata['RangeMax'] = ord(dat[127])
        pdata['Accuracy'] = ord(dat[123])
        pdata['Radius'] = ord(dat[128])
        pdata['SpecialType'] = ord(dat[160])
        pdata['MaxInstances'] = ord(dat[152])
        AttackFlags = ( ord(dat[157]) << 8 ) + ord(dat[156])
        pdata['AttackFlags'] = AttackFlags
    pdata['notForCustom'] = ord(dat[158])
    pdata['powerID'] = struct.unpack('l',dat[168:172])[0]
    pdata['_junk'] = ord(dat[176])
    return pdata

def CreateTemplateCharFromCustom(custom,charName=None,power_prepend=None,tier_a_start=None, tier_b_start=None, attrib_start=None):
    temp = custom.copy()
    # fill in the missing fields and make adjustments
    if power_prepend==None:
        power_prepend = ''
        for i in range(len(temp['charName'])):
            if temp['charName'][i] not in ' 01234567890abcdefghijklmnopqurstuvwxyzABCDEFGHIJKLMNOPQURSTUVWXYZ_':
                power_prepend = power_prepend + '_'
            else:
                power_prepend = power_prepend + temp['charName'][i]
    if charName:
        temp['charName'] = charName
    else:
        cname2 = ''
        for i in range(len(temp['charName'])):
            if temp['charName'][i] not in '01234567890abcdefghijklmnopqurstuvwxyzABCDEFGHIJKLMNOPQURSTUVWXYZ_':
                cname2 = cname2 + '_'
            else:
                cname2 = cname2 + temp['charName'][i]
        temp['charName'] = cname2
    temp['HDR'] = 'HDR4\x05\x00\x00\x00sCDef\x0E\x00\x00\x00\x21\x04'
    temp['camp_only'] = 1
    if attrib_start==None:
        temp['attrib_start'] = attrib_start or temp['activeAttributes']
    temp['tier_a'] = []
    temp['tier_b'] = []
    pposinvert = {}
    for k in temp['powerPositions'].keys():
        pposinvert[temp['powerPositions'][k]] = k
    for ppos in range(5):
        if pposinvert.get(ppos):
            temp['tier_a'].append(power_prepend+pposinvert.get(ppos))
    for ppos in range(5,10):
        if pposinvert.get(ppos):
            temp['tier_b'].append(power_prepend+pposinvert.get(ppos))
    temp['tier_a_start'] = tier_a_start or len(temp['tier_a'])
    temp['tier_b_start'] = tier_b_start or len(temp['tier_b'])
    return temp
    

class DATblock:     # define how the data in a DAT block or sub-block is written
    def __init__(self, size, fields, type='genDict', repeatlen=0):
        self.size = size
        self.fields = fields
        self.type = type
        self.repeatlen = repeatlen

catemp = DATblock(150, ('30s',), type='list', repeatlen=30)       # character attributes
tiertemp = DATblock(150, ('30s',), type='list', repeatlen=30)       # tier power list
oatemp = DATblock(540, ('32s', 'f'), type='dictkeylist', repeatlen=36)   # object attributes
pltemp = DATblock(10, ('B',), type='dictvallist', repeatlen=1)           # power-levels
ptemp = DATblock(184, {'FX': (90, '30s'), 'notForCustom': (158, 'B'), 'BlockType': (182, 'B'),
                    'Stun': (124, 'B'), 'Accuracy': (123, 'B'), 'PowerType': (132, 'B'),
                    'DamageTypesBlocked': (176, 'H'), 'SubType': (136, 'B'),
                    'animation': (60, '30s'), 'DefenceFlags': (156, 'B'),
                    'Knockback': (125, 'B'), 'Success': (180, 'B'), 'AttackFlags': (156, 'H'),
                    'Magnitude': (140, 'B'), 'SpecialType': (160, 'B'), 'RangeMax': (127, 'B'),
                    'AttackModesBlocked': (172, 'B'), 'PowerName': (0, '30s'),
                    'MaxInstances': (152, 'B'), 'DamageType': (144, 'B'), 'Radius': (128, 'B'),
                    'RangeMin': (126, 'B'), 'EPCost': (121, 'B'), 'Duration': (181, 'B'),
                    'Speed': (122, 'B')}, type='gendict')
hftemp = DATblock(5211, {'energy': (289, 'B'), 'skin': (3519, '30s'), 'speed': (286, 'B'),
                      'strength': (285, 'B'), 'endurance': (287, 'B'),
                      'camp_only': (791, 'B'),
                      'agility': (290, 'B'), 'alterEgo': (801, '30s'), 'AI': (754, '30s'),
                      'objectAttributes': (2949, oatemp),
                      'characterAttributes': (291, catemp),
                      'CSBase': (30, '255s'), 'XP': (788, 'H'), 'CP': (784, 'H'),
                      'class': (2918, '30s'), 'activeAttributes': (790, 'B'),
                      'VID': (751, '2s'), 'powerLevels': (741, pltemp),
                      'NIF': (2663, '255s'), 'BaseTemplateName': (2633, '30s'),
                      'name': (4917, '30s'), 'powers': (793, ptemp)}, type='gendict')
chartemp = DATblock(1078, {'HDR': (0,'19s'),
                      'energy': (310, 'B'), 'speed': (307, 'B'),  # FFEdit character
                      'strength': (306, 'B'), 'endurance': (308, 'B'),
                      'camp_only': (812, 'B'),
                      'agility': (311, 'B'), 'alterEgo': (822, '30s'), 'AI': (775, '30s'),
                      'characterAttributes': (312, catemp),
                      'CSBase': (51, '255s'), 
                      'attrib_start': (811, 'B'),
                      'VID': (772, '2s'), 
                      'tier_a_start': (807,'B'), 'tier_b_start': (808,'B'), 
                      'charName': (21, '30s'),
                      'movementRadius': (818, 'f'),
                      'tier_a': (462, tiertemp), 'tier_b': (612, tiertemp)}, type='gendict')
objtemp = DATblock(5211, {'skin': (3519, '30s'), 
                      'objectAttributes': (2949, oatemp),
                      'class': (2918, '30s'),
                      'NIF': (2663, '255s'),
                      }, type='gendict')
# *** Might want to add some check so that the name field uses the template name if we are writing a built-in character to a hero file.

def CreateDATBlock(pdata,template,echo=0):
    dat = '\00'*template.size
    if template.type == 'gendict':
        for k in pdata.keys():
            if echo: print 'CreateDATBlock: gendict, pdata[%s] = <%s>' % (k,repr(pdata[k]))
            if k in template.fields.keys():
                if isinstance(template.fields[k][1],types.StringType):    # direct struct data
                    dat = overwrite(dat,struct.pack(template.fields[k][1],pdata[k]),template.fields[k][0])
                else:                       # a dictionary of other data
                    if echo: print 'CreateDATBlock: need to create a sub-block'
                    # *** Not working for powers yet.  There is one extra level as the powers are passed as a dictionary of power dictionaries, not just a power dictionary
                    if (template.fields[k][1].type=='dictkeylist'): # this is specific to objectAttrib lists
                        subdata = {}
                        for item in pdata[k]:
                            subdata[item] = pdata[item] 
                        newblock = CreateDATBlock(subdata,template.fields[k][1],echo=echo)
                        dat = overwrite(dat,newblock,template.fields[k][0])
                    elif k == 'powers':
                        pkeys = pdata[k].keys()
                        pkeys.sort()
                        pos = template.fields[k][0]
                        for p in pkeys:
                            subdata = pdata[k][p]
                            newblock = CreateDATBlock(subdata,template.fields[k][1],echo=echo)
                            dat = overwrite(dat,newblock,pos)
                            pos = pos + template.fields[k][1].size
                    else:
                        subdata = pdata[k]
                        newblock = CreateDATBlock(subdata,template.fields[k][1],echo=echo)
                        dat = overwrite(dat,newblock,template.fields[k][0])
                        
                    ###pass
            else:
                if echo: print 'CreateDATBlock: key <%s> in pdata not found in blockdata' % k
    elif template.type == 'list':
        pos = 0
        for pd in pdata:
            if echo: print 'CreateDATBlock: list, pdata[i] = <%s>' % (pd,)
            dat = overwrite(dat,struct.pack(template.fields[0],pd),pos)
            pos = pos + template.repeatlen
    elif template.type == 'dictvallist':
        pos = 0
        pkeys = pdata.keys()
        pkeys.sort()
        for k in pkeys:
            if echo: print 'CreateDATBlock: dictvallist, pdata[%s] = <%s>' % (k,str(pdata[k]))
            dat = overwrite(dat,struct.pack(template.fields[0],pdata[k]),pos)
            pos = pos + template.repeatlen
    elif template.type == 'dictkeylist':
        offset = template.repeatlen - struct.calcsize(template.fields[1])
        pos = 0
        for k in pdata.keys():
            if echo: print 'CreateDATBlock: dictkeylist, pdata[%s] = <%s>' % (k,str(pdata[k]))
            dat = overwrite(dat,struct.pack(template.fields[0],k),pos)
            dat = overwrite(dat,struct.pack(template.fields[1],pdata[k]),pos+offset)
            pos = pos + template.repeatlen
    else:
        print 'CreateDATBlock: unrecognized template type %s for pdata = %s' % \
              (template.type,pdata)

    return dat

##pblockdata = { 'PowerName' : (0,'30s'),  'PowerType' : (132,'B'),  'BlockType' : (182,'B'),
##               'DamageTypesBlocked' : (176,'H'),  'AttackModesBlocked' : (172,'B'),
##               'DefenceFlags' : (156,'B'), 'Success' : (180,'B'), 'EPCost' : (121,'B'),
##               'Duration' : (181,'B'), 'animation' : (60,'30s'), 'FX' : (90,'30s'),
##               'SubType' : (136,'B'), 'Magnitude' : (140,'B'), 'DamageType' : (144,'B'),
##               'Speed' : (122,'B'), 'Stun' : (124,'B'), 'Knockback' : (125,'B'),
##               'RangeMin' : (126,'B'), 'RangeMax' : (127,'B'), 'Accuracy' : (123,'B'),
##               'Radius' : (128,'B'), 'SpecialType' : (160,'B'), 'MaxInstances' : (152,'B'),
##               'AttackFlags' : (156,'H'), 'notForCustom' : (158,'B'),}
### Given a dictionary with power data, create the binary chunk associated with it.
##def CreatePowerBlock(pdata,blockdata,blocklen,echo=0):
##    dat = '\00'*blocklen
##    for k in pdata.keys():
##        if echo: print 'CreatePowerBlock: pdata[%s] = <%s>' % (k,repr(pdata[k]))
##        if k in blockdata.keys():
##            if len(blockdata[k]==2):    # direct struct data
##                dat = overwrite(dat,struct.pack(blockdata[k][1],pdata[k]),blockdata[k][0])
##            else:                       # a dictionary of other data
##                pass
##        else:
##            print 'CreatePowerBlock: key <%s> in pdata not found in blockdata' % k
##    return dat

# Take string block orig and, starting at position pos, overwrite it with the contents of text 
def overwrite(orig,text,pos):
    return orig[:pos] + text + orig[pos+len(text):]

# for debugging
def ShowHero(hero,ShowPowers=1):
    fields = ['name', 'charName', 'isCustom', 'BaseTemplateName',
    'strength', 'speed', 'agility', 'endurance', 'energy','VID', 'AI',
    'NIF', 'material', 'mass', 'skin', 'alterEgo','attrib_start',
    'activeAttributes', 'characterAttributes','objectAttributes',
    'movementRadius', 'class', 'CSBase', 'XP', 'CP','tier_a_start',
    'tier_a', 'tier_b_start', 'tier_b', 'powerLevels', 'arg2', 'arg3',
    'arg4', 'explosion', 'alwaysdetect', 'ambient', 'animated',
    'animRandomOffset', 'avoid', 'baseForce', 'canPickUp',
    'cantIngnite', 'castShadow', 'complex', 'complexFade', 'crusher',
    'customRadius', 'damagedOnMove', 'debris', 'defaultToAttack',
    'detectangleHeight', 'detectangleWidth', 'detectradius',
    'detectradiusMin', 'door', 'elasticity', 'energyPoints',
    'EPRechargeRate', 'flyover', 'health', 'heroPoints',
    'ignoredByVehicle', 'ladder', 'maxEnergyPoints', 'minForce',
    'minHealth', 'mock', 'mp_targettable', 'noFade', 'notInDB',
    'pathBudget', 'physical', 'pickupDistance', 'prestige',
    'rangedZOffset', 'scary', 'semiPhysical', 'shadowType', 'standon',
    'throwable', 'turretMaxAngle', 'turretUseLOS', 'undetectable',
    'unpassable', 'useVoice', 'visibleInEditor', 'wieldable',
    'zombieState', 'powers']    
    debug = []
    keylist = filter(lambda k, hero=hero: k in hero.keys(), fields)
    for k in keylist:
        try:
            if k == 'powers':
                print '%20s : %s' % (k+'.keys()',hero[k].keys())
                if ShowPowers:
                    for p in hero[k].keys():
                        if hero[k][p]:
                            print '%s:' % p
                            test = ShowPower(hero[k][p])
                            if test:
                                debug.append((hero,p,test))
            else:
                print '%20s : %s' % (k,hero[k])
        except:
            debug.append((hero,sys.exc_info()[0]))
    return debug

def ShowPower(pdata):

    fieldnames = ['PowerName', 'PowerType', 'SubType', 'BlockType',
                  'DamageTypesBlocked', 'AttackModesBlocked',
                  'DefenceFlags', 'Success', 'EPCost', 'Duration',
                  'animation', 'FX', 'Magnitude', 'DamageType',
                  'Speed', 'Stun', 'Knockback', 'RangeMin',
                  'RangeMax', 'Accuracy', 'Radius', 'SpecialType',
                  'MaxInstances', 'AttackFlags', 'notForCustom',]
                  ##'powerPositions', 'powerID', '_junk']

    fields = {'PowerName' : '', 'PowerType' : (PT_POWER_TYPES, 0), 'SubType' : (PT_ATTACK_SUBTYPES,0),
              'BlockType' : (PT_BLOCK_TYPES,0), 'DamageTypesBlocked' : (PT_DAMAGES_BLOCKED,1),
              'AttackModesBlocked' : (PT_MODES_BLOCKED,1), 'DefenceFlags' : (PT_DEFENCE_FLAGS,1),
              'Success' : (PT_BLOCK_SUCCESSES,0), 'EPCost' : (PT_EP_COST_NAMES,0),
              'Duration' : (PT_RANGE_NAMES,0), 'animation' : '', 'FX' : '', 'Magnitude' : (PT_DAMAGE_MAGNITUDE_NAMES,0),
              'DamageType' : (PT_DAMAGE_TYPES,0), 'Speed' : (PT_SPEED_NAMES,0), 'Stun' : (PT_STUN_NAMES,0),
              'Knockback' : (PT_KB_NAMES,0), 'RangeMin' : (PT_RANGE_NAMES,0), 'RangeMax' : (PT_RANGE_NAMES,0),
              'Accuracy' : (PT_ACCURACY_NAMES,0), 'Radius' : (PT_RADIUS_NAMES,0), 'SpecialType' : (PT_SPECIAL_TYPES,0),
              'MaxInstances' : '', 'AttackFlags' : (PT_ATTACK_FLAGS,1), 'notForCustom': '',}
              ##'powerPositions': '', 'powerID': '', '_junk': ''}

    keylist = filter(lambda k, pdata=pdata: k in pdata.keys(), fieldnames)
    debug = []
    for k in keylist:
        try:
            if isinstance(fields[k],types.TupleType):
                ShowIndexMatches(pdata,k,fields[k][0],fields[k][1])
            else:
                print '\t%s = %s' % (k,pdata[k])
        except (KeyError, IndexError):
            print '\t***possible invalid key-value pair***  %s = %s' % (k,pdata[k])
            debug.append((k,sys.exc_info()[0]))
    return debug

def ShowIndexMatches(data,key,dict,multiple=0):
    s = '\t%s = ' % key
    if multiple:
        for k in dict.keys():
            if k & data[key]:  s = s + ('%s ' % dict[k])
    else:
        s = s + ('%s' % dict[data[key]])
    print s
            

# Testing: Can we just assume that the first record is always at
# FIRSTITEMPOS?  Let's do and decide which FX version (FF1 or FF2) we
# have from that.
def FindFXDATVersion(dat):
    global _FXData
    FIRSTITEMPOS = 2048
    ITEMSIZE = 1024
    hdr = 'HDR4\x06\x00\x00\x00sFXDef'
    hdrlen = len(hdr)
    datlen = len(dat)
    cword = GetDatWords(dat,FIRSTITEMPOS,1,'\x1f')[0]
    cwordpos = GetDatWordPos(dat,FIRSTITEMPOS,1,'\x1f')
    cwordlen = len(cword)
    ##print 'checking <%s> vs cword=<%s>' % (repr(dat[cwordpos+cwordlen+1:cwordpos+cwordlen+cwordlen+1]),repr(cword))
    if dat[cwordpos+cwordlen+1:cwordpos+cwordlen+cwordlen+1] == cword:
        ##print 'found FF1 first FX entry <%s> at %d' % (cword,cwordpos)
        return 1
    else:
        ##print 'checking <%s> vs cword=<%s>' % (repr(dat[cwordpos+cwordlen+1:cwordpos+cwordlen+1+hdrlen]),repr('HDR4\x06\x00\x00\x00sFXDef'))
        if ( dat[cwordpos+cwordlen+1:cwordpos+cwordlen+1+hdrlen] == 'HDR4\x06\x00\x00\x00sFXDef' ):
            nword = GetDatWords(dat,cwordpos+cwordlen+hdrlen+9,1,'\x1f')[0]
            ##print 'checking nword=<%s> vs cword=<%s>' % (repr(nword),repr(cword))
            if nword == cword:
                ##print 'found FF2 first FX entry <%s> at %d' % (cword,cwordpos)
                return 2
        else:
            print 'FindFXDATVersion: **** NO FX ENTRY FOUND AT %d ****' % FIRSTITEMPOS
            return 0

# Return the whole binary blob of the next fx.dat fx record that
# starts at startpos or later.
def GetNextFXRecord(dat,startpos):
    powerrec = ''
    # get the first matching pair
    cword = ''
    cwordpos = startpos
    nword = GetDatWords(dat,startpos,1,'\x1f')[0]
    nwordpos = GetDatWordPos(dat,startpos,1,'\x1f')
##    print 'cword=<%s> cwordpos=%d nword=<%s> nwordpos=%d dat[nwordpos-1]=<%s>  (cwordpos + len(cword) + 1)=%d' % \
##          (cword,cwordpos,nword,nwordpos,repr(dat[nwordpos-1]),cwordpos + len(cword) + 1)
    # check that the two words are equal and separated by one zero-blank
    while ( nword ) and not ( ( nword == cword ) and ( dat[nwordpos-1] == '\x00' ) and ( nwordpos == cwordpos + len(cword) + 1 ) ):
        cword = nword
        cwordpos = nwordpos
        nword = GetDatWords(dat,cwordpos+len(cword),1,'\x1f')[0]
        nwordpos = GetDatWordPos(dat,cwordpos+len(cword),1,'\x1f')
    # track the starting position
    ##print 'GetNextFXRecord: first match found at %d, cword=<%s> nword=<%s>' % (cwordpos,cword,nword)
    recstartpos = cwordpos
    duppos = nwordpos
    # now find the next pair
    cword = nword
    cwordpos = nwordpos
    nword = GetDatWords(dat,cwordpos+len(cword),1,'\x1f')[0]
    nwordpos = GetDatWordPos(dat,cwordpos+len(cword),1,'\x1f')
    while ( nword ) and not ( ( nword == cword ) and ( dat[nwordpos-1] == '\x00' ) and ( nwordpos == cwordpos + len(cword) + 1 ) ):
        cword = nword
        cwordpos = nwordpos
        nword = GetDatWords(dat,cwordpos+len(cword),1,'\x1f')[0]
        nwordpos = GetDatWordPos(dat,cwordpos+len(cword),1,'\x1f')
    # now check that and return the chunk of dat
    ##print 'GetNextFXRecord: second match or EOF found at %d, cword=<%s> nword=<%s>' % (cwordpos,cword,nword)
    if ( nword ):
        recstoppos = cwordpos
    else:
        recstoppos = len(dat)
    powerrec = dat[recstartpos:recstoppos]
    return powerrec, recstartpos

# Return the next wordnum words in dat as a list, starting at postition pos.
# Reads past initial empty space.
# "Words" are separated by one or more bytes with ordinal value == 0, or
# if minord is specified, by any bye with an ordinal value less than
# or equal to minord. Similarly, maxord is the highest ordinal value that
# a non-delimiter may have.
def GetDatWords(dat,pos,wordnum=1,minord='\x00',maxord='\xC8'):
    words = []
    datlen = len(dat)
    for i in range(wordnum):
        w = ''
        while ( pos < datlen ) and ( ( dat[pos] <= minord ) or ( dat[pos] > maxord ) ):
            pos = pos+1
        while ( pos < datlen ) and ( dat[pos] > minord ) and ( dat[pos] <= maxord ):
            w = w + dat[pos]
            pos = pos+1
        ##print 'GetDatWords: just got <%s> ending at pos=%d' % (w,pos)
        words.append(w)
    return words

# Return the absolute position of the wordnum word in dat, starting in
# dat a postition pos.  If no wordnum word is found, returns -1.
# Reads past initial empty space.
# "Words" are separated by one or more bytes with ordinal value == 0, or
# if minord is specified, by any bye with an ordinal value less than
# or equal to minord. Similarly, maxord is the highest ordinal value that
# a non-delimiter may have.
def GetDatWordPos(dat,pos,wordnum=1,minord='\x00',maxord='\xC8'):
    startpos = -1
    datlen = len(dat)
    for i in range(wordnum):
        while ( pos < datlen ) and ( ( dat[pos] <= minord ) or ( dat[pos] > maxord ) ):
            pos = pos+1
        while ( pos < datlen ) and ( dat[pos] > minord ) and ( dat[pos] <= maxord ):
            if ( i == wordnum-1 ):
                startpos = pos
                break
            pos = pos+1
    return startpos

# I stole this from string.py because FF's old version of python
# doesn't support find for strings with null embedded characters (e.g.
# DAT data!). Find substring, return -1 if not found *** Check that
# this is really needed if we are careful about declaring strings that
# start/end with null chars.
def find(s, sub, i = 0, last=None):
	"""find(s, sub [,start [,end]]) -> in

	Return the lowest index in s where substring sub is found,
	such that sub is contained within s[start,end].  Optional
	arguments start and end are interpreted as in slice notation.

	Return -1 on failure.

	"""
	Slen = len(s)  # cache this value, for speed
	if last is None:
		last = Slen
	elif last < 0:
		last = max(0, last + Slen)
	elif last > Slen:
		last = Slen
	if i < 0: i = max(0, i + Slen)
	##f = open('test.log','w')
	n = len(sub)
	m = last + 1 - n
	while i < m:
		matched = 1
		pos = 0
		while matched and ( pos < n ):
			##f.write('sub[%d]=%s (%x) s[%d]=%s(%x)\n' % (pos,sub[pos],ord(sub[pos]),i+pos,s[i+pos],ord(s[i+pos])))
			matched = matched and (sub[pos] == s[i+pos])
			pos = pos + 1
		if matched:
			##f.close()
			return i
		i = i+1
	##f.close()
	return -1

def eventeval(fn,event):
    if '.' in fn:   # module was specified
        try:
            eval(fn+'(event)')
        except:
            try:    # maybe we haven't imported it
                dotpos = len(fn)-2
                while fn[dotpos]!='.':  dotpos=dotpos-1
                module = fn[:dotpos]
                exec('import '+module)
                eval(fn+'(event)')
            except:
                print 'eventeval: problem evaluating string %s' % repr(fn+'(event)')
                raise
    else:   # no module specified: take some guesses then try them all
        success = 0
        tryfirst = ['ff.','cshelper.','ffx.','m25ai.']
        for module in tryfirst: # try the preferred modules
            try:
                eval(module+fn+'(event)')
                success = 1
                break
            except:
                try:    # maybe we haven't imported it
                    exec('import '+module[:-1])
                    eval(module+fn+'(event)')
                    success = 1
                    break
                except:
                    pass
        if not success:
            for module in sys.modules.keys():   # try all the imported modules
                try:
                    eval(module+'.'+fn+'(event)')
                    success = 1
                    break
                except:
                    try:    # maybe we haven't imported it
                        exec('import '+module)
                        eval(module+'.'+fn+'(event)')
                        success = 1
                        break
                    except:
                        pass
        if not success:
            print 'eventeval: unable to evaluate string %s in any module.' \
                  % repr(fn+'(event)')
            raise
    
#################################
### Guessing Files and Paths ####
#################################

# All of these functions return absolute paths.

# Determine if the code is being run from within the game
def inGame():
    return 'ff' in sys.modules.keys()

# Determine which version of FF is running.  Returns default if not run from within the
# game.  Otherise returns 1 for Freedom Force and 2 for Freedom Force vs. The Third Reich.
def ffver(default = 0):
    if sys.argv[0] == 'FreedomForce':
        return 1
    elif sys.argv[0] == 'Freedom Force vs the 3rd Reich':
        return 2
    else:
        print 'ffver: unrecognized game version, returning %s' % str(default)
        return default

# Returns FF directory when run from within the game. If run from outside the game,
# gameVersion must be specified as either 1 or 2 and it will grab the FF directory
# from the Windows registry.
# WARNING: If run from within the game, gameVersion is ignored and the version of
# the running game is returned.
def GetFFPath(gameVersion=-1):
    if inGame():  # from within game
        gameVersion = ffver()
        if gameVersion >  0:
            return os.path.dirname(sys.executable)
    elif gameVersion == 1:
        import _winreg
        return _winreg.QueryValueEx(_winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,'SOFTWARE\\Irrational Games\\Freedom Force\\'),'InstallDir')[0]
        return FFdir
    elif gameVersion == 2:
        import _winreg
        return _winreg.QueryValueEx(_winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,'SOFTWARE\\Irrational Games\\FFVTTR\\'),'InstallDir')[0]
    return ''

# Find the path where the basic mod DATs (e.g. characters, powers,
# objects, etc.) are when run during a mission.  This path is returned
# lower-case. Added some I/O to get this working with FFvtTR and added
# caching so that frequent calls aren't impeded by the I/O.  Update:
# I/O not needed, keeping caching anyway.
def GetModPath(ForceCheck=0):
    global _ModPath
    if ( not ForceCheck ) and ('_ModPath' in globals().keys() ):
        return _ModPath
    _ModPath = ''
    syspath = map(string.lower,sys.path)
    workingdir = string.lower(os.getcwd())
    if ffver() == 1:     # for v1 we can get this from the system path.
        tailstring = 'temp'
    elif ffver() == 2:
        tailstring = 'missions\\scripts'
    else:   # likely not in game
        return ''
    for p in syspath:
        if ( p[:len(workingdir)]==workingdir ) and ( p[-len(tailstring):]==tailstring ):
            _ModPath = os.path.normpath(p[:-len(tailstring)])
    return _ModPath

# Find the path where temporary files are saved.  This path is
# returned lower-case.  This is NOT TESTED for the first game!
# Intended for use with the second game where this stuff is stored in
# places like 'C:\Documents and Settings\vivalalibertad\Local
# Settings\Application Data\Irrational Games\Freedom Force vs the 3rd
# Reich\' *** Need to see if this works on pre-Win2000 machines.
# If AssumeVersion is not 0, we assume that to be the desired version
# and return a path accordingly.  This makes some assumptions about
# the location of the install (for version 2, for version 1 it just
# returns the mod path).
def GetWinTempPath(ForceCheck=0,ModPath=''):
    global _WinTempPath
    if ( not ForceCheck ) and ('_WinTempPath' in globals().keys() ):
        return _WinTempPath
    _WinTempPath = ''
    if inGame():
        if ffver() == 1:     # NOT TESTED.  Just returns mod path
            if ModPath:
                _WinTempPath = ModPath
            else:
                _WinTempPath = GetModPath(ForceCheck)
            return _WinTempPath
        # now assuming we are in FF2
        if os.environ.get('OS') == 'Windows_NT':
            tailstring = 'temp'
            syspath = map(string.lower,sys.path)
            headstr = string.lower(os.environ['USERPROFILE'])
            for p in syspath:
                if ( p[:len(headstr)]==headstr ) and ( p[-len(tailstring):]==tailstring ):
                    _WinTempPath = os.path.normpath(p[:-len(tailstring)])
        else:   # Win9X/ME machine
            _WinTempPath = GetFFPath()
    else:   # we're runing outside the game environement, need a ModPath for this
        if string.find(string.lower(ModPath),'reich') >= 0:
            if os.environ.get('OS') == 'Windows_NT':
                _WinTempPath = os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                                            'Local Settings','Application Data',
                                            'Irrational Games','Freedom Force vs the 3rd Reich')
            else:   # Win9X/ME machine
                _WinTempPath = GetFFPath(2)
        else:
            _WinTempPath = ModPath
    return _WinTempPath

# Try to guess where the current mission.dat file is (when run during a mission).
# Note: Apparently, skirmish missions run from a mod can use maps and mission data
# from the main campaign directory, so we return that if we can't find anything in
# the corresponding mod directory.
# NOTE: FFvsT3R supports the zipfile module for scripting, so this function checks
# for the mission.dat file and, if it does not find it, attempts to unzip the files
# for that entire mission.  I have found instability if only the mission.dat file
# is unzipped.
def GetMissionDatFile(ForceCheck=0):
    global _DatPath
    if ( not ForceCheck ) and ('_DatPath' in globals().keys() ):
        return _DatPath
    _DatPath = ''
    ModPath = GetModPath(ForceCheck)
    # look through the ff.log file
    lf = open('ff.log','r')
    lines = lf.readlines()
    lf.close()
    lines.reverse()
    keep=''
    matchstr='CMission::init('
    for line in lines:
        splitline = string.split(line)
        if ( len(splitline) > 2 ) and ( splitline[2][:len(matchstr)] == matchstr ):
            keep = line
            break
    if keep == '':
        return _DatPath
    MissionPath = string.split(keep,',')[-1][1:-2]
    ##print 'GetMissionDatFile: MissionPath = <%s>' % MissionPath
    _DatPath = os.path.join(ModPath,MissionPath)
    ##print 'GetMissionDatFile: before 1st isfile check _DatPath = <%s>' % _DatPath
    if os.path.isfile(_DatPath):
        return _DatPath
    # try looking in the Data directory (hopefully this works for rumble room maps)
    _DatPath = os.path.join(os.path.dirname(ModPath),'Data',MissionPath)
    ##print 'GetMissionDatFile: before 2nd isfile check _DatPath = <%s>' % _DatPath
    if os.path.isfile(_DatPath):
        return _DatPath
    # See if we can read the file from the zipped archive
    zipname = os.path.join(ModPath,'missions.ff')
    if os.path.isfile(zipname):
        try:
            import zipfile
            print 'GetMissionDatFile: <%s> not found; trying to unzip from <%s>' % (MissionPath,zipname)
            zf = zipfile.ZipFile(zipname,'r')
            MPlower = string.lower(MissionPath)
            zipnamelist = zf.namelist()
            zipnormlist = map(os.path.normpath,map(string.lower,zipnamelist))
            if MPlower in zipnormlist:
                # Get all of the zipped files for this mission
                missiondir = os.path.dirname(MissionPath)                    # in Win format
                fullmissiondir = os.path.dirname(_DatPath)                   # in Win format
                if not os.path.isdir(fullmissiondir):
                    os.makedirs(fullmissiondir)
                missiondirlower = string.lower(missiondir)                  # in Win format, lowercase
                missionzips = filter(lambda f, mdl=missiondirlower: f[:len(mdl)]==mdl, zipnormlist)     # in Win format, lowercase
                for f in missionzips:
                    unzippedname = os.path.join(fullmissiondir,os.path.basename(f)) # Win Format
                    if not os.path.isfile(unzippedname):
                        zipfname = zipnamelist[zipnormlist.index(f)]            # unix format
                        dat = zf.read(zipfname)
                        print '    GetMissionDatFile: unzipping  %s' % (unzippedname,)
                        f = open(unzippedname,'wb')
                        f.write(dat)
                        f.close()
                zf.close()
                return _DatPath
        except ImportError:
            pass
    _DatPath = ''   # we were unable to unzip
    return _DatPath

# For the basic mod DATs (e.g. characters, powers, objects, etc.), try to determine
# the full filepath (when run during a mission).  If it doesn't find a current modpath
# tries to use the main campaign (the Data mod).  fname is the local DAT file name, e.g. 
# 'objects.dat', without any path information.
##def GetDatFile(fname):
##    # find the base path
##    ModPath = GetModPath()
##    if (ModPath == ''):
##        ModPath = os.path.join(GetFFPath(),'Data')
##    DatPath = os.path.join(ModPath,fname)
##    if os.path.isfile(DatPath):
##        return DatPath
##    return ''
def GetDatFile(fname):
    # find the base path
    ModPath = GetModPath()
    DatPath = os.path.join(ModPath,fname)
    if os.path.isfile(DatPath):
        return DatPath
    ModPath = os.path.join(GetFFPath(),'Data')
    DatPath = os.path.join(ModPath,fname)
    if os.path.isfile(DatPath):
        return DatPath
    return ''

def GetLangPath(language=''):
    if not language:
        if inGame():
            import ff
            language = ff.Language
        else:
            language = 'english'
    LangPath = os.path.abspath(os.path.join(GetModPath(),'lang',language))
    if os.path.isdir(LangPath):
        return LangPath
    else:
        print 'GetLangPath: couldn\'t determine language path'
        return ''

# Try to determine the full filepath (when run during a mission) of
# the most recently accessed saved game file that is from a mission
# that comes no later in the campaign than the current mission.  NB: I
# am assuming every DAT file in the SaveGames and SaveGames/quicksave
# directories is a saved game.
#
# Unfortunately, we probably should not cache this, even though it
# might be slow.  If the player saves during a mission, we want to
# catch that because he might not have saved after levelling up his
# heroes at the base.  We can cache the ff.log read for
# _currentMission, though.
#
# Note that, in general, we might not always want the most recent
# saved game.  We might have saved a game and then went back and
# played another mission from the start (e.g. by loading the saved
# base).  So, in those cases, we are just guessing the most recent
# saved game before the current one in the mission list is what we
# want.  But, if we are playing from the middle of a saved mission
# that we loaded, we always want the name of the loaded saved game, so
# we grab it from GetLoadedSavedGame() and return that.
##import time
def GetMostRecentSavedGameFile(currentMission=''):
    import js
    global _currentMission
    try:    # make this base-safe
        if js.Mission_GetAttr('MLOG_SavedGameLoaded') == 1:
            return GetLoadedSavedGame()
    except RuntimeError:
        pass
    currentMission = string.upper(currentMission)
    # find the base path
    ##tstart = time.clock()
    ModPath = GetModPath()
    if ffver() == 1:
        SavePath = os.path.join(ModPath,'SaveGames\\')
    else:
        TempPath = GetWinTempPath()
        SavePath = os.path.join(TempPath,'user','SaveGames\\')
    if not ( currentMission or ( '_currentMission' in globals().keys() ) ):
        ##print 'GetMostRecentSavedGameFile: readling ff.log'
        lf = open('ff.log','r')
        lines = lf.readlines()
        lines.reverse()
        lf.close()
        keep=''
        matchstr='CMission::init('
        for line in lines:
            splitline = string.split(line)
            if ( len(splitline) > 2 ) and ( splitline[2][:len(matchstr)] == matchstr ):
                keep = line
                break
        _currentMission = string.upper(string.split(string.split(keep,',',1)[0],'(')[1])
    elif currentMission:
        ##print 'GetMostRecentSavedGameFile: currentMission given'
        _currentMission = currentMission
    currentMission = _currentMission
    ##print 'GetMostRecentSavedGameFile currentMission = %s' % currentMission
    if string.upper(os.path.basename(ModPath)) == 'DATA':  # main campaign has no campdef.dat
        missions = filter(lambda name: (name[0] in string.digits) and (name[1] in string.digits),os.listdir(os.path.join(ModPath,'Missions\\')))
        missions = map(string.upper,missions)
    else:        
        missions = map(string.upper,Campaign_ReadCampaignDefinitions(os.path.join(ModPath,'campdef.dat'))['missions'])
    if currentMission not in missions:      # skirmish match or bad argument passed
        return ''
    missionsbefore = missions[:missions.index(string.upper(currentMission))+1]
    filelist = [os.path.join(SavePath,'quicksave', 'game.dat'),
                os.path.join(SavePath,'autosave', 'autosave - in base.dat'),
                os.path.join(SavePath,'autosave', 'autosave - in mission.dat')]
    filelist = filter(os.path.isfile,filelist)
    saved = filter(lambda f, SavePath=SavePath: os.path.isfile(os.path.join(SavePath,f)),os.listdir(SavePath))
    saved = filter(lambda f: f[-len('.dat'):]=='.dat',saved)
    saved = map(lambda f, SavePath=SavePath: os.path.join(SavePath,f), saved)
    filelist.extend(saved)
    filetimes = map(os.path.getmtime,filelist)
    files = map(lambda i,filetimes=filetimes,filelist=filelist: (filetimes[i],filelist[i]),range(len(filelist)))
    files.sort()
    # now open the first 4k of these files, get the mission name and see if its string.upper is in missionsbefore
    while  files:
        file = files.pop()
        f = open(file[1],'rb')
        f.seek(3072)
        dat = f.read(256)
        f.close()
        fnamelen = ord(dat[16])
        SavedMission = string.split(GetDatWords(dat,17,2)[1],os.sep)[1]
        if string.upper(SavedMission) in missionsbefore:
            ##print 'GetMostRecentSavedGameFile: elapsed time = %2.4f' % (time.clock()-tstart,)
            return file[1]
    ##print 'GetMostRecentSavedGameFile: elapsed time = %2.4f' % (time.clock()-tstart,)
    return ''

# Try to determine the full filepath (when run during a mission) of
# the saved game file that was just loaded.  This only works when used
# on a saved game file saved while using mlogreader 1.0 alpha 14 or
# later, though that module need not be imported here.
#
# The passed parameter is the game time at which the saved-game load
# was detected.  If not given, it will be assumed to be the time
# currently in the mission attribute 'MLOG_GameTime'.  Generally, this
# function will only be called without passing that value when it is
# being asked to return its cached file name (that it got when it
# *was* called with the GameTime).
#
# We *can* cache this, since we would have restarted Python fresh if
# we loaded another saved game.
#
# This is an FFvT3R-only function, as the first game did not have the
# needed mlogreader functionality.
def GetLoadedSavedGame(GameTime=-1.0):
    import _missionvar
    global SavedGameName
    try:        # Is it cached?
        return SavedGameName
    except:     # No. Get to work!
        pass
    if GameTime == -1.0:    # not called by mlogreader
        import js
        if not js.Mission_GetAttr('MLOG_SavedGameLoaded'):
            return ''
        GameTime = js.Mission_GetAttr('MLOG_GameTime')
    # get this with a direct call to _missionvar because caching isn't clean yet
    currentMission = _missionvar._Mission_GetVar('MLOG_MissionName')
    TempPath = GetWinTempPath()
    SavePath = os.path.join(TempPath,'user','SaveGames\\')
    filelist = [os.path.join(SavePath,'quicksave', 'game.dat'),
                os.path.join(SavePath,'autosave', 'autosave - in base.dat'),
                os.path.join(SavePath,'autosave', 'autosave - in mission.dat')]
    filelist = filter(os.path.isfile,filelist)
    saved = filter(lambda f, SavePath=SavePath: os.path.isfile(os.path.join(SavePath,f)),os.listdir(SavePath))
    saved = filter(lambda f: f[-len('.dat'):]=='.dat',saved)
    saved = map(lambda f, SavePath=SavePath: os.path.join(SavePath,f), saved)
    filelist.extend(saved)
    filetimes = map(os.path.getmtime,filelist)
    files = map(lambda i,filetimes=filetimes,filelist=filelist: (filetimes[i],filelist[i]),range(len(filelist)))
    files.sort()
    # now open the first 4k of these files, get the mission name and check matches for their stored game time
    SavedGameName = ''
    BestFileTimeDiff = 1.0e6
    while  files:
        file = files.pop()
        f = open(file[1],'rb')
        f.seek(3072)
        dat = f.read(256)
        fnamelen = ord(dat[16])
        SavedMission = string.split(GetDatWords(dat,17,2)[1],os.sep)[1]
        if string.upper(SavedMission) == currentMission:
            f.seek(0)
            dat = f.read()
            pos = string.find(dat,'MLOG_GameTime')
            if pos >= 0:
                FileTime = struct.unpack('f',dat[pos+13:pos+17])[0]
                FileTimeDiff = GameTime - FileTime
                if ( FileTimeDiff >= 0.0 ) and ( FileTimeDiff < BestFileTimeDiff ):
                    BestFileTimeDiff = GameTime - FileTime
                    SavedGameName = file[1]
                    if ( FileTimeDiff == 0.0 ): # it won't get better than this
                        f.close()
                        break
        f.close()
    return SavedGameName

# ModPath parameter to be passed when run outside the game.
def GetHeroFilePath(ModPath='',ForceCheck=0):
    global _HeroFilePath
    if ( not ForceCheck ) and ('_HeroFilePath' in globals().keys() ):
        return _HeroFilePath
    _HeroFilePath = ''
    if inGame():
        if ffver() == 1:
            ModPath = GetModPath(ForceCheck=ForceCheck)
            _HeroFilePath = os.path.join(ModPath,'Heroes')
        elif ffver() == 2:
            TempPath = GetWinTempPath(ModPath=ModPath, ForceCheck=ForceCheck)
            _HeroFilePath = os.path.join(TempPath,'Custom','Heroes')
    else:
        if string.find(string.lower(ModPath),'reich') >= 0:
            TempPath = GetWinTempPath(ModPath=ModPath, ForceCheck=ForceCheck)
            _HeroFilePath = os.path.join(TempPath,'Custom','Heroes')
        else:
            _HeroFilePath = os.path.join(ModPath,'Heroes')
    return _HeroFilePath

############################################
### Working with FFX Control Centre ########
############################################

# Returns all SaveGames paths for the given FF version.  It will check
# that the file or directory given by checkfor is in the mod folder.
# If checkfor is a list of filenames, it will check that all files in
# the list is in each mod folder.
def GetAllModPaths(gameVersion=-1,checkfor=['characters.dat','objects.dat','powers.dat','attributes.dat']):
    if ( gameVersion == -1 ) and ( inGame() ):
        gameVersion = ffver()
    if ( gameVersion == 1 ) or ( gameVersion == 2 ):
        if not isinstance(checkfor,types.ListType):
            checkfor = [checkfor]
        FFsubdirs = filter(os.path.exists,map(lambda f, checkfor=checkfor[0], FFdir=GetFFPath(gameVersion=gameVersion):
                                              os.path.join(FFdir,f,checkfor), os.listdir(GetFFPath(gameVersion=gameVersion))))
        FFsubdirs = map(os.path.dirname,FFsubdirs)
        for cf in checkfor[1:]:
            FFsubdirs = filter(os.path.exists,map(lambda f, checkfor=cf, FFdir=GetFFPath(gameVersion=gameVersion):
                                                  os.path.join(FFdir,f,checkfor), FFsubdirs))
            FFsubdirs = map(os.path.dirname,FFsubdirs)
        return FFsubdirs

# Returns all SaveGames paths for the given FF version.
def GetAllSaveGamePaths(gameVersion=-1):
    if ( gameVersion == -1 ) and ( inGame() ):
        gameVersion = ffver()
    if gameVersion == 1:
        FFsubdirs = filter(os.path.isdir,map(lambda f, FFdir=GetFFPath(gameVersion=gameVersion):
                                             os.path.join(FFdir,f,'SaveGames'), os.listdir(GetFFPath(gameVersion=gameVersion))))
        return FFsubdirs
    if gameVersion == 2:
        FFSGdirs = [os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                                 'Local Settings','Application Data',
                                 'Irrational Games','Freedom Force vs the 3rd Reich',
                                 'User','SaveGames')]
        FFSGdirs.append(os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                                     'Local Settings','Application Data',
                                     'Irrational Games','Freedom Force vs the 3rd Reich',
                                     'User','SaveGames','quicksave'))
        FFSGdirs.append(os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                                     'Local Settings','Application Data',
                                     'Irrational Games','Freedom Force vs the 3rd Reich',
                                     'User','SaveGames','autosave'))
        return FFSGdirs

# Returns all Hero paths for the given FF version.
def GetAllHeroFilePaths(gameVersion=-1):
    if ( gameVersion == -1 ) and ( inGame() ):
        gameVersion = ffver()
    if gameVersion == 1:
        FFsubdirs = filter(os.path.isdir,map(lambda f, FFdir=GetFFPath(gameVersion=gameVersion):
                                             os.path.join(FFdir,f,'Heroes'), os.listdir(GetFFPath(gameVersion=gameVersion))))
        return FFsubdirs
    if gameVersion == 2:
        FFHdir = os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                              'Local Settings','Application Data',
                              'Irrational Games','Freedom Force vs the 3rd Reich',
                              'Custom','Heroes')
        return [FFHdir]

def MakeHeroIndices(gameVersion=-1):
    if ( gameVersion == -1 ) and ( inGame() ):
        gameVersion = ffver()
    if ( gameVersion == 1 ) or ( gameVersion == 2 ):
        for mp in GetAllModPaths(gameVersion=gameVersion):
            Campaign_MakeHeroIndex(ModPath=mp, indexFileName='', ForceRead=1)

# Generate the batch file wrapper for the FFX Control Centre so that
# it will save the file times before CC is run, restore them after,
# and run the Campaign_MakeHeroIndex() for all of that game version's
# mods. Since CC has no registry entries, I am going to assume it is
# installed in the base game directory.
def GenerateCCWrapper(gameVersion=-1):
    if ( gameVersion == -1 ) and ( inGame() ):
        gameVersion = ffver()
    if gameVersion == 1:
        scriptpath = os.path.join(GetFFPath(gameVersion=gameVersion),'Data','Missions','Scripts')
        wrapfname = os.path.join(scriptpath,'ControlCentreWrapper.bat')
        wrapf = open(wrapfname,'w')
        wrapf.write('"%s" --gameVersion 1\n' % os.path.join(scriptpath,'SaveFileTimeInfo.py'))
        wrapf.write('"%s"\n' % os.path.join(GetFFPath(gameVersion=gameVersion),'FFXEdit.exe'))
        wrapf.write('"%s" --gameVersion 1\n' % os.path.join(scriptpath,'RestoreFileTimeInfo.py'))
        wrapf.write('"%s" --MakeHeroIndices 1\n' % os.path.join(scriptpath,'datfiles.py'))
        wrapf.close()

#################################
### Writing Functions ###########
#################################

# Read a objects.dat file and change the minForce values of all the
# generic objects (that don't already have it explicitly set) to be
# MassFactor*mass.
# MassFactor defaults to 10.
# Makes a backup of the previous DAT as FileName+'.bak' in the same
# directory that FileName (probably "objects.dat") was in.
def Campaign_ChangeObjectMinforce(FileName='',MassFactor=10.0,verbose=0):
    if FileName=='':        # try to guess the current DAT file
        FileName = GetDatFile('objects.dat')
    FileName = os.path.abspath(FileName)
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    f = open(FileName,'rb')
    alldat = f.read()
    f.close()
    datlen = len(alldat)
    newdat = alldat[0:0]
    pos = 0
    changed = 0
    if verbose:
        print '# Campaign_ChangeObjectMinforce called at %s on file <%s>.  The following is a list of the changed objects' \
              % (time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime(time.time())), FileName)
        print '['
    while pos < datlen:
        gotobj = 0
        if ord(alldat[pos]) > 8:                        # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            if words[0] == words[1]:                    # check that it's a legit name
                gotobj = 1
                objName = words[0]
                objClass = words[3]
                if ( objClass == 'GAME_OBJ_GENERIC' ) or ( objClass == 'GAME_OBJ_VEHICLE' ):
                    offset = GetDatWordPos(dat,0,2) + 316   # first attrib
                    attribs = []
                    throwable = 0.0
                    wieldable = 0.0
                    for count in range(15):
                        if dat[offset+count*36] != '\x00':
                            attrib = GetDatWords(dat,offset+36*count,1)[0]
                            attribs.append(attrib)
                            if attrib == 'mass':
                                mass = struct.unpack('f',dat[offset+36*count+32:offset+36*count+36])[0]
                            if attrib == 'throwable':
                                throwable = struct.unpack('f',dat[offset+36*count+32:offset+36*count+36])[0]
                            if attrib == 'wieldable':
                                wieldable = struct.unpack('f',dat[offset+36*count+32:offset+36*count+36])[0]
                    if ( 'minForce' not in attribs ) and \
                       ( 'mass' in attribs ) and \
                       ( ( 'throwable' not in attribs ) or ( throwable > 0.0 ) or ( wieldable > 0.0 ) ):
                        for count in range(15):     # find first empty attrib slot
                            if dat[offset+count*36] == '\x00':
                                break
                        if count < 15:
                            if verbose: print '%s, # minForce=%f' % (objName,mass*MassFactor)
                            minForce = '\x00'*(32-len('minForce'))
                            minForce = minForce+struct.pack('f',mass*MassFactor)
                            dat = dat[:offset+count*36] + 'minForce' + minForce + dat[offset+(count+1)*36:]
                            changed = changed + 1
                        else:
                            if verbose: print '# ERROR: %s no empty attrib slots found.' % (objName,)
        if gotobj:
            newdat = newdat + dat
            pos = pos + ITEMSIZE
        else:
            newdat = newdat + alldat[pos:pos+SKIPSIZE]
            pos = pos + SKIPSIZE
    if verbose: print ']'
    if changed > 0:
       print ' ... changed %d objects' % (changed,)
       try:
           i = 1
           backupname = '%s.bak' % (FileName,)
           while os.path.isfile(backupname):
               i = i + 1   # first numbered .bak will be '.bak2'
               backupname = '%s.bak%d' % (FileName,i)
           os.rename(FileName,backupname)
           f=open(FileName,'wb')
           f.write(newdat)
           f.close()
       except OSError:
           print 'Campaign_ChangeObjectMinforce: Unable to create backup of file <%s>, possible Read-Only problem.  Quitting' % (FileName,)
           raise
       except:
           print 'Campaign_ChangeObjectMinforce: Unable to create backup of file <%s>.  Quitting' % (FileName,)
           raise

# Read a objects.dat file and remove the 'maxEnergyPoints' template
# attribute from any character templates for whom it has been set to
# something elss than 100.
# Makes a backup of the previous DAT as FileName+'.bak' in the same
# directory that FileName (probably "objects.dat") was in.
def Campaign_RemoveObjectMaxEP(FileName='',verbose=0):
    if FileName=='':        # try to guess the current DAT file
        FileName = GetDatFile('objects.dat')
    FileName = os.path.abspath(FileName)
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    f = open(FileName,'rb')
    alldat = f.read()
    f.close()
    datlen = len(alldat)
    newdat = alldat[0:0]
    char_classes = ['GAME_OBJ_MINION','GAME_OBJ_HERO', 'GAME_OBJ_POLICE', 'GAME_OBJ_CIVILIAN', 'GAME_OBJ_VILLIAN', 'GAME_OBJ_TANK']
    pos = 0
    changed = 0
    if verbose:
        print '# Campaign_RemoveObjectMaxEP called at %s on file <%s>.  The following is a list of the changed objects' \
              % (time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime(time.time())), FileName)
        print '['
    while pos < datlen:
        gotobj = 0
        if ord(alldat[pos]) > 8:                        # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            if words[0] == words[1]:                    # check that it's a legit name
                gotobj = 1
                objName = words[0]
                objClass = words[3]
                if ( objClass in char_classes ):
                    offset = GetDatWordPos(dat,0,2) + 316   # first attrib
                    attribs = []
                    maxEnergyPoints = 100
                    for count in range(15):
                        if dat[offset+count*36] != '\x00':
                            attrib = GetDatWords(dat,offset+36*count,1)[0]
                            attribs.append(attrib)
                            if attrib == 'maxEnergyPoints':
                                maxEnergyPoints = struct.unpack('f',dat[offset+36*count+32:offset+36*count+36])[0]
                                if maxEnergyPoints < 100:   # then blank out that attrib slot
                                    if verbose:
                                        print 'blanking %s attrib (former value = %3.2f) for object <%s>' % (attrib,maxEnergyPoints,objName)
                                    dat = dat[:offset+count*36] + ('\x00'*36) + dat[offset+(count+1)*36:]
                                    changed = changed + 1
        if gotobj:
            newdat = newdat + dat
            pos = pos + ITEMSIZE
        else:
            newdat = newdat + alldat[pos:pos+SKIPSIZE]
            pos = pos + SKIPSIZE
    if verbose: print ']'
    if changed > 0:
       print ' ... changed %d objects' % (changed,)
       try:
           i = 1
           backupname = '%s.bak' % (FileName,)
           while os.path.isfile(backupname):
               i = i + 1   # first numbered .bak will be '.bak2'
               backupname = '%s.bak%d' % (FileName,i)
           os.rename(FileName,backupname)
           f=open(FileName,'wb')
           f.write(newdat)
           f.close()
       except OSError:
           print 'Campaign_RemoveObjectMaxEP: Unable to create backup of file <%s>, possible Read-Only problem.  Quitting' % (FileName,)
           raise
       except:
           print 'Campaign_RemoveObjectMaxEP: Unable to create backup of file <%s>.  Quitting' % (FileName,)
           raise

# Read an objects DAT and a characters DAT file and make sure every
# character in the latter file that has an object template in the
# former file also has a 'complex' attribute set.  If it does not,
# assign it a unique value.
#
# If ObjectFileName and CharacterFileName are left as the defaults, we
# try and get the ones for the current mod.
#
# Makes a backup of the previous object file as FileName+'.bak' in the
# same directory that FileName (probably "objects.dat") was in.
#
#
# USAGE NOTES: This will typically be run AFTER the FFX Control Centre
# has run.  The FFX CC adds the 'complex' attribute to most
# characters, but misses some, like civilians and tanks.  This adds
# the attribute to those character types.
#
# The easiest way for most people to use it is from the game console.
# From the console, enter
#   import datfiles
#   datfiles.Campaign_SetMissingComplexAttribs()
# and that should do it.
#
# Since the FFX CC does not remove the 'complex' attribute from
# characters that have it, this should only need to be run when a new
# civilian or tank is added via FFEdit.

def Campaign_SetMissingComplexAttribs(ObjectFileName='',CharacterFileName='',verbose=1):
    # if not given DAT file names, try and guess them (for in-game use)
    if ObjectFileName=='':
        ObjectFileName = GetDatFile('objects.dat')
    ObjectFileName = os.path.abspath(ObjectFileName)
    if CharacterFileName=='':
        CharacterFileName = GetDatFile('characters.dat')
    if verbose:
        print 'Campaign_SetMissingComplexAttribs: looking at objects in %s.' % (ObjectFileName,)
    # make a list of the chars with missing complex attribs
    missings = []
    ComplexValList = []
    chars = Campaign_ReadCharacters(CharacterFileName, ForceRead=1)
    templates = Campaign_ReadObjects(ObjectFileName, ForceRead=1)
    for char in chars.keys():
        if templates.has_key(char):
            complexVal = templates[char].get('complex')
            if complexVal == None :
                ##if verbose: print 'Campaign_SetMissingComplexAttribs: Character <%s> has no complex value set.' % (char,)
                missings.append(char)
            else:
                ComplexValList.append(complexVal)
        else:
            if char == 'narrator':
                print 'Campaign_SetMissingComplexAttribs: Character <%s> has no corresponding object template.  This is okay.' % (char,)
            else:
                print 'Campaign_SetMissingComplexAttribs: Possible trouble: character <%s> has no corresponding object template. ' % (char,)
    if not missings:
        print 'Campaign_SetMissingComplexAttribs: no characters found with unset complex attribute in %s' % (ObjectFileName,)
        return
    # now read and hunt through the objects DAT, making changes as needed
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    f = open(ObjectFileName,'rb')
    alldat = f.read()
    f.close()
    datlen = len(alldat)
    newdat = alldat[0:0]
    pos = 0
    changed = 0
    if verbose:
        print '# Campaign_SetMissingComplexAttribs called at %s on file <%s>.  The following is a list of the changed objects' \
              % (time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime(time.time())), ObjectFileName)
        print '['
    NewComplexVal = -1000.0    # start here and go down to get unique complex values
    while pos < datlen:
        gotobj = 0
        if ord(alldat[pos]) > 8:                            # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            if words[0] == words[1]:                        # check that it's a legit name
                gotobj = 1
                objName = words[0]
                if ( objName in missings ):                 # this object needs its complex attrib set
                    offset = GetDatWordPos(dat,0,2) + 316   # first attrib
                    for count in range(15):                 # find first empty attrib slot
                        if dat[offset+count*36] == '\x00':
                            break
                    if count < 15:
                        while NewComplexVal in ComplexValList:
                            NewComplexVal = NewComplexVal - 1
                        if verbose: print '%s being assigned complex=%f' % (objName,NewComplexVal)
                        complexstr = 'complex' + ('\x00'*(32-len('complex')))
                        complexstr = complexstr+struct.pack('f',NewComplexVal)
                        dat = dat[:offset+count*36] + complexstr + dat[offset+(count+1)*36:]
                        changed = changed + 1
                        ComplexValList.append(NewComplexVal)
                    else:
                        if verbose: print '# ERROR: %s has no empty attrib slots.' % (objName,)
        if gotobj:
            newdat = newdat + dat
            pos = pos + ITEMSIZE
        else:
            newdat = newdat + alldat[pos:pos+SKIPSIZE]
            pos = pos + SKIPSIZE
    if verbose: print ']'
    if changed > 0:
        print ' ... changed %d objects' % (changed,)
            
        try:
            i = 1
            backupname = '%s.bak' % (ObjectFileName,)
            while os.path.isfile(backupname):
                i = i + 1   # first numbered .bak will be '.bak2'
                backupname = '%s.bak%d' % (ObjectFileName,i)
            os.rename(ObjectFileName,backupname)
            f=open(ObjectFileName,'wb')
            f.write(newdat)
            f.close()
        except OSError:
            print 'Campaign_SetMissingComplexAttribs: Unable to create backup of file <%s>, possible Read-Only problem.  No changes saved.' % (ObjectFileName,)
            raise
        except:
            print 'Campaign_SetMissingComplexAttribs: Unable to create backup of file <%s>.  No changes saved.' % (ObjectFileName,)
            raise
        # probably not necessary, but let's re-read the objects DAT to update the cache
        templates = Campaign_ReadObjects(ObjectFileName,ForceRead=1)

# Read an objects DAT remove from each object the named attribute.  If
# CharactersOnly is set (as it is by default) this only acts on
# objects with character classes.
def Campaign_RemoveObjectAttrib(attribute,ObjectFileName='',CharactersOnly=1,verbose=1):
    # if not given DAT file names, try and guess them (for in-game use)
    if ObjectFileName=='':
        ObjectFileName = GetDatFile('objects.dat')
    ObjectFileName = os.path.abspath(ObjectFileName)

    # now read and hunt through the objects DAT, making changes as needed
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    f = open(ObjectFileName,'rb')
    alldat = f.read()
    f.close()
    datlen = len(alldat)
    newdat = alldat[0:0]
    pos = 0
    changed = 0
    if verbose:
        print '# Campaign_RemoveObjectAttrib called at %s on file <%s>.  The following is a list of the changed objects' \
              % (time.strftime('%a, %d %b %Y %H:%M:%S',time.localtime(time.time())), ObjectFileName)
        print '['
    attribstr = '\x00'*36
    char_classes = ['GAME_OBJ_MINION','GAME_OBJ_HERO', 'GAME_OBJ_POLICE', 'GAME_OBJ_CIVILIAN', 'GAME_OBJ_VILLIAN', 'GAME_OBJ_TANK']
    while pos < datlen:
        gotobj = 0
        if ord(alldat[pos]) > 8:                            # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            if words[0] == words[1]:                        # check that it's a legit name
                gotobj = 1
                objName = words[0]
                objClass = words[3]
                if ( not CharactersOnly ) or ( objClass in char_classes ):
                    attribs = []
                    offset = GetDatWordPos(dat,0,2) + 316   # first attrib
                    for count in range(15):
                        if dat[offset+count*36] != '\x00':
                            attrib = GetDatWords(dat,offset+36*count,1)[0]
                            attribs.append(attrib)
                    if attribute in attribs:
                        for count in range(15):
                            if dat[offset+count*36] != '\x00':
                                attrib = GetDatWords(dat,offset+36*count,1)[0]
                                if attrib == attribute:
                                    if verbose: print '%s having attribute <%s> removed' % (objName,attribute)
                                    dat = dat[:offset+count*36] + attribstr + dat[offset+(count+1)*36:]
                                    changed = changed + 1
        if gotobj:
            newdat = newdat + dat
            pos = pos + ITEMSIZE
        else:
            newdat = newdat + alldat[pos:pos+SKIPSIZE]
            pos = pos + SKIPSIZE
    if verbose: print ']'
    if changed > 0:
        print ' ... changed %d objects' % (changed,)
            
        try:
            i = 1
            backupname = '%s.bak' % (ObjectFileName,)
            while os.path.isfile(backupname):
                i = i + 1   # first numbered .bak will be '.bak2'
                backupname = '%s.bak%d' % (ObjectFileName,i)
            os.rename(ObjectFileName,backupname)
            f=open(ObjectFileName,'wb')
            f.write(newdat)
            f.close()
        except OSError:
            print 'Campaign_RemoveObjectAttrib: Unable to create backup of file <%s>, possible Read-Only problem.  No changes saved.' % (ObjectFileName,)
            raise
        except:
            print 'Campaign_RemoveObjectAttrib: Unable to create backup of file <%s>.  No changes saved.' % (ObjectFileName,)
            raise
        # probably not necessary, but let's re-read the objects DAT to update the cache
        templates = Campaign_ReadObjects(ObjectFileName,ForceRead=1)

# Epimethee 2006-09-07
# ACTIVE DEFENCES AND DURATION In FFEdit2, active defences duration
# has no UI. Because of that, whenever a specific active defence is
# saved, its Duration is set to 0 (short).

# This returns defence powers names with their duration as a list of
# each [stringPowerName, intDuration] and prints out the equivalent to
# the console.
def ShowAllActiveDefenceDurations(FileName=''):
    powers = Campaign_ReadPowers(FileName)
    defences = []
    for k in powers.keys():
        if powers[k].has_key('Duration'):
            defences.append( [ k, powers[k]['Duration'] ] )
    defences.sort()
    for d in defences:
        print "Defence '%s' duration: %s" % ( d[0], PT_RANGE_NAMES[d[1]] )
    return defences

# Rewrite the selected active defence power to change the duration
# Valid durations are:
# 0 (short)
# 1 (medium)
# 2 (long)
# 3 (infinite)
# Note that this function does not reload powers.dat
def Campaign_ChangeActiveDefenceDuration(stringPower, intDuration, FileName='', ForceRead=0):
    powers = Campaign_ReadPowers(FileName, ForceRead, 0)
    if not powers.has_key(stringPower):
        print "Campaign_ChangeActiveDefenceDuration: '%s' not found in 'powers.dat'" % stringPower
        return
    if not powers[stringPower].has_key('Duration'):
        print "Campaign_ChangeActiveDefenceDuration: '%s' doesn't have a 'Duration' key" % stringPower
        return
    if not intDuration in (0, 1, 2, 3):
        print "Campaign_ChangeActiveDefenceDuration: '%s' can't be set to a duration of %i" % (stringPower, intDuration)
        return
    dictDefence = powers[stringPower]
    dictDefence['Duration'] = intDuration
    Campaign_RewritePowers(FileName, [dictDefence], ForceRead)

    # Take two character DAT files (a source and a target) and a list of items to change and this
# function will find those items in each DAT and then copy the items from source
# into target and save the new target.  BACK UP TARGET BEFORE DOING THIS!
def UpdateCharactersFromDAT(source, target, items):
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    headermark = 'HDR4'
    unfounditems = items
    # first, read the source data
    fsource = open(source,'rb')
    dat = fsource.read()
    fsource.close()
    sourcedata = {}
    datlen = len(dat)
    pos = FIRSTITEMPOS
    while pos < datlen:
        if ord(dat[pos]) > 8:   # we have a character name (potentially)
            words = GetDatWords(dat,pos,2)
            if words[1][:len(headermark)] == headermark:
                if words[0] in items:
                    print 'UpdateCharactersFromDAT: found %s in %s' % (words[0],source)
                    sourcedata[words[0]] = dat[pos:pos+ITEMSIZE]
                    unfounditems.remove(words[0])
        pos = pos + ITEMSIZE
    if unfounditems:  print 'UpdateCharactersFromDAT: %s NOT found in %s' % (unfounditems,source)
    # now read the target data, making replacements as we go
    ftarget = open(target,'rb')
    dat = ftarget.read()
    ftarget.close()
    datlen = len(dat)
    pos = FIRSTITEMPOS
    while pos < datlen:
        if ord(dat[pos]) > 8:   # we have a character name (potentially)
            words = GetDatWords(dat,pos,2)
            if words[1][:len(headermark)] == headermark:
                if words[0] in sourcedata.keys():
                    print 'UpdateCharactersFromDAT: replacing %s in %s' % (words[0],target)
                    dat = dat[:pos]+sourcedata[words[0]]+dat[pos+len(sourcedata[words[0]]):]
        pos = pos + ITEMSIZE
    # now write the new data to target
    ftarget = open(target,'wb')
    ftarget.write(dat)
    ftarget.close()

# Take two object DAT files (a source and a target) and a list of items to change and this
# function will find those items in each DAT and then copy the items from source
# into target and save the new target.  BACK UP TARGET BEFORE DOING THIS!
def UpdateObjectsFromDAT(source, target, items):
    SKIPSIZE = 1024
    ITEMSIZE = 3072
    headermark = 'HDR4'
    unfounditems = items
    # first, read the source data
    fsource = open(source,'rb')
    alldat = fsource.read()
    fsource.close()
    sourcedata = {}
    datlen = len(alldat)
    pos = 0
    while pos < datlen:
        ##print 'pos = %d' % pos
        gotobj = 0
        if ord(alldat[pos]) > 8:                        # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            ##print 'UpdateObjectsFromDAT: words=%s' % (words,)
            if words[0] == words[1]:                    # check that it's a legit name
                gotobj = 1
                if words[0] in items:
                    print 'UpdateObjectsFromDAT: found %s in %s' % (words[0],source)
                    sourcedata[words[0]] = dat
                    unfounditems.remove(words[0])
        if gotobj:
            pos = pos + ITEMSIZE
        else:
            pos = pos + SKIPSIZE
    if unfounditems:  print 'UpdateObjectsFromDAT: %s NOT found in %s' % (unfounditems,source)
    # now read the target data, making replacements as we go
    ftarget = open(target,'rb')
    alldat = ftarget.read()
    ftarget.close()
    datlen = len(alldat)
    pos = 0
    while pos < datlen:
        gotobj = 0
        if ord(alldat[pos]) > 8:                        # we have an object name (potentially)
            dat = alldat[pos:pos+3072]
            words = GetDatWords(dat,0,4)
            if words[0] == words[1]:                    # check that it's a legit name
                gotobj = 1
                if words[0] in sourcedata.keys():
                    print 'UpdateObjectsFromDAT: replacing %s in %s' % (words[0],target)
                    alldat = alldat[:pos]+sourcedata[words[0]]+alldat[pos+len(sourcedata[words[0]]):]
        if len(alldat) != datlen:
            print 'pos = %d:  length just changed!' % pos
        if gotobj:
            pos = pos + ITEMSIZE
        else:
            pos = pos + SKIPSIZE
    # now write the new data to target
    ftarget = open(target,'wb')
    ftarget.write(alldat)
    ftarget.close()


# unfinished

def ChangeMarkerType(MissionFileName='',MarkerName='',NewType=MT_POSITIONAL,
                     TemplateName='',NIFpath='',GameObjectType='GAME_OBJECT_GENERIC'):
    global _StartingMarkers
    if ( not ForceRead ) and ( '_StartingMarkers' in globals().keys() ):
        return _StartingMarkers
    FIRSTITEMPOS = 1024
    ITEMSIZE = 1024
    _StartingMarkers = {}
    if MissionFileName=='': # try to guess the current temp directory
        MissionFileName = GuessDatPath()
    f = open(MissionFileName,'rb')
    dat = f.read()
    f.close()
    datlen = len(dat)
    pos = FIRSTITEMPOS
    while pos < datlen:
        if ord(dat[pos]) > 8:
            words = GetDatWords(dat,pos,3)
            if MarkerName == words[0]:
                dat[pos+len(itemName)+1] = NewType
                if NewType == MT_GENERIC:
                    dat = dat[:pos+2*len(MarkerName)+34] + \
                          TemplateName + \
                          dat[pos+2*len(MarkerName)+len(TemplateName)+35:]
                    dat = dat[:pos+2*len(MarkerName)+34] + \
                          TemplateName + \
                          dat[pos+2*len(MarkerName)+len(TemplateName)+35:]
            if verbose: print '%6d\t%30s\t%d\t%s' % (pos,itemName,itemType,words)
        pos = pos + ITEMSIZE
    return _StartingMarkers

#################################
### Testing Functions ###########
#################################

# trying out the gdbm module
try:
    if sys.version[:5] == '1.5.2':
        import gdbm
except:
    pass
GDBM_MAGIC_NUM = '\xCE\x9A\x57\x13'
DAT_MAGIC_NUM = '\xAD\xDB\xDE\xFA'

# read a DAT file as a gdbm file and return its gdbm "dictionary"
def ReadDAT(FileName):
    gdbmFileName = FileName+'.gdbm'
    f = open(FileName,'rb')
    dat = f.read()
    f.close()
    f = open(gdbmFileName,'wb')
    gdbmdat = GDBM_MAGIC_NUM+dat[4:]
    f.write(gdbmdat)
    f.close()
    db = gdbm.open(gdbmFileName,'r')
    dbdict = {}
    key = db.firstkey()
    while key:
        dbdict[key[:-1]] = db[key]
        key =  db.nextkey(key)
    db.close()
    os.remove(gdbmFileName)
    return dbdict

# change the magic number so that a DAT file can be read in as a GDBM file
def DAT2GDBM(FileName):
    '''convert an FF DAT file into a GDBM file that can be read by standard GDBM libraries'''
    gdbmFileName = FileName+'.gdbm'
    f = open(FileName,'rb')
    dat = f.read()
    f.close()
    f = open(gdbmFileName,'wb')
    gdbmdat = GDBM_MAGIC_NUM+dat[4:]
    f.write(gdbmdat)
    f.close()

# change the magic number so that a GDBM file can be read in as a DAT file
# NB the FileName is of the original DAT file.  E.g. 'power.dat'
def GDBM2DAT(FileName):
    '''convert a GDBM file into a DAT file that can be read by the FF game'''
    gdbmFileName = FileName+'.gdbm'
    f = open(gdbmFileName,'rb')
    dat = f.read()
    f.close()
    f = open(FileName,'wb')
    datdat = DAT_MAGIC_NUM+dat[4:]
    f.write(datdat)
    f.close()

# write out the records as individual files. Takes a path argument to
# allow putting these files in their own folder. matchstring specifies
# a string that must be found in the key for a given item to be
# written.  It is case-sensitive.
OC_CHARACTER = 256000
def WriteDATRecords(db,path,matchstring=None):
    if not os.path.isdir(path):
        os.mkdir(path)
    node = db.firstkey()
    while node:
        if ( matchstring!=None ) and (string.find(node,matchstring) < 0 ):
            node = db.nextkey(node)
            continue
        if node[-1]=='\x00':
            nodename = node[:-1]
        else:
            nodename = node
        filename = os.path.join(path,nodename)
        f = open(filename,'wb')
        f.write('%s' % db[node])
        f.close()
        node = db.nextkey(node)

# credit Alex Martelli, "Python Cookbook" p. 440
def whoami():
    "returns name of calling function"
    return sys._getframe(1).f_code.co_name

def WritePowerRecordPowers(PowerFileName='',MatchPowerName='',SaveFileName=''):
    f = open(PowerFileName,'rb')
    dat = f.read()
    f.close()
    FIRSTITEMPOS = 2048
    positions = GetPowerRecordPositions(dat,FIRSTITEMPOS)
    for pos in positions:
        npr = dat[pos:pos+184]
        pname = string.split(npr,'\x00',1)[0]
        if pname == MatchPowerName:
            print 'WritePowerRecordPowers: writing record for <%s> to <%s>' % (pname,SaveFileName)
            f = open(SaveFileName,'wb')
            f.write(npr)
            f.close()

def ComparePowerRecordPowers(PowerFileName='',MatchPowerName='',SaveSinglePowerFileName=''):
    f = open(PowerFileName,'rb')
    dat = f.read()
    f.close()
    f = open(SaveSinglePowerFileName,'rb')
    comppower = f.read()
    f.close()
    FIRSTITEMPOS = 2048
    positions = GetPowerRecordPositions(dat,FIRSTITEMPOS)
    for pos in positions:
        npr = dat[pos:pos+184]
        pname = string.split(npr,'\x00',1)[0]
        if pname == MatchPowerName:
            print 'ComparePowerRecordPowers: comparing <%s> in %s vs. %s' % (MatchPowerName,PowerFileName,SaveSinglePowerFileName)
            for i in range(len(npr)):
                if npr[i] != comppower[i]:
                    print '@%4d (%4x):\tin DAT = <%s> %3d (%2x)\tbase = <%s> %3d (%2x)' % (i,i,npr[i],ord(npr[i]),ord(npr[i]),comppower[i],ord(comppower[i]),ord(comppower[i]))
##                    print '%4d (offset=%d):\tin DAT = <%s> %3d (%2x)\tbase = <%s> %3d (%2x)' % (i,i-offset,npr[i],ord(npr[i]),ord(npr[i]),comppower[i],ord(comppower[i]),ord(comppower[i]))
            break


# for command-line testing
if __name__ == '__main__':
    verbose=0
    import profile
##    chars = Campaign_ReadCharacters(FileName='C:\\Program Files\\games\\Freedom Force\\strangers\\characters.dat',ForceRead=1,verbose=verbose)
##    print 'len(chars) = %d' % len(chars)
##    charlist = chars.keys()
##    charlist.sort()
##    for char in charlist:  print 'char = %s' % char
##    markers=Mission_ReadStartingMarkers('C:\\Program Files\\games\\Freedom Force\\Data\\Missions\\01_begin\\mission.dat',1,verbose=verbose)
##    print 'len(markers) = %d' % len(markers)
##    print 'len(Campaign_ReadObjects()) = %d' % \
##          len(Campaign_ReadObjects('C:\\Program Files\\games\\Freedom Force\\strangers\\objects.dat',ForceRead=1,verbose=verbose))
##    profile.run('print len(Campaign_ReadObjects(\'C:\\Program Files\\games\\Freedom Force\\strangers\\objects.dat\'))','Campaign_ReadObjects.profile')
##    print len(Campaign_ReadObjects('C:\\Program Files\\games\\Freedom Force\\strangers\\objects.dat'))
##    powers = Campaign_ReadPowers('C:\\Program Files\\games\\Freedom Force\\strangers\\powers.dat',ForceRead=1,verbose=verbose)
##    print 'len(powers) = %d' % len(powers)
##    for power in powers.keys():
##        print '\n<%s>' % power
##        for attrib in powers[power].keys():
##            print '%20s : %s' % (attrib,str(powers[power][attrib]))
##    WritePowerRecordPowers('C:\\Program Files\\games\\Freedom Force\\brokow_mods\\powers.dat','querldox 12th level intellect','C:\\Program Files\\games\\Freedom Force\\brokow_mods\\power-fftrap.dat')
##    ComparePowerRecordPowers('C:\\Program Files\\games\\Freedom Force\\brokow_mods\\powers.dat','querldox 12th level intellect','C:\\Program Files\\games\\Freedom Force\\brokow_mods\\power-fftrap.dat')
##    FX = Campaign_ReadFX('C:\\Program Files\\games\\Freedom Force\\Data\\fx.dat',ForceRead=1,verbose=verbose)
##    for fx in FX.keys():
##        print '\n<%s>' % fx
##        for attrib in FX[fx].keys():
##            print '%20s : %s' % (attrib,str(FX[fx][attrib]))
##    attribs = Campaign_ReadAttributes('C:\\Program Files\\games\\Freedom Force\\Data\\attributes.dat',ForceRead=1,verbose=verbose)
##    for attrib in attribs.keys():
##        print '%s: %5d' % (attrib,attribs[attrib])
##
##    print 'len(Campaign_ReadCharactersFromSavedGame()) = %d' %\
##          len(Campaign_ReadCharactersFromSavedGame(FileName='C:\\Program Files\\games\\Freedom Force\\Strangers\\SaveGames\\A Concert In The Park - Sp,Fly,DD,AZ - nrm late.dat',ForceRead=0,verbose=verbose))
##    saveddata = Campaign_ReadCharactersFromSavedGame(FileName='C:\\Program Files\\games\\Freedom Force\\Strangers\\SaveGames\\quicksave\\game.dat',ForceRead=0,verbose=verbose)
##    profile.run('Campaign_ReadCharactersFromSavedGame(FileName=\'C:\\Program Files\\games\\Freedom Force\\Strangers\\SaveGames\\quicksave\\game.dat\')','Campaign_ReadCharactersFromSavedGame.profile')
##    saveddata = Campaign_ReadCharactersFromSavedGame(FileName='c:\Program Files\games\Freedom Force\Strangers\SaveGames\gardens end - AZ,JJ,MP,KZ - nrm buyup2.dat',ForceRead=0,verbose=verbose)
##    print 'len(Campaign_ReadCharactersFromSavedGame()) = %d' % len(saveddata)
##    saveddata = Campaign_ReadCharactersFromSavedGame(FileName='c:\Program Files\games\Freedom Force\Strangers\SaveGames\gardens end - AZ,JJ,MP,KZ - nrm buyup2.dat',ForceRead=0,verbose=verbose)
##    print 'len(Campaign_ReadCharactersFromSavedGame()) = %d' % len(saveddata)
##    somechars = saveddata.keys()[0:5]
##    somechars.append('custom_template_179')
##    print somechars
##    for char in somechars:
##        print '%s powers:' % char
##        for p in saveddata[char]['powers'].keys():
##            print '\t%s at level %d' % (p,saveddata[char]['powerLevels'][p])
##    tstart = time.clock()
##    Campaign_ReadCharactersFromSavedGame(FileName='C:\\Program Files\\games\\Freedom Force\\Strangers\\SaveGames\\A Concert In The Park - Sp,Fly,DD,AZ - nrm late.dat',ForceRead=0,verbose=verbose)
##    chars=Campaign_ReadCharactersFromSavedGame(FileName='C:\\Program Files\\games\\Freedom Force\\Strangers\\SaveGames\\partners in wonder 1 - lots - nrm done.dat',ForceRead=0,verbose=verbose)
##    f = open('chardata','w')
##    f.write('%s' % repr(chars))
##    f.close()
##    Campaign_ReadHeroFile('C:\\Program Files\\games\\Freedom Force\\Data\\Heroes\\shell shock2.hero',0,1)
##    print 'Campaign_ReadCharactersFromSavedGame: time = %3.4f' % (time.clock()-tstart)
##campdef = Campaign_ReadCampaignDefinitions(FileName='C:\\Program Files\\games\\Freedom Force\\Strangers\\campdef.dat',ForceRead=0,verbose=0)
##print 'len(campdef)=%d,  campdef = %s' % (len(campdef),campdef)
##for mission in campdef['missions']:
##    print 'mission = <%s> isSubmission=%d' % (mission,campdef[mission]['isSubmission'])
##    print '\tnumAllowed = %d, required = %s' % (campdef[mission]['numAllowed'],campdef[mission]['required'])
##    print '\tdisallowed = %s' % (campdef[mission]['disallowed'],)
##Campaign_MakeHeroIndex(ModPath = 'C:\\Program Files\\games\\Freedom Force\\Strangers\\')
##    Campaign_ChangeObjectMinforce('C:\\Program Files\\games\\Freedom Force\\Data\\objects.dat',verbose=1)
##    powers = Campaign_ReadPowers('C:\\Program Files\\games\\Freedom Force vs The 3rd Reich\\Data\\powers.dat')
##    problem_powers = []
##    pkeys = powers.keys()
##    pkeys.sort()
####    ShowPower(powers[pkeys[0]])
##    for p in pkeys:
##        try:
##            ShowPower(powers[p])
##        except:
##            problem_powers.append(p)
##    print problem_powers
##    chars=Campaign_ReadCharacterData('c:\\Program Files\\games\\Freedom Force vs The 3rd Reich\\data',1)
##    problems = []
##    for c in chars.keys():
##        prob = ShowHero(chars[c])
##        if prob:
##            problems.append(prob)
##    print 'problems = %s' % problems
##    # testing the psyco module
##    for i in range(10):
##        chars=Campaign_ReadCharacterData('c:\\Program Files\\games\\Freedom Force vs The 3rd Reich\\data',1)
##    for i in range(10):
##        chars=Campaign_ReadCharacterData('c:\\Program Files\\games\\Freedom Force vs The 3rd Reich\\data',0)

    import getopt
    optlist, args = getopt.getopt(sys.argv[1:], '',['MakeHeroIndices='])
    if optlist[0][0] == '--MakeHeroIndices':
        MakeHeroIndices(gameVersion=int(optlist[0][1]))
