
print "Loading m25aiopt.py v.3.2.0 final; branch = swingman 4"

#May 14
#m25: changed Opt_AddAndCalcSubtype and Opt_CalcAutoSubtype to respect any entries in the AI definition
#m25: changed Opt_ConvertPct to not override the AI definition

import ff
import js

import string
import time
import types

import chardata
import datfiles

import ffxcustom

import m25custompowers
import m25distance
import m25energy
import m25report
import m25string
import m25template

PRECALC_PCT_ALWAYS = 100
PRECALC_PCT_HIGH = 80
PRECALC_PCT_MEDIUM = 60
PRECALC_PCT_LOW = 40
PRECALC_PCT_VERYLOW = 20

AIOPTPRINTCODE = 0

opt_powerprecalc = {
	'TActiveDefense': {
		'exec': 'AIGenericActiveDefense',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_HIGH)',
		'situation': 'Opt_AddSituation(tactic,"not displaced")',
		'subtype': 'Opt_AddDefenseSubtype(power,subtype)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'defense_time': 'Opt_ConvertDefenseTime(power)',
		'category': 'defense',
	},
	
	'TAfterburner': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_SHORT,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",0.5)',
		'category': 'move',
	},
	
	'TAirSuperiority': {
		'exec': 'AIGenericMoveTypeOpt',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_ALWAYS)',
		'energy': 0,
		'mode': 'hover',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,100)',
		'situation': 'Opt_AddSituation(tactic,"not flying")',
		'category': 'flee',
	},

	'TArea': {
		'exec': 'AIGenericPowerOnSelf',
		'situation': 'Opt_AddNotInRadius(tactic)',
		'subtype': 'Opt_CalcAutoSubtype(char,tactic)',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_LOW)',
		'mindist': 0,
		'maxdist': 'Opt_GetAreaMaxDist(tactic)',
		'energy': 'Opt_GetEnergy(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TChargeUp': {
		'exec': 'AIChargeUp',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 0,
		'maxdist': 'Opt_GetMaxDist(tactic)',
	},
	
	'TCloak': {
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not invisible")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'category': 'defense',
	},
	
	'TCloneSelf': {
		'exec': 'TCloneSelfOpt',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_ALWAYS)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'category': 'defense',
	},
	
	'TCloneTarget': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
	},
	
	'TDecoy': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,["not blind","not neardecoy"])',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'end_function': 'AIDecoyDone',
		'category': 'defense',
	},
	
	'TDirect': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'subtype': 'Opt_CalcAutoSubtype(char,tactic)',		
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 0,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TDirectExplosive': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'subtype': 'Opt_CalcAutoSubtype(char,tactic)',		
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 0,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TDisplaceImage': {
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'situation': 'Opt_AddSituation(tactic,"not displaced")',
		'category': 'defense',
	},
	
	'TDisruptObject': {
		'exec': 'AIDisruptObject',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not nearenemy 80")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_SHORT,
		'maxdist': 'Opt_GetMaxDist(tactic)',
	},
	
	'TElectrify': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,["not nearenemy 100","not flying"])',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': m25distance.DISTANCE_LONG,
		#'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TEmpathy': {
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'situation': 'Opt_AddSituation(tactic,"not empathy")',
		'category': 'defense',
	},
	
	'TFleeMelee': {
		'exec': 'AIGenericFleeOpt',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 0,
		'mindist': 'Opt_ConvertMinDist(tactic,50)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,200)',
		'category': 'flee',
	},
	
	'TLand': {
		'exec': 'AIGenericMoveTypeOpt',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_ALWAYS)',
		'energy': 0,
		'mode': 'land',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,100)',
		'situation': 'Opt_AddSituation(tactic,"flying")',
		'category': 'move',
	},
	
	'TLivingFires': {
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
	},
	
	'TMelee': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"not flying")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_GetMeleeMinDist(tactic)',
		'maxdist': 'Opt_GetMeleeMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TMeleeArea': {
		'exec': 'AIGenericPowerOnTarget',
		'situation': 'Opt_AddSituation(tactic,"not nearally 100")',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"not flying")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_GetMeleeMinDist(tactic)',
		'maxdist': 'Opt_GetMeleeMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TMeleeAir': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"flying")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_GetMeleeMinDist(tactic)',
		'maxdist': 'Opt_GetMeleeMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	
	'TMoveTo': {
		'exec': 'AIGenericMoveOpt',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_ALWAYS)',
		'energy': 0,
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,400)',
		'category': 'move',
	},
	
	'TPolymorph': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TPossession': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not alone")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TRally': {
		'exec': 'TRally',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TRanged': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"none")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TRangedCone': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"not nearenemy 100")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TRangedExplosive': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_ConvertPct(tactic,PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddAndCalcSubtype(char,tactic,"not nearenemy 120")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_SHORT,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'carrier': 'Opt_GetCarrierIndex(char, tactic)',
	},
	
	'TSelfHeal': {
		'exec': 'AIGenericPowerOnSelf',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_VISIBLE)',
		'situation': 'Opt_AddSituation(tactic,"lowerhealthpct 40")',
		'category': 'defense',
	},
	
	'TSonicBoom': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'situation': 'Opt_AddSituation(tactic,"not nearally 120")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",4)',
	},
	
	'TSpeedingBullet': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,["not nearenemy 80","not flying"])',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",0.75)',
	},
	
	'TSprint': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",0.75)',
		'category': 'move',
	},
	
	'TSprintAway': {
		'exec': 'AIGenericTravelPower',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",0.75)',
		'category': 'flee',
	},
	
	'TTeleport': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",5)',
		'category': 'move',
	},
	
	'TTeleportAway': {
		'exec': 'AIGenericTravelPower',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_HIGH)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",5)',
		'category': 'flee',
	},
	
	'TThrow':{
		'mindist': 'Opt_ConvertMinDist(tactic,0)',
		'maxdist': 'Opt_ConvertMaxDist(tactic,m25distance.DISTANCE_LONG)',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not nearenemy 80")',
	},
	
	'TTumble': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",5)',
		'category': 'move',
	},
	
	'TTumbleAway': {
		'exec': 'AIGenericTravelPower',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",2)',
		'category': 'flee',
	},
	
	'TTunnelTravel': {
		'exec': 'AIGenericPowerOnTarget',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'subtype': 'Opt_AddSubtype(tactic,"not flying")',
		#'situation': 'Opt_AddSituation(tactic,"not blind")',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': m25distance.DISTANCE_MELEE,
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",5)',
		'category': 'move',
	},
	
	'TTunnelTravelAway': {
		'exec': 'AIGenericTravelPower',
		'pct': 'Opt_Convert(tactic,"pct",PRECALC_PCT_MEDIUM)',
		'energy': 'Opt_GetEnergy(tactic)',
		'mindist': 'Opt_ConvertMinDist(tactic,m25distance.DISTANCE_MELEE)',
		'maxdist': 'Opt_GetMaxDist(tactic)',
		'time': 'Opt_Convert(tactic,"time",2)',
		'category': 'flee',
	},
	
	'TUseRemedy': {
		'category': 'remedy',
	},
}


opt_filename = ''
def Opt_GetPowerData():
	global opt_filename
	if opt_filename == '':
		opt_filename = datfiles.GetDatFile('powers.dat')
	return datfiles._PowerData[opt_filename]


def Opt_CalcParameters(char,tacticstring):
	tactic = Opt_GetTacticInfo(tacticstring)
	tactic_name = tactic['name']
	power, subtype = None, None
	if tactic.has_key('power'): power = tactic['power']
	if tactic.has_key('subtype'): subtype = tactic['subtype']
	
	precalc = {}
	
	if not tactic_name in opt_powerprecalc.keys(): 
		try:
			#see if there is an optimization function out there somewhere
			precalc = eval('ff.AIOpt_%s(%s,%s)'%(tactic_name,`char`,`tactic`))
		except:
			return tactic  #return it as-is
		#go through the precalc dictionary and calculate the new parameters
	else:
		precalc = opt_powerprecalc[tactic_name]
		
	tactic_newname = tactic_name
	if precalc.has_key('exec'):
		tactic_newname = precalc['exec']
		tactic['name'] = precalc['exec']
		
# 		if tactic.has_key('type') and tactic['type'] == 'friend':
# 			tactic_newname = tactic_newname + 'Friend'
	for param in precalc.keys():		
		if param == 'exec': continue
		fn = precalc[param]
		value = ''
		if (param != 'power') and (type(fn) == types.StringType) and ('(' in fn):
			value = eval(fn)  #exec 'value = ' + fn  #evaluate function
		else:
			value = fn  #no function, just take straight value
		tactic[param] = value
			
	#generate code for certain parameters
	for param in tactic.keys():
		value = None
		if param in ['subtype','situation']: 
			if tactic[param] in ['none',None,'',[]]: value = None
			else: 
				value = Opt_GenerateSubtypeCode(tactic[param])
			tactic[param] = value
		
	return tactic			
			
# 	#generate the new string		
# 	newtactic = tactic_newname + '('
# 	comma = ''
# 	for t in tactic.keys():
# 		if t != 'name':
# 			newtactic = newtactic + comma + t + '=' + `tactic[t]`
# 			if comma == '': comma = ', '
# 	newtactic = newtactic + ')'
# 	return newtactic
	
def Opt_GetTacticInfo(tactic):
	tacticinfo = {}
	num = 0
	param = None
	for t in tactic:
		num = num + 1
		#first bit is the name
		if t == '(': 
			tacticinfo['name'] = string.strip(tactic[:num-1])
			#tacticinfo.append(['name',string.strip(tactic[:num-1])])
			break
	#get the power if present
	start = num
	for t in tactic[num:]:
		num = num + 1
		if t in [',',')']:
			value = string.strip(tactic[start:num-1])
			try:
				exec 'value = Opt_GetStringValue('+value+')'
			except:
				pass
			if value: tacticinfo['power'] = value
			#tacticinfo.append(['power',string.strip(tactic[start:num-1])])
			break
		else:
			if t == '=':
				param = string.strip(tactic[start:num-1])
				break  #no power present
	#get other parameters
	start = num
	depth = 0
	for t in tactic[num:]:
		num = num + 1
		if (t == ',' and depth == 0) or (t == ')' and param != ''):
			value = string.strip(tactic[start:num-1])
			try:
				exec 'value = Opt_GetStringValue('+value+')'
			except:
				pass
			tacticinfo[param] = value
			#tacticinfo.append([param,value])
			param = ''
			start = num
		elif t == '=':
			param = string.strip(tactic[start:num-1])
			start = num
		elif t == '[': depth = depth + 1
		elif t == ']': depth = depth -1
		else:
			pass
	return tacticinfo
	

	
	
	
	
def Opt_ReplaceOrAddTactic(tactic,param,value):
	index = Opt_FindTacticParameterPosition(tactic,param)
	if index == None:
		if string.strip(tactic)[-1] == '(':
			return tactic + param + '=' + `value`
		else:
			return tactic + ', ' + param + '=' + `value`
	else:
		lefttactic = tactic[:index]
		righttactic = string.strip(m25string.GetRightString(tactic[index+len(param):],','))
		if righttactic != '':
			return lefttactic + righttactic + ', ' + param + '=' +`value`
		else:
			return lefttactic + param + '=' + `value`
	
def Opt_FindTacticName(tactic):
	return m25string.GetLeftString(string.strip(tactic),'(')
	

	
def Opt_FindTacticParameter(tactic, param, default=None):
	num = 0
	for t in tactic:
		end = num + len(param)
		if end > len(tactic) -1: return default
		if tactic[num:end] == param:
			#found it
			containingstring = m25string.GetRightString(tactic[num:],'=')
			value = string.strip(Opt_GetParam(containingstring))
			try:
				exec 'name = Opt_GetStringValue('+value+')'
				return name
			except:
				return value
		else:
			num = num + 1
	return default
	
	
def Opt_GetParam(tactic):
	depth = 0
	num = 0
	for t in tactic:
		if t == '[': depth = depth + 1
		if t == ']': depth = depth - 1
		if t == ',' and depth == 0:
			return tactic[:num]
		num = num + 1	
	return tactic
	
def Opt_FindTacticParameterPosition(tactic, param, default=None):
	num = 0
	for t in tactic:
		end = num + len(param)
		if end > len(tactic) -1: return default
		if tactic[num:end] == param:
			return num
		else:
			num = num + 1
	return default	
	
def Opt_FindTacticPower(tactic):
	containingstring = m25string.GetRightString(tactic,'(')
	possible = string.strip(m25string.GetLeftString(containingstring,','))
	name = ''
	try:
		exec 'name = Opt_GetStringValue('+possible+')'
	except:
		name = possible
	return name	
	
	
#-----------------------------------------------------------------------------
# Individual precalc functions	
	

def Opt_AddDefenseSubtype(power, subtype):
	power_data = Opt_GetPowerData()[power]
	modes = {datfiles.PT_MELEE_BLOCKED:'melee', datfiles.PT_RANGED_BLOCKED:'ranged', 
		datfiles.PT_AREA_BLOCKED:'area', datfiles.PT_DIRECT_BLOCKED:'direct', 
		datfiles.PT_SPECIAL_BLOCKED:'special'}
	subtype = ''
	for m in modes.keys():
		if power_data['AttackModesBlocked'] & m:
			subtype = Opt_AddOred(subtype, Opt_ConvertDefenceMode(power_data, modes[m]))
	if subtype == '': return None
	else: return subtype

def Opt_AddNotInRadius(tactic):
	radius = 100
	if tactic.has_key('radius'):
		radius = m25distance.DistanceConvert(tactic['radius'])
	else:
		if tactic.has_key('power'):
			radius = m25distance.DistanceConvert('area_%s'%datfiles.PT_RADIUS_NAMES[Opt_GetPowerData()[tactic['power']]['Radius']])
	return Opt_AddSituation(tactic,'not nearally %i'%radius)	
	
def Opt_AddSituation(tactic,addition):
	#situation = Opt_FindTacticParameter(tactic,'situation')
	situation = ''
	if tactic.has_key('situation'): situation = tactic['situation']
	newlist = Opt_AddToList(situation,addition)
	return newlist
	
def Opt_AddSubtype(tactic,addition):
	#subtype = Opt_FindTacticParameter(tactic,'subtype')
	subtype = ''
	if tactic.has_key('subtype'): subtype = tactic['subtype']
	newlist = Opt_AddToList(subtype,addition)
	return newlist	
	
def Opt_AddAndCalcSubtype(char,tactic,addition):
	#subtype = Opt_FindTacticParameter(tactic,'subtype')
	#subtype = ''
	#if tactic.has_key('subtype'): subtype = tactic['subtype']
	#newlist = Opt_AddToList(subtype,addition)
	newlist = Opt_AddToList([],addition)
	calclist = Opt_CalcAutoSubtype(char,tactic)
	if calclist not in [None, '', []]: newlist = Opt_AddToList(newlist,calclist)
	return newlist
	
def Opt_CalcAutoSubtype(char,tactic):
	#powername = Opt_FindTacticPower(tactic)
	subtype = ''
	if tactic.has_key('subtype'): subtype = tactic['subtype']
	powername = ''
	if tactic.has_key('power'):	powername = tactic['power']
	calclist = Opt_GetSpecialSubtype(char, powername)
	if subtype != '':	newlist = Opt_AddToList(calclist,subtype)
	else:	newlist = Opt_GetSpecialSubtype(char, powername)
	return newlist
	
def Opt_ConvertDefenceMode(power, mode):
	damtypes = {datfiles.PT_DAMAGE_BLOCKED_CRUSH:'crushing', datfiles.PT_DAMAGE_BLOCKED_PIERCE:'piercing', datfiles.PT_DAMAGE_BLOCKED_ENERGY:'energy',
		datfiles.PT_DAMAGE_BLOCKED_FIRE:'fire', datfiles.PT_DAMAGE_BLOCKED_RADIATION:'radiation', datfiles.PT_DAMAGE_BLOCKED_ELECTRICAL:'electrical',
		datfiles.PT_DAMAGE_BLOCKED_ACID:'acid', datfiles.PT_DAMAGE_BLOCKED_COLD:'cold', datfiles.PT_DAMAGE_BLOCKED_WATER:'water',
		datfiles.PT_DAMAGE_BLOCKED_MENTAL:'mental', datfiles.PT_DAMAGE_BLOCKED_MYSTICAL:'mystical'}
	subtype = ''
	for dt in damtypes.keys():
		if power['DamageTypesBlocked'] & dt:
			subtype = Opt_AddOred(subtype, mode + '_' + damtypes[dt])
	return subtype	
	
def Opt_ConvertDefenseTime(power):
	dt = Opt_GetPowerData()[power]['Duration']
	vals = [10,20,30]
	return vals[dt]	
	
def Opt_Convert(tactic, param, default=0):
	#value = Opt_FindTacticParameter(tactic, param)
	if tactic.has_key(param): return tactic[param]
	else: return default
	
def Opt_ConvertEnergy(tactic,default=0):
	if tactic.has_key('energy'): 
		return m25energy.ConvertEnergy(tactic['energy'])
	else: return default	
		
def Opt_ConvertMaxDist(tactic,default=0):
	if tactic.has_key('maxdist'):
		return m25distance.DistanceConvert(tactic['maxdist'])
	else: return default
	
def Opt_ConvertMinDist(tactic,default=0):
	if tactic.has_key('mindist'):
		return m25distance.DistanceConvert(tactic['mindist'])
	else: return default

def Opt_ConvertPct(tactic,default=0):
	energy = Opt_GetEnergy(tactic)
	if tactic.has_key('pct'):
		#should respect the AI writer's wishes
		#if tactic['pct'] < PRECALC_PCT_HIGH and energy == 0: return PRECALC_PCT_HIGH
		#else: return tactic['pct']
		return tactic['pct']
	else:
		if energy == 0: return PRECALC_PCT_HIGH
		else: return default
			
	
def Opt_GetCarrierIndex(char, tactic):
	if tactic.has_key('power'):
		return m25custompowers.GetCarrierIndex(char,tactic['power'])
	else: return -1
		
def Opt_GetEnergy(tactic):
	if tactic.has_key('energy'):
		return m25energy.ConvertEnergy(tactic['energy'])
	else:
		if tactic.has_key('power'):
			try:
				return m25energy.ConvertEnergy(datfiles.PT_EP_COST_NAMES[Opt_GetPowerData()[tactic['power']]['EPCost']])
			except:
				return 0
		else:
			return 0

	
def Opt_GetMaxDist(tactic):
	if tactic.has_key('maxdist'):
		return m25distance.DistanceConvert(tactic['maxdist'])
	else:
		if tactic.has_key('power'):
			return m25distance.DistanceConvert(datfiles.PT_RANGE_NAMES[Opt_GetPowerData()[tactic['power']]['RangeMax']])
		else: return 0

		
def Opt_GetAreaMaxDist(tactic):
	if tactic.has_key('maxdist'):
		return m25distance.DistanceConvert(tactic['maxdist'])
	else:
		if tactic.has_key('power'):
			return m25distance.DistanceConvert('area_%s'%datfiles.PT_RADIUS_NAMES[Opt_GetPowerData()[tactic['power']]['Radius']])
		else: return 0
		
def Opt_GetMeleeMaxDist(tactic):
	if tactic.has_key('maxdist'):
		return m25distance.DistanceConvert(tactic['maxdist'])
	else:
		if tactic.has_key('power'):
			return m25distance.DistanceConvert(Opt_GetPowerData()[tactic['power']]['RangeMax'] * 10)
		else: return 0

			

def Opt_GetMinDist(tactic):
	if tactic.has_key('mindist'):
		return m25distance.DistanceConvert(tactic['mindist'])
	else:
		if tactic.has_key('power'):
			return m25distance.DistanceConvert(datfiles.PT_RANGE_NAMES[Opt_GetPowerData()[tactic['power']]['RangeMin']])
		else: return 0

		
				
def Opt_GetMeleeMinDist(tactic):
	if tactic.has_key('mindist'):
		return m25distance.DistanceConvert(tactic['mindist'])
	else:
		if tactic.has_key('power'):
			mindist = m25distance.DistanceConvert(Opt_GetPowerData()[tactic['power']]['RangeMin'] * 10)
			if mindist <= 30: mindist = 0
			return mindist
		else: return 0

		
		
#--------------------------------------------------------
# generics	
	
def Opt_AddToList(vlist, value):
	newlist = vlist
	if type(vlist) == types.ListType:
		newlist = vlist[0:]
		del vlist
		if type(value) == types.ListType:
			for v in value:
				newlist.append(v)
		else:
			newlist.append(value)
	elif vlist == 'none' or vlist == None or vlist == '':
		newlist = value
	else:
		if type(value) == types.ListType:
			newlist = value[0:]
			del value
			newlist.append(vlist)
		else:
			newlist = [vlist,value]
	return newlist
	
def Opt_AddOred(string, addendum):
	if string == '':
		return addendum
	if addendum == '':
		return string
	return string + ' or ' + addendum	
	
	
def Opt_GetStringValue(st):
	return st	
	
def Opt_TabDepth(depth=0):
	if depth <= 0: return ''
	st = ''
	for i in range(0,depth):
		st = st + '\t'
	return st	

	
#----------------------------------------------------------------
# subtype optimization	
		
stcodedict = {}
subtypeindex = {}
subtypeindexcurrent = 0
		
def Opt_generate_stcodedict():
	global stcodedict
	stlist = []

	curfunction = ''
	curcode = []
		
	modpath = datfiles.GetModPath()
	fname = open(modpath+'\Missions\Scripts\m25ai.py','r')
	line = fname.readline()
	while line != '':
		tag = m25string.GetLeftString(line,' ')
		leftover = m25string.GetRightString(line,' ')
		if tag == 'def':
			
			#add subtype code to a dictionary for later processing and optimization
			if (curcode != []) and (curfunction != ''):
				stcodedict[curfunction] = curcode
				curcode = []
				curfunction = ''
		
			if leftover[0:3] == "ST_":
				stlist.append(m25string.GetLeftString(leftover,'('))
				curfunction = m25string.GetLeftString(leftover,'(')
				curcode = []
		else: #not a function header
			l2 = string.replace(line, 'return ', 'codeval = ')
			line3 = string.lstrip(l2)
			if line3 != '' and (line[0] != '#'):
				curcode.append(l2[1:])
		line = fname.readline()
	fname.close()

def Opt_GenerateSubtypeCode(subtype, depth=0):
	global stcodedict
	global subtypeindex
	if stcodedict == {}: Opt_generate_stcodedict()
	code = ''
	#handle 'and'
	n = 1
	if type(subtype) == types.ListType:	
		for t in subtype:
			code = code + Opt_GenOneSubtypeCode(t, depth)
			if n != len(subtype):
				code = code + Opt_TabDepth(depth) + 'if codeval == 1: \n'
			n = n + 1
			depth = depth + 1
	else:
		code = Opt_GenOneSubtypeCode(subtype, depth)
	compiled_code = compile(code,'<string>','exec')
	index = Opt_GetNextSubtypeIndex()
	subtypeindex[index] = compiled_code
	return index
			
def Opt_GenOneSubtypeCode(subtype, depth=0):
	global stcodedict
	code = ''
	trylist = m25string.SplitAroundOr(subtype)
	if trylist[0] == 1:
		code = code + Opt_GenSingleSubtype(trylist[1], depth)
		code = code + Opt_TabDepth(depth) + 'if codeval == 0:\n'
		code = code + Opt_GenOneSubtypeCode(trylist[2], depth+1)
		return code
	else:
		return Opt_GenSingleSubtype(subtype, depth)
		
def Opt_GenSingleSubtype(subtype, depth=0):
	global stcodedict
	code = ''
	notval = 0
	st = subtype
	if m25string.GetLeftString(subtype,' ') == 'not': 
		notval = 1
		st = m25string.GetRightString(subtype, ' ')
	stype = m25string.GetLeftString(st,' ')
	p = m25string.GetRightString(st, ' ')
	param = m25string.GetLeftString(p, ' ')
	try:
		if int(param) != 0: param = int(param)
	except: 
		pass
	param2 = m25string.GetRightString(p, ' ')
	try: 
		if int(param2) != 0: param2 = int(param)
	except: 
		pass
	if stcodedict.has_key('ST_' + stype):
		if param != '':
			code = code + Opt_TabDepth(depth) + 'param = ' + `param` + '\n'
		if param2 != '':
			code = code + Opt_TabDepth(depth) + 'param2 = ' + `param2` + '\n'
		for line in stcodedict['ST_' + stype]:
			code = code + Opt_TabDepth(depth) + line
	else:
		code = Opt_TabDepth(depth) + 'codeval = TryOneSubtype(char, ' + `st` + ', target)\n'
	if notval:
		code = code + Opt_TabDepth(depth) + 'codeval = not codeval\n'
	return code

def Opt_GetNextSubtypeIndex():
	global subtypeindexcurrent
	subtypeindexcurrent = subtypeindexcurrent + 1
	return subtypeindexcurrent
	
def Opt_GetSubtypeCode(index):
	global subtypeindex
	if subtypeindex.has_key(index):
		return subtypeindex[index]
	else:
		return None
		
		
#---------------------------------------------------------------------

opt_targetsubtypespecial = {datfiles.PT_SPECIAL_AFTERBURNER : [],
	datfiles.PT_SPECIAL_ALTERATION : ['not blind','not panicked','not hypnotized','not staticked'],
	datfiles.PT_SPECIAL_ALTER_GENETIC_STRUCTURE : [],
	datfiles.PT_SPECIAL_ANIMATE_DEAD : [],
	datfiles.PT_SPECIAL_ANIMATE_DEAD_MINIONS :[],
	datfiles.PT_SPECIAL_BEZERKER :[],
	datfiles.PT_SPECIAL_BUOYANCY : ['not buoyant'],
	datfiles.PT_SPECIAL_BURROW : [],
	datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION : ['blind or panicked or hypnotized or enraged'],
	datfiles.PT_SPECIAL_CLOAK : [],
	datfiles.PT_SPECIAL_CLONE_SELF : [],
	datfiles.PT_SPECIAL_CLONE_TARGET : [],
	datfiles.PT_SPECIAL_DEEP_SENSOR : [],
	datfiles.PT_SPECIAL_DENSITY_MAXIMIZE : [],
	datfiles.PT_SPECIAL_DIMENSION_SHIFT : ['not exiled'],
	datfiles.PT_SPECIAL_DISPLACE_IMAGE : [],
	datfiles.PT_SPECIAL_DISRUPT_OBJECT : [],
	datfiles.PT_SPECIAL_DOPPELGANGER : [],
	datfiles.PT_SPECIAL_ELECTRIFY_SELF : [],
	datfiles.PT_SPECIAL_EMPATHY : [],
	datfiles.PT_SPECIAL_ENERGIZE : [],
	datfiles.PT_SPECIAL_ENERGIZE_AREA : [],
	datfiles.PT_SPECIAL_EP_DRAW : ['higherenergy 33'],
	datfiles.PT_SPECIAL_EP_TRANSFER : ['lowerenergy 33'],
	datfiles.PT_SPECIAL_HP_BALANCE : ['higherhealththan'],
	datfiles.PT_SPECIAL_IGNITE : [],
	datfiles.PT_SPECIAL_IMAGE_PROJECTOR : [],
	datfiles.PT_SPECIAL_LAWORDER : [],
	datfiles.PT_SPECIAL_LIFE_GIFT : ['lowerhealthpct 50'],
	datfiles.PT_SPECIAL_LIVING_FIRES : [],
	datfiles.PT_SPECIAL_MOLECULAR_EXCITE : [],
	datfiles.PT_SPECIAL_MORPH : [],
	datfiles.PT_SPECIAL_NONE : [],
	datfiles.PT_SPECIAL_NORMALISE : ['blind or panicked or hypnotized or enraged'],
	datfiles.PT_SPECIAL_PATRIOT_CHARGE : [],
	datfiles.PT_SPECIAL_POSSESSION : ['not possessed'],
	datfiles.PT_SPECIAL_POWER_NULLIFICATION : ['usesenergy'],
	datfiles.PT_SPECIAL_RALLY : [],
	datfiles.PT_SPECIAL_REMOVE_DEFENCE : [],
	datfiles.PT_SPECIAL_SACRIFICE : [],
	datfiles.PT_SPECIAL_SONIC_BOOM : [],
	datfiles.PT_SPECIAL_SPEEDING_BULLET : [],
	datfiles.PT_SPECIAL_SPRINT : [],
	datfiles.PT_SPECIAL_TELEPORT_SELF : [],
	datfiles.PT_SPECIAL_TELEPORT_SWAP : [],
	datfiles.PT_SPECIAL_TUMBLE : [],
	datfiles.PT_SPECIAL_TUNNEL : [],
	datfiles.PT_SPECIAL_UNCLOAK : []}	
	
opt_targetsubtypedamage = {datfiles.PT_DAMAGE_ACID : [],
	datfiles.PT_DAMAGE_ACID_BURN : ['not acid_burnt'],
	datfiles.PT_DAMAGE_BLIND : ['not blind'],
	datfiles.PT_DAMAGE_COLD : [],
	datfiles.PT_DAMAGE_CRUSH : [],
	datfiles.PT_DAMAGE_ELECTRICAL : [],
	datfiles.PT_DAMAGE_ENERGY : [],
	datfiles.PT_DAMAGE_FIRE : [],
	datfiles.PT_DAMAGE_FREEZE : ['not frozen'],
	datfiles.PT_DAMAGE_HEX : ['not hexed'],
	datfiles.PT_DAMAGE_HYPNOSIS : ['not hypnotized'],
	datfiles.PT_DAMAGE_IRRADIATE : ['not irradiated'],
	datfiles.PT_DAMAGE_MENTAL_BLANK : ['not blanked'],
	datfiles.PT_DAMAGE_OBJECTIFY : ['not objectified'],
	datfiles.PT_DAMAGE_PANIC : ['not panicked'],
	datfiles.PT_DAMAGE_PIERCE : [],
	datfiles.PT_DAMAGE_POLYMORPH : ['not alone long'],
	datfiles.PT_DAMAGE_RADIATION : [],
	datfiles.PT_DAMAGE_RAGE : ['not enraged','nearfriend'],
	datfiles.PT_DAMAGE_SPECIAL : [],
	datfiles.PT_DAMAGE_SPEED_DECREASE : ['not speed_decreased'],
	datfiles.PT_DAMAGE_SPEED_INCREASE : ['not speed_increased'],
	datfiles.PT_DAMAGE_STASIS : ['not staticked'],
	datfiles.PT_DAMAGE_TOADIFY : ['not toadified'],
	datfiles.PT_DAMAGE_WATER : [],
	datfiles.PT_DAMAGE_WEAKEN : ['not weakened']}	
		
opt_targetsubtypeffx = {'absorb': ['not cloth','not flesh', 'not samematerial'],
'castle': [],
'cloak' : ['not cloaked'],
'crystallise': ['not caged'],
'firecage': ['not caged'],
'gravityincrease': ['higherstrength 0'],
'heal': ['lowerhealthpct 50'],
'icepatch': ['not ice_control', 'not acrobatic', 'not flying'],
'leadership': [],  #needs situation and subtype!
'lifedrain': [],
'magneticpull': ['metal'],
'magneticthrow': ['metal'],
'magnetise': [],
'mentaldamage': [],
'mimic': [],
'staticcage': ['not caged'],
'strengthdrain': ['higherstrength 0'],
'phase': ['not displaced'],  #the ai shouldn't be throwing this one around
'pull': [],
'selfheal': ['lowerhealthpct 50'],
'shrink': ['not shrunk'],
'speedboost': ['lowerspeed 5'],
'strobe': [],
'summon': [],
'taunt': [],
'teleport': [],
'throw': [],
'transmute': [],
'vertigo': ['not vertigo'],
'weaken': ['not ffx_weakened'],
'zerog': ['not flying'],
}	
	

opt_specialtypestring = {datfiles.PT_SPECIAL_AFTERBURNER : 'afterburner',
	datfiles.PT_SPECIAL_ALTERATION : 'alteration',
	datfiles.PT_SPECIAL_ALTER_GENETIC_STRUCTURE : '',
	datfiles.PT_SPECIAL_ANIMATE_DEAD : '',
	datfiles.PT_SPECIAL_ANIMATE_DEAD_MINIONS :'',
	datfiles.PT_SPECIAL_BEZERKER : '',
	datfiles.PT_SPECIAL_BUOYANCY : 'buoyancy',
	datfiles.PT_SPECIAL_BURROW : '',
	datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION : '',
	datfiles.PT_SPECIAL_CLOAK : 'cloak',
	datfiles.PT_SPECIAL_CLONE_SELF : '',
	datfiles.PT_SPECIAL_CLONE_TARGET : '',
	datfiles.PT_SPECIAL_DEEP_SENSOR : '',
	datfiles.PT_SPECIAL_DENSITY_MAXIMIZE : 'density',
	datfiles.PT_SPECIAL_DIMENSION_SHIFT : 'exile',
	datfiles.PT_SPECIAL_DISPLACE_IMAGE : 'displace',
	datfiles.PT_SPECIAL_DISRUPT_OBJECT : '',
	datfiles.PT_SPECIAL_DOPPELGANGER : '',
	datfiles.PT_SPECIAL_ELECTRIFY_SELF : '',
	datfiles.PT_SPECIAL_EMPATHY : 'empathy',
	datfiles.PT_SPECIAL_ENERGIZE : 'energize',
	datfiles.PT_SPECIAL_ENERGIZE_AREA : '',
	datfiles.PT_SPECIAL_EP_DRAW : '',
	datfiles.PT_SPECIAL_EP_TRANSFER : 'energysurge',
	datfiles.PT_SPECIAL_HP_BALANCE : '',
	datfiles.PT_SPECIAL_IGNITE : '',
	datfiles.PT_SPECIAL_IMAGE_PROJECTOR : 'limeylure',
	datfiles.PT_SPECIAL_LAWORDER : '',
	datfiles.PT_SPECIAL_LIFE_GIFT : '',
	datfiles.PT_SPECIAL_LIVING_FIRES : '',
	datfiles.PT_SPECIAL_MOLECULAR_EXCITE : '',
	datfiles.PT_SPECIAL_MORPH : '',
	datfiles.PT_SPECIAL_NONE : '',
	datfiles.PT_SPECIAL_NORMALISE : '',
	datfiles.PT_SPECIAL_PATRIOT_CHARGE : '',
	datfiles.PT_SPECIAL_POSSESSION : '',
	datfiles.PT_SPECIAL_POWER_NULLIFICATION : 'powernull',
	datfiles.PT_SPECIAL_RALLY : 'rally',
	datfiles.PT_SPECIAL_REMOVE_DEFENCE : '',
	datfiles.PT_SPECIAL_SACRIFICE : '',
	datfiles.PT_SPECIAL_SONIC_BOOM : '',
	datfiles.PT_SPECIAL_SPEEDING_BULLET : '',
	datfiles.PT_SPECIAL_SPRINT : '',
	datfiles.PT_SPECIAL_TELEPORT_SELF : '',
	datfiles.PT_SPECIAL_TELEPORT_SWAP : '',
	datfiles.PT_SPECIAL_TUMBLE : '',
	datfiles.PT_SPECIAL_TUNNEL : '',
	datfiles.PT_SPECIAL_UNCLOAK : ''}	
	
opt_damagetypestring = {datfiles.PT_DAMAGE_ACID : '',
	datfiles.PT_DAMAGE_ACID_BURN : 'acidburn',
	datfiles.PT_DAMAGE_BLIND : 'blind',
	datfiles.PT_DAMAGE_COLD : '',
	datfiles.PT_DAMAGE_CRUSH : '',
	datfiles.PT_DAMAGE_ELECTRICAL : '',
	datfiles.PT_DAMAGE_ENERGY : '',
	datfiles.PT_DAMAGE_FIRE : '',
	datfiles.PT_DAMAGE_FREEZE : 'frozen',
	datfiles.PT_DAMAGE_HEX : 'hex',
	datfiles.PT_DAMAGE_HYPNOSIS : 'hypnosis',
	datfiles.PT_DAMAGE_IRRADIATE : 'irradiate',
	datfiles.PT_DAMAGE_MENTAL_BLANK : 'blank',
	datfiles.PT_DAMAGE_OBJECTIFY : '',
	datfiles.PT_DAMAGE_PANIC : '',
	datfiles.PT_DAMAGE_PIERCE : '',
	datfiles.PT_DAMAGE_POLYMORPH : 'polymorph',
	datfiles.PT_DAMAGE_RADIATION : '',
	datfiles.PT_DAMAGE_RAGE : 'rage',
	datfiles.PT_DAMAGE_SPECIAL : '',
	datfiles.PT_DAMAGE_SPEED_DECREASE : 'speeddecrease',
	datfiles.PT_DAMAGE_SPEED_INCREASE : 'speedincrease',
	datfiles.PT_DAMAGE_STASIS : 'stasis',
	datfiles.PT_DAMAGE_TOADIFY : 'toadify',
	datfiles.PT_DAMAGE_WATER : '',
	datfiles.PT_DAMAGE_WEAKEN : 'weaken'}

def Opt_GetSpecialSubtype(char, powername):
	power = Opt_GetPowerData()[powername]
	target = []
	if (not (power['PowerType'] in [datfiles.PT_PASSIVE_DEFENCE,datfiles.PT_ACTIVE_DEFENCE])):
		sp = Opt_GetFFXSubtype(char,power['PowerName'])
		if sp == []:
			sp = opt_targetsubtypespecial[power['SpecialType']]
		if sp == []:
			sp = opt_targetsubtypedamage[power['DamageType']]
		if sp == []:
			if datfiles.PT_DAMAGE_MAGNITUDE_NAMES[power['Magnitude']] == 'none':
				if datfiles.PT_STUN_NAMES[power['Stun']] != 'none':
					sp = ['not stunned']
		if sp != []:
			for s in sp:
				if not (s in target):
					target.append(s)
	if target == []:
		target = 'none'
	return target

def Opt_GetFFXSubtype(char, powername):
	carrier = Opt_FindFFXCarrier(char,powername)
	if carrier == None: return []
	if opt_targetsubtypeffx.has_key(carrier):
		return opt_targetsubtypeffx[carrier]
	else:
		return []

def Opt_FindFFXCarrier(char,powername):
	template = m25template.GetComplexTemplate(char)
	power = Opt_GetPowerData()[powername]
	truepowername = m25custompowers.GetTruePowerName(char,powername,default=powername)
	if power.has_key('SpecialType'): specialtype = power['SpecialType']
	else:
		specialtype = datfiles.PT_SPECIAL_NONE
	if power.has_key('DamageType'): damagetype = power['DamageType']
	else: 
		damagetype = datfiles.PT_DAMAGE_WATER
	for e in ffxcustom.FFX_CARRIERS:
		if e[0] == template:
			if e[3] == truepowername:
				return e[2]
			else:
				if e[3] == 'all':
					if (e[1] == opt_specialtypestring[specialtype]) or (e[1] == opt_damagetypestring[damagetype]):
						return e[2]
	return None
		
	

#--------------------------------------------------------------------------------

# def testopt():
# 	tactics = [
# 	]
# 	Optimize_GenerateOptimizedTactics('test', tactics)


def Optimize_GenerateOptimizedTactics(char, tactics, energy_reserve=0):
	#start = ff.clock()
	codedict = {}
	Opt_FleshOutCodeDict(codedict)
	#tactics = GetAITimer(char)[1:]
	for t in tactics:
		#tactic_info = Opt_GetTacticInfo(t)
		tactic_info = Opt_CalcParameters(char,t)
		if tactic_info == None: 
			m25report.Report('error', 'Optimize_GenerateOptimizedTactics for %s failed on %s'%(char,`t`))
			continue
		
		#find type, energy cost and range of the ability
		if tactic_info.has_key('category'): tac_type = tactic_info['category']
		else: tac_type = 'attack'
		tac_range = Opt_FindOptRange(tactic_info)
		tac_energy = Opt_FindOptEnergy(tactic_info)	
		Opt_CreateEntries(codedict, tactic_info, tac_type, tac_range, tac_energy)

	#generate code from codedict
	code = Opt_GenerateTacticCode(char,codedict, energy_reserve)
	if AIOPTPRINTCODE:	
		print '====================================='
		print 'Code for %s'%char
		print '-----'
		print code
		print '====================================='
	#compile the code
	compiled_code = compile(code, '<string>', 'exec')
	return compiled_code

def Opt_FleshOutCodeDict(dict):
	dict['melee'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['short'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['medium'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['long'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['defense'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['move'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['flee'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['remedy'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['none'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['pre'] = {'high':[],'medium':[],'low':[],'none':[]}
	dict['post'] = {'high':[],'medium':[],'low':[],'none':[]}
	
def Opt_CreateEntries(dict, tactic, ttype, trange, tenergy):
# 	if (ttype == 'none') or (trange == 'none'):
# 		dict['none'].append(tactic)
	if ttype != 'attack':
		dict[ttype][tenergy].append(tactic)
	else:
		dict[trange][tenergy].append(tactic)
		if trange == 'long':
			dict['medium'][tenergy].append(tactic)
			dict['short'][tenergy].append(tactic)
		if trange == 'medium':
			dict['short'][tenergy].append(tactic)
		if trange == 'short':
			dict['melee'][tenergy].append(tactic)
		

def Opt_GenerateTacticCode(char,dict,reserve):
	code = 'energy=js.Object_GetAttr(char,"energyPoints")\ntarget=Target_GetCurrentTarget(char)\nrange=Target_GetDistance(char)\ndone=0\nreserve='+`m25energy.ConvertEnergy(reserve)`+'\n'
	code = code + 'if js.Object_GetSecondaryStates(char) & js.SCSTATE_POWER_NULLIFICATION: energy=0\n'
	tempcode = Optimize_LowMediumHigh(dict['remedy'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	tempcode = Optimize_LowMediumHigh(dict['defense'],1)
	if tempcode != '':
		code = code + 'if not HasRole(char, "active_defense"):\n' + tempcode
	tempcode = Optimize_LowMediumHigh(dict['flee'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	tempcode = Optimize_LowMediumHigh(dict['pre'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	elstr = ''
	tempcode = Optimize_LowMediumHigh(dict['long'],1)
	if tempcode != '': 
		code = code + 'if (not done) and range >= 206 and range <= 320:\n' + tempcode
		elstr = 'el'
	tempcode = Optimize_LowMediumHigh(dict['medium'],1)
	if tempcode != '': 
		code = code + elstr+'if (not done) and range >= 106 and range <= 205:\n' + tempcode
		if elstr == '': elstr = 'el'
	tempcode = Optimize_LowMediumHigh(dict['short'],1)
	if tempcode != '': 
		code = code + elstr+'if (not done) and range >= 61 and range <= 105:\n' + tempcode
		if elstr == '': elstr = 'el'
	tempcode = Optimize_LowMediumHigh(dict['melee'],1)
	if tempcode != '': 
		code = code + elstr+'if (not done) and range <= 60:\n' + tempcode
	code = code + 'else: pass\n'
	tempcode = Optimize_LowMediumHigh(dict['post'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	tempcode = Optimize_LowMediumHigh(dict['move'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	tempcode = Optimize_LowMediumHigh(dict['none'],1)
	if tempcode != '':
		code = code + 'if (not done):\n' + tempcode
	#if dict['none'] != []:
		#for d in dict['none']:
		#	code = code + Optimize_CreateTacticCall(d) + '\n'
	return code

def Optimize_GenLine(dict, depth=0):
	code = ''
	start = 1
	for d in dict:
		if start:
			code = Opt_TabDepth(depth) + ' done = ' + Optimize_CreateTacticCall(d)
			start = 0
		else:
			code = code + ' or ' + Optimize_CreateTacticCall(d)
	code = code + '\n'
	return code
		
def Optimize_LowMediumHigh(dict, depth=0):
	code = ''
	if dict['high'] != []:
		code = code + Opt_TabDepth(depth) + 'if (not done) and energy>=99:\n' + Optimize_GenLine(dict['high'],depth+1)
	if dict['medium'] != []:
		code = code + Opt_TabDepth(depth) + 'if (not done) and energy>=66 and energy>=reserve:\n' + Optimize_GenLine(dict['medium'],depth+1)
	if dict['low'] != []:
		code = code + Opt_TabDepth(depth) + 'if (not done) and energy>=33 and energy>=reserve:\n' + Optimize_GenLine(dict['low'],depth+1)
	if dict['none'] != []:
		code = code + Opt_TabDepth(depth) + 'if (not done):\n'  + Optimize_GenLine(dict['none'],depth+1)
	return code
	
def Optimize_CreateTacticCall(tactic):
	newtactic = 'ff.%s(tactic_id,event.object,event.string,event.float,event.user'%(tactic['name'])
	for t in tactic.keys():
		if t not in ['name','category','radius']:
			if type(tactic[t]) == types.StringType:
				newtactic = '%s, %s = "%s"'%(newtactic,t,tactic[t])
			else:
				newtactic = '%s, %s = %s'%(newtactic,t,tactic[t])
	newtactic = newtactic + ')'
	return newtactic


	

def Opt_FindOptEnergy(tactic):
	energy_value = 'none'
	tac_energy = None
	if tactic.has_key('energy'): tac_energy = tactic['energy']
	if tac_energy == None:
		energy_value = 'none'
	elif m25energy.ConvertEnergy(tac_energy,push=0) >= 100:
		energy_value = 'high'
	elif m25energy.ConvertEnergy(tac_energy,push=0) >= 66:
		energy_value = 'medium'
	elif m25energy.ConvertEnergy(tac_energy,push=0) >= 33:
		energy_value = 'low'
	else:
		energy_value = 'none'
	return energy_value

def Opt_FindOptRange(tactic):
	#translate to the actual value
	range_value = 'none'
	tac_range = None
	if tactic.has_key('maxdist'): tac_range = tactic['maxdist']
	if tac_range == None:
		range_value = 'none'
	elif m25distance.DistanceConvert(tac_range) >= 201:
		range_value = 'long'
	elif m25distance.DistanceConvert(tac_range) >= 101:
		range_value = 'medium'
	elif m25distance.DistanceConvert(tac_range) >= 61:
		range_value = 'short'
	elif m25distance.DistanceConvert(tac_range) <= 60:
		range_value = 'melee'
	else: 
		range_value = 'none'
	return range_value	
