#####################################################################################
#
# Module: Distance
# Author: m25
# 
# Description: defines distance constants and determines distance between objects
#
#
# Version 3.052
#
#####################################################################################

#print "Loading m25distance.py v.3.2.0 final; branch = m25 branch 1"

#from ff import *
import ff
import js
import types
import math

#import m25timestamp
#from m25timestamp import *
#import m25getset

DISTANCE_OUTOFRANGE = 1000000
DISTANCE_WHOLEMAP = 1000000
DISTANCE_INFINITE = 1000000
DISTANCE_VISIBLE = 400
DISTANCE_LONG = 300
DISTANCE_MEDIUM = 200
DISTANCE_SHORT = 100
DISTANCE_MELEE = 50
DISTANCE_AREA_SMALL = 60
DISTANCE_AREA_MEDIUM = 90
DISTANCE_AREA_LARGE = 120
DISTANCE_AREA_EXTREME = 150
DISTANCE_AREA_OVERPOWER = 300
DISTANCE_CONE_NARROW = 20
DISTANCE_CONE_MEDIUM = 40
DISTANCE_CONE_WIDE = 60
DISTANCE_CONE_EXTREME = 80

def GetDistance(object1, object2, recalc=0):
	#check cache first
	#val = m25timestamp.Timestamp_GetMissionCachedAttr(`object1`+'_'+`object2`+'_dist', delta=1)
	#if (val != None) and not recalc: return val
	
	if not (js.Object_Exists(object1)): 
		return DISTANCE_OUTOFRANGE
	if type(object2) not in [types.StringType, types.TupleType]:
		return DISTANCE_OUTOFRANGE
	if type(object2) == types.StringType and (not js.Object_Exists(object2)):
		return DISTANCE_OUTOFRANGE
	p1 = js.Get_ObjectPos(object1)
	if type(object2) == types.TupleType:
		p2 = object2
	else:		
		p2 = js.Get_ObjectPos(object2)
	v = js.Vector_Sub(p1,p2)
	try:
		val = math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
	except:
		return DISTANCE_OUTOFRANGE
	
	#cache the distance so it doesn't have to be recalculated immediately
	#m25timestamp.Timestamp_SetMissionCachedAttr(`object1`+'_'+`object2`+'_dist', val)
	#m25timestamp.Timestamp_SetMissionCachedAttr(`object2`+'_'+`object1`+'_dist', val)
	return val
    

def GetDistanceToMarker(object, marker):
	if not js.Object_IsAlive(object):
		return DISTANCE_OUTOFRANGE
	p1 = js.Get_ObjectPos(object)
	p2 = js.Get_MarkerPos(marker)
	v = js.Vector_Sub(p1,p2)
	return math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])

def GetDistanceBetweenMarkers(marker1, marker2):
	if not js.Marker_Exists(marker1) or not js.Marker_Exists(marker2):
		return DISTANCE_OUTOFRANGE
	p1 = js.Get_MarkerPos(marker1)
	p2 = js.Get_MarkerPos(marker2)
	v = js.Vector_Sub(p1,p2)
	return math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])	
	
def GetTupleDistance(t1, t2):
	v = js.Vector_Sub(t1, t2)
	return math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
	

_distanceconvert ={
	'melee':DISTANCE_MELEE,
	'long':DISTANCE_LONG,
	'medium':DISTANCE_MEDIUM,
	'short':DISTANCE_SHORT,
	'small':DISTANCE_AREA_SMALL,
	'area_small':DISTANCE_AREA_SMALL,
	'area_medium':DISTANCE_AREA_MEDIUM,
	'area_large':DISTANCE_AREA_LARGE,
	'large':DISTANCE_AREA_LARGE,
	'none':0,
	'area_none':0,
	'cone_narrow':DISTANCE_CONE_NARROW,
	'cone_medium':DISTANCE_CONE_MEDIUM,
	'cone_wide':DISTANCE_CONE_WIDE,
	'cone_extreme':DISTANCE_CONE_EXTREME,
	'map':DISTANCE_WHOLEMAP,
	'wholemap':DISTANCE_WHOLEMAP,
	'infinite':DISTANCE_INFINITE,
}
		
def DistanceConvert(dist,char='',default=0):
	if type(dist) is types.StringType:
		if dist == 'visible':
			if js.Object_Exists(char):
				#print 'DistanceConvert(): ' + char + ' has detect radius ' + `js.Object_GetAttr(char,'detectradius')`
				return js.Object_GetAttr(char,'detectradius') * 10
			return DISTANCE_VISIBLE
		elif dist == 'ranged_min':
			return ff.RANGED_MIN * 10
		else:
			if dist in _distanceconvert.keys():
				return _distanceconvert[dist]
		try:
			d = int(dist)
			return d
		except:
			print 'DistanceConvert(): bad distance value ' + dist
			#raise
			return default
	else:
		return dist	
		
def DistanceCompare(dist1, dist2, compare_type):
	d1 = DistanceConvert(dist1)
	d2 = DistanceConvert(dist2)
	val = 0
	exec 'val = ' + `d1` + compare_type + `d2`
	return val