#####################################################################
#
# EZ script reader (pretty stories)
# 
# Take an ez script mission description and turn it into something
# that m25story can run.
#
# 
# Version 3.052
#
# This is an ad-hoc parser (hopefully to be converted to a grammar-based one)
#
#####################################################################

print "Loading m25pstory.py v.3.2.0 final; May 27, 2008"

#May 27,2008 (m25): Fixed win, lose, end conditions so that they can be in any case and can be used in condition Next: statements

#BUGS:   
# - proper names are always made lower case and thus won't work correctly in cutscenes


import js
import string
import datfiles
import os
import types
import shutil

import m25string
#import m25cutscene
import m25report
import m25event

import m25gametype

import m25enccustom

import datfiles
import chardata

from m25config import *

PSTORY_WRITE_CAPTIONS = 0

m25report.Report_SetChannel('PSTORY', 1)


pstory_replacements_list = ['hero','villain','minion','police','ally','fallen','object','buddy', 'interloper', 'rival', 'cohort', 'vehicle', 'switch', 'item']


#####################################################################################
#
# Main command for Pstory
# 
# Description: 
#
#####################################################################################	
	
pstory_required_markers = ['encounter1', 'skirmishspawn']
pstory_has_random = 0

def PStory_GetRequiredMarkers(filename):
	global pstory_required_markers
	pstory_required_markers = ['encounter1', 'skirmishspawn']
	PStory_MakeStoryFromFile(filename)
	return tuple(pstory_required_markers)

def PStory_GetRandomRequired(filename):
	global pstory_has_random
	pstory_has_random = 0
	PStory_MakeStoryFromFile(filename)
	return pstory_has_random	
	

	
def PStory_MakeStoryFromFile(filename):
	struct = PStory_ReadStoryFromFile(filename)	
	if struct != None:
		if PStory_IsAnthology(struct):
			return PStory_FlattenList(PStory_MakeAnthology(struct))
		if PStory_IsStory(struct):
			return [PStory_MakeStory(struct)]
	return []

def PStory_MakeAnthology(struct):
	anthology = []
	if struct != {}:
		for line in struct:
			line = string.lower(string.strip(line))
			if (line != '') and (line[0]!='#') and (line[0:len('anthology')]!= 'anthology'):
				anthology.append(PStory_MakeStoryFromFile(line))
	return anthology

def PStory_FlattenList(struct):
	newlist = []
	for s in struct:	
		if type(s) != types.ListType:
			newlist.append(s)
		else:
			templist = PStory_FlattenList(s)
			for t in templist:
				newlist.append(t)
	return newlist		
		
def PStory_ReadStoryFromFile(filename):
	struct = PStory_MakeAnthologyFromDirectory(filename)
	if struct != ['Anthology']: return struct
	struct = ''
	filepath = ''
	found = 0
	path,file = os.path.split(filename)
	if path != '':
		filepath = filename
		found = 1
	#see if we can find the story file in the same directory as the mission file inside a campaign
	if (not found) and chardata.isCampaign():
		try:
			filepath = os.path.join(os.path.dirname(datfiles.GetMissionDatFile()),filename+'.txt')
			if os.path.isfile(filepath):
				found = 1
		except:
			pass
	#look in the stories directory under the mod
	if not found:
		scriptpath = os.path.join(datfiles.GetModPath(),'stories')
		filepath = os.path.join(scriptpath,filename+'.txt')
	try:
		f = open(filepath,'r')
		struct = f.readlines()
		f.close()
	except:
		m25report.Report('ERROR', 'The story file ' + `filename` + ' is not in your stories directory!')
		return None
	return struct
	
def PStory_MakeAnthologyFromDirectory(dirname):
	struct = ['Anthology']
	modpath = datfiles.GetModPath()
	scriptpath = os.path.join(modpath,dirname)
	if os.path.isdir(scriptpath):
		files = os.listdir(scriptpath)
		for f in files:
			if f[-4:] == '.txt':
				struct.append(os.path.join(scriptpath,f))
	return struct

	
def PStory_GetStoryNameFromFile(filename):
	struct = PStory_ReadStoryFromFile(filename)
	if struct != {}:
		for line in struct:
			if string.lower(string.lstrip(line)[0:len('story:')]) == 'story:':
				name = string.strip(m25string.GetRightString(line,' '))
				if name != '':
					return name
	return None
			
def PStory_IsAnthology(struct):
	#struct = PStory_ReadStoryFromFile(filename)
	if struct != {}:
		for line in struct:
			if string.lower(string.lstrip(line)[0:len('anthology')]) == 'anthology':
				return 1
	return 0

def PStory_IsStory(struct):
	#struct = PStory_ReadStoryFromFile(filename)
	if struct != {}:
		for line in struct:
			if string.lower(string.lstrip(line)[0:len('story:')]) == 'story:':
				return 1
	return 0

#####################################################################################
#
# Convert EZ script story description into a temporary form (Tempform)
# 
# Description: 
#
#####################################################################################	
	
storycut = 'default'


def PStory_MakeStory(struct):

	story = {}
	story['name'] = 'default_story'
	story['start'] = []
	story['encounters'] = []

	for line in struct:
		if PStory_IsBlankLine(line): continue
	
		#ignore comments
		if string.lstrip(line)[0] == '#':
			continue
	
		if string.lower(string.lstrip(line)[0:len('story:')]) == 'story:':
			story['name'] = string.strip(m25string.GetRightString(line,' '))
			#story cut is used to create captions in captions.txt later on
			global storycut
			storycut = PStory_RemoveSpaces(string.lower(story['name']))
			
		#if starting encounters specified - these are the encounters that will fire at mission start
		if string.lower(string.lstrip(line)[0:len('starting encounters')]) == 'starting encounters':
			story['start'] = PStory_CommaToList(string.strip(m25string.GetRightString(line,':')))
			
	#get the required elements
	#get each encounter
	index = 0
	for line in struct:
		if string.lower(string.lstrip(line[0:len('Encounter')])) == 'encounter':
			story['encounters'].append(PStory_GetEncounter(struct[index:]))
		index = index + 1
		
	#write out any spoken lines from cutscenes to the captions.txt file if that option is selected
	if M25_MAKE_CAPTIONS or PSTORY_WRITE_CAPTIONS:
		PStory_WriteCaptions()
	
	#convert the structure to an actual story
	m25report.Report('PSTORY',repr(story))
	return PStory_TranslateStory(story)
	

	
#-------------------------------------------------------------------
#
#  Read in Encounters
#
#-------------------------------------------------------------------

#read in the actual encounters (encounter: x)
def PStory_GetEncounter(struct):
	#set up default values so the game will do something even with incomplete information
	encounter = {}
	encounter['cutscenes'] = []
	encounter['parameters'] = []
	encounter['marker'] = 'encounter'
	encounter['type'] = 'Fight'
	encounter['next'] = []
	encounter['trigger'] = []
	encounter['objectives'] = []
	encounter['required_map'] = []
	encounter['timeofday'] = 'none'
	encounter['pre'] = []
	index = 0
	have_encounter = 0
	
	#check each line for keywords - those having a ':' after them
	for line in struct:
		#ignore comments
		if not PStory_IsBlankLine(line) and string.lstrip(line)[0] == '#':
			index = index + 1
			continue
		#ignore the line if it isn't a keyword
		if string.find(line,':') == -1: 
			index = index + 1
			continue
		#get the keyword
		command = string.lower(string.lstrip(m25string.GetLeftString(line,':')))
		#encounter name
		if command == 'encounter':
			if have_encounter: # we've hit the start of another encounter, so end this one
				return encounter
			encounter['name'] = string.strip(m25string.GetRightString(line,':'))
			have_encounter = 1
		#type of encounter
		elif command == 'type':
			encounter['type'] = string.strip(m25string.GetRightString(line,':'))
		elif command == 'map':
			encounter['required_map'] = PStory_AddToList(encounter['required_map'],PStory_CommaToList(string.lower(string.strip(m25string.GetRightString(line,':')))))
		elif command == 'time of day':
			encounter['timeofday'] = string.lower(string.strip(m25string.GetRightString(line,':')))
		#villains used
		elif command == 'villains':
			encounter['villains'] = PStory_CommaNamedList(m25string.GetRightString(line,':'))
		#minions used
		elif command == 'minions':
			encounter['minions'] = PStory_CommaNamedList(m25string.GetRightString(line,':'))
		elif command == 'play cutscene in order shown':
			encounter['cutscenes'].append('no_randomize_%s'%string.lower(string.strip(m25string.GetRightString(line,':'))))
		#found a cutscene - invoke the cutscene reader
		elif string.find(command,'cutscene') != -1:
			encounter['cutscenes'].append(PStory_GetCutscene(struct[index:]))
		#next - the encounter that will fire after the current encounter completes
		elif command == 'next':
			encounter['next'] = PStory_AddToList(encounter['next'],PStory_GetNextEncounter(string.strip(m25string.GetRightString(line,':'))))
			#encounter['next'].append(PStory_GetNextEncounter(string.strip(m25string.GetRightString(line,':'))))
		#what marker, if any, that the encounter will start from
		elif command == 'marker':
			encounter['marker'] = string.strip(m25string.GetRightString(line,':'))
			pstory_required_markers.append(encounter['marker']+'1')
		elif command == 'starts when':
			params = PStory_CommaToList(string.strip(m25string.GetRightString(line,':')))
			for p in params:
				encounter['trigger'].append(p)
# 			
# 				if string.lower(p) == 'briefing':
# 					encounter['trigger'].append('briefing')
# 				else:
# 					temp = m25string.SplitAroundPhrase(p,' at ')
# 					if temp[0] == 1:
# 						encounter['trigger'].append([temp[1],temp[2]])
# 					else:
# 						encounter['trigger'].append(temp)
		elif command == 'preconditions':
			params = PStory_CommaToList(string.strip(m25string.GetRightString(line,':')))
			for p in params:
				encounter['pre'].append(p)
		elif command == 'primary objective':
			params = string.strip(m25string.GetRightString(line,':'))
			# the objective itself is in quotes ("")
			startdesc = m25string.GetRightString(params, '"')
			description = m25string.GetLeftString(startdesc, '"')
			temp = m25string.GetRightString(startdesc, '"')
			temp1 = m25string.SplitAroundPhrase(string.strip(temp), 'for')
			if temp1[0] == 1:
				#description = temp1[1]
				temp2 = m25string.SplitAroundPhrase(string.strip(temp1[2]),'prestige')
				if temp2[0] == 1:
					prestige = int(temp2[1])
					temp3 = m25string.SplitAroundPhrase(string.strip(temp2[2]), 'and')
					if temp3[0] == 1:
						temp4 = m25string.SplitAroundPhrase(string.strip(temp3[2]), 'xp')
						if temp4[0] == 1:
							cp = int(temp4[1])
							encounter['objectives'].append([0,description,prestige,cp])
					else:
						encounter['objectives'].append([0,description,prestige,0])
		elif command == 'secondary objective':
			params = string.strip(m25string.GetRightString(line,':'))
			# the objective itself is in quotes ("")
			startdesc = m25string.GetRightString(params, '"')
			description = m25string.GetLeftString(startdesc, '"')
			temp = m25string.GetRightString(startdesc, '"')
			temp1 = m25string.SplitAroundPhrase(string.strip(temp), 'for')
			if temp1[0] == 1:
				#description = temp1[1]
				temp2 = m25string.SplitAroundPhrase(string.strip(temp1[2]),'prestige')
				if temp2[0] == 1:
					prestige = int(temp2[1])
					temp3 = m25string.SplitAroundPhrase(string.strip(temp2[2]), 'and')
					if temp3[0] == 1:
						temp4 = m25string.SplitAroundPhrase(string.strip(temp3[2]), 'xp')
						if temp4[0] == 1:
							cp = int(temp4[1])
							encounter['objectives'].append([1,description,prestige,cp])
					else:
						encounter['objectives'].append([1,description,prestige,0])
		#bonus to prestige and xp - these are treated as silent objectives internally
		elif command == 'bonus':
			params = string.strip(m25string.GetRightString(line,':'))
			temp1 = m25string.SplitAroundPhrase(string.strip(params),'prestige')
			if temp1[0]:
				prestige = int(temp1[1])
				temp3 = m25string.SplitAroundPhrase(string.strip(temp1[2]), 'and')
				if temp3[0] == 1:
					temp4 = m25string.SplitAroundPhrase(string.strip(temp3[2]), 'xp')
					if temp4[0] == 1:
						cp = int(temp4[1])
						encounter['objectives'].append([-1,'',prestige,cp])
		#other parameters are placed in a list to be passed to the encounter function
		else:  #we have an unrecognized parameter
			if (command != line) and (command != ''):
				#if command in pstory_name_lists: encounter['parameters'].append(PStory_CommaNamedList(m25string.GetRightString(line,':')))
				#else:
				encounter['parameters'].append([string.strip(string.lower(PStory_RemoveSpaces(command))),string.strip(m25string.GetRightString(line,':'))])
		index = index + 1
	return encounter

	
#-------------------------------------------------------------------
#
#deal with Next: keywords.  this can be:
#1. a single name of an encounter
#2. a list of encounters seperated by commas
#3. if condition: encounter list - condition must be met before encounter list will fire
#4. choose x from encounter list - pick x encounters from the list at random
#
#-------------------------------------------------------------------
		
def PStory_GetNextEncounter(line):
	nextstate = 'None'
	if string.lower(string.lstrip(line)[0:len('if')]) == 'if':
		expr = string.strip(m25string.GetLeftString(line[2:],':'))
		remainder = m25string.GetRightString(line,':')
		if string.lower(string.lstrip(remainder)[0:len('choose')]) == 'choose':
			nextstate = [[PStory_TranslateExpression(expr), PStory_TranslateChooseFrom(string.strip(remainder)[len('choose'):])]]
		else:
			nextstate = [[PStory_TranslateExpression(expr), PStory_TransEndStates(PStory_CommaToList(m25string.GetRightString(line,':')))]]
	elif string.lower(string.lstrip(line)[0:len('choose')]) == 'choose':
		nextstate = PStory_TranslateChooseFrom(string.strip(line)[len('choose'):])
	else:
		nextstate = PStory_TransEndStates(PStory_CommaToList(line))
	return nextstate

def PStory_TransEndStates(nextstate):
	trans = {'win':'__win__', 'lose':'__lose__', 'end':'__end__'}
	index = 0
	if type(nextstate) == types.ListType:
		for n in nextstate:
			if string.lower(n) in trans.keys(): nextstate[index] = trans[string.lower(n)]
			index = index + 1
	else:
		if string.lower(nextstate) in trans.keys(): nextstate = trans[string.lower(n)]
	return nextstate
	
def PStory_TranslateChooseFrom(expr):
	temp = m25string.SplitAroundPhrase(string.strip(expr), 'from')
	if temp[0] == 1:
		actuallist = []
		count = int(temp[1])
		nlist = PStory_TransEndStates(PStory_CommaToList(temp[2]))
		num = 1
		while num <= count:
			actuallist.append('Story_RandomListElement(' + `nlist` + ')')
			num = num + 1
		return actuallist
	else:
		return []
	
def PStory_TranslateExpression(expr):
	trans = ''
	test = m25string.SplitAroundAnd(expr)
	if test[0] == 1:
		return PStory_TranslateExpression(test[2]).append(test[1])
	else:
		return [expr]
			


	
#-------------------------------------------------------------------
#
#  Read in cutscenes
#
#-------------------------------------------------------------------	
	
def PStory_LookAhead():
	global story_struct
	global story_index	
	if story_index < len(story_struct):
		line = story_struct[story_index]
		story_index = story_index + 1
		return line
	else:
		return ''

def PStory_PushBack():
	global story_index
	story_index = story_index - 1

def PStory_GetCutscene(struct):
	global story_struct
	global story_index
	
	story_index = 0
	story_struct = struct
	
	cs = {}
	cs['replace'] = []
	cs['required'] = []
	cs['cutscene'] = []
	cs['priority'] = 1
	cs['name'] = 'default'
	cinematic = 0  #cinematic cutscenes need the letterbox mode to prevent the player from messing them up
	arrownum = 0
	#for line in struct:
	while (story_index < len(story_struct)):
		line = story_struct[story_index]
		story_index = story_index + 1
		
		if PStory_IsBlankLine(line):
			#check to see if cinematic and if so insert the letterboxing commands
			if cinematic:
				new_cs = ['CS_Start()']
				for c in cs['cutscene']:
					new_cs.append(c)
				new_cs.append('CS_End()')
				cs['cutscene'] = new_cs
			return cs
		value = m25event.Event_TriggerFirst('pstory_cutsceneline', event=m25event.UserEvent('pstory_cutsceneline',string.strip(line)))
		if value:
			command, actors = value
			for a in actors:
				cs['replace'], cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],a)			
			cs['cutscene'].append(command)
		#ignore comments
		elif string.lstrip(line)[0] == '#':
			pass
		elif string.find(string.lower(line),'cutscene') != -1:
			cs['name'] = PStory_RemoveSpaces(string.strip(line[0:string.find(string.lower(line),'cutscene')]))
# 		elif string.find(string.lower(line),'says telepathically') != -1:
# 			speaker = string.strip(line[0:string.find(string.lower(line),'says telepathically')])
# 			speech = m25string.GetRightString(string.strip(line)[:-1],'"')	
# 			if M25_MAKE_CAPTIONS or PSTORY_WRITE_CAPTIONS:
# 				PStory_AddCaption(speech)	
# 			cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],speaker,force=1)
# 			cs['cutscene'].append('CS_Speak(' + `string.lower(speaker)` + ', ' + `speech` + ',force=1)')
		elif string.find(string.lower(line),'says') != -1:
			telepathic = (string.find(string.lower(line),'says telepathically') != -1)
			trait = PStory_IsTraitSpeech(line)
			speaker, speech, defaultspeech = '', '', ''
			if trait:
				speaker, speech, defaultspeech = PStory_GetTraitSpeech(line)
				if M25_MAKE_CAPTIONS or PSTORY_WRITE_CAPTIONS:
					for s in speech:
						PStory_AddCaption(s[2])
					PStory_AddCaption(defaultspeech)
			else:
				speaker = string.strip(line[0:string.find(string.lower(line),'says')])
				speech = m25string.GetRightString(string.strip(line)[:-1],'"')	
				if M25_MAKE_CAPTIONS or PSTORY_WRITE_CAPTIONS:
					PStory_AddCaption(speech)	
			cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],speaker, force=telepathic)
			if trait:
				cs['cutscene'].append('CS_TraitSpeak("%s", %s, "%s", force=%i)'%(string.lower(speaker), repr(speech), defaultspeech, telepathic))
			else:
				cs['cutscene'].append('CS_Speak(' + `string.lower(speaker)` + ', ' + `speech` + ', force=' + `telepathic` + ')')
		elif string.find(string.lower(line),'turns to') != -1:
			temp = m25string.SplitAroundPhrase(line,'turns to')
			if temp[0] == 1:
				actor = temp[1]
				target = temp[2]
				cs['replace'], cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_TurnTo(' + `string.lower(actor)` + ', ' + `string.lower(target)` + ')')
				#cinematic = 1
		elif string.find(string.lower(line),'moves to') != -1:
			temp = m25string.SplitAroundPhrase(line,'moves to')
			if temp[0] == 1:
				actor = temp[1]
				target = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_MoveTo(' + `string.lower(actor)` + ', ' + `string.lower(target)` + ')')
				cinematic = 1
		elif string.find(string.lower(line),'flies to') != -1:
			temp = m25string.SplitAroundPhrase(line,'flies to')
			if temp[0] == 1:
				actor = temp[1]
				target = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_MoveTo(' + `string.lower(actor)` + ', ' + `string.lower(target)` + ', command="fly")')
				cinematic = 1
		elif string.find(string.lower(line),'walks to') != -1:
			temp = m25string.SplitAroundPhrase(line,'walks to')
			if temp[0] == 1:
				actor = temp[1]
				target = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_WalkTo(' + `string.lower(actor)` + ', ' + `string.lower(target)` + ')')
				cinematic = 1
		elif string.find(string.lower(line),'teleports to') != -1:
			temp = m25string.SplitAroundPhrase(line,'teleports to')
			if temp[0] == 1:
				actor = temp[1]
				target = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_TeleportTo(' + `string.lower(actor)` + ', ' + `string.lower(target)` + ')')
				cinematic = 1
		elif string.lower(string.lstrip(line)[0:len('teleport all')]) == 'teleport all':
			success, teleport_class, target = m25string.SplitAroundPhrase(string.lstrip(line[len('teleport all'):]), 'to')
			if success:
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
				cs['cutscene'].append('CS_TeleportAllTo(%s, "%s")'%(PStory_Translate_SpecialRequired(teleport_class), target))
				cinematic = 1
		elif string.lower(string.strip(line)) == 'shake camera':
			cs['cutscene'].append('CS_Shake()')
			cinematic = 1 
		elif string.find(string.lower(line),'camera') != -1:
			if string.find(string.lower(line),'cinematic') != -1:
				temp = m25string.SplitAroundPhrase(line, 'Cinematic camera on')
				actor,target = '', ''
				if temp[0] == 1:
					camera_data = temp[2]
					temp2 = m25string.SplitAroundPhrase(camera_data, ' to ')
					if temp2[0] == 1:
						actor = temp2[1]
						target = temp2[2]
					else:
						actor = camera_data
					if actor != '':
						cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
						if target != '':
							cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],target)
							cs['cutscene'].append('CS_CinematicCamera(' + `string.lower(actor)` + ',' + `string.lower(target)` + ')')
							cinematic = 1
						else:
							cs['cutscene'].append('CS_CinematicCamera(' + `string.lower(actor)` + ')')
							cinematic = 1
			else:
				temp = m25string.SplitAroundPhrase(line,'Camera on')
				if temp[0] == 1:
					camera_data = temp[2]
					actor, distance, yaw = PStory_GetCameraCommands(camera_data)
					cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
					if type(actor) == types.StringType:
						actor = string.lower(actor)
					cs['cutscene'].append('CS_Camera(%s, dist=%s, yaw=%s)'%(repr(actor),repr(distance),repr(yaw)))
		
					#cs['cutscene'].append('CS_Camera(' + `string.lower(actor)` + ', dist='+`distance` + ', yaw='+`yaw` + ')')
					cinematic = 1
		elif string.find(string.lower(line),'plays animation') != -1:
			temp = m25string.SplitAroundPhrase(line,'plays animation')
			if temp[0] == 1:
				actor = temp[1]
				anim = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['cutscene'].append('CS_Animate(' + `string.lower(actor)` + ',' + `anim` + ')')
				cinematic = 1
		elif string.lower(string.strip(line)) == 'revive all heroes':
			cs['cutscene'].append('CS_ReviveAllHeroes()')
		elif string.find(string.lower(line),'is revived') != -1:
			temp = m25string.SplitAroundPhrase(line,'is revived')
			if temp[0] == 1:
				actor = temp[1]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['cutscene'].append('CS_Revive(' + `string.lower(actor)` + ')')
		elif string.find(string.lower(line),'is killed') != -1:
			temp = m25string.SplitAroundPhrase(line,'is killed')
			if temp[0] == 1:
				actor = temp[1]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['cutscene'].append('CS_Kill(' + `string.lower(actor)` + ')')
		elif string.find(string.lower(line),'is destroyed') != -1:
			temp = m25string.SplitAroundPhrase(line,'is destroyed')
			if temp[0] == 1:
				actor = temp[1]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['cutscene'].append('CS_Destroy(' + `string.lower(actor)` + ')')
		elif string.find(string.lower(line),'is hurt for') != -1:
			temp = m25string.SplitAroundPhrase(line,'is hurt for')
			if temp[0] == 1:
				actor = temp[1]
				damage = temp[2]
				cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'],actor)
				cs['cutscene'].append('CS_Hurt(' + `string.lower(actor)` + ',' + damage + ')')
		elif string.find(string.lower(line),'arrow') != -1:
			temp = m25string.SplitAroundPhrase(line,'arrow on')
			if temp[0] == 1:
				arrow_type = None
				if string.lower(temp[1]) == 'red':
					arrow_type = 'js.ARROW_RED'
				elif string.lower(temp[1]) == 'yellow':
					arrow_type = 'js.ARROW_YELLOW'
				else:
					if string.lower(temp[1]) == 'remove':
						object = temp[2]
						cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'], object)
						cs['cutscene'].append('CS_RemoveArrow(' + repr(string.lower(temp[2])+'arrow') + ')')
					
				if arrow_type != None:
					object = temp[2]
					cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'], object)
					cs['cutscene'].append('CS_AddArrow('+repr(string.lower(temp[2])+'arrow') + ', ' + `string.lower(temp[2])` + ', arrow_type=' + arrow_type + ')')
			arrownum = arrownum + 1
		elif string.lower(string.lstrip(line)[0:len('wait for')]) == 'wait for':
			temp = m25string.SplitAroundPhrase(string.lstrip(line[len('wait for'):]), 'seconds')
			if temp[0] == 1:
				temp2 = m25string.SplitAroundPhrase(temp[1],'to')
				if temp2[0]:
					cs['cutscene'].append('CS_Wait(%s,%s)'%(temp2[1],temp2[2]))
				else:
					seconds = temp[1]
					cs['cutscene'].append('CS_Wait(' + seconds + ')')
				#cinematic = 1
		elif string.lower(string.lstrip(line)[0:len('fade for')]) == 'fade for':
			temp = m25string.SplitAroundPhrase(string.lstrip(line[len('fade for'):]), 'seconds')
			if temp[0] == 1:
				seconds = temp[1]
				cs['cutscene'].append('CS_Fade(' + seconds + ')')
				cinematic = 1
		elif string.lower(string.lstrip(line)[0:len('unfade for')]) == 'unfade for':
			temp = m25string.SplitAroundPhrase(string.lstrip(line[len('unfade for'):]), 'seconds')
			if temp[0] == 1:
				seconds = temp[1]
				cs['cutscene'].append('CS_UnFade(' + seconds + ')')
				cinematic = 1
		elif string.lower(string.lstrip(line)[0:len('fade in')]) == 'fade in':
			cs['cutscene'].append('CS_UnFade(2)')
		elif string.lower(string.lstrip(line)[0:len('fade out')]) == 'fade out':
			cs['cutscene'].append('CS_Fade(2)')
		elif string.lower(string.lstrip(line)[0:len('fade')]) == 'fade':
			cs['cutscene'].append('CS_Fade(2)')
		elif string.lower(string.lstrip(line)[0:len('unfade')]) == 'unfade':
			cs['cutscene'].append('CS_UnFade(2)')
		elif string.lower(string.lstrip(line)[0:len('play transition')]) == 'play transition':
			cs['cutscene'].append('CS_Transition()')
			cinematic = 1
		elif string.lower(string.lstrip(line)[0:len('set lighting to')]) == 'set lighting to':
			temp = m25string.SplitAroundPhrase(string.lower(string.strip(line)),'to')
			if temp[0] == 1:
				lightingvar = temp[2]
				cs['cutscene'].append('CS_Lighting(' + `lightingvar` + ')')
		elif string.lower(string.lstrip(line)[0:len('play music')]) == 'play music':
			temp = m25string.SplitAroundPhrase(string.strip(line),'play music')
			if temp[0] == 1:
				music = temp[2]
				cs['cutscene'].append('CS_Music(' + `music` + ')')
		elif string.lower(string.lstrip(line)[0:len('play sound')]) == 'play sound':
			temp = m25string.SplitAroundPhrase(string.strip(line),'play sound')
			if temp[0] == 1:
				param = temp[2]
				vals = m25string.SplitAroundPhrase(param,' at ')
				if vals[0] == 1:
					sound = vals[1]
					location = vals[2]
					cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'], location)
					cs['cutscene'].append('CS_Sound(' + `sound` + ',' + string.lower(`location`) + ')')
		elif string.lower(string.lstrip(line)[0:len('play effect')]) == 'play effect':
			temp = m25string.SplitAroundPhrase(string.strip(line),'play effect')
			if temp[0] == 1:
				param = temp[2]
				vals = m25string.SplitAroundPhrase(param,' at ')
				if vals[0] == 1:
					effect = vals[1]
					location = vals[2]
					cs['replace'],cs['required'] = PStory_GetReplace(cs['replace'],cs['required'], location)
					cs['cutscene'].append('CS_Effect(' + `effect` + ',' + `string.lower(location)` + ')')
		else:
			pass
	#we get here if the file ends unexpectantly - wrap up the cutscene
	if cinematic:
		new_cs = ['CS_Start()']
		for c in cs['cutscene']:
			new_cs.append(c)
		#new_cs.append('CS_Wait(2)')
		new_cs.append('CS_End()')
		cs['cutscene'] = new_cs
	return cs
	
pstory_special_replace = {
	'villains': '__allvil__',
	'minions': '__allmin__',
	'villains and minions': '__allvilmin__',
	'minions and villains': '__allvilmin__',
	'heroes': '__allherbud__',
	'allies': '__allall__',
}
	
def PStory_Translate_SpecialRequired(value):
	global pstory_special_replace
	if value in pstory_special_replace.keys():
		return pstory_special_replace[value]
	else:
		return value
	

	
def PStory_AddToReplaceList(value):
	global pstory_replacements_list
	if value not in pstory_replacements_list:
		pstory_replacements_list.append(value)

def PStory_GetReplace(replace, required, actor, force=0):
	global pstory_replacements_list
	if type(actor) != types.StringType:
		return [replace, required]
	replacements = pstory_replacements_list
	found = 0
	for r in replacements:
		if (string.find(string.lower(actor),r) != -1):
			found = 1
			if ([string.lower(actor), r] not in replace):
				replace.append([string.lower(actor),r])
	if (not force) and (not found) and (string.lower(actor) not in ['random','narrator']) and (actor not in required):
		required.append(string.lower(actor))
	return [replace, required]

			
def PStory_GetCameraCommands(camdata):
	#get actor
	actor = ''
	commandlist = ['zoom', 'heading']
	distance = -160
	yaw = 0
	at_commands = 0
	words = string.split(camdata)
	num = 0
	for c in words:
		c = string.lower(c)
		if c in commandlist:
			at_commands = 1
			if c == 'zoom':
				distance = string.atof(words[num+1])
			if c == 'heading':
				yaw = string.atof(words[num+1])
		else:
			if not at_commands:
				if actor != '':
					actor = actor + ' ' + c
				else:
					actor = c
		num = num + 1
	#check to see if it is a tuple instead of a name
	if actor != '' and actor[0] == '(':
		exec 'actor = PStory_EvalString(' + actor + ')'
	return [actor,distance,yaw]
				
	
def PStory_IsTraitSpeech(line):
	success, leftstring, remainder = m25string.SplitAroundPhrase(line, '"')
	if not success: 
		return 0
	if string.find(string.lower(leftstring),' if ') != -1: 
		return 1
	return 0
	
def PStory_GetTraitSpeech(line):
	done = 0
	speechlist = []
	defaultspeech = 'The mod maker forgot to put in a default speech line.'
	success, actor, remainder = m25string.SplitAroundPhrase(line, ' says,')
	if not success: return [None, None, None]
	line = remainder
	while not done:
		name, trait, speech = None, None, None
		if string.strip(string.lower(line))[:3] == 'if ':
			success, leftstring, rightstring = m25string.SplitAroundPhrase(line, 'if ')
			if success: success, name, rightstring = m25string.SplitAroundPhrase(rightstring, ' is ')
			if success: success, trait, rightstring = m25string.SplitAroundPhrase(rightstring, ',')
			if success: speech = m25string.GetRightString(string.strip(rightstring)[:-1],'"')
			if name and trait and speech:
				speechlist.append([string.lower(name),string.lower(trait),speech])
				line = PStory_LookAhead()
		elif string.strip(string.lower(line))[0] ==	'"':
			defaultspeech = m25string.GetRightString(string.strip(line)[:-1],'"')
			done = 1
		else:
			PStory_PushBack()
			done = 1
	return [actor, speechlist, defaultspeech]

	
	
#####################################################################################
#
# Tempform to Story translator
# 
# Description: 
#
#####################################################################################	
	
#-------------------------------------------------------------------
#
# Translate from temporary form into story structure that can be
# interpreted by the m25story.py module
#
#-------------------------------------------------------------------	

precache = []

def PStory_AddToPrecache(var):
	global precache
	if type(var) == types.ListType:
		if not (var[1] in precache): precache.append(var[1])
	else:
		if not (var in precache): precache.append(var)

def Pstory_DoPrecache():
	global precache
	#precache any needed templates
	for p in precache:
		try:
			js.Mission_PrecacheTemplateObj(p, 0, 0)
			m25report.Report('PSTORY', 'precaching %s'%(p))
		except:
			m25report.Report('PSTORY', 'unable to precache %s'%(p))		
		
			
pstory_groups_list = ['allies','objects','heroes', 'interlopers', 'rivals', 'cohorts', 'switches', 'vehicles']

def Pstory_AddToGroupsList(value):
	global pstory_groups_list
	if value not in pstory_groups_list:
		pstory_groups_list.append(value)		

def PStory_TranslateStory(story):
	global pstory_groups_list
	#precache = []
	
	truestory = {}
	encdict = {}
	encdict['encounters'] = {}
	
	start = 1
	num = 0
	for enc in story['encounters']:
		temp = {}
		
		#determine the trigger
		
		if story['start'] != []:
			if enc['name'] in story['start']:
				temp['trigger'] = ['start']
		else:
			if start:
				temp['trigger'] = ['start']
				start = 0
				
		if enc.has_key('trigger') and (enc['trigger'] != []):
			temp['trigger'] = enc['trigger']
			
		if enc.has_key('pre') and (enc['pre'] != []):
			temp['pre'] = enc['pre']
		
		#determine the game type (for freeroam missions)
		if enc.has_key('required_map'):
			temp['required_map'] = enc['required_map']
			
		if enc.has_key('timeofday'):
			temp['timeofday'] = enc['timeofday']
			
		#determine action
		fn = ''
		villainlist = []
		minionlist = []
		params = ''
		cutscenes = 'cutscenes=[]'
		minions = ''
		villains = ''
		marker = 'marker = encounter'
		objective = ''
		for key in enc.keys():
			if key == 'villains':
				villainlist = PStory_TranslateCharacterType(PStory_AddToList(enc['villains'],villainlist))
				for v in villainlist:
					PStory_AddToPrecache(v)
					#if not v in precache: precache.append(v)
			elif key == 'minions':
				minionlist = PStory_TranslateCharacterType(PStory_AddToList(minionlist, enc['minions']))
				for v in minionlist:
					PStory_AddToPrecache(v)
					#if not v in precache: precache.append(v)
			elif key == 'type':
				# check the custom list to see if there is a custom expansion for this type of encounter
				custom_key = string.strip(string.lower(enc['type']))
				if m25enccustom.ENC_CUSTOM_TYPES.has_key(custom_key):
					fn = 'DoEncCustom('
					if params == '':
						params = 'actions="%s"'%(m25enccustom.ENC_CUSTOM_TYPES[custom_key])
					else:
						params = ', actions="%s"'%(m25enccustom.ENC_CUSTOM_TYPES[custom_key])
				else:
					fn = 'DoEnc' + PStory_RemoveSpaces(string.capitalize(enc['type']) + '(')
			elif key == 'cutscenes':
				cutscenes = 'cutscenes = ' + `PStory_TranslateCutscenes(enc['cutscenes'])`
			elif key == 'marker':
				marker = 'marker = ' + `enc['marker']`
			elif key == 'objectives':
				if enc['objectives'] != []:
					objective = 'objectives = ' + `enc['objectives']`
			elif key == 'parameters':
				for p in enc['parameters']:
					if params != '': params = params + ', '
					if p[0] in pstory_groups_list:
						params = params + p[0] + ' = ' + `PStory_CommaNamedList(p[1])`
						for v in PStory_CommaNamedList(p[1]):
							PStory_AddToPrecache(v)
							#if not v in precache: precache.append(v)
					else:
						params = params + p[0] + ' = ' + `p[1]`
			else:
				pass
		#build the function call
		#skip if encounter type is None
		if fn != 'DoEncNone(':
			if fn == '': fn = 'DoEncFight('
			if villainlist != []: villains = 'villains = ' + `villainlist`
			if minionlist != []: minions = 'minions = ' + `minionlist`
			usecomma = 0
			if villains != '':
				fn = fn + villains
				usecomma = 1
			if minions != '':
				if usecomma:
					fn = fn + ', ' + minions
				else:
					fn = fn + minions
					usecomma = 1
			if cutscenes != '':
				if usecomma:
					fn = fn + ', ' + cutscenes
				else:
					fn = fn + cutscenes
					usecomma = 1
			if marker != '':
				if usecomma: fn = fn + ', ' + marker
				else:
					fn = fn + marker
					usecomma = 1
			if objective != '':
				if usecomma: fn = fn + ', ' + objective
				else:
					fn = fn + objective
					usecomma = 1
			if params != '':
				if usecomma: fn = fn + ', ' + params
				else: fn = fn + params
			fn = fn + ')'
			temp['action'] = [fn]
		
		#determine next encounter
		if enc.has_key('next') and enc['next'] != []:
# 			if string.lower(enc['next'][0]) == 'end':
# 				temp['next'] = ['__end__']
# 			elif string.lower(enc['next'][0]) == 'lose':
# 				temp['next'] = ['__lose__']
# 			elif string.lower(enc['next'][0]) == 'win':
# 				temp['next'] = ['__win__']
# 			else:
# 				temp['next'] = enc['next']
			temp['next'] = enc['next']
		else:
			if enc != story['encounters'][-1]:
				temp['next'] = [story['encounters'][num+1]['name']]
			else:
				temp['next'] = ['__end__']
	
		#put the completed entry into the dictionary
		encdict['encounters'][enc['name']] = temp
		num = num + 1
		
	#create the end state so that the mission ends properly when all encounters are completed
	#if m25gametype.GameType_IsFreeRoam():
		#don't win the mission in freeroam campaigns - it messes up the technology
	encdict['encounters']['__end__'] = {'action':['ff.Story_SetStoryEnded('+`story['name']`+')']}
	encdict['encounters']['__lose__'] = {'action':['ff.Story_LoseMission()']}
	encdict['encounters']['__win__'] = {'action':['ff.Story_WinMission()']}
	#else:
	#	encdict['encounters']['__end__'] = {'action':['ff.SetMissionOver()','Mission_Win()']}	
	
	
	#put it all together
	truestory[story['name']]=encdict
	
	m25report.Report('PSTORY',repr(truestory))
	return truestory
	
	
def PStory_TranslateCharacterType(charlist):
	global pstory_has_random
	newlist = []
	for c in charlist:
		if type(c) == types.StringType:
			val, command, chartype = m25string.SplitAroundPhrase(c, ' ')
			if string.lower(command) == 'random':
				pstory_has_random = 1
				if val:
					newlist.append('Story_GetRandomUserEnemy("%s")'%(chartype))
				else:
					newlist.append('Story_GetRandomUserEnemy()')
			else:
				newlist.append(c)
		else:
			newlist.append(c)
	return newlist
	
def PStory_TranslateCutscenes(cutscenes):
	cscomplete = []
	for cs in cutscenes:
		if cs != None:
			if type(cs) == types.StringType:
				if cs[0:len('no_randomize_')] == 'no_randomize_':
					cscomplete.append(cs)
					continue
			cslist = []
			ingamename = string.lower(cs['name']) + 'cs'
			cslist.append(PStory_AddToList(['priority'],cs['priority']))
			if len(cs['replace']) >= 1:
				cslist.append(PStory_AddToList(['replace'], cs['replace']))
			if len(cs['required']) >= 1:
				cslist.append(PStory_AddToList(['required'], cs['required']))
			cslist.append(PStory_AddToList(['cutscene'], cs['cutscene']))
			
			#check to see if there is already a cutscene with the same name
			if cscomplete != []:
				num = 0
				for c in cscomplete:
					if c[0] == ingamename:
						#check to see if there are already multiple cutscenes in the list
						cscomplete.remove(c)
						if c[1][0] == 'cslist':
							cslist = PStory_AddToList(c[1],[cslist])
						else:
							cslist = ['cslist',c[1],cslist]
					num = num + 1
							
			cscomplete.append([ingamename,cslist])
	return cscomplete
	


	
#####################################################################################
#
# Handle Captions.txt to provide support for talking heads
# 
# Description: 
#
#####################################################################################	
	
#add things to the captions.txt

pstory_captions = []
pstory_captionnum = 1

def PStory_ResetCaptions():
	global pstory_captions
	global pstory_captionnum
	pstory_captions = []
	pstory_captionnum = 1

def PStory_AddCaption(text):
	global pstory_captionList
	global pstory_reverseCaptionList
	global pstory_captions
	global pstory_captionnum
	PStory_getCaptions()
	if not PStory_checkReverseCaption(text):
		num = pstory_captionnum
		done = 0
		name = ''
		while not done:
			name = storycut + `num` + '_01'
			#name = 'm25tag' + `num` + '_01'
			if not PStory_checkCaption(name):
				done = 1
			num = num + 1
		pstory_captions.append(name + ', neutral, ' + text + '\n')
		pstory_captionList.append(name)
		pstory_reverseCaptionList[string.strip(text)] = name
		pstory_captionnum = num
	
def PStory_WriteCaptions():
	global pstory_captions
	m25report.Report('PSTORY', 'writing captions file.')
	#print pstory_captions
	modpath = datfiles.GetModPath()
	fname = open(modpath+'\Lang\English\captions.txt','a')
	if pstory_captions != []:
		fname.write('\n')
	for p in pstory_captions:
		fname.write(p)
	fname.close()
			
		
def PStory_getCaptions():
	global pstory_captionList
	global pstory_reverseCaptionList
	
	if 'pstory_captionList' in globals().keys():
		if pstory_captionList != []: return
	
	pstory_captionList=[]
	pstory_reverseCaptionList = {}
	modpath = datfiles.GetModPath()
	
	try:
		fname = open(modpath+'\Lang\English\captions.txt','r')
	except:
		m25report.Report('ERROR', 'PStory_getCaptions unable to open \Lang\English\captions.txt')
		return
		#file doesn't exist, so let's make a copy
# 		try:
# 			shutil.copy(modpath+'\Lang\English\captions.txt',modpath+'\Lang\captions.txt')
# 			shutil.copy(modpath+'\Lang\English\speech.txt',modpath+'\Lang\speech.txt')
# 			shutil.copy(modpath+'\Lang\English\strings.txt',modpath+'\Lang\strings.txt')
# 			fname = open(modpath+'\Lang\captions.txt','r')
# 		except:
# 			m25report.Report('ERROR', 'Pstory_getCaptions unable to copy captions, strings and speech.txt files.')
# 			m25report.Report('ERROR', 'Make sure these files are in <yourmod>\Lang\English')
# 			return
	
	line = fname.readline()
	while line != '':
		tag = m25string.GetLeftString(line,',')
		if tag != '\n':
			pstory_captionList.append(tag)
			pstory_reverseCaptionList[string.strip(m25string.GetRightString(m25string.GetRightString(line,','), ','))] = tag
		line = fname.readline()
	fname.close()

def PStory_checkCaption(tag):
	global pstory_captionList
	if pstory_captionList == []:
		PStory_getCaptions()
	if tag in pstory_captionList:
		return 1
	return 0

def PStory_checkReverseCaption(tag):
	global pstory_reverseCaptionList
	if pstory_reverseCaptionList.has_key(tag):
		return pstory_reverseCaptionList[tag]
	return None		
		


#####################################################################################
#
# Helper functions
# 
# Description: 
#
#####################################################################################	
	
def PStory_IsBlankLine(line):
	return (string.strip(line) == '')	
	
def PStory_AddToList(vlist, value):
	newlist = vlist
	if type(vlist)  == types.ListType:
		newlist = vlist[0:]
		del vlist
		if type(value) == types.ListType:
			for v in value:
				newlist.append(v)
		else:
			newlist.append(value)
	elif vlist == 'none' or vlist == None:
		newlist = value
	else:
		if type(value) == types.ListType:
			newlist = value[0:]
			del value
			newlist.append(vlist)
		else:
			newlist = [vlist,value]
	return newlist	
	
def PStory_CommaToList(line):
	templist = string.split(line,',')
	for index in range(0,len(templist)):
		templist[index] = string.strip(templist[index])
	return templist

def PStory_CommaNamedList(line):
	templist = PStory_CommaToList(line)
	newlist = []
	for t in templist:
		plist = m25string.SplitAroundPhrase(t, 'named')
		if plist[0] == 1:
			newlist.append([plist[2],plist[1]])
		else:
			newlist.append(t)
	return newlist
	
def PStory_RemoveSpaces(line):
	whitespace = ['\n',' ','\t']
	newline = ''
	for c in line:
		if c in whitespace:
			pass
		else:
			newline = newline + c
	return newline	
		
def PStory_EvalString(string):
	return string	
