# -*- coding: latin-1 -*-

######################################################################
## 
## FFX Programming Utilities
## By pimthe
## 
######################################################################
print "Starting FFX Programming Utilities 1.3"

# DESCRIPTION
# A small collection of helper functions related to debugging FFX Python scripts.

# Feel free to change or add to the code here. Just make sure you maintain compatibility,
# so external modules which could be using these functions keep working. :)

# DEPENDENCIES
# - FFX 3.2 +

# VERSION HISTORY
# 1.0 2006-08-16:
# - Added debugging helper functions previously found in ffx.py: startTimer(), FFQ_GetAttrForAll()
#   and variants, whoIsThat(), startFX(), printRemovableAttributes() and getRemovableAttributes().
# - Minor update to FFQ_GetAttrForAll() and variants and to getRemovableAttributes().
# - Added documentPyFile(), to generate preliminary documentation from a Python file.
# - First release.
# 1.1 2006-08-22:
# - Added OnEventTest(event).
# - Minor changes to startTimer()
# 1.2 2007-02-17:
# - Added encoding (cf. first line) for compatibility with Python 2.5.
# - Updated documentPyFile():
#   - use inline CSS only if it is specificed that there is no external CSS file
#   - removed first-line encoding comment from HTML docs
#   - added basic support for HTML entities
# - Improved documentation.
# 1.3 2007-03-17:
# - Replaced manual filename extraction from path with os.path.split()
# - Updated the import for when outside the game
# - Changed HTML extension to .htm for consistency with the FFX Manual. :/
# 1.4 2008-04-23:
# - normalized all line endings to CR/LF.

if __name__ != '__main__':  #from the game engine or another interpreter
    try:
        import js
        from js import *
        print "FFX Programming Utilities called from the game."
        try:
            import cshelper 
            from cshelper import *
        except:
            print "FFX Programming Utilities error:"
            print "    Cannot import 'cshelper' module"
        try:
            import ffx
            from ffx import *
        except:
            print "FFX Programming Utilities error:"
            print "    Cannot import 'ffx' module"
    except:
        print "FFX Programming Utilities running outside the game."

try:
    import sysutils
except:
    print "FFX Programming Utilities error:"
    print "    Cannot import 'sysutils' module"

import string
import os



######################################################################
## 
## Events
## 
######################################################################

# A very simple timer, displaying the countdown in the console and, optionally,
# call immediately a function. A sound will play when the time expires.
# Requires a from progrutils import *
# Works only with integer values for the seconds.
def startTimer(countdown = 10, increment = 1, verbose = 1, funct = ''):
    print "-----------------------------------"
    if funct != '':
        print "... startTimer() calling function '%s'"%funct
        exec funct
    print "... startTimer()... the countdown begins at %s"%(countdown)
    countdown = int(abs(countdown)) - 1
    RegTimer('loopTimer',increment,countdown,str(verbose))

def loopTimer(event):
    increment = event.float
    countdown = int(event.user)
    verbose = int(event.object)
    if verbose:
        print "... startTimer()... tic-tac...  %s"%(countdown)
    countdown = countdown - 1
    if countdown == 0:
        print "... startTimer()... the time expired"
        print "-----------------------------------"
        Sound_PlaySound('ui_move_splat')
    else:
        RegTimer('loopTimer',increment,countdown,str(verbose))


#For testing results of an event function call. This will print out the value of various event parameters.
def OnEventTest(event):
    print "-------------------------"
    print "OnEventTest(event) called"
    print "    event.event = ",
    try:
        print event.event
    except:
        "?"
    print "    event.data = ",
    try:
        print event.data
    except:
        "?"
    print "    event.object = ",
    try:
        print event.object
    except:
        "?"
    print "    event.string = ",
    try:
        print event.string
    except:
        "?"
    print "    event.user = ",
    try:
        print event.user
    except:
        "?"
    print "    event.float = ",
    try:
        print event.float
    except:
        "?"


######################################################################
## 
## Object attributes
## 
######################################################################

# attribute tests
# get the value of a selected object attribute for everyone on the level
# use with care, since if the attribute is not set for a character, FF will pop an error message box!
def FFQ_GetAttrForAll(whichAttribute):
    print "--Attribute to check=%s"%(whichAttribute)
    for obj in getAllCharacters():
        print "----I am=%s"%(obj)
        try:
            print "-----my attribute's value is=%s"%(Object_GetAttr(obj,whichAttribute))
        except:
            print "-----my attribute's value is=not set!"

# set the value of a selected attribute for everyone on the level
def FFQ_SetAttrForAll(whichAttribute,newValue):
    print "--Attribute to check=%s"%(whichAttribute)
    for obj in getAllCharacters():
        print "----I am=%s"%(obj)
        try:
            print "-----my attribute's old value was=%s"%(Object_GetAttr(obj,whichAttribute))
            attributeHolder="attrHolder-" + whichAttribute
            Object_SetAttr(obj,attributeHolder,Object_GetAttr(obj,whichAttribute))
            newValue=float(newValue)
            Object_SetAttr(obj,whichAttribute,newValue)
            print "------my attribute's new value is=%s"%(Object_GetAttr(obj,whichAttribute))
        except:
            print "-----my attribute's value is=not set!"

# set the value of a selected attribute to a random value for everyone on the level
def FFQ_SetAttrForAllRandom(whichAttribute,rangeStart,rangeEnd):
    print "--Attribute to check=%s"%(whichAttribute)
    for obj in getAllCharacters():
        print "----I am=%s"%(obj)
        try:
            print "-----my attribute's old value was=%s"%(Object_GetAttr(obj,whichAttribute))
            attributeHolder="attrHolder-" + whichAttribute
            Object_SetAttr(obj,attributeHolder,Object_GetAttr(obj,whichAttribute))
            newValue=randint(rangeStart,rangeEnd)
            newValue=float(newValue)
            Object_SetAttr(obj,whichAttribute,newValue)
            print "------my attribute's new random value is=%s"%(Object_GetAttr(obj,whichAttribute))
        except:
            print "-----my attribute's value is=not set!"

# restore the value of a selected attribute for everyone on the level
def FFQ_RestoreAttrForAll(whichAttribute):
    print "--Attribute to check=%s"%(whichAttribute)
    for obj in getAllCharacters():
        print "----I am=%s"%(obj)
        print "-----my attribute's temporary value was=%s"%(Object_GetAttr(obj,whichAttribute))
        attributeHolder="attrHolder-" + whichAttribute
        try:
            Object_SetAttr(obj,whichAttribute,Object_GetAttr(obj,attributeHolder))
        except:
            print "was the attribute named '%s' changed before?"%(whichAttribute)
        print "------my attribute's true value is=%s"%(Object_GetAttr(obj,whichAttribute))



######################################################################
## 
## Information
## 
######################################################################

#Based on the WHO AM I idea used in FFX 1.3
#Version 1.0
#Updated FFX 2.5
#Attach a "WHO IS" custom command to every hero. When using the custom command on an object, info is printed out to the console/script.log.
def whoIsThat(onoffvalue=1):
    if onoffvalue == 1:
        for hero in GetAllHeroes():
            Mission_CustomAction('CUSTOM_WHOAMI',hero,'','FFQ_OnWho',500,0)
    if onoffvalue == 0:
        for hero in GetAllHeroes():
            Mission_RemoveCustomAction('CUSTOM_WHOAMI',hero,'')
    print "FFQ_whoIsThat(number):"
    print "value should be 1 to activate, 0 to disactivate."

def FFQ_OnWho(target,dummy):
    objclass = Object_GetClass(target)
    thisIs='object is '
    if (objclass & FFX_CHARACTER):
        thisIs=thisIs+'a character '
        print 'This is %s of template %s (%s)'%(target,FFX_GetTemplate(target),Object_GetTemplate(target))
        templateFFXAttr = getFFXAttributesForTemplate(target)
        if len(templateFFXAttr) > 0:
            print "its template's FFX attributes are:"
            print templateFFXAttr
        charFFXAttr = getFFXAttributesForCharacter(target)
        if len(charFFXAttr) > 0:
            print "its character-specific FFX attributes are:"
            print charFFXAttr
    elif (objclass & OC_POLICE):
        thisIs=thisIs+'a police '
    elif (objclass & OC_CIVILIAN):
        thisIs=thisIs+'a civilian '
    elif (objclass & OC_MINION):
        thisIs=thisIs+'a minion '
    elif (objclass & OC_VILLIAN):
        thisIs=thisIs+'a villain '
    elif (objclass & OC_VEHICLE):
        thisIs=thisIs+'a vehicule '
    elif (objclass & OC_BUILDING):
        thisIs=thisIs+'a building '
    elif (objclass & OC_TERRAIN):
        thisIs=thisIs+'a terrain '
    elif (objclass & OC_POWERUP):
        thisIs=thisIs+'a powerup '
    else:
        thisIs=thisIs+'something else (a prop) of type %d'%(objclass)
    if (objclass & OC_CONTROLLABLE):
        thisIs=thisIs+'and is player-controlled'
    print thisIs
    if (objclass & FFX_CHARACTER):
        print "My gender is ",
        print getGender(target)
    currentPos = Get_ObjectPos(target)
    print 'currentPos ='
    print currentPos


######################################################################
## 
## Object manipulation
## 
######################################################################

# a simple method to quickly test effects in-game
def startFX(char, sparks = "effect_ffxabsorbstrength", pos = "Bip01 Spine"):
    print 'format = startFX(char, sparks = "effect_ffxabsorbstrength", pos = "Bip01 Spine")'
    if not pos == '':
        if sparks != '':
            handle = Object_PlayEffect(char,sparks,'',FX_LOOP|FX_TRACK_OBJECT_POS,0,pos)
            Object_SetAttr(char,'ffqstartFX-handle',handle)
    else:
        if sparks != '':
            handle = Object_PlayEffect(char,sparks,'',FX_LOOP|FX_TRACK_OBJECT_POS,0)
            Object_SetAttr(char,'ffqstartFX-handle',handle)

def stopFX(char):
    handle = Object_GetAttr(char,'ffqstartFX-handle')
    Object_StopEffect(char,handle)



######################################################################
## 
## Character attributes
## 
######################################################################

# (Can be run outside the game)
# Prints to script.log (or to the console if used outside the game) a list of which FFX attributes are removable and which are not.
# added FFX 3.2.0 alpha
def printRemovableAttributes():
    removableAttribs = getRemovableAttributes(remove = 1)
    nonRemovableAttribs = getRemovableAttributes(remove = 0)
    removableAttribs.sort()
    nonRemovableAttribs.sort()
    print "********************************************"
    print "  %-2i Removable attributes" % len(removableAttribs)
    print "********************************************"
    for a in removableAttribs:
        print a
    print "********************************************"
    print "  %-2i Non-removable attributes" % len(nonRemovableAttribs)
    print "********************************************"
    for a in nonRemovableAttribs:
        print a

# (Can be run outside the game)
# Returns a list of removable (or non-removable if 'remove' parameter set to 0) FFX attributes.
# Note that it is quite slow, as it need to parse thousands of items; too bad we can't use inspect, introduced in Python 2.1...
# Notes/bugs:
# - A function whose name starts with 'init' and which is all in lowercase will falsely register as an FFX attribute.
# added FFX 3.2.0 alpha build 0, updated progrutils 1.0 to be run from outside the game.
def getRemovableAttributes(remove = 1, basePath = "C:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\ffx3"):
    ffxAttribs = []
    currentAttr = ''
    try:
        try:
            path = '%s\\missions\\scripts/ffx.py' % sysutils.getModFolder()
        except:
            path = '%s\\missions\\scripts/ffx.py' % basePath
        f = open(path, 'r')
        name = f.readline()
        while name != '':
            name = f.readline()
            if (name[0:4] == 'def ') and (name[4:8] == 'init'): # weirdest bug ever: FFX Control Centre crashes if using the more straightforward version %P
                #quick trick so bolean parameter is applied in conjunction with the -1 (not found) result of string.find()
                if string.find(name, 'remove') - remove > -2:
                    thisAttrib = string.split(name, '(')[0]
                    thisAttrib = thisAttrib[8:]
                    if thisAttrib == string.lower(thisAttrib): #mixed-case init... functions are never FFX attributes, so remove them
                        ffxAttribs.append(thisAttrib)
        f.close()
    except:
        print 'getRemovableAttributes() fatal error:'
        print '    cannot open %s' % path
    return ffxAttribs


######################################################################
## 
## Documentation
## 
######################################################################

# (Can be run outside the game)
# Create a rough plain text or HTML file listing top-level functions, variables and classes, plus the top-level #-style comments above them.

# How it works:
# The function reads your source .py file and looks for the following items: 
# - Section Headers: Any block starting  a number sign ("#") followed by at least four identical characters (which can be "#" too).
# - Object Class Definition: a line beginning with the word "class ". Comments above this line are considered to be related to the current class.
# - Function Definition: a line beginning with the word "def ". Comments above this line are considered to be related to the current function.
# - Variable Definition: a line beginning with a non-reserved word. Comments above this line are considered to be related to the current variable.

# Example:
# From an external console or the FF built-in one:
# >>> import sys
# >>> sys.path.append("C:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\ffx3\\Missions\\Scripts")
# Starting FFX Programming Utilities 1.3
# FFX Programming Utilities running outside the game.
# <module 'progrutils' from 'C:\Program Files\Irrational Games\Freedom Force vs The 3rd Reich\ffx3\Missions\Scripts\progrutils.py'>
# >>> progrutils.documentPyFile("C:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\ffx3\\Missions\\Scripts\\myfile.py")
# That would create a file named myfile_doc.htm under C:\Program Files\Irrational Games\Freedom Force vs The 3rd Reich\ffx3\Missions\Scripts\.
# Moving the file to C:\Program Files\Irrational Games\Freedom Force vs The 3rd Reich\ffx3\Manual\ would have it use the manual.css formatting
# style used in the FFX manual.

# Notes/bugs/limitations/etc.:
# - This function attempts to keep things very basic, especially since each FFX coder has a very different coding style. If you need something
#   fancier, look into the PyDoc module, which is part of the Python distro. Another option might be the third-party Doxygen (http://www.stack.nl/~dimitri/doxygen/).
# - The output file WILL need to be edited manually afterward, as, for example, any comment before recognized code is supposed to belong
#   to it.
# - Second-level functions (i.e., inside classes), variables and comments aren't supported.
#   Exception: Comments nested inside functions or classes will still be considered first-level if the # is put at the begining of the line
#   (as when using an editor's automatic commenting out feature) instead of following the current nesting level. Ex.:
#       def helloWorld():
#           print "hello, world!'
#           #print "indented comment"
#       ##    print "non-indented comment"
#   will output
#       helloWorld()
#       print "non-indented comment"
# - Triple quote-style comments aren't supported either, since they're not much used in FF/FFX scripts and detecting them correctly
#   would have required a wee bit more work. (Yes, I know that they are the correct way to document functions automatically...)
# - Only the first line of variables is recognized. Again, it would require more work; beside the value of documenting the content of
#   long lists, tupples or dictionnary outside the source code file itself is rather limited. In fact, variables are documented mostly
#   so the comments belonging to them can be recognized as such when editing the output file.
# - A sequence of four identical characters (except spaces or tabs) after the initial number-sign character(s) is considered as being
#   the start of a new section.
# - Only the most basic special characters are converted to HTML entities ("<", ">", "&" and some "\" variants), so don't expect to get valid
#   XHTML 1.1 syntax out of the box. :P
# - If there are special Python characters in your comments, they might be erroneously converted.
# - Multiple consecutive spaces are rendered as a single space (this is a feature of HTML). Indenting in commented-out code isn't preserved
#   either.
# - This function won't win any prize for coding elegance...

# arguments:
# - file: full path name of the source file
# - output: extension name-format, either 'txt' or 'htm'
# - css: if output is set to 'htm', source of the external cascading style sheet file to be linked to; if set to any value except '',
#   the line
#   <link rel="stylesheet" href="css filename" type="text/css" />
#   will be added to the HTML file header. It defaults to FFX 3.2 documentation's manual.css called from a subfolder.
# - txtSeparator: if output is set to 'txt', the text to be used to visually separate the various functions, classes and variables.
def documentPyFile(file, output = 'htm', css = '../manual.css', txtSeparator = '\n\n\n'):
    comments = []
    doc = []
    f = open(file, 'r')
    line = 'dummy'
    while line !='':
        line = f.readline()
        if len(line) < 1: #weird...  "while line !='':" should make that unnecessary!
            break
        #the line starts with a #-comment but doesn't seem to announce a new section
        if (line[0] == '#') and not \
            ( (len(line) > 5) and (line[1] == line[2] == line[3] == line[4]) and (line[1] not in (' ', '\t')) ):
            while line[0] in ('#', ' ', '\t',):
                line = line[1:]
            if output == 'htm':
                line = convertLineEntities(line[:-1])
                line = '%s<br />\n' % line
            comments.append(line)
        #the line is empty
        elif isLineEmpty(line):
            if output == 'htm':
                comments.append('<br />\n')
            else:
                comments.append('\n')
        #the line contains a function, class or variable definition or a new section indicator
        else:
            typeName = ''
            if line[0:6] == 'class ':
                typeName = 'className'
                line = line[:-2]
            elif line[0:4] == 'def ':
                typeName = 'functionName'
                line = line[4:-2]
            elif (line[0] not in ('#', ' ', '\t', '\n', '(', ')', '[', ']', '{', ']', '"', "'")) and \
                (line[0:3] != 'if ') and \
                (line[0:4] not in ('del ', 'for ', 'try:')) and \
                (line[0:5] not in ('from ', 'else:', 'elif ', 'exec ', 'eval ', 'eval(')) and \
                (line[0:6] not in ('while ', 'yield ', 'print ')) and \
                (line[0:7] not in ('lambda ', 'global ', 'import ', 'except:', 'except ')) and \
                (line[0:8] != 'finally:'):
                typeName = 'variableName'
            elif (line[0] == '#') and \
                (  (len(line) > 5) and (line[1] == line[2] == line[3] == line[4]) and (line[1] not in (' ', '\t')) ):
                typeName = 'blockName'
                while line[0] == '#':
                    line = line[1:]
                if  (len(line) > 4) and (line[0] == line[1] == line[2] == line[3]):
                    repeater = line[0]
                    while line[0] == repeater:
                        line = line[1:]
                while line[0] == ' ':
                    line = line[1:]
                if line == '\n':
                    if output == 'htm':
                        line = '<hr />'
                    else:
                        line =  '#####\n'
            #add the function, class, variable definition or new section indicator and append the comments above it
            if typeName != '':
                if ( len(comments) > 0 ) and ( (comments[0] == '\n') or (comments[0] == '<br />\n') ):
                    del comments[0]
                if output == 'htm':
                    if typeName == 'blockName':
                        line = '<h3>%s</h3>\n' % line
                    else:
                        line = '<h4 class="%s">%s</h4>\n' % (typeName, line)
                    i = 0
                    while i < len(comments):
                        if comments[i] == '<br />\n':
                            del comments[0]
                        else:
                            break
                        i = i + 1
                    if len(comments) > 0:
                        comments[0] = '<p>%s' % comments[0]
                        comments.append('</p>\n\n')
                else:
                    line = '%s%s' % (txtSeparator, line)
                doc.append(line)
                doc.extend(comments)
                comments = []
    f.close()
    #post-processing for html: cleanup and additional formatting
    if output == 'htm':
        #remove encoding declaration from HTML version
        if string.find(doc[1], 'coding:') > -1:
            del doc[1]
        i = 1
        while i < len(doc):
            if doc[i - 1] == '<h3><hr /></h3>\n':
                if (doc[i][:3] == '<p>') and (doc[i] != '<p><br />\n'):
                    doc[i] = doc[i][3:-7]
                    doc[i] = '<h3>%s</h3>\n<p>' % doc[i]
                    del doc[i - 1]
            elif doc[i:-2] == '</p>':
                previousLine = doc[i - 1]
                print previousLine
                if previousLine[-7:] == '<br />\n':
                    previousLine = '%s\n' % previousLine[:-7]
            i = i + 1
    #create the file from the 'doc' list
    docFile = "%s_doc.%s" % ( file[ : file.rindex('.') ], output )
    f = open(docFile, 'w')
    if output == 'htm':
        f.write('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">\n')
        f.write('<html xmlns="http://www.w3.org/1999/xhtml">\n')
        f.write('<head>\n')
        f.write('  <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />\n')
        f.write('  <meta name="generator" content="progrutils.documentPyFile()" />\n')
        f.write('  <title>Documentation for %s</title>\n' % os.path.split(file)[1])
        if css !='':
            f.write('<link rel="stylesheet" href="%s" type="text/css" />\n' % css)
        else:
            f.write('  <style type="text/css">\n')
            f.write('h4.className {\n')
            f.write('  color:#3300cc;\n')
            f.write('}\n')
            f.write('h4.functionName {\n')
            f.write('  color:#000099;\n')
            f.write('}\n')
            f.write('h4.variableName {\n')
            f.write('  color:#007700;\n')
            f.write('}\n')
            f.write('  </style>\n')
        f.write('</head>\n')
        f.write('<body>\n')
        f.write('<h2 style="text-transform:uppercase;">Documentation for %s</h2>\n<p>&nbsp;</p>\n\n' % os.path.split(file)[1])
    else:
        f.write('Documentation for %s\n\n\n' % os.path.split(file)[1])
    f.writelines(doc)
    if output == 'htm':
        f.write('</body>\n')
        f.write('</html>\n')
    f.close()
    print "documentPyFile(): %s created." % docFile


# convert problematic characters within a line to HTML entities and escaped characters
def convertLineEntities(line):
    searchChars = ["<",   ">",    "&",     "\a",  "\b",  "\f",  "\n",  "\r",  "\t",  "\v"]
    convertChars =["lt;", "&gt;", "&amp;", "\\a", "\\b", "\\f", "\\n", "\\r", "\\t", "\\v"]
    newLine = ""
    for c in line:
        if c not in searchChars:
            newLine = newLine + c
        else:
            newLine = newLine + convertChars[searchChars.index(c)]
    return newLine
                        


# return 1 if the line is empty (the string contains only a return and possibly spaces or tabs)
def isLineEmpty(line):
    for char in line:
        if char not in (' ', '\t', '\n'):
            return 0
    return 1

