# -*- coding: latin-1 -*-

######################################################################
## 
## System Utilities for Freedom Force
## By pimthe
## 
######################################################################
print "Starting System Utilities 1.5"

# DESCRIPTION
# A small collection of functions for Freedom Force and FFX related to file, folder, mod, etc.

# Feel free to change or add to the code here. Just make sure you maintain compatibility,
# so external modules which could be using these functions keep working. :)

# DEPENDENCIES
# - Tested only with FFvsTTR with a branch of FFX 3.2 alpha.

# VERSION HISTORY
# 0.1:
# - Added getModFolder() to ffx.py.
# 1.0 2006-06-11:
# - Made into a module, as there isn't anything FFX-specific in it.
# 1.1 2006-06-22:
# - Added saveModFolder().
# - getModFolder() should now work when the module is imported from a default non-mod game
#   (i.e., using the 'data' folder).
# - Added file management utilities formerly in heightcheck.py,  updated to use getModFolder()
# 1.2 2006-08-03:
# - Added exception handling to writeFile().
# 1.3 2006-08-03 (stumpy):
# - Removed imports of datfiles (not used), chardata (not used), and
#   missionobjvar (faster solution just to cache getModFolder() and
#   call it when we want the mod folder).
# 1.4 2007-02-17:
# - Added encoding (cf. first line) for compatibility with Python 2.5.
# 1.5 2007-10-25:
# - Added readFileInZip() to read a file archived in .ff and other .zip files.

import string
import sys
import zipfile

######################################################################
## 
## Mod folder management
## 
######################################################################

############ getModFolder()
# Get the current mod's main folder name (for writing text files, etc.)
# Replaces the old MODFOLDER='my folder' manual entry in ffxextras.py.
# Note: just after updating to v.1.3, I realised Stumpy had already done
# GetModPath() in the DATFiles module... :P
# v. 1.3
# stumpy 20061013: modified slightly to use caching
def getModFolder():
    global sysutil_modfolder
    try:                # if it's cached, just return it
        return sysutil_modfolder
    except NameError:   # first call to this fn
        pass
    systemPath = []
    thisPath = ""
    foundPath = ""
    foundNbr = []
    for p in sys.path:
        thisPath = string.lower(p)
        systemPath.append(thisPath)
    for p in systemPath:
        if p[-17:] == "\\missions\\scripts":
            if p[-22:] != "\\data\\missions\\scripts":
                foundPath = p
                foundNbr.append(p)
    if len(foundNbr) == 0:
        print "getModFolder() non-fatal error:"
        print "    no potential mod path found in sys.path"
        print "    returning the default 'data' folder"
        sysutil_modfolder = 'data'
        return sysutil_modfolder
    #if multiple paths are found, check that they're not identical
    #limitations: doesn't recognize non-root paths or paths using dot notation as identical to the same path using a standard full path;
    #idem for forward slash versus double backslash
    if len(foundNbr) > 1:
        comparedPaths = 0
        while comparedPaths < len(foundNbr) - 1:
            if foundNbr[comparedPaths] != foundNbr[comparedPaths + 1]:
                print foundNbr[comparedPaths]
                print foundNbr[comparedPaths + 1]
                print "getModFolder() fatal error:"
                print "    more than one potential mod path found in sys.path"
                sysutil_modfolder = ''
                return sysutil_modfolder
            comparedPaths = comparedPaths + 1
            print "getModFolder() warning:"
            print "    the same path name is present %i times in FFvsTTR's Python sys.path" % len(foundNbr)
            print foundPath
    #strips the path to get the folder name
    foundPath = foundPath[:-17] #strips the end
    sysutil_modfolder = foundPath[string.rfind(foundPath, "\\") + 1:] #strips until after the \ before the path
    return sysutil_modfolder

# stumpy 20061013: just call the getModFolder() function
############## Save Mod Folder
### Saves the current mod's folder in a few Mission_SetAttr() (using Mission_SetVar()) so it will get
### embedded in the save game itself. This is called automatically on startup.
##
### Note: this is only for forward compatibility with an eventual save game management utility which
### would read the info in the save game itself. However, this shall require that any mod either
### import the current module directly or uses FFX 3.2+.
### v. 0.1
##
##missionobjvar.Mission_SetVar('sysutil_modfolder', getModFolder())



######################################################################
## 
## File input/output
## 
######################################################################

############ updateFile(), writeFile() and readFile()
# These were originally released in heightcheck.py.
# They've been updated to use getModFolder() instead of the MODFOLDER variable. The path has
# also been made customizable.
# Note the use of double backslashes in subFolder parameter

# Adds or updates a variable to a text file.
# A variable cannot be completely removed, only emptied.
# Multilines variables are not supported.
# ChangedVar needs to be entered as a string!
# If the variable is not found in the file, it will be added at the end of the file.
# To use directly in the currently active directory, employ the dot notation:
# sysutils.readFile("name of the file.txt", baseFolder=".", subFolder=".\\.")
def updateFile(fileName, changedVar, changedValue, baseFolder = "", subFolder = "missions\\scripts"):
    if baseFolder == "":
        baseFolder = getModFolder()
    fileContent = readFile(fileName, baseFolder, subFolder)
    updatedLine = "%s = %s\n"%(changedVar, changedValue)
    if fileContent == ['']:
        print "File '%s' is empty or doesn't exist: creating file with variable %s"%(fileName, changedVar)
        fileContent = [updatedLine]
    else:
        found = -1
        for thisLine in fileContent:
            if (thisLine[0:len(changedVar)] == changedVar) and ((thisLine[len(changedVar) + 1] == '=') or (thisLine[len(changedVar):len(changedVar) + 2] == ' =')):
                    print "content found in '%s': %s"%(fileName, thisLine)
                    found = fileContent.index(thisLine)
                    break
        if found != -1:
            fileContent[found] = updatedLine
        else:
            print "variable %s not found: adding variable"%changedVar
            fileContent.append('\n')
            fileContent.append(updatedLine)
    writeFile(fileName, fileContent, baseFolder, subFolder)

def readFile(fileName, baseFolder = "", subFolder = "missions\\scripts"):
    if baseFolder == "":
        baseFolder = getModFolder()
    filePath = '%s\\%s\\%s' % (baseFolder, subFolder, fileName)
    fileContent = []
    try:
        f = open(filePath, 'r')
        thisLine = 'dummy'
        while thisLine != '':
            thisLine = f.readline()
            fileContent.append(thisLine)
        f.close()
    except:
        print "readFile() warning:"
        print "    %s\\%s\\%s not found"%(baseFolder, subFolder, fileName)
    return fileContent

# content is a list
def writeFile(fileName, content, baseFolder = "", subFolder = "missions\\scripts"):
    if baseFolder == "":
        baseFolder = getModFolder()
    filePath = '%s\\%s\\%s' % (baseFolder, subFolder, fileName)
    try:
        f = open(filePath, 'w')
        print "writing to file %s"%fileName
        f.writelines(content)
        f.close()
    except IOError:
        print "writeFile() fatal error:"
        print "    cannot write to %s\\%s\\%s"%(baseFolder, subFolder, fileName)
    except:
        print "writeFile() fatal error:"
        print "     unknown error while trying to write to %s\\%s\\%s"%(baseFolder, subFolder, fileName)



# Return the content of a file archived in a .ff or .zip file 
# filePath is the name of the file plus eventual archived folders path
# zipPath is relative to the ..\Irrational Games\Freedom Force vs The 3rd Reich\ folder.
# You can use getModFolder() to get the current mod's root folder.
# Content is read in binary mode and returned as a string; if the file cannot be found, the returned value is set to "".
# Notes:
# - There is no cacheing mechanism, since it should be much more efficient to extract the relevant data and then
#   cache it rather than cache the whole file (which might take a meg or more RAM for a mesh).
# - Processing large files can be long and shouldn't be used in real-time. For example, on my (old) computer, opening
#   ant_soldier.dds from art.ff (183MB) can take 8 sec. If you *really* need to access a large archive during a mission,
#   do it at startup and add some mission status text. As an aside, doing the same from the Python interpreter 2.5 outside
#   the game takes a little more than .3 sec...
# - To improve performances a bit, you should use the correct file case. (The paths inside a zip aren't actually represented
#   as and folders hierarchies by the zipfile module, but simply as keys to the archive object's NameToInfo dictionary. 
#   As such, the filePath parameter should respect the exact  path's case usage and only the "/" should be used as separator,
#   not "\\" or os.path.join(); otherwise, readFileInZip() may need to convert your filePath and every path inside the archive.)
#
# Usage example:
# You create a FFX plugin with a various subfolders containing text files and archive these subfolders
# inside "myarchive.zip"
# to get data inside a myplugin folder installed in ffx3, you'd use
# modFolder = sysutils.getModFolder()
# myPath = "%s/missions/scripts/ffxplugins/active/myplugin/myarchive.zip"
# sysutils.readFileInZip("someFolder/myFile.txt", myPath)
# Cf. testReadZip() below for another example
def readFileInZip(filePath, zipPath):
    #convert the filepath to use / rather than \\
    if string.find(filePath, "\\") > -1:
        words = string.split(filePath, "\\")
        filePath = string.join(words, "/")
    fileContent = ''
    try:
        z = zipfile.ZipFile(zipPath, 'r')
        try:
            fileContent = z.read(filePath)
        except:
            #if not found, do a comparison ignoring the case to see if we can find the entry this way
            try:
                filePath = string.lower(filePath)
                zDir = z.namelist()
                for dir in zDir:
                    if string.lower(dir) == filePath:
                        # now, try again
                        fileContent = z.read(dir)
                        break
            except:
                print "readFileInZip() warning:"
                print "    %s not found in %s" % (filePath, zipPath)
        z.close()
    except:
        print "readFileInZip() warning:"
        print "    cannot open %s" % zipPath
    return fileContent

# readFileInZip() test:
# extract the "art\\library\\characters\\ant_soldier\\skins\\nightmare\\ant_soldier.dds" file
# and copy it to the FFvsTTR's root folder as testReadZip.dds.
# for use only from inside the game
def testReadZip():
    import time
    t1 = time.time()
    dds = readFileInZip("Art\\library\\characters\\ant_soldier\\skins\\nightmare\\ant_soldier.dds", "C:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\data\\art.ff")
    t2 = time.time()
    try:
        f = open('testReadZip.dds', 'wb')
        print "writing to file testReadZip.dds"
        f.write(dds)
        f.close()
    except IOError:
        print "writeFile() fatal error:"
    print "ZIP archive was read in %f seconds" % (t2-t1)
