#####################################################################
#
# Encounters for use in mods and skirmishes
# contains a high degree of cusomizability
# Version 3.2
#
#
#####################################################################

print "Loading m25enc.py v.3.2.0 July 5, 2008"

# Basic encounter methodology:
#	1. Input parameters:  
#		1.1 A unique number that identifies the encounter (for save game purposes) - enc_id
#		1.2 A list of characters to spawn/use in the encounter
#		1.3 The names of cutscenes (or lists of cutscenes) to use
#		1.4 The name of a function to call when the encounter completes - fn
#
#	2. Search the map for markers that denote where to spawn the characters using getMarkerListByName
#
#	3. Spawn the characters and/or move them to their start locations using m25_spawn
#      this will ensure that the AI is setup correctly
#
#	4. Disable the AI on the characters (so that the encounter starts when the hero arrives on the scene)
#      using m25ai.AIDisable - not all encounters do this though.
#
#	5. Set up the triggers for various actions within the encounter... examples:
#		- play a cutscene when a character is ko'd
#		- remove a character when they esacape from the map
#		- play a cutscene and call the return function when the encounter is finished
#		- start up the AI for all characters when a hero arrives on the scene
#
#	7. Output parameters:
#		7.1 The unique id of the encounter - enc_id
#		7.2 A return code (string value) that states how the encounter ended (such as "all ko'd") - return_value
#		7.3 A list that can contain anything you'd like - return_list




#May 27, 2007:  Fixed bug in EncCustom_allies_fight_villains (custom end function not submitted properly)
#July 5, 2008: Fixed bug in EncCreateMarker and EncCreateAllMarkers (was adding None to the marker list, which caused spawn problems)

import ff
from ff import *

import time

import thread

from m25aiconstants import *

import m25ai
#from m25ai import *
import m25getset
#from m25getset import *
import m25report
#from m25report import *
import m25distance
#from m25distance import *
import m25name
import m25string
import m25distance
from m25event import *
import m25string

import m25cutscene
from m25cutscene import *

import js
import cshelper
import random
import traceback
import string
#import ai

import datfiles

import m25event
import ffx

import chardata
import skXMapInfo

import m25randomenc
import m25gametype
import m25objective

import m25freeroamally
import m25freeroamdata

import m25enccustom

#import m25pstory


#m25pstory.PStory_AddToReplaceList('switch')
#m25pstory.PStory_AddToReplaceList('vehicle')


#import m25ezplugin1
#from m25ezplugin1 import *

from ffxTraits import *

ENC_CHANNEL = 'ENCOUNTER'
m25report.Report_SetChannel(ENC_CHANNEL,1)


#Special attribute - this is set to zero for the PopulateBase encounter to ensure that 
# the same character isn't spawned twice.
Mission_SetAttr('enc_spawnduplicates',1)

def getStringVariable(string):
	val = ''
	exec 'val = ' + string
	return val
	
	
# a dictionary containing all active and completed encounters in the current session
_encounter_dict = {}	
	

#---------------------------------------------------------------------
#
# Standardized encounter functions
#
#---------------------------------------------------------------------

# Start and End trigger - routes to custom functions

# def EncStart(event):
# 	CallEncFunction(int(event.user),'start')


def IsEncounterComplete(enc_id):
	return GetEncVar(enc_id,'-encounter_complete')
	
def SetEncounterComplete(enc_id):
	SetEncVar(enc_id,'-encounter_complete',1)
	
def SetEncounterNotComplete(enc_id):
	SetEncVar(enc_id,'-encounter_complete',0)

def EncCheckEnd(enc_id):
	if not IsEncounterComplete(int(enc_id)):
		CallEncFunction(int(enc_id),'checkend')

def EncEnd(enc_id,return_value='',return_list=[], complete=1):
	if not IsEncounterComplete(enc_id):
		m25report.Report(ENC_CHANNEL, 'Encounter ended ' + `enc_id` + ' with ' + `return_value`)
		if complete: 
			SetEncounterComplete(enc_id)
		else:
			m25report.Report(ENC_CHANNEL, 'Encounter %i set to continue past end.'%(enc_id))
		try:
			EncStartFree(GetEncVar(enc_id,'start'))
		except:
			m25report.Report(ENC_CHANNEL, 'Could not free start location for %i'%(enc_id))
		CallEncounterReturnFunction(enc_id,return_value,return_list)	

# def EncEnd_CallPost(enc_id, encounter):
# 	enc = GetEncVar(enc_id, 'on_end', default=None)
# 	if enc != None:
# 		for s in enc:
# 			try:
# 				exec 'ff.' + s + '(' + `enc_id` + ')'				
# 			except:
# 				m25report.Report(ENC_CHANNEL, 'm25enc.EncEnd_CallPost - error in cleanup function ff.%s(%i)'%(s,enc_id))
# 				print sys.last_type, sys.last_value, sys.last_traceback				
		
# def EncEndTimer(event):
# 	EncEnd(int(event.user),event.object)	


#--------------------------------------------------------------
#
# Store and reload encounters
# Used in freeroam campaigns
#
#--------------------------------------------------------------


def Enc_StoreEncounters():
	global _encounter_dict
	#purge completed encounters so they don't clog up the files
	new_dict = {}
	for d in _encounter_dict.keys():
		if _encounter_dict[d].has_key('type'):
			# for now, remove anything that isn't an alliance encounter
			if ('alliance' in _encounter_dict[d]['type']) and (GetEncVar(d,'allies_saved',default=0)):
				new_dict[d] = _encounter_dict[d]
	m25getset.SetMissionVar('_encounter_dict',new_dict,campaign=1)
	
def Enc_LoadEncounters():
	global _encounter_dict
	new_dict = m25getset.GetMissionVar('_encounter_dict',default={})	
	for d in new_dict.keys():
		_encounter_dict[d] = new_dict[d]
	
#--------------------------------------------------------------
#
# Allow for the callbacks when an encounter completes
#
# Return the encounter id, the return code from the encounter,
# and a list of unspecified values.
#
#--------------------------------------------------------------
			
def SetEncounterReturnFunction(enc_id, return_function=None):
	SetEncVar(enc_id, 'returnfunction', return_function)
	
def CallEncounterReturnFunction(enc_id, return_code, return_list=[]):
	function_name = GetEncFunctionName(enc_id,'returnfunction')
	if function_name == None:
		return
	if function_name == []:
		m25report.Report('Bad return function for ' + enc_name + `enc_id`)
		return
	if type(function_name) == types.ListType:
		function_name = function_name[0]
	if type(return_code) != types.ListType:
		return_code = [return_code]
	function_string = 'ff.' + function_name + '(' + `enc_id` + ',' + `return_code` + ',' + `return_list` +')'
	exec function_string

	
#--------------------------------------------------------------
#
# Special spawn code (to allow for difficulty levels - pending),
# and to ensure that a unique tag is assigned to the new spawn
#
#--------------------------------------------------------------

def m25_spawn(name,template,marker,teleport=1,objclass='',disable=0):
	print name, template, marker
	massfactors = [1.0,3.0, 2.5, 1.25, 0.75, 1.0, 0, 1.5, 2.5, 1.5, 1.0]
	if not Object_Exists(name):
		#do not spawn duplicates for special base missions
		if not Mission_GetAttr('enc_spawnduplicates'):
			objects = Mission_GetDynamicObjects()
			for obj in objects:
				if Object_GetTemplate(obj) == template: return
		try:			
			cshelper.spawn(name,template,marker)
			
		except:
			newtemplate = m25string.String_ConvertToTemplate(template)
			try:
				cshelper.spawn(name, newtemplate, marker)
			except:
				m25report.Report('ERROR', 'The character %s with template %s is missing a mesh or a hero file.'%(name, template))
				return
		if js.Object_GetAttr(name,'complex'):
			chdata = chardata.GetCharacterData(name)
			if chdata.has_key('energy'):
				Object_SetAttr(name,'energy',chdata['energy'])
			if chdata.has_key('material') and chdata.has_key('mass'):
				Object_SetAttr(name,'mass',chdata['mass']*massfactors[int(chdata['material'])])
		try:
			if objclass != 'hero':
 				if ffx.FFX_HasPortrait(name):
 					EncDestroyPortrait(name)
				ai = cshelper.findAI(name)
				ai.removeNaturalKillGoal()
			if disable: m25ai.AIDisable(name)
			m25ai.SetupAI(name)
			if objclass != 'hero':
				m25ai.SetBrawler(name)
		except:
			raise


def EncDestroyPortrait(char):
	js.AI_DestroyHeroAI(char)
	ffx.FFX_ObjectSetAttr(char, 'fakehero', 1)
	js.AI_MakeHeroAI(char)		
		
	
# spawning	
#this is in serious need of cleaning up
def EncSpawnGroup(enc_id, marker, chars, savename='', tag='', disable=1, teleport=1, marker_type='', create_markers=1):
	start_markers = getMarkerListByName(marker)
	if start_markers == [] and create_markers: 
		start_markers = CreateMarkerListByName(enc_id, marker, marker_type)
	num = 0
	name = ''
	template = ''
	spawned = []
	spawned_template = []
	spawnmaxsize = len(start_markers)-1
	newchars = []
	if m25string.IsString(chars):  #chars is simply a template
		for s in start_markers:
			newchars.append(chars) #create an entry for each marker found so they can be spawned
	else:
		newchars = chars
	
	for c in newchars:
		mark = time.clock()
		alreadyspawned = 0
		if num <= spawnmaxsize:
			if type(c) == types.ListType: 
				#check for sets of random characters (generated by m25randomenc) 
				if c[0] == '_set':  
					for element in c[1:]:
						if num <=spawnmaxsize:
							name = m25name.GetNextObjectName()
							template = element
							m25_spawn(name,template,start_markers[num],teleport=teleport,objclass=savename,disable=disable)
							#if disable:
							#	m25ai.AIDisable(name)
							spawned.append(name)
							spawned_template.append(template)
							num = num + 1
							alreadyspawned = 1
				else:
					# have both name and template
					name = c[0]
					#named critters are encounter independent by default
					Enc_SetEncounterIndependent(name)
					template = c[1]
					if m25string.IsFunction(template):
						try:
							exec 'template = ff.'+template
						except:
							m25report.Report(ENC_CHANNEL, 'EncSpawn problem with spawn function ' + `c`)
							template = 'thug_with_bat'
			else:						# only have template
				name = m25name.GetNextObjectName()
				if m25string.IsFunction(c):
					try:
						exec 'template = ff.'+c
						if template[0] == '_set':
							templist = template[1:]  
							for element in templist:
								if num <=spawnmaxsize:
									name = m25name.GetNextObjectName()
									template = element
									m25_spawn(name,template,start_markers[num],teleport=teleport,objclass=savename,disable=disable)
									#if disable:
									#	m25ai.AIDisable(name)
									spawned.append(name)
									spawned_template.append(template)
									num = num + 1
									alreadyspawned = 1
					except:
						m25report.Report(ENC_CHANNEL, 'EncSpawn problem with spawn function ' + `c`)
						raise
						template = 'thug_with_bat'
				else:
					#check for 'building' and 'object' specials
					if string.lower(c) == 'building':
						try:
							name = Enc_GetNearestBuilding(start_markers[num])
							template = js.Object_GetTemplate(name)
							m25getset.SetObjectVar(name, 'ai_spawn_teleport', 0)
						except:
							m25report.Report('ERROR', 'EncSpawn no nearby buildings ' + `c`)
							template = c
					elif string.lower(c) == 'object':
						try:
							name = Enc_GetNearestObject(start_markers[num])
							template = js.Object_GetTemplate(name)
							m25getset.SetObjectVar(name, 'ai_spawn_teleport', 0)
						except:
							m25report.Report('ERROR', 'EncSpawn no nearby objects ' + `c`)
							template = c
					else:
						#none of the above, take it as a template
						template = c
			if not alreadyspawned:
				m25_spawn(name,template,start_markers[num],teleport=teleport,objclass=savename,disable=disable)
				#if disable:
				#	m25ai.AIDisable(name)
				spawned.append(name)
				spawned_template.append(template)
		else: #if others already exist, add them to the group too
			if type(c) == types.ListType:
				if js.Object_Exists(c[0]):
					spawned.append(c[0])
					spawned_template.append(js.Object_GetTemplate(c[0]))
			else:
				if js.Object_Exists(c):
					spawned.append(c)
					spawned_template.append(js.Object_GetTemplate(c))
		num = num + 1
	
	if savename != '':
		#SetEncVar(enc_id,savename,spawned,save=0)
		SetEncVar(enc_id,savename,spawned,save=0)
		SetEncVar(enc_id,savename+'_template',spawned_template,save=0)
		# use groups
		if tag == '':
			tag = savename
		encgroupname = `enc_id` + '_' + tag
		m25ai.AddToGroup(encgroupname,spawned)
		for s in spawned:
			#m25ai.AddToGroup(encgroupname,s)
			SetEncGroupPrefix(s,tag)  #group is the AI group the character belongs to
			SetEncClassPrefix(s,savename) #class is dependent on the encounter (minions, villain, ally, etc.)
			Enc_SetCharEnc(s,enc_id) #make sure we are able to find out which encounter this is a part of
		m25ai.SetGroupVar(encgroupname,'enc_id',enc_id)
#	print 'setting spawn variables took: %f'%(time.clock()-start)
	return spawned	
	
	
	
#--------------------------------------------------------------
#
# Find a set of markers and return a random one from the set.
# markers must start with 'name' and end with a number 
# in order to be found.
#
#--------------------------------------------------------------
	
_enc_markers = {}

def getMarkerListByName(name):
	global _enc_markers
	if _enc_markers.has_key(name): return _enc_markers[name]
	
	mlist = []
	try:
		pos = Get_MarkerPos(name)
		mlist.append(name)
	except:
		pass #marker of name not found, but we'll still try the number markers
	for num in range(1,100):
		try:
			pos = Get_MarkerPos(name+`num`)
			mlist.append(name+`num`)
		except:
			if mlist != []:  # we don't want to save an empty list as that would prevent new markers from being spawned correctly
				_enc_markers[name] = mlist
			return mlist
	

def DetermineMarkerNum(name):
	start_locs = getMarkerListByName(name)
	if start_locs == []:
		return -1
	else:
		return random.randint(1,len(start_locs))	
			

#-----------------------------------------------------------------------
#
# Take a list and a type of character (type is any user-defined string)
# and find an example of that type in a list.
# 
# The list can be passed in as minionlist = [[char,type],...] or
# it can be a chartype and end_id combination which is used to 
# read an object list saved in Mission variables.
#
#------------------------------------------------------------------------


def findCharFromList(minionlist=[],chartype='',enc_id=0,excludelist=[],live=1,subtype=None):
	charlist = []
	#special chartypes
	if minionlist == []:
		if chartype == 'hero':
			heroes = Enc_GetAllHeroes()
			for h in heroes:
				if not (h in excludelist):
					charlist.append(h)
		elif (chartype == 'buddy') or (chartype=='ally' and GetEncVar(enc_id,chartype)==[]):
			buddies = m25freeroamally.GetFreeroamAllies()
			for b in buddies:
				if not (b in excludelist):
					charlist.append(b)
		#get the characters from saved variables
		else:
			charlist = GetEncVar(enc_id, chartype)
	else:
		#ferret out all of the elements of minionlist that match chartype
		for m in minionlist:
			if m[1] == chartype:
				charlist.append(m[0])
		if charlist == []:
			if chartype == 'hero':
				heroes = Enc_GetAllHeroes()
				for h in heroes:
					if not (h in excludelist):
						charlist.append(h)
			elif (chartype == 'buddy') or (chartype=='ally' and GetEncVar(enc_id,chartype)==[]):
				buddies = m25freeroamally.GetFreeroamAllies()
				for b in buddies:
					if not (b in excludelist):
						charlist.append(b)
			#get the characters from saved variables
			else:
				charlist = GetEncVar(enc_id, chartype)
	for c in charlist:
		if not (c in excludelist) and Enc_CharIsSubtype(c,subtype):
				if live and (chartype != 'fallen'):
					if Object_Exists(c) and Object_IsAlive(c):
						return c
				else:
					return c
	return ''


def Enc_FerretOutSubtype(line):
	words = string.splitfields(string.lower(string.strip(line)), ' ')
	if (len(words) > 1) and string.lower(words[-1]) in ['hero','villain','minion','ally','object']:
		trait = ''
		for s in words[:-1]:
			trait = trait + ' ' + s
		return string.strip(trait)
	return None
	
def Enc_CharIsSubtype(char,subtype):
	if subtype == None: return 1
	try:
		if m25ai.IsSubtype(char, subtype): 
			return 1
	except:
		pass
	for traitlist in ffxTraits:
		if subtype == string.lower(traitlist[0]):
			try:
				if js.Template_GetSpeechID(js.Object_GetTemplate(char)) in traitlist[1:]:
					return 1
			except:
				pass
	return 0
	
def Enc_GetAllHeroes():
	heroes = []
	dynamic = js.Mission_GetDynamicObjects()
	for name in dynamic:
		if (js.Object_GetClass(name)&js.OC_CONTROLLABLE) and not (js.Mission_TestAttr('%s_%s'%(name,'fakehero'))):
			heroes.append(name)
	return heroes	

#----------------------------------------------------------
#
# Save and Restore encounter information within saved games
#
#----------------------------------------------------------
	
def Enc_GetDict(enc_id):
	global _encounter_dict
	if not _encounter_dict.has_key(enc_id):
		_encounter_dict[enc_id] = m25getset.GetMissionVar('enc_%i'%(enc_id),default={})
	return _encounter_dict[enc_id]
	
def Enc_SaveDict(enc_id):
	m25report.Report(ENC_CHANNEL, 'Saving encounter data for %i'%(enc_id))
	dict = Enc_GetDict(enc_id)
	m25getset.SetMissionVar('enc_%i'%(enc_id),dict)

def SetEncCutscenes(enc_id, cslist):
	for c in cslist:
		SetEncVar(enc_id, c[0], c[1], save=0)

def SetEncVar(enc_id, varname, value, save=1):
	enc_id = int(enc_id)
	dict = Enc_GetDict(enc_id)
	dict[varname] = value
	if save:
		m25getset.SetMissionVar('enc_%i'%(enc_id),dict)
	#m25getset.SetMissionVar('enc_' + `int(enc_id)` + '_' + varname, value)
#	m25getset.SetMissionVar('enc_%i_%s'%(enc_id,varname), value)
			
def SetEncStateVars(enc_id, varlist, save=0):
	for v in varlist:
		SetEncVar(enc_id, v[0], v[1], save=save)
		#SetEncStateVar(enc_id, v[0], v[1])

def SetEncVars(enc_id, varlist, save=0):
	for v in varlist:
		SetEncVar(enc_id, v[0], v[1], save=save)		
		
def SetEncFunctions(enc_id, funclist):
	for f in funclist:
		SetEncVar(enc_id, f[0], f[1], save=0)
		#m25getset.SetMissionVar('enc_'+`int(enc_id)` + '_' + f[0], f[1])		
				
def GetEncCutscene(enc_id, csname, default=None):
	return GetEncVar(enc_id, csname, default)

def GetEncVar(enc_id, varname,default=None,integer=0):
	enc_id = int(enc_id)
	dict = Enc_GetDict(enc_id)
	if _encounter_dict[enc_id].has_key(varname):
		return _encounter_dict[enc_id][varname]
	return default
	#return m25getset.GetMissionVar('enc_' + `int(enc_id)` + '_' + varname, default=default,integer=integer)
	#return m25getset.GetMissionVar('enc_%i_%s'%(enc_id,varname), default=default, integer=integer)	
	
def GetEncFunctionName(enc_id, funcname, default=None):
	return GetEncVar(enc_id, funcname, default)
	#return m25getset.GetMissionVar('enc_' + `int(enc_id)` + '_' + funcname, default=default)

def CallEncFunction(enc_id, funcname, values=[]):
	function_name = GetEncFunctionName(int(enc_id), funcname)
	if type(function_name) == types.ListType:
		function_name = function_name[0]
	if function_name == [] or function_name == None:
		m25report.Report('Error in CallEncFunction(): Bad function for ' + 'enc_' + `int(enc_id)` + ' ' + funcname)
		return
	if values != []:
		function_string = 'ff.' + function_name + '(' + `int(enc_id)` + ',' + `values` + ')'
	else:
		function_string = 'ff.' + function_name + '(' + `int(enc_id)` + ')'
	exec function_string		
	

_enc_character_vars = {}
	
def Enc_LoadCharacterVars():
	global _enc_character_vars
	if _enc_character_vars == {}:
		_enc_character_vars = m25getset.GetMissionVar('_enc_character_vars',default={})

def Enc_SaveCharacterVars():
	global _enc_character_vars
	#clean out outdated entries to speed up the save
	for char in _enc_character_vars.keys():
		if not js.Object_Exists(char):
			del _enc_character_vars[char]
	m25getset.SetMissionVar('_enc_character_vars', _enc_character_vars)

def GetEncGroupPrefix(char):
	global _enc_character_vars
	Enc_LoadCharacterVars()
	if _enc_character_vars.has_key(char):
		if _enc_character_vars[char].has_key('group_prefix'):
			return _enc_character_vars[char]['group_prefix']
	return None
	#return m25getset.GetObjectVar(char,'__enc_group_prefix',	default=None)
	
def SetEncGroupPrefix(char, prefix):
	global _enc_character_vars
	if not _enc_character_vars.has_key(char): _enc_character_vars[char] = {}
	_enc_character_vars[char]['group_prefix'] = prefix
	#m25getset.SetObjectVar(char, '__enc_group_prefix', prefix)	
	
def GetEncClassPrefix(char):
	global _enc_character_vars
	Enc_LoadCharacterVars()
	if _enc_character_vars.has_key(char):
		if _enc_character_vars[char].has_key('class_prefix'):
			return _enc_character_vars[char]['class_prefix']
	return None
	#return m25getset.GetObjectVar(char, '__enc_class_prefix', default=None)
	
def SetEncClassPrefix(char, prefix):
	global _enc_character_vars
	if not _enc_character_vars.has_key(char): _enc_character_vars[char] = {}
	_enc_character_vars[char]['class_prefix'] = prefix
	#m25getset.SetObjectVar(char, '__enc_class_prefix', prefix)
		
def GetEncEncounter(enc_id):
	return GetEncVar(enc_id, '_encounter', default = {})	
	

def EncGetGroupName(enc_id, enc_group):
	if enc_group == None: return None
	return `enc_id` + '_' + enc_group	

	
Global_RegisterAttr('_enc_id',0)
	
def Enc_SetCharEnc(char, enc_id):
	js.Object_SetAttr(char, '_enc_id', enc_id)
	#m25getset.SetObjectVar(char, '_enc_id', enc_id)
	
def Enc_GetCharEnc(char):
	return int(js.Object_GetAttr(char, '_enc_id'))
	#return m25getset.GetObjectVar(char, '_enc_id', default=0)		
	
#-----------------------------------------------------------------------
#
# Custom cutscene manipulators. 
#
#------------------------------------------------------------------------		
	
def Enc_StripWhiteSpace(line):
	newline = ''
	for s in line:
		if not (s in string.whitespace):
			newline = newline + s
	return newline

def PlayEncCutscene(enc_id, csname, charlist=[], return_function=None, live=1, in_order=0):
	start = time.clock()
	csname = string.lower(Enc_StripWhiteSpace(csname))
	m25report.Report(ENC_CHANNEL,'PlayEncCutscene(): ' + `csname` + ' : ' + `return_function`)
	if GetEncVar(enc_id, '_cs_no_randomize_%s'%csname, default=0): in_order = 1
	value = PlayCutsceneFromList(GetEncCutscene(enc_id, csname),enc_id, charlist, return_function, live, in_order)
	m25report.ReportTime('PlayEncCutscene', (time.clock()-start))
	return value

def PlayCutsceneFromList(cslist, enc_id=0, charlist=[], return_function=None, live=1, in_order=0):
	m25report.Report(ENC_CHANNEL,'PlayCutsceneFromList(): ' + `cslist` + ' : ' + `return_function` + ' ... ' + `charlist`)
	if cslist == [] or cslist == None or cslist == 0:  #no cutscene to play
		if return_function != None:
			try:
				exec 'ff.' + return_function
			except:
				m25report.Report('ERROR', 'problem with return function %s'%return_function)
				raise
		return 0
	newcslist = []
	if m25string.IsString(cslist):
		#list passed in as a string - get the corresponding list from the string
		try:
			newcslist = getStringVariable('ff.' + cslist)
		except:
			m25report.Report(ENC_CHANNEL, cslist + ' does not exist.')
			if return_function != None:
				try:
					exec 'ff.' + return_function	
				except:
					m25report.Report('ERROR', 'problem with return function %s'%return_function)
					raise
			return 0 #the cutscene list doesn't exist, exit gracefully
	else:
		#list passed in directly
		newcslist = cslist
	
	# now, check to make sure the the first entry in the list is a string 'cslist'
	if newcslist[0] == 'cslist':
		cstoplay = []
		if in_order: cstoplay = newcslist[1]
		else: cstoplay = newcslist[random.randint(1,len(newcslist)-1)]
		cutscene_ok = EncPlayCutscene(cstoplay, enc_id, charlist=charlist, return_function=return_function, live=live)
		if not cutscene_ok:
			#try them all in order
			for cs in newcslist[1:]:
				cutscene_ok = EncPlayCutscene(cs,enc_id,charlist,return_function)
				if cutscene_ok:
					return 1
		else:
			return 1
		#even if a cutscene doesn't play, make sure the return function gets executed
		if return_function!=None: exec 'ff.' + return_function
		return 0 #none of the cutscenes in the list were able to be played!
		
	# if the word 'cslist' is not the first element, then this is a single cutscene
	# and needs to be played directly
	else:
		if not EncPlayCutscene(newcslist,enc_id,charlist, return_function, live):
			if return_function!=None: exec 'ff.' + return_function
			return 0
		return 1

		

def EncPlayCutscene(cstoplay, enc_id, charlist=[], return_function=None, live=1):
	newcs = CSGetElement(cstoplay, 'cutscene')
	extra_replacements, exclude_list = CSGetRequired(cstoplay, charlist, enc_id)
	if extra_replacements == 'error': return 0
	replace = CSGetReplacements(cstoplay, charlist, enc_id, live, exclude=exclude_list)
	if replace == 'error': return 0
	replace = m25ai.AddToList(replace, extra_replacements)
	if replace != None:
		newcs = CSDoReplacements(newcs, replace)
	newcs = CSSpecialReplacements(newcs, enc_id)
	#if return_function != None:
	#	newcs.append('CS_ReturnFunction(0.5,' + `return_function` + ',' + `return_value` + ',' + `return_value2` + ',' + `enc_id` + '),')
	priority = CSGetElement(cstoplay,'priority')
	if priority != None: priority = priority[0]
	m25cutscene.PlayCutscene(newcs,priority,return_function,enc_id)
	return 1

def CSGetRequired(cs, charlist=[], enc_id=0):
	block = CSGetElement(cs, 'required')
	if not block: return [[],[]]
	extra_replacements = []
	exclude_list = []
	for n in block:
		if not (js.Object_Exists(n) and js.Object_IsAlive(n)) and (not CSCheckTemplate(n)):
			#character doesn't exist, try to get a subtype or speech id
			new_replacement = CSGetTempReplacement(n, charlist, exclude_list, enc_id)
			if not new_replacement: return ['error','error']  #could not find a required character
			extra_replacements.append(new_replacement)
			exclude_list.append(new_replacement[1])
	return [extra_replacements, exclude_list]
	
# a required character in a cutscene can also be
# an actual character or template, but the user
# has written it with spaces and capitalization that doesn't exist in the dat files
#  In this case, set it up as a replacement
# and add it to the excluded characters list.
def CSGetTempReplacement(subtype, charlist=[], exclude=[], enc_id=0):
	#first check for an actual character template
	#convert a subtype to lowercase with spaces changed to underscore characters
	test_template = m25string.String_ConvertToTemplate(subtype)
	objects = js.Mission_GetDynamicObjects()
	for obj in objects:
		if (not (obj in exclude)) and ((obj==test_template) or (m25template.GetComplexTemplate(obj)==test_template)):
			return [subtype, obj]
	return None

def CSCheckTemplate(char):
	if js.Marker_Exists(char): return 1
	objects = js.Mission_GetDynamicObjects()
	actualtemplate = m25template.GetComplexTemplate(char)
	for obj in objects:
		if (m25template.GetComplexTemplate(obj) == actualtemplate):
			return 1
	return 0
		
def CSGetReplacements(cs, charlist, enc_id, live=1, exclude=[]):
	block = CSGetElement(cs, 'replace')
	if block == None: return None
	replacements = []
	for n in block:
		if CSCheckForForce(n): continue
		#if live == 1:
		#	live = CSCheckForLive(n)
		subtype = Enc_FerretOutSubtype(n[0])
		char = findCharFromList(charlist,n[1],enc_id,excludelist=exclude,live=live,subtype=subtype)
		if char == '':
			return 'error'
		exclude.append(char)
		replacements.append([n[0],char])
	return replacements

_encspecialreplace = [
	['__allherbud__', "m25ai.AddToList(ffx.GetAllHeroes(),m25freeroamally.GetFreeroamAllies())"],
	['__allvilmin__', "m25ai.AddToList(GetEncVar(enc_id, 'minion'), GetEncVar(enc_id, 'villain'))"],
	['__allvil__', "GetEncVar(enc_id, 'villain')"],
	['__allmin__', "GetEncVar(enc_id, 'minion')"],
	['__allall__', "GetEncVar(enc_id, 'ally')"],
	['__allhe__', "ffx.GetAllHeroes()"],
	['__allbud__', "m25freeroamally.GetFreeroamAllies()"],
]	
		
def CSSpecialReplacements(cs, enc_id):
	global _encspecialreplace
	newcs = []
	newreplace = []
	for e in _encspecialreplace:
		newreplace.append([e[0],eval(e[1])])
	for c in cs:
		newstr = c
		for n in newreplace:
			newstr = string.replace(newstr, n[0], repr(n[1]))
		newcs.append(newstr)
	return newcs
			

def CSCheckForForce(element):
	try:
		return element[2]
	except:
		return 0
		
def CSDoReplacements(cs, replacements):
	newcs = []
	for c in cs:
		newstr = c
		special = 0
		if c[:9] == 'CS_Speak(': 
			val, newstr, remainder = m25string.SplitAroundPhrase(c[9:],',')
			special = val
		for r in replacements:
			newstr = string.replace(newstr, r[0], r[1])
		if special:
			newcs.append('CS_Speak(%s,%s'%(newstr,remainder))
		else:
			newcs.append(newstr)
	return newcs
				
def CSGetElement(cs, element):
	for c in cs:
		try:
			if c[0] == element:
				return c[1:]
		except:
			m25report.Report('ERROR', 'm25enc.CSGetElement - %s ... %s'%(repr(cs),repr(element)))
	return None
	
	
	
	
		
	
	
		
# Triggers	

	
# def EncMinionFleeMap(event):
# 	enc_id = int(event.user)
# 	char = event.object
# 	if IsEncounterComplete(enc_id):
# 		return
# 	m25ai.AIDisable(char)
# 	SetEncVar(enc_id,'minionsescaped', GetEncVar(enc_id,'minionsescaped')+1)
# 	if not PlayCutsceneFromList(GetEncCutscene(enc_id,'minionescapecs'),enc_id,charlist=[[char,'minion']],return_function='EncMinionFleeMap2', return_value=char):
# 		EncMinionFleeMap2(enc_id,char)
# 		
# def EncMinionFleeMap2(event):
# 	enc_id = int(event.user)
# 	char = event.object
# 	js.Object_Destroy(char)
# 	EncCheckEnd(enc_id)

# def EncMinionHurtSpeak(event):
# 	enc_id = int(event.user)
# 	char = event.object
# 	if IsEncounterComplete(enc_id):
# 		return
# 	hurt = Object_GetAttr(char,'health') / Object_GetAttr(char,'maxHealth')
# 	if random.randint(1,100) <= 25:
# 		if hurt <= 0.1:
# 			cs = 'minionneardeathcs'
# 		elif hurt <= 0.25:
# 			cs = 'minionseriouslyhurtcs'
# 		elif hurt <= 0.5:
# 			cs = 'minionbadlyhurtcs'
# 		elif hurt <= 0.75:
# 			cs = 'minionhurtcs'
# 		elif hurt <= 0.9:
# 			cs = 'minionbarelyhurtcs'
# 		else:
# 			cs = 'minionnothurtcs'
# 		PlayCutsceneFromList(GetEncCutscene(enc_id,cs),enc_id,charlist=[[char,'minion']])
# 		
# 		
# def EncHeroHurtSpeak(event):
# 	enc_id = int(event.user)
# 	char = event.object
# 	if IsEncounterComplete(enc_id):
# 		return
# 	hurt = Object_GetAttr(char,'health') / Object_GetAttr(char,'maxHealth')
# 	if random.randint(1,100) <= 25:
# 		if hurt <= 0.1:
# 			cs = 'heroneardeathcs'
# 		elif hurt <= 0.25:
# 			cs = 'heroseriouslyhurtcs'
# 		elif hurt <= 0.5:
# 			cs = 'herobadlyhurtcs'
# 		elif hurt <= 0.75:
# 			cs = 'herohurtcs'
# 		elif hurt <= 0.9:
# 			cs = 'herobarelyhurtcs'
# 		else:
# 			cs = 'heronothurtcs'
# 		PlayCutsceneFromList(GetEncCutscene(enc_id,cs),enc_id,charlist=[[char,'hero']])		

# 			
def EncHeroPickup(target, char):
	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
	Object_Destroy(target)
	SetEncVar(enc_id, 'objects_pickedup', GetEncVar(enc_id, 'objects_pickedup',default=0)+1)
	PlayCutsceneFromList(GetEncCutscene(enc_id,'heropickupcs'),enc_id,charlist=[[char,'hero']])
	EncEventCallback(enc_id, 'on_hero_pickup', [target,char])
		
def EncMakeObject(enc_id, var, template, pos, indestructible=0):
	name = m25name.GetNextObjectName()
	m25_spawn(name,template,pos)
	SetEncVar(enc_id, var, name)
	if indestructible:
		Object_SetAttr(name,'minHealth',1)
	m25getset.SetObjectVar(name,'encounter',enc_id)
	Enc_SetCharEnc(name, enc_id)	
	return name	

		
# setting up state
def StateCount(enc_id, name):
	try:
		val = len(GetEncVar(enc_id,name))
		return val
	except:
		return 0
	
	
# setting up triggers

# def GetEncObjectList(enc_id, name):
# 	return GetEncVar(enc_id, name)
# 	#return readObjectList('enc_' + `enc_id` + '_' + name)


		
def OnEachKO(enc_id, groupname, function):
	chars = GetEncVar(enc_id, groupname)
	print 'ONEACHKO'
	print chars
	print '----'
	for c in chars:
		cshelper.regDeath(c,function,user=enc_id)
	
# def OnEachEnterMarker(enc_id, groupname, markername, function):
# 	chars = GetEncVar(enc_id, groupname)
# 	completemarkername = GetEncVar(enc_id,markername)
# 	#completemarkername = markername + `int(GetEncVar(enc_id,'start'))` + `int(GetEncVar(enc_id, markertype))`
# 	for c in chars:
# 		cshelper.regMarkerEnter(c,completemarkername,function,user=enc_id)

def Enc_OnEachHurt(enc_id, groupname, function):
 	chars = GetEncVar(enc_id, groupname)
 	for c in chars:
 		cshelper.regDamage(c,function,enc_id,persistent=1)		
		
def OnGroupMemberKO(enc_id, enc_groupname, function):
	groupname = EncGetGroupName(enc_id, enc_groupname)
	m25ai.RegGroupMemberKO(groupname, function)
	
def OnGroupLeaderKO(enc_id, enc_groupname, function):
	groupname = EncGetGroupName(enc_id, enc_groupname)
	m25ai.RegGroupLeaderKO(groupname, function)

def OnGroupLeaderSet(enc_id, enc_groupname, function):
	groupname = EncGetGroupName(enc_id, enc_groupname)
	m25ai.RegGroupLeaderSet(groupname, function)			
		
def OnHeroHurt(enc_id, function):
	heroes = ffx.GetAllHeroes()
	for h in heroes:
		cshelper.regDamage(h,function,enc_id,persistent=1)		

def OnRandomTimer(enc_id, mintime, maxtime, function):
	SetTimer(random.randint(mintime,maxtime),function,user=enc_id)

def OnTimer(enc_id, time, function):
	SetTimer(time, function, '', `time`, user=enc_id)
				
# setting up goals		


		
# helpers

def EncEnableAllChars(enc_id, listname):
	chars = GetEncVar(enc_id, listname, default=[])
	for c in chars:
		m25ai.AIEnable(c)
		


		
# start and end markers

#--------------------------------------------------------------
#
# Find a set of markers and return a random one from the set.
# markers must start with 'name' and end with a number 
# in order to be found.
#
#--------------------------------------------------------------
	
def EncFindRandomMarker(name):
	start_locs = getMarkerListByName(name)
	if start_locs == []:
		return None
	else:
		done = 0
		tries = 0
		val = 0
		while tries < 10 and not done:
			#val = random.randint(1,len(start_locs))
			#if not EncMarkerTooClose(name + `val`) and not EncMarkerTooFar(name+`val`) and not EncStartTaken(name+`val`):
			val = start_locs[random.randint(0,len(start_locs)-1)]
			if not EncMarkerTooClose(val) and not EncMarkerTooFar(val) and not EncStartTaken(val):
				done = 1
			tries = tries + 1
		if done: return val
	#random again, but this time don't worry about far distance
	tries = 0
	while tries < 10:
		#val = random.randint(1,len(start_locs))
		#if not EncStartTaken(name+`val`) and not EncMarkerTooClose(name + `val`): return val
		val = start_locs[random.randint(0,len(start_locs)-1)]
		if not EncMarkerTooClose(val) and not EncStartTaken(val): return val
		tries = tries + 1
	#random pick didn't work so pick the first one not taken
	val = 0
	for s in start_locs:
		#val = val + 1
		#if not EncStartTaken(name +`val`): return val
		val = s
		if not EncStartTaken(val): return val
	#nothing exists, so return none
	return None

def EncFindMarkers(enc_id, mlist):
	markers = []
	for m in mlist:
		name = EncFindRandomMarker(m[1])
		if name == None:
			m25report.Report(ENC_CHANNEL, 'No markers for ' + `enc_id` + ' ' + `mlist`)
			#start create a new marker
			num = 1
			name = m[1] + `int(num)`
			start = GetEncVar(enc_id, 'start')
			marker_type = ''
			try:
				marker_type = m[2]
			except:
				pass
			if not EncCreateMarker(name, js.Get_MarkerPos(start), marker_type=marker_type, num=num):
				return None
			#end create a new marker
			# return None
		#SetEncVar(enc_id,m[0],m[1]+`int(num)`,save=0)
		SetEncVar(enc_id,m[0],name,save=0)
		markers.append(name)
	return markers
	
def EncCreateAllMarkers():
	global _enc_markers
	encmarkers = getMarkerListByName('encounter')
	dotype = ['villain','ally','minion','object','interloper']
	for e in encmarkers:
		mlist = []
		pos = js.Get_MarkerPos(e)
		for d in dotype:
			mlist = []
			for n in range(1,24):
				marker = EncCreateMarker('%s_%s%i'%(e,d,n), pos, num =n)
				if marker: mlist.append(marker)
			_enc_markers['%s_%s'%(e,d)] = mlist
		mlist = []
		for i in range(1,4):
			marker = EncCreateMarker('%s_allyleave%i'%(e,i), pos, marker_type='dist500')
			if marker: mlist.append(marker)
		_enc_markers['%s_allyleave'%(e)] = mlist
			
		
	
def EncCreateMarker(name, loc, randomize=1, marker_type='', num=-1):
	if js.Marker_Exists(name): return name
	newloc = loc
	heading = 0
	if marker_type == 'edge':
		minx, miny, minz, maxx, maxy, maxz = 0,0,0, 0,0,0
		try:
			minx, miny, minz, maxx, maxy, maxz = skXMapInfo.MapInfo['mapExtents']
		except:
			maxx, maxy, maxz, minx, miny, minz = EncFindMapBounds()
		done, numtries = 0, 0
		while not done and (numtries < 10):
			numtries = numtries + 1
			i = randint(1,4)
			if i == 1:
				newloc = (minx+30, loc[1], loc[2])
			elif i == 2:
				newloc = (maxx-30, loc[1], loc[2])
			elif i == 3:
				newloc = (loc[0], miny+ 30, loc[2])
			else:
				newloc = (loc[0], minx-30, loc[2])
			if m25distance.GetTupleDistance(loc,newloc) >= m25distance.DISTANCE_LONG:
				done = 1
	elif marker_type == 'dist100':
		newloc = (loc[0]+100 + randint(-6,6)*5, loc[1]+100 + randint(-6,6)*5, loc[2])
	elif marker_type == 'dist300':
		newloc = (loc[0]+300 + randint(-6,6)*5, loc[1]+300 + randint(-6,6)*5, loc[2])
	elif marker_type == 'dist500':
		newloc = (loc[0]+500 + randint(-6,6)*5, loc[1]+500 + randint(-6,6)*5, loc[2])
	elif randomize:
		grid = [
			[0,0], [0,-30], [20,-30], [-20,-30], 
			[0,-10], [20,-10], [-20,-10],
			[40,-30], [-40,-30], [40, -10], [-40, -10],
			[0, 10], [20, 10], [-20, 10], [40, 10], [-40, 10],
			[0, 30], [20, 30], [-20, 30], [40, 30], [-40, 30],
			[0, 50], [20, 50], [-20, 50], [40, 50], [-40, 50]
		]
		#group standard types
		if string.find(name,'villain') != -1:
			if num != -1 and len(grid) > num:
				newloc = (loc[0] + grid[num][0], loc[1]+30+grid[num][1], loc[2])
			else:
				newloc = (loc[0]  + randint(-6,6)*5, loc[1] + 30 + randint(-6,6)*5, loc[2])
		elif string.find(name,'minion') != -1:
			if num != -1 and len(grid) > num:
				newloc = (loc[0] + grid[num][0], loc[1]+60+grid[num][1], loc[2])
			else:
				newloc = (loc[0] + randint(-6,6)*5, loc[1] + 60 + randint(-6,6)*5, loc[2])
		elif string.find(name,'ally') != -1:
			heading = 135
			if num != -1 and len(grid) > num:
				newloc = (loc[0] + grid[num][0], loc[1]-30+grid[num][1], loc[2])
			else:
				newloc = (loc[0] + randint(-6,6)*5, loc[1] - 30 + randint(-6,6)*5, loc[2])
		elif string.find(name,'object') != -1:
			heading = randint(0,359)
			if num != -1 and len(grid) > num:
				newloc = (loc[0]-60 - grid[num][1], loc[1] - grid[num][0], loc[2])
			else:
				newloc = (loc[0] + randint(-6,6)*5, loc[1] + randint(-6,6)*5, loc[2])
		elif string.find(name,'rival') != -1:
			heading = 90
			if num != -1 and len(grid) > num:
				newloc = (loc[0]+60+ grid[num][1], loc[1] + grid[num][0], loc[2])
			else:
				newloc = (loc[0] + randint(-6,6)*5, loc[1] + randint(-6,6)*5, loc[2])
		elif string.find(name,'cohort') != -1:
			heading = 90
			if num != -1 and len(grid) > num:
				newloc = (loc[0]+90+ grid[num][1], loc[1] + grid[num][0], loc[2])
			else:
				newloc = (loc[0] + randint(-6,6)*5, loc[1] + randint(-6,6)*5, loc[2])
		else:
			newloc = (loc[0] + randint(-6,6)*5, loc[1] + randint(-6,6)*5, loc[2])
	else:
		newloc = loc
 	try:
 		newloc = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding(newloc) 		
 	except: 
 		pass
	try:
		js.Marker_Spawn(name, newloc, heading)
# 		if string.find(name,'object') != -1:
# 			cshelper.spawn(name+'t','cub_market_crate1',name)
# 		if string.find(name,'villain') != -1:
# 			cshelper.spawn(name+'t','cub_palm1',name)
		return name
	except:
		return None

		
def EncOutsideMapBounds(point):
	bounds = EncFindMapBounds()
	if (point[0] > bounds[0]) or (point[0] < bounds[3]):
		print 'x dimension too high or too low'
	if (point[1] > bounds[1]) or (point[1] < bounds[4]):
		print 'y dimension too high or too low'

def EncFindMapBounds():
	return datfiles.Mission_GetMapExtents()
		
def EncFindStartMarker(enc_id, name):
	num = EncFindRandomMarker(name)
	if num == None: 
		return None
	else:
		#SetEncVar(enc_id,'start',name+`num`,save=0)
		SetEncVar(enc_id,'start',num,save=0)
		return num
		
# 	if EncStartTaken(name+`num`):
# 		start_locs = getMarkerListByName(name)
# 		num = 0
# 		for s in start_locs:
# 			if not EncStartTaken(name+`num+1`):
# 				SetEncVar(enc_id,'start',name + `num+1`)
# 				return num+1
# 			num = num + 1
# 	else:
# 		SetEncVar(enc_id,'start',name+`num`)
# 		return num
# 	return None


def EncMarkerTooClose(marker):
	#if we are in a base screen, then don't worry about too close
	if m25gametype.GameType_InBase():
		return 0
	heroes = ffx.GetAllHeroes()
	for h in heroes:
		if m25distance.GetDistanceToMarker(h,marker) < 400:
			return 1
	return 0
	
def EncMarkerTooFar(marker):
	heroes = ffx.GetAllHeroes()
	for h in heroes:
		if m25distance.GetDistanceToMarker(h,marker) > 2000:
			return 1
	return 0	
	
	
#
#  Ensure that two encounters don't spawn on the same spot
#
	
def EncStartTaken(marker):
	return js.Mission_TestAttr(marker+'_taken')
	#return m25getset.m25getset.GetMissionVar(marker+'_taken',default=0)
	
def EncStartFree(marker):
	js.Mission_SetAttr(marker+'_taken',0)
	#m25getset.SetMissionVar(marker+'_taken',0)	
	
def EncStartTake(marker):
	js.Mission_SetAttr(marker+'_taken',1)
	#m25getset.SetMissionVar(marker+'_taken',1)	
	
	
def CreateMarkerListByName(enc_id, marker, marker_type=''):
	marker_list = []
	loc = js.Get_MarkerPos(GetEncVar(enc_id, 'start'))
	for i in range(1,24):
		marker_list.append(EncCreateMarker(marker+`i`, loc, num=i, marker_type=marker_type))
	return marker_list
	
def Enc_GetNearestBuilding(marker):
	objects = skXMapInfo.MapInfo_GetObjects (js.Get_MarkerPos(marker),1000)
	for o in objects:
		if js.Object_GetClass(o) & js.OC_BUILDING:
			return o
	return None
	
def Enc_GetNearestObject(marker):
	objects = skXMapInfo.MapInfo_GetObjects (js.Get_MarkerPos(marker))
	for o in objects:
		if m25ai.ST_object(o):
			return o
	return None
	
	
	
	
	
	


	

#---------------------------------------------------------------------
#
# Custom Encounter Callbacks
#
#---------------------------------------------------------------------	


def EncAddReg(enc_id, function_name='', var=''):
	glist = GetEncVar(enc_id, var, default=[])
	glist.append(function_name)
	SetEncVar(enc_id, var, glist)	
	m25report.Report(ENC_CHANNEL, '%i - registering %s on variable %s'%(enc_id, function_name, var))

	
def EncEventCallback(enc_id, var, values=[]):
	glist = GetEncVar(enc_id, var, default=[])
	m25report.Report(ENC_CHANNEL, '%i - event callback %s has value %s'%(enc_id, var, repr(glist)))
	for g in glist:
		try:
			m25report.Report(ENC_CHANNEL, '%i - event callback %s calling %s'%(enc_id, var, g))
			exec 'ff.' + g + '(' + `enc_id` + ',' + `values` + ')'
		except:
			m25report.Report(ENC_CHANNEL, g + ' problem with registered function ' + str(g))
			print sys.exc_type, sys.exc_value	
			raise
	
#---------------------------------------------------------------------
#
# Standardized Encounter functions
#
#---------------------------------------------------------------------

def SetupEncounter(enc_id, encounter):
	start = time.clock()
	m25report.Report(ENC_CHANNEL, 'Setting up encounter ' + `enc_id`)
	#SetEncVar(enc_id, '_encounter', encounter)
	start_num = SetupEnc_Start(enc_id, encounter)
	if start_num == None: return 0
	if not SetupEnc_Markers(enc_id, start_num, encounter): return 0
	if not SetupEnc_Spawn(enc_id, start_num, encounter): return 0
	m25report.Report(ENC_CHANNEL, 'Encounter spawn complete for ' + `enc_id`)
	SetupEnc_Cutscenes(enc_id, encounter)
	m25report.Report(ENC_CHANNEL, 'Encounter cutscenes complete for ' + `enc_id`)
	if not SetupEnc_Functions(enc_id, encounter): return 0
	if not SetupEnc_State(enc_id, encounter): return 0
	SetupEnc_Goals(enc_id, encounter)
	m25report.Report(ENC_CHANNEL, 'Encounter goals complete for ' + `enc_id`)
	SetupEnc_Objectives(enc_id, encounter)
	m25report.Report(ENC_CHANNEL, 'Encounter objectives complete for ' + `enc_id`)
	SetupEnc_Parameters(enc_id, encounter)
	m25report.Report(ENC_CHANNEL, 'Encounter parameters complete for ' + `enc_id`)
	SetEncounterNotComplete(enc_id)
	#mark the start marker as taken, so other encounters don't try to use it
	EncStartTake(GetEncVar(enc_id,'start'))
	SetupEnc_CallPost(enc_id, encounter)
	m25report.Report(ENC_CHANNEL, 'Encounter post complete for ' + `enc_id`)
	Enc_SaveCharacterVars()
	Enc_SaveDict(enc_id)
	m25ai.AI_SaveGroupData()
	PlayEncCutscene(enc_id, 'alertcs')
	#use banter in all encounters by default
	Enc_Banter(enc_id)
	Enc_BuddyBanter(enc_id)
	m25report.Report(ENC_CHANNEL, 'Encounter setup complete for ' + `enc_id`)
	m25report.ReportTime('Encounter Setup for %i'%(enc_id), (time.clock()-start))
	#m25report.Report('TIME', 'Encounter setup for %i took %f seconds'%(enc_id,time.clock()-start))
	return 1

def SetupEnc_Start(enc_id, encounter):
	if encounter.has_key('start'):
		start_num = EncFindStartMarker(enc_id, encounter['start'][0])
		m25report.Report(ENC_CHANNEL, 'Start marker found for ' + `encounter['start']` + ' ' + `start_num`)
		if start_num == None:
			m25report.Report(ENC_CHANNEL, 'No start marker found for ' + `encounter['start']`)
			return None
		else:
			return start_num
	else:
		m25report.Report(ENC_CHANNEL, 'Start marker not defined')
		return None
		
def SetupEnc_Markers(enc_id, start_num, encounter):
	if encounter.has_key('markers'):
		mlist = []
		for m in encounter['markers']:
			marker_type = ''
			try: marker_type = m[2]
			except: pass
			#mlist.append([m[0],encounter['start'][0] + `start_num` + '_' + m[1],marker_type])
			mlist.append([m[0],start_num + '_' + m[1],marker_type])
		found_markers = EncFindMarkers(enc_id, mlist)
		if found_markers == None:
			m25report.Report(ENC_CHANNEL, 'Unable to find other markers for ' + `encounter['start']`)
			return 0
		else:
			return 1
	else:
		#m25report.Report(ENC_CHANNEL, 'No additional markers defined for ' + `encounter['start']`)
		return 1
	
		
def SetupEnc_Spawn(enc_id, start_num, encounter):
	if encounter.has_key('spawn'):	
		for s in encounter['spawn']:
			#EncSpawnGroup(enc_id,encounter['start'][0]+`start_num` +'_'+ s[1] ,s[2],s[0],s[3],marker_type=s[4],create_markers=s[5])
			EncSpawnGroup(enc_id,start_num +'_'+ s[1] ,s[2],s[0],s[3],marker_type=s[4],create_markers=s[5])
		#post spawn hook
		if encounter.has_key('postspawn'):
			for sp in encounter['postspawn']:
				try:
					exec 'ff.' + sp + '(' + `enc_id` + ')'
				except:
					m25report.Report('ERROR', 'problem with post spawn function for %i'%(enc_id))
					pass
		return 1
	else:
		return 1
	
def SetupEnc_Cutscenes(enc_id, encounter):
	if encounter.has_key('cutscenes'):
		cs = []
		for c in encounter['cutscenes']:
			if type(c) == types.StringType:
				if c[0:len('randomize_')] == 'no_randomize_':
					SetEncVar(enc_id, '_cs_%scs'%c, 1, save=0)
			else:
				cs.append(c)
		SetEncCutscenes(enc_id, cs)

def SetupEnc_Functions(enc_id, encounter):
	if encounter.has_key('check_end') and encounter.has_key('return_function'):
		SetEncFunctions(enc_id, [['checkend', encounter['check_end']], ['returnfunction', encounter['return_function']]])
		return 1
	else:
		m25report.Report('ERROR', 'Encounter has no end and/or return function defined ' + `encounter['start']`)
		print encounter
		return 0

def SetupEnc_State(enc_id, encounter):
	if encounter.has_key('state'):
		newstate = []
		for s in encounter['state']:
			value = 0
			if m25string.IsFunction(s[1]):
				try:
					exec 'value = ff.' + m25string.GetLeftString(s[1],'(') + '(' + `enc_id` + ',' + m25string.GetRightString(s[1],'(')
				except:
					m25report.Report('ERROR', 'Error in encounter state')
					print 'value = ff.' + m25string.GetLeftString(s[1],'(') + '(' + `enc_id` + ',' + m25string.GetRightString(s[1],'(')
					return 0
			else:
				value = s[1]
			newstate.append([s[0],value])
		SetEncStateVars(enc_id,newstate,save=0)		
	return 1
	
#save any additional encounter parameters so they can be retrieved later	
def SetupEnc_Parameters(enc_id, encounter):
	for param in encounter.keys():
		if param not in ['cutscenes','state','triggers','goals','start','markers','spawn','objectives','setup','check_end','return_function', 'postspawn']:
			SetEncVar(enc_id,param,encounter[param],save=0)	
	
def SetupEnc_Goals(enc_id, encounter):
	if encounter.has_key('triggers'):
		for t in encounter['triggers']:
			try:
				exec 'ff.' + m25string.GetLeftString(t,'(') + '(' + `enc_id` + ',' + m25string.GetRightString(t,'(')
			except:
				m25report.Report(ENC_CHANNEL, 'Error in encounter triggers')
				print 'ff.' + m25string.GetLeftString(t,'(') + '(' + `enc_id` + ',' + m25string.GetRightString(t,'(')
				raise
	if encounter.has_key('goals'):
		for g in encounter['goals']:
			try:
				exec 'ff.' + m25string.GetLeftString(g,'(') + '(' + `enc_id` + ',' + m25string.GetRightString(g,'(')
			except:
				m25report.Report('ERROR', 'Error in encounter goals')
				print 'ff.' + m25string.GetLeftString(g,'(') + '(' + `enc_id` + ',' + m25string.GetRightString(g,'(')
				raise

def SetupEnc_Objectives(enc_id, encounter):
	if encounter.has_key('objectives'):
		SetEncVar(enc_id,'objectives',encounter['objectives'],save=0)
		#num = 1
		for e in encounter['objectives']:
			# [type, description, prestige, cp]
			#check for silent objectives (used in freeroam to give xp and prestige without creating actual new objectives)
			if (e[0] > -1) and (e[1] != ''):
				m25objective.Objective_AddObjective(objective=e[1], prestige=e[2], objectivetype=e[0])
				#js.Objective_Add(`enc_id`+'objective'+`num`, e[1], e[0], prestige)	
			#num = num + 1

def SetupEnc_CallPost(enc_id, encounter):
	if encounter.has_key('setup'):
		for s in encounter['setup']:
			try:
				m25report.Report(ENC_CHANNEL, 'Encounter post function ' + `s`)
				exec 'ff.' + s + '(' + `enc_id` + ')'				
			except:
				m25report.Report('ERROR', 'Error in encounter post function ' + `s`)
				print 'ff.' + s + '(' + `enc_id` + ')'
				print sys.exc_type, sys.exc_value
				raise

	
#---------------------------------------------------------------------
#
# Putting encounters together
#
#---------------------------------------------------------------------
	
def CombineEncounters(enc1, enc2):
	return CombineEncounterData(enc1, enc2)

def CombineEncounterData(enc1, enc2):
	encounter = {}
	for key in enc1.keys():
		if enc2.has_key(key):
			encounter[key] = EncRemoveDuplicates(m25ai.AddToList(enc1[key], enc2[key]))
		else:
			encounter[key] = EncRemoveDuplicates(m25ai.AddToList(enc1[key],[]))
	for key in enc2.keys():
		if not enc1.has_key(key):
			encounter[key] = EncRemoveDuplicates(m25ai.AddToList(enc2[key],[]))
	return encounter

def EncRemoveDuplicates(list):
	newlist = []
	for item in list:
		if item not in newlist: newlist.append(item)
	return newlist
	
def EncGetNextEncID():
	global _encounter_dict
	current = m25getset.GetMissionVar('__current_enc_id', default=0, integer=1) + 1
	done = 0
	while not done:
		if _encounter_dict.has_key(current): current = current + 1
		else: done = 1	
	m25getset.SetMissionVar('__current_enc_id',current)
	return current
	
	
def PrintEncounter(enc):
	print '---------------------'
	for k in enc.keys():
		print `k` + ' : ' + `enc[k]`
	print '---------------------'
	
#####################################################################################
#
# Encounter Building Blocks
# 
# Description: 
#
#####################################################################################	

#---------------------------------------------------------------------
#
# Examine Object
#
#---------------------------------------------------------------------	

def EncExamine(objectgroup='', command='EXAMINE', cutscene='examinecs', allow_enemies=0):
	encounter = {}
	encounter['state'] = [
		['examine_objectgroup', objectgroup],
		['examine_command', command],
		['examine_complete', 0],
		['examine_object', ''],
		['examine_cutscene', cutscene],
		['examine_allow_enemies', allow_enemies],
	]
	encounter['setup'] = [
		'Examine_Setup'
	]
	return encounter

def Examine_Setup(enc_id):
	group = GetEncVar(enc_id, 'examine_objectgroup')
	members = GetEncVar(enc_id, group)
	chosen = members[0]
	js.Object_SetAttr(chosen,'minHealth',js.Object_GetAttr(chosen,'maxHealth')+1)
	#fix it so that objects with 'damagedOnMove' don't get destroyed when they are hit (and thus can't be examined)
	if js.Object_GetAttr(chosen, 'damagedOnMove'):
		js.Object_SetAttr(chosen, 'damagedOnMove', 0)
	SetEncVar(enc_id, 'examine_object', chosen)
	cshelper.addCustomCommand(GetEncVar(enc_id, 'examine_command'),chosen,'Examine_Try')
	examined = GetEncVar(enc_id,'on_examined',default=[])
	for e in examined: EncAddReg(enc_id, e, '_examined')
	
def Examine_Try(object, hero):
	enc_id = Enc_GetCharEnc(object)
	if GetEncVar(enc_id, 'examine_allow_enemies') or CanInterrogate(object, hero):
		cshelper.removeAttention(object)
		PlayCutsceneFromList(GetEncVar(enc_id,GetEncVar(enc_id,'examine_cutscene')), enc_id, charlist=[[hero,'hero']], return_function = 'Examine_End(' + `enc_id` + ')')
	else:
		cshelper.addCustomCommand(GetEncVar(enc_id, 'examine_command'),object,'Examine_Try')
	
def Examine_End(enc_id):
	if GetEncVar(enc_id, 'examine_complete', default=0): return
	SetEncVar(enc_id, 'examine_complete', 1)
	EncEventCallback(enc_id, '_examined', GetEncVar(enc_id,'examine_object'))
	EncCheckEnd(enc_id)
	
	
#---------------------------------------------------------------------
#
# Flee To Marker
#
#---------------------------------------------------------------------		
	
def EncFleeToMarker(group='minion',marker_id='end',marker_prefix='end',mass_flee=0):
	encounter = {}
	encounter['markers'] = [ [marker_id,marker_prefix,'edge'], ]
	encounter['state'] = [
		[group+'s_escaped',0],
		['leader_fled', 0],
		['flee_marker', marker_id],
		['mass_flee', mass_flee],
	]
	
	encounter['setup'] = [
		'FleeToMarker_Setup'
	]
	if mass_flee:
	 	encounter['goals'] = [
 			'Enc_EachFleeToMarker('+`group`+','+`marker_id`+',"FleeToMarker_FleeMap()")',
	 	]
	else:
		encounter['triggers'] = [
			'OnGroupLeaderSet(' + `group` + ',"FleeToMarker_LeaderSet")',
		]
	return encounter


def FleeToMarker_Setup(enc_id):	
	m25ai.SetGroupStayWithTeam(EncGetGroupName(enc_id,'minion'),on=0)
	fled = GetEncVar(enc_id, 'on_fled', default=[])
	for e in fled: 
		EncAddReg(enc_id, e, '_fled')
	
def FleeToMarker_LeaderSet(groupname, leader):
	enc_id = m25ai.GetGroupVar(groupname, 'enc_id')
	FleeToMarker_ResumeMoveGoal(enc_id, groupname)
	#m25ai.AddGoal_MoveToMarker(leader,marker=GetEncVar(enc_id,GetEncVar(enc_id, 'flee_marker')),priority=m25ai.TACTIC_PRIORITY_MEDIUM,pct=60,fn='FleeToMarker_FleeMap('+`enc_id`+')')
	#m25ai.AddGoal_MoveToMarker(leader, GetEncVar(enc_id,GetEncVar(enc_id, 'flee_marker')), pct='DHealthWatch()', fn='FleeToMarker_FleeMap('+`enc_id`+')')

def FleeToMarker_PauseMoveGoal(enc_id, groupname):
	leader = m25ai.Group_GetLeader(groupname)
	if not leader: return
	goal_id = m25ai.AIHasGoal(leader,'Goal_MoveToMarker(')
	if goal_id:
		m25ai.AIRemoveGoalByID(goal_id,leader,'Goal_MoveToMarker(')
	
def FleeToMarker_ResumeMoveGoal(enc_id, groupname):
	leader = m25ai.Group_GetLeader(groupname)
	if not leader: return
	m25ai.AddGoal_MoveToMarker(leader,marker=GetEncVar(enc_id,GetEncVar(enc_id, 'flee_marker')),priority=m25ai.TACTIC_PRIORITY_MEDIUM,pct=80,fn='FleeToMarker_FleeMap('+`enc_id`+')')
	
def FleeToMarker_FleeMap(char, enc_id=None):
	if enc_id == None:
		enc_id = Enc_GetCharEnc(char)
	m25ai.AIDisable(char)
	gp = GetEncClassPrefix(char)
	#set min health so the character isn't inadvertenly wiped out while playing the cutscene
	Object_SetAttr(char,'minHealth',1)
	SetEncVar(enc_id,gp+'s_escaped', GetEncVar(enc_id,gp+'s_escaped',default=0)+1)
	#try to play a cutscene for the escapee
	charlist = [[char,gp]]  #force the cutscene to be played for the character that escaped
	cutscene_name = gp+'escapescs'
	if GetEncVar(enc_id, char+'escapescs',default=None):
		cutscene_name = char+'escapescs'
	else:
		if GetEncVar(enc_id, m25template.GetComplexTemplate(char)+'escapescs',default=None):
			cutscene_name = m25template.GetComplexTemplate(char)+'escapescs'		
	PlayEncCutscene(enc_id, cutscene_name, charlist=charlist, return_function='FleeToMarker_FleeMap2('+`enc_id`+','+`char`+')')


def FleeToMarker_FleeMap2(enc_id,char):
	if char == m25ai.CharGetGroupLeader(char):
		gp = GetEncGroupPrefix(char)
		groupname = EncGetGroupName(enc_id,gp)
		m25event.SetTimer(1,'FleeToMarker_FleeMap3',groupname,user=enc_id)
		js.Object_Destroy(char)
	else:
		EncEventCallback(enc_id, '_fled', [char,'not leader'])
		js.Object_Destroy(char)
		EncCheckEnd(enc_id)
	
		
def FleeToMarker_FleeMap3(event):
	enc_id = int(event.user)
	groupname = event.object
	SetEncVar(enc_id, 'leader_fled', 1)
	m25ai.Group_PromoteLeader(groupname)
	EncEventCallback(enc_id, '_fled', ['', 'leader'])
	EncCheckEnd(enc_id)
	
#---------------------------------------------------------------------
#
# Guard object
#
#---------------------------------------------------------------------	
	
def EncGuardObject(group_id='minion', guard_object=None):
	encounter = {}
	encounter['state'] = [
		['guard_object',guard_object],
	]
	encounter['triggers'] = [
		'OnGroupLeaderSet('+`group_id` + ', "GuardObject_LeaderSet")',
	]
	return encounter

def GuardObject_LeaderSet(event):
	m25ai.GroupCommand_Guard(EncGetGroupName(int(event.user),GetEncGroupPrefix(event.object)), GetEncVar(int(event.user),'guard_object'))	
	
	
#---------------------------------------------------------------------
#
# Capture The Object
#
#---------------------------------------------------------------------	
		
def EncCapture(object='cub_market_crate2', groups=['minion'], on_ground=0):
	encounter = {}
	sObState, objgroup = 'carried', groups[0]
	if on_ground: sObState, objgroup = 'dropped', None
	encounter['state'] = [
		['object_state', sObState],
		['object_template', object],
		['object_group', 'minion'],
		['loot', None],
		['capture_groups', groups],
	]
	encounter['setup'] = [ 'Enccapture_Setup']
	encounter['markers'] = [ ['item','item','']]
	triglist = []
	for g in groups:
		triglist.append('OnGroupLeaderKO("%s", "Enccapture_LeaderKO")'%(g))
		triglist.append('OnGroupLeaderSet("%s","Enccapture_LeaderSet")'%(g))
	encounter['triggers'] = triglist
	encounter['on_fled'] = ['Enccapture_Escaped']
	return encounter

def Enccapture_Setup(enc_id):
	pickup = GetEncVar(enc_id, 'hero_pickup',default=None)
	if pickup != None:
		for p in pickup: EncAddReg(enc_id, p, 'on_hero_pickup')
	dropped = (GetEncVar(enc_id,'object_state') == 'dropped')
	if dropped:
		EncMakeObject(enc_id, 'loot', GetEncVar(enc_id,'object_template'), GetEncVar(enc_id,'item'), 1)
		for g in GetEncVar(enc_id, 'capture_groups', default=[]):
			m25ai.GroupCommand_RetrieveItem(g, GetEncVar(enc_id,'loot'), priority=TACTIC_PRIORITY_HIGH, function='Enccapture_ObjectRetrieved('+`g`+')')
	else:
		carrier_group = EncGetGroupName(enc_id, GetEncVar(enc_id, 'object_group'))
		for g in GetEncVar(enc_id, 'capture_groups', default=[]):
			if g != carrier_group:
				m25ai.GroupCommand_Attack(g, m25ai.Group_GetLeader(carrier_group))
				m25ai.GroupCommand_MoveToTarget(g, m25ai.Group_GetLeader(carrier_group), priority=TACTIC_PRIORITY_HIGH)
						
def Enccapture_LeaderKO(groupname, leader):
	enc_id = Enc_GetCharEnc(leader)
	if groupname == EncGetGroupName(enc_id, GetEncVar(enc_id, 'object_group')):   #leader was carrying the flag
		#drop the flag
		name = EncMakeObject(enc_id, 'loot', GetEncVar(enc_id,'object_template'), js.Get_ObjectPos(leader), 1)
		#add a command so the player can retrieve it
		cshelper.addCustomCommand('take',name,'Enccapture_HeroPickup')
		#make everyone chase the flag
		for g in GetEncVar(enc_id, 'capture_groups', default=[]):
			m25ai.GroupCommand_RetrieveItem(g, GetEncVar(enc_id,'loot'), priority=TACTIC_PRIORITY_HIGH, function='Enccapture_ObjectRetrieved('+`g`+')')
		#play a cutscene
		PlayEncCutscene(enc_id, 'droppeditemcs', charlist=[[leader,'dropper']])
	
def Enccapture_ObjectRetrieved(char, groupname):
	enc_id = Enc_GetCharEnc(char)
	SetEncVar(enc_id,'object_state','carried')
	SetEncVar(enc_id, 'object_group', GetEncGroupPrefix(char))
	money = GetEncVar(enc_id,'loot')
	js.Object_Destroy(money)
	#EncPlayCutscene('itemretrievedcs',enc_id,charlist=[[char,GetEncClassPrefix(char)]])
	PlayEncCutscene(enc_id, 'itemretrievedcs')
	for g in GetEncVar(enc_id, 'capture_groups', default=[]):
		if g != groupname:
			m25ai.GroupCommand_Attack(g, [char])

def Enccapture_Escaped(enc_id, values=['','not leader']):
	if values[1] == 'leader':
		if GetEncVar(enc_id, 'object_state') == 'carried' and (GetEncVar(enc_id, 'object_group') == GetEncGroupPrefix(values[0])):
			SetEncVar(enc_id, 'object_state', 'lost')
			PlayEncCutscene(enc_id, 'itemlostcs')

def Enccapture_HeroPickup(object, hero):
	enc_id = Enc_GetCharEnc(object)
	SetEncVar(enc_id, 'object_state', 'hero carried')
	SetEncVar(enc_id, 'object_group', None)
	EncHeroPickup(object, hero)
	for g in GetEncVar(enc_id, 'capture_groups', default=[]):
		m25ai.GroupCommand_MoveToTarget(g, hero, priority=TACTIC_PRIORITY_HIGH)

def Enccapture_LeaderSet(groupname, leader):
	enc_id = Enc_GetCharEnc(leader)
	if GetEncVar(enc_id, 'object_state') == 'dropped':
		m25ai.GroupCommand_RetrieveItem(groupname, GetEncVar(enc_id,'loot'), function='Enccapture_ObjectRetrieved('+`groupname`+')')
			
	

#---------------------------------------------------------------------
#
# Leader Drop Object
#
#---------------------------------------------------------------------	
	
def EncLeaderDropObject(object='crate', group='minion', flee_marker=None, retrieve=0):
	encounter = {}
	encounter['state'] = [
		['object_state','carried'],			#0:carried, 1:dropped, 2:lost, 3:hero carried, 4:returned
		['object_template',object],
	]
	encounter['setup'] = [ 'LeaderDropObject_Setup']
	encounter['triggers'] =['OnGroupLeaderKO('+`group`+',"LeaderDropObject_LeaderKO")',]
	encounter['on_fled'] = ['LeaderDropObject_Escaped']
	if retrieve:
		encounter['triggers'].append('OnGroupLeaderSet("minion","LeaderDropObject_LeaderSet")')
	return encounter

def LeaderDropObject_Setup(enc_id):
	pickup = GetEncVar(enc_id, 'hero_pickup',default=None)
	if pickup != None:
		for p in pickup: EncAddReg(enc_id, p, 'on_hero_pickup')
	
def LeaderDropObject_LeaderKO(groupname, leader):
	enc_id = m25ai.GetGroupVar(groupname, 'enc_id', integer=1)
	if GetEncVar(enc_id, 'object_state') == 'carried':
		SetEncVar(enc_id,'object_state','dropped')
		name = EncMakeObject(enc_id, 'loot', GetEncVar(enc_id,'object_template'), js.Get_ObjectPos(leader), 1)
		SetEncVar(enc_id, 'loot', name)
		cshelper.addCustomCommand('take',name,'LeaderDropObject_HeroPickup')
		PlayCutsceneFromList(GetEncCutscene(enc_id,'droppedlootcs'),enc_id,charlist=[[leader,'minion']])				

def LeaderDropObject_Escaped(enc_id, values=['','not leader']):
	#char = event.object
	#if char == CharGetGroupLeader(char):
	if values[1] == 'leader':
		if GetEncVar(enc_id, 'object_state') == 'carried':
			SetEncVar(enc_id, 'object_state', 'lost')
			
def LeaderDropObject_HeroPickup(target, char):
	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
	SetEncVar(enc_id, 'object_state', 'hero carried')
	EncHeroPickup(target,char)
			
def LeaderDropObject_LeaderSet(groupname, leader):
	enc_id = m25ai.GetGroupVar(groupname, 'enc_id')
	if GetEncVar(enc_id, 'object_state') == 'dropped':
		m25ai.GroupCommand_RetrieveItem(groupname, GetEncVar(enc_id,'loot'), function='LeaderDropObject_ObjectRetrieved('+`groupname`+')')
	
def LeaderDropObject_ObjectRetrieved(char, groupname):
	enc_id = m25ai.GetGroupVar(groupname,'enc_id')
	SetEncVar(enc_id,'object_state','carried')
	money = GetEncVar(enc_id,'loot')
	Object_Destroy(money)
	PlayCutsceneFromList(GetEncCutscene(enc_id,'moneyretrievedcs'),enc_id,charlist=[[char,'minion']])		
		
	
#---------------------------------------------------------------------
#
# Hunt enemies
#
#---------------------------------------------------------------------	
	
def EncHuntEnemy(group_id='minion'):
	encounter = {}
	encounter['triggers'] = [
		'OnGroupLeaderSet('+`group_id` + ', "HuntEnemy_LeaderSet")',
	]
	return encounter

def HuntEnemy_LeaderSet(groupname, val):
	m25ai.GroupCommand_HuntEnemies(groupname)
	
	
#---------------------------------------------------------------------
#
# Hunt heroes
#
#---------------------------------------------------------------------		
	
def EncHuntHero(group_id='minion'):
	encounter = {}
	encounter['triggers'] = [
		'OnGroupLeaderSet('+`group_id` + ', "HuntHero_LeaderSet")',
	]
	return encounter

def HuntHero_LeaderSet(groupname, val):
	m25ai.GroupCommand_HuntHeroes(groupname)	

	
#---------------------------------------------------------------------
#
# Interrogate
#
#---------------------------------------------------------------------	
	
def EncInterrogate(group='minion', rand=0):
	encounter ={}
	encounter['state'] = [
		['interrogate_group', group],
		['interrogate_random', rand],
	]
	encounter['setup'] = ['Interrogate_Setup']
	return encounter

def Interrogate_Setup(enc_id):
	#choose a character
	group = GetEncVar(enc_id, 'interrogate_group')
	members = GetEncVar(enc_id, group)
	#make sure the list isn't empty
	if members == None:
		if group == 'villain':
			group = 'minion'
			members = GetEncVar(enc_id, group)
			if members == None:
				return
	rand = GetEncVar(enc_id, 'interrogate_random')
	if rand:
		chosen_one = members[randint(0,len(members)-1)]
	else:
		chosen_one = members[0]
	cshelper.regAlmostDead(chosen_one, 'Interrogate_MakeItSo')
	m25getset.SetObjectVar(chosen_one, 'enc_id', enc_id)
	
def Interrogate_MakeItSo(event):
	char = event.object
	js.Object_SetAttr(char, 'minHealth', 1)
	cshelper.setInterrogateState(char)
	cshelper.dynamicSpeakCS(char, 'INTEROGATEME')
	cshelper.addCustomCommand("CUSTOM_INTERROGATE", char, 'TryInterrogate', 0, hero = '')
	m25ai.AIDisable(char)
	
def TryInterrogate(char, hero):
	enc_id = m25getset.GetObjectVar(char,'enc_id',integer=1)
	if CanInterrogate(char, hero):
		cshelper.removeAttention(char)
		PlayCutsceneFromList(GetEncVar(enc_id,'interrogationcs'),enc_id,return_function='DoneInterrogation('+`enc_id`+',' + `char` + ')')
	
def DoneInterrogation(enc_id, char):
	js.Object_SetAttr(char, 'minHealth', 0)
	cshelper.kill(char)
	EncCheckEnd(enc_id)
	
def CanInterrogate(char, hero):
	all = m25ai.FindAllOfType(hero,reaction=m25ai.TEAM_ENEMIES,maxdist=350, default=[])
	alive = 0
	for a in all:
		if not m25ai.IsAIDisabled(a):
			arrowName = '_' + a + '_int_arrow'
			cshelper.addArrow(arrowName, a, delete = 1)
			alive = 1
	if alive:
		cshelper.dynamicSpeakCS(hero, 'CANTINTEROGATE')
		return 0
	else:
		return 1

	
#---------------------------------------------------------------------
#
# Patrol
#
#---------------------------------------------------------------------		
		
def EncPatrol(patrol_group='minion'):
	encounter = {}
	encounter['state'] = [
		['current_patrol',1],
		['patrol_group',patrol_group],
	]
	encounter['setup'] = ['Patrol_Setup']
	return encounter

def Patrol_Setup(enc_id):
	EncEnableAllChars(enc_id, GetEncVar(enc_id,'patrol_group'))
	groupname = EncGetGroupName(enc_id,GetEncVar(enc_id,'patrol_group'))
	start = GetEncVar(enc_id,'start')
	patrol_marker = start + '_path1'
	if js.Marker_Exists(patrol_marker):
		m25ai.GroupCommand_MoveToMarker(groupname, patrol_marker,pct=20,priority=m25ai.TACTIC_PRIORITY_LOW-1,patrol=1,function='Patrol_NextMarker('+`groupname`+')')
	else:
		m25report.Report(ENC_CHANNEL,'Patrol_setup failed.  No marker ' + patrol_marker + ' present.')
	
def Patrol_NextMarker(char, groupname):
	enc_id = m25ai.GetGroupVar(groupname,'enc_id',integer=1)
	current = GetEncVar(enc_id, 'current_patrol',integer=1)+1
	start = GetEncVar(enc_id,'start')
	patrol_marker = start + '_path' + `current`
	if not js.Marker_Exists(patrol_marker):
		current = 1
		patrol_marker = start + '_path1'
	SetEncVar(enc_id, 'current_patrol', current)
	m25ai.GroupCommand_MoveToMarker(groupname, patrol_marker,pct=20,priority=m25ai.TACTIC_PRIORITY_LOW-1,patrol=1,function='Patrol_NextMarker('+`groupname`+')')
		
	
#---------------------------------------------------------------------
#
# Rescue Ally
#
#---------------------------------------------------------------------		
		
def EncRescueally(group_id='ally', marker_id='end', follow=0):
	encounter = {}
	encounter['state'] = [
		[group_id+'s_escaped', 0],
		['rescue_marker', marker_id],
		['rescue_follow', follow],
	]
	encounter['markers'] = [ [marker_id, marker_id, 'edge'], ]
	encounter['freed'] = ['RescueAlly_Freed']
	#doesn't work for some odd reason
#   	encounter['triggers'] = [
#   		'OnEachEnterMarker('+`group_id` + ',' + `marker_id` + ', "Rescue_allyEscapes")',
#   	]
	return encounter

def RescueAlly_Freed(enc_id, ally):
	m25ai.AIEnable(ally)
	rescue_marker = GetEncVar(enc_id, GetEncVar(enc_id,'rescue_marker'))
	if GetEncVar(enc_id,'rescue_follow'):
		m25ai.AddGoal_MoveToHero(ally, priority=22, followdist=20)
		m25ai.AddGoal_MoveToMarker(ally, rescue_marker, priority=21, pct=10, mindist=100, fn='Rescue_allyEscapes('+`enc_id`+ ')')
		js.Mission_AddArrowPos(`enc_id` + '_rescue_arrow',js.Get_MarkerPos(rescue_marker),js.ARROW_RED)
		js.Mission_CustomAction('wait here', '', ally, 'RescueAlly_Wait', 1000, 1)
	else:
		m25ai.AddGoal_MoveToMarker(ally, rescue_marker, pct=100, fn='Rescue_allyEscapes('+`enc_id`+ ')')
	PlayEncCutscene(enc_id, 'allyfreedcs')

def RescueAlly_Wait(char, hero):
	m25ai.AIDisable(char)
	js.Mission_CustomAction('follow me', '', char, 'RescueAlly_Follow', 1000, 1)

def RescueAlly_Follow(char, hero):
	m25ai.AIEnable(char)
	js.Mission_CustomAction('wait here', '', char, 'RescueAlly_Wait', 1000, 1)

def Rescue_allyEscapes(ally, enc_id):
	#enc_id = int(event.user)
	if GetEncVar(enc_id, 'rescue_follow'):
		try:
			js.Mission_RemoveArrow(`enc_id` + '_rescue_arrow')
		except:
			pass
	#ally = event.object
	PlayEncCutscene(enc_id, 'allyescapes')
	js.Object_Destroy(ally)
	gp = GetEncGroupPrefix(ally)
	SetEncVar(enc_id,gp+'s_escaped',GetEncVar(enc_id,gp+'s_escaped')+1)
	EncCheckEnd(enc_id)	

#---------------------------------------------------------------------
#
# Spawn
#
#---------------------------------------------------------------------		
	
def EncSpawn(group_id='minion', marker_prefix='minion', var=[], tag='', marker_type='', create_markers=1):
	encounter = {}
	# minions scale to meet hero squad size
	if group_id in ['minion','cohort']:
		var = Enc_MakeMinionList(var)
	encounter['spawn'] = [ [group_id, marker_prefix, var, tag, marker_type, create_markers] ]
	encounter['state'] = [
		['total'+group_id+'s','StateCount('+`group_id`+')'],
		[group_id+'s_koed',0],
	]
	encounter['triggers'] = [
		'OnGroupMemberKO('+`group_id`+',"EncMinionKO")',
	]
	encounter['spawn_groups'] = [group_id]
	#encounter['on_end'] = ['Spawn_OnEnd']
	return encounter
	
# def Spawn_OnEnd(enc_id):
# 	for a in GetEncVar(enc_id, 'spawn_groups'):
# 		m25ai.SetGroupVar(`enc_id` + '_' + a, '_timerset', 0)
	
#---------------------------------------------------------------------
#
# Start On Enter Marker
#
#---------------------------------------------------------------------		
		
def EncStartOnEnterMarker(start_function=''):
	encounter = {}
	encounter['start_function'] = [start_function]
	#encounter['state']=[
	#	['start_function',start_function],
	#]
	encounter['setup'] = [
		'EncStartOnEnterMarker_Setup',
	]
	return encounter
	
def EncStartOnEnterMarker_Setup(enc_id):
	m25event.SetTimer(0.1, 'Enc_CheckStart', user=enc_id)
	
#---------------------------------------------------------------------
#
# Waves
#
#---------------------------------------------------------------------			

def EncMinionWaves(group_id='minion', marker='minion', num_waves=10, wave_type='fill-in', wave_timer=0):
	encounter = {}
	encounter['state'] = [
		['waves_left', num_waves],
		['wave_type',wave_type],
		['wave_timer',wave_timer],
		['wave_group',group_id],
		['wave_marker', marker],
	]
	encounter['setup'] = ['MinionWaves_Setup']
	if wave_timer > 0:
		encounter['triggers'] = [
			'OnTimer('+`wave_timer` + ', "MinionWaves_NextWave")',
		]
	#encounter['check_end'] = [
	#	'MinionWaves_CheckEnd'
	#]
	return encounter

def MinionWaves_Setup(enc_id):
	wave = GetEncvar(enc_id,'on_wave',default=[])
	for w in wave: EncAddReg(enc_id, w, '_wave')
# 	encounter = GetEncVar(enc_id, '_encounter')
# 	if encounter.has_key('on_wave'):
# 		for a in encounter['on_wave']: EncAddReg(enc_id, a, 'on_wave')
# 	pass	
	
def MinionWaves_NextWave(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): 
		return
	wave_type = GetEncVar(enc_id, 'wave_type')
	gp = GetEncVar(enc_id, 'wave_group')
	nw = GetEncVar(enc_id,'waves_left')
	if wave_type == 'fill-in':
		encgroup = GetEncVar(enc_id, gp)
		encgroup_template = GetEncVar(enc_id,gp+'_template')
		newgroup = []
		newspawn = []
		num = 0
		retry=0
		for c in encgroup:
			if not js.Object_IsAlive(c) and nw > 0: 
				retry=1
				newname = m25name.GetNextObjectName()
				newspawn.append(newname)
				newgroup.append([newname,encgroup_template[num]])
				nw = nw - 1
			else:
				newgroup.append([c,encgroup_template[num]])
			num = num + 1
		if retry:
			m25report.Report(ENC_CHANNEL, 'Fill-in wave.')
			EncSpawnGroup(enc_id, GetEncVar(enc_id,'start')+'_'+GetEncVar(enc_id,'wave_marker'), newgroup, savename=gp, disable=0, teleport=0)
			#for n in newspawn:
			#	m25ai.ResetAI(n)
			SetEncVar(enc_id,'waves_left',nw)
			EncEventCallback(enc_id, '_wave', newspawn)
	timer = GetEncVar(enc_id,'wave_timer')
	if timer > 0:
		m25event.SetTimer(timer, 'MinionWaves_NextWave', user=enc_id)
	
def MinionWaves_CheckEnd(enc_id, return_code, return_list):
	pass
	
	
	
#####################################################################################
#
# Encounters
# 
# Description:  Usable encounters.
#
#####################################################################################

#---------------------------------------------------------------------
#
# Rescue the guarded character in distress
#
#---------------------------------------------------------------------	

# def DoEncRescue(enc_id, villains=[], minions=['thug_with_bat'], allies=['civilian_female'], cutscenes=[], marker='rescue', rescue_type='caged', return_function=''):
# 	encounter = EncRescueGuarded(enc_id, villains, minions, allies, cutscenes, marker, rescue_type, return_function)
# 	return SetupEncounter(enc_id, encounter)

# def EncRescueGuarded(enc_id, minions=['thug_with_bat'], allies=['civilian_female'], cutscenes = [], marker='rescue', rescue_type='caged', return_function = ''):
# 	encounter = {}
# 	if rescue_type == 'caged':
# 		encounter = CombineEncounters(encounter,EncCaged(enc_id, allies, cutscenes, marker, return_function))
# 	encounter = CombineEncounters(encounter, EncSpawn(group_id='villain', marker_prefix='villain', var=villains, tag='minion'))
# 	encounter = CombineEncounters(encounter, EncSpawn(group_id='minion', marker_prefix='minion', var=minions))
# 	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('RescueGuarded_Start'))
# 	return encounter

# def RescueGuarded_Start(event):
# 	enc_id = int(event.user)
# 	EncEnableAllChars(enc_id, 'minion')
# 	m25ai.Group_PromoteLeader(EncGetGroupName(int(event.user),'minion'))
# 	PlayEncCutscene(enc_id, 'startcs')
# 	
#---------------------------------------------------------------------
#
# Caged
#
#---------------------------------------------------------------------		

# def DoEncCaged(enc_id, allies=['civilian_male'], cutscenes=[], marker='rescue', return_function=''):
# 	encounter = EncCaged(enc_id, allies, cutscenes, marker, return_function)
# 	return SetupEncounter(enc_id, encounter)

# def EncCaged(enc_id, allies=[], cutscenes=[], marker='rescue', return_function=''):
# 	encounter = {}
# 	encounter['start'] = marker
# 	encounter['triggers'] = [
# 		'OnEachKO("cage", "Caged_allyFreed")',
# 	]
# 	encounter['cutscenes'] = cutscenes
# 	encounter['postspawn'] = ['Caged_PostSpawn']
# 	encounter['check_end'] = ['Caged_CheckEnd']
# 	encounter['return_function'] = return_function
# 	#encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Caged_Start'))
# 	encounter = CombineEncounters(encounter, EncSpawn(group_id='ally',marker_prefix='ally',var=allies))
# 	encounter = CombineEncounters(encounter, EncRescueally(group_id='ally', marker_id='end', follow=0))
# 	if encounter.has_key('freed'):
# 		for a in encounter['freed']: EncAddReg(enc_id, a, '_freed')
# 	return encounter

# def Caged_PostSpawn(enc_id):
# 	allies = GetEncVar(enc_id,'ally')
# 	cages = []
# 	for v in allies:
# 		name = m25name.GetNextObjectName()
# 		m25_spawn(name,'nazihq_killarilla_cage',Get_ObjectPos(v))
# 		cages.append(name)
# 	SetEncVar(enc_id,'cage',cages)

# def Caged_Start(event):
# 	#if not Enc_ValidHero(event.object): return
# 	if GetEncVar(enc_id,'encounter_started',default=0): return
# 	SetEncVar(enc_id,'encounter_started',1)
# 	PlayEncCutscene(int(event.user), 'startcs')
# 	
# def Caged_allyFreed(event):
# 	#print 'ally freed'
# 	cage = event.object
# 	ally = None
# 	enc_id = int(event.user)
# 	cages = GetEncVar(enc_id, 'cage')
# 	allies = GetEncVar(enc_id,'ally')
# 	num = 0
# 	for c in cages:
# 		if c == cage:
# 			ally = allies[num]
# 		num = num + 1
# 	if ally != None:
# 		EncEventCallback(enc_id, '_freed', ally)
# 		
# def Caged_CheckEnd(enc_id):
# 	total = GetEncVar(enc_id, 'totalallies',default=0)
# 	escaped = GetEncVar(enc_id, 'allies_escaped',default=0)
# 	koed = GetEncVar(enc_id,'allies_koed',default=0)
# 	if escaped + koed >= total:
# 		PlayCutsceneFromList(GetEncCutscene(enc_id,'donecs'),enc_id, return_function='EncEnd('+`enc_id`+',"ally flees")')

# 		


# 		
# #---------------------------------------------------------------------
# #
# # Fire Fighter
# #
# #---------------------------------------------------------------------			
# 			
# def DoEncFire(enc_id, minions=[], fire="m25_fire", allies="civilian_female", props=[], cutscenes=[], marker='fire', return_function=''):
# 	encounter = EncFire(enc_id, minions, fire, allies, props, cutscenes, marker, return_function)
# 	return SetupEncounter(enc_id, encounter)

# def EncFire(enc_id, minions=[], fire=[], allies=[], props=[], cutscenes=[], marker='fire', return_function=''):
# 	encounter = {}
# 	encounter['start'] = marker
# 	encounter['setup'] = ['Fire_Setup']
# 	encounter['cutscenes'] = cutscenes
# 	encounter['return_function'] = return_function
# 	encounter['check_end'] = ['Fire_CheckEnd']
# 	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Fire_Start'))
# 	if minions != []:
# 		encounter = CombineEncounters(encounter, EncSpawn(group_id='minion',marker_prefix='minion',var=minions))
# 	encounter = CombineEncounters(encounter, EncSpawn(group_id='fire', marker_prefix='fire', var=fire))
# 	encounter = CombineEncounters(encounter, EncSpawn(group_id='ally', marker_prefix='ally',var=allies))
# 	encounter = SpawnProps(encounter, props)
# 	return encounter

# def Fire_Setup(enc_id):
# 	fires = GetEncVar(enc_id, 'fire')
# 	for f in fires:
# 		Object_PlayEffect(f,'effect_ffx_flames2','',FX_LOOP)
# 	
# def Fire_Start(event):
# 	#if not Enc_ValidHero(event.object): return
# 	enc_id = int(event.user)
# 	if GetEncVar(enc_id,'encounter_started',default=0): return
# 	SetEncVar(enc_id,'encounter_started',1)
# 	EncEnableAllChars(enc_id, 'minion')	
# 	EncEnableAllChars(enc_id, 'fire')
# 	allies = GetEncVar(enc_id, 'ally')
# 	for v in allies:
# 		cshelper.setInterrogateState(v)
# 		cshelper.dynamicSpeakCS(v, 'INTEROGATEME')
# 		cshelper.addCustomCommand("CUSTOM_SAVE", v, 'Fire_Save', 0, hero = '')
# 	PlayEncCutscene(enc_id, 'startcs')
# 	
# def Fire_Save(char, hero):
# 	pass
# 	
# def Fire_CheckEnd(enc_id):
# 	if GetEncVar(enc_id,'minions_koed') >= GetEncVar(enc_id,'totalminions'):
# 		PlayCutsceneFromList(GetEncCutscene(enc_id,'allkocs'),enc_id, return_function='EncEnd('+`enc_id`+',"all_ko")')
# 			
# def SpawnProps(encounter, props):
# 	if props == []: return encounter
# 	for i in props:
# 		encounter = CombineEncounters(encounter, EncSpawn(i[0],i[1],i[2]))
# 	return encounter
# 		


#---------------------------------------------------------------------
#
# Gang War
#
#---------------------------------------------------------------------				

def DoEncGangwar(enc_id, villains=[], minions=[], allies=[], objects=[], interlopers=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncGangWar(enc_id, villains, minions, allies, objects, interlopers, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncGangWar(enc_id, villains=[], minions=[], allies=[], objects=[], interlopers=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = {}
	encounter['start_function'] = ['GangWar_Start']
	encounter['goals'] = [
		"EachFightForTeam('interloper', 3)",
	]
	encounter['setup'] = ['GangWar_Setup']
	encounter['state'] = [
		['minionsdown', 0],
		['interlopersdown', 0],
	]
	#interlopers = m25string.CommaToList(interlopers)
	encounter = CombineEncounters(encounter, EncFight(enc_id, villains,minions,allies,objects,cutscenes,return_function,marker,objectives))
	encounter['check_end'] = ['GangWar_CheckEnd']
	encounter = CombineEncounters(encounter, EncSpawn(group_id='interloper', marker_prefix='interloper', var=interlopers))
	return encounter
	
def GangWar_Setup(enc_id):
	interlopers = GetEncVar(enc_id, 'interloper')
	for i in interlopers:
		m25ai.AIDisable(i)
	
def GangWar_Start(event):
	enc_id = int(event.user)
	if GetEncVar(enc_id,'encounter_started',default=0): return
	EncEnableAllChars(enc_id, 'interloper')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'interloper'))
		
def GangWar_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['villain', 'minion']) and not GetEncVar(enc_id, 'minionsdown'):
		if not Enc_IsAllGone(enc_id, ['interlopers']):
			PlayEncCutscene(enc_id, "villainsdownfirstcs")
		else:
			PlayEncCutscene(enc_id, "villainsdownlastcs")
		SetEncVar(enc_id, 'minionsdown', 1)
	if Enc_IsAllGone(enc_id, ['interloper']) and not GetEncVar(enc_id, 'interlopersdown'):
		if not Enc_IsAllGone(enc_id, ['villain', 'minion']):
			PlayEncCutscene(enc_id, "interlopersdownfirstcs")
		else:
			PlayEncCutscene(enc_id, "interlopersdownlastcs")
		SetEncVar(enc_id, 'interlopersdown', 1)
	if Enc_IsAllGone(enc_id, ['villain','minion','interloper']):
		Enc_CompleteObjective(enc_id)
		PlayEncCutscene(enc_id, "end", return_function='Enc_CleanUp(%i,"end")'%(enc_id))
		

#---------------------------------------------------------------------
#
# Heist
#
#---------------------------------------------------------------------			
		
def DoEncHeist(enc_id, villains=[], minions=[], allies=[], objects=[], item='cub_market_crate2', cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = EncHeist(enc_id, villains, minions, allies, objects, item, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)

def EncHeist(enc_id, villains=[], minions=[], allies=[], objects=[], item='cub_marked_crate2', cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = {}
	encounter['hero_pickup'] = ['Heist_HeroPickup']
	encounter['on_fled'] = ['Heist_CheckEnd']
	encounter = CombineEncounters(encounter, EncFleeTheScene(enc_id, villains,minions,allies,objects,cutscenes,return_function,marker,objectives,immediate=0,mass_flee=0))
	encounter = CombineEncounters(encounter, EncCapture(object=item, groups=['minion'], on_ground=1))
	#overrides
	encounter['check_end'] = ['Heist_CheckEnd']
	return encounter

def Heist_HeroPickup(enc_id, values):
	object, hero = values
	PlayEncCutscene(enc_id, 'herogetsitemcs', charlist=[[hero,'hero']], return_function="EncCheckEnd(%i)"%(enc_id))

def Heist_CheckEnd(enc_id, values=''):
	if Enc_IsAllGone(enc_id, ['minion','villain']):
		if GetEncVar(enc_id, 'object_state') != 'lost':
			Enc_CompleteObjective(enc_id)
			PlayEncCutscene(enc_id, 'itemrecoveredcs',return_function='Enc_CleanUp('+`enc_id`+',["item recovered","end"])')
		else:
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'itemlostcs',return_function='Enc_CleanUp('+`enc_id`+',["item lost","end"])')
		
		

#---------------------------------------------------------------------
#
# Portal Enemies
#
#---------------------------------------------------------------------			

def DoEncPortalSpawn(enc_id, portal='', minions=[], cutscenes=[], return_function=''):
	encounter = EncPortalSpawn(enc_id, portal, minions, cutscenes, return_function)
	return SetupEncounter(enc_id, encounter)
	
def EncPortalSpawn(enc_id, portal=[], minions=[], cutscenes=[], return_function=''):
	encounter = {}
	encounter['start'] = 'portal'
	encounter['cutscenes'] = cutscenes
	encounter['return_function'] = return_function
	encounter['check_end'] = ['PortalSpawn_CheckEnd']
	encounter['on_wave'] = ['PortalSpawn_OnWave']
	encounter['triggers'] = [
		'OnGroupMemberKO("portal","PortalSpawn_PortalDown")',
	]
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('PortalSpawn_Start'))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='portal',marker_prefix='portal',var=portal))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='minion',marker_prefix='minion',var=minions))
	encounter = CombineEncounters(encounter, EncMinionWaves(group_id='minion', marker='minion', num_waves=10, wave_type='fill-in', wave_timer=1))
	return encounter
	
def PortalSpawn_Start(event):
	#if not Enc_ValidHero(event.object): return
	enc_id = int(event.user)
	if GetEncVar(enc_id,'encounter_started',default=0): return
	SetEncVar(enc_id,'encounter_started',1)
	EncEnableAllChars(enc_id, 'minion')
	m25ai.Group_PromoteLeader(`enc_id` + '_minion')
	PlayEncCutscene(enc_id, 'startcs')
	
def PortalSpawn_OnWave(enc_id, newchars):
	m25ai.Group_PromoteLeader(`enc_id` + '_minion')
	
def PortalSpawn_PortalDown(groupname, char):
	enc_id = m25ai.GetGroupVar(groupname,'enc_id')
	SetEncVar(enc_id, 'waves_left', 0)  #stop any more from spawning
	
def PortalSpawn_CheckEnd(enc_id):
	if GetEncVar(enc_id,'minions_koed') >= GetEncVar(enc_id,'totalminions')+10:
		PlayCutsceneFromList(GetEncCutscene(enc_id,'all_kocs'),enc_id, return_function='EncEnd('+`enc_id`+',"all_ko")')
	
				

		
	
#---------------------------------------------------------------------
#
# Base Housekeeping Encounter
#
#---------------------------------------------------------------------	
	
def DoEncBase(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0, recruit=None, disallow=None, require=None, add=None, clearoverrides=None):
	#if clearoverrides:
	js.Campaign_Recruit_ClearOverrides()
	if recruit:
		for a in m25string.CommaToList(recruit):
			if not js.Campaign_IsRecruited(a): 
				js.Campaign_Recruit(a)
				if m25gametype.GameType_IsFreeRoam():
					m25freeroamdata.AddFreeRoamRecruited([a])
	if add:
		for a in m25string.CommaToList(add):
			if not js.Campaign_IsRecruited(a): 
				js.Campaign_AddHero(a)
				if m25gametype.GameType_IsFreeRoam():
					m25freeroamdata.AddFreeRoamAvailableHeroes([a])
	if disallow:
		for a in m25string.CommaToList(disallow):
			if not js.Campaign_IsRecruited(a): js.Campaign_Recruit(a)
			js.Campaign_Recruit_Disallow(a)
	if require:
		for a in m25string.CommaToList(require):
			if not js.Campaign_IsRecruited(a): js.Campaign_Recruit(a)
			js.Campaign_Recruit_Require(a)
	for v in villains:
		if type(v) == types.ListType:
			js.Campaign_AddVillain(v[1])
		else:
			js.Campaign_AddVillain(v)
	for m in minions:
		if type(m) == types.ListType:
			js.Campaign_AddMinion(m[1])
		else:
			js.Campaign_AddMinion(m)
	js.Campaign_AddPrestige(int(prestige))
# 	for a in allies:
# 		if type(a) == types.ListType:
# 			if not js.Campaign_IsRecruited(a[1]): js.Campaign_Recruit(a[1])
# 			js.Campaign_Recruit_Disallow(a[1])		
# 		else:
# 			if not js.Campaign_IsRecruited(a): js.Campaign_Recruit(a)
# 			js.Campaign_Recruit_Disallow(a)
	js.Campaign_Save()
	SetEncVar(enc_id,'returnfunction',return_function)
	SetEncCutscenes(enc_id, cutscenes)
	PlayEncCutscene(enc_id, 'startcs', return_function='BaseDone(%i)'%enc_id)
	#PlayCutsceneFromList(cutscenes, enc_id, return_function= return_function)
	#m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)

def BaseDone(enc_id):
	CallEncounterReturnFunction(enc_id,'end',[])
	

#---------------------------------------------------------------------
#
# Disallow Heroes from the next mission
#
#---------------------------------------------------------------------	
	
def DoEncCampaigndisallow(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0):
	for a in allies:
		if type(a) == types.ListType:
			if not js.Campaign_IsRecruited(a[1]): js.Campaign_Recruit(a[1])
			js.Campaign_Recruit_Disallow(a[1])		
		else:
			if not js.Campaign_IsRecruited(a): js.Campaign_Recruit(a)
			js.Campaign_Recruit_Disallow(a)
	js.Campaign_Save()
	SetEncVar(enc_id,'returnfunction',return_function)
	m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)
	

	
#---------------------------------------------------------------------
#
# Require heroes for the next mission
#
#---------------------------------------------------------------------	
	
def DoEncCampaignrequire(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0):
	for a in allies:
		if type(a) == types.ListType:
			if not js.Campaign_IsRecruited(a[1]): js.Campaign_Recruit(a[1])
			js.Campaign_Recruit_Require(a[1])		
		else:
			if not js.Campaign_IsRecruited(a): js.Campaign_Recruit(a)
			js.Campaign_Recruit_Require(a)
	js.Campaign_Save()
	SetEncVar(enc_id,'returnfunction',return_function)
	m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)
	

#---------------------------------------------------------------------
#
# Clear any campaign overrides
#
#---------------------------------------------------------------------	
	
def DoEncCampaignclear(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0):
	js.Campaign_Recruit_ClearOverrides()
	js.Campaign_Save()
	SetEncVar(enc_id,'returnfunction',return_function)
	m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)
	
	
	
#---------------------------------------------------------------------
#
# Recruit Heroes into a Campaign
#
#---------------------------------------------------------------------	
	
def DoEncCampaignrecruit(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0):
	recruits = []
	for a in allies:
		if type(a) == types.ListType:
			#js.Campaign_AddHero(a[1])
			js.Campaign_Recruit(a[1])	
			recruits.append(a[1])
			#js.Campaign_AddCP(a[1],300)		
		else:
			#js.Campaign_AddHero(a)
			js.Campaign_Recruit(a)
			recruits.append(a)
			#js.Campaign_AddCP(a,300)
	for v in villains:
		if type(v) == types.ListType:
			js.Campaign_AddVillain(v[1])
		else:
			js.Campaign_AddVillain(v)
	for m in minions:
		if type(m) == types.ListType:
			js.Campaign_AddMinion(m[1])
		else:
			js.Campaign_AddMinion(m)
	js.Campaign_AddPrestige(int(prestige))
	js.Campaign_Save()
	
	if m25gametype.GameType_IsFreeRoam():
		ff.FreeRoam_PrestigeBonus(int(prestige))
		m25freeroamdata.AddFreeRoamRecruited(recruits)

	SetEncVar(enc_id,'returnfunction',return_function)
	m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)
	
def RecruitDone(event):
	CallEncounterReturnFunction(int(event.user),'end',[])

	
#---------------------------------------------------------------------
#
# Add Heroes into a Campaign, but don't recruit them
#
#---------------------------------------------------------------------	
	
def DoEncCampaignaddheroes(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', prestige=0):
	recruits = []
	for a in allies:
		if type(a) == types.ListType:
			js.Campaign_AddHero(a[1])
			recruits.append(a[1])
			#js.Campaign_Recruit(a[1])	
			#js.Campaign_AddCP(a[1],300)		
		else:
			js.Campaign_AddHero(a)
			recruits.append(a)
			#js.Campaign_Recruit(a)
			#js.Campaign_AddCP(a,300)
	for v in villains:
		if type(v) == types.ListType:
			js.Campaign_AddVillain(v[1])
		else:
			js.Campaign_AddVillain(v)
	for m in minions:
		if type(m) == types.ListType:
			js.Campaign_AddMinion(m[1])
		else:
			js.Campaign_AddMinion(m)
	js.Campaign_AddPrestige(int(prestige))
	js.Campaign_Save()
	
	if m25gametype.GameType_IsFreeRoam():
		ff.FreeRoam_PrestigeBonus(int(prestige))
		m25freeroamdata.AddFreeRoamAvailableHeroes(recruits)
	
	SetEncVar(enc_id,'returnfunction',return_function)
	m25event.SetTimer(0.1, 'RecruitDone', user=enc_id)

#---------------------------------------------------------------------
#
# Spawn Heroes into a Mission
#
#---------------------------------------------------------------------		
	
def DoEncPopulatebase(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', heroes=[], objectives=[]):
	Mission_SetAttr('enc_spawnduplicates',0)
	num = 0
	if heroes == []:
		herodict = Enc_GetBaseHeroes()
		if herodict == {}: herodict = {'hero':['minute_man'],}
	else: herodict = {'hero':heroes,}
		
	encounter = {}
	encounter['check_end'] = ['PopulateBase_CheckEnd']		
	encounter['setup'] = ['PopulateBase_Setup']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	#encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	for h in herodict.keys():
		if herodict[h] != []:
			encounter = CombineEncounters(encounter, EncSpawn(group_id='hero',marker_prefix=h,var=herodict[h],create_markers=0))
	return SetupEncounter(enc_id, encounter)
		
# 	for h in heroes:
# 		name = 'hero_' + `num`
# 		try:
# 			if js.Marker_Exists(name):
# 				cshelper.spawn(name,h,name)
# 			else:
# 				cshelper.spawn(name,h,'skirmishspawn')
# 			Enc_SetHeroAI(name)
# 		except:
# 			pass
# 		num = num + 1
# 	SetEncVar(enc_id, 'returnfunction', return_function)
# 	SetEncCutscenes(enc_id, cutscenes)
# 	PlayEncCutscene(enc_id, 'startcs', return_function='CallEncounterReturnFunction('+`enc_id` + ',"end", [])')
# 	#CallEncounterReturnFunction(enc_id, 'end', [])
	
def PopulateBase_Setup(enc_id):
	Mission_SetAttr('enc_spawnduplicates',1)
	PlayEncCutscene(enc_id, 'startcs', return_function='EncCheckEnd('+`enc_id`+')')

def PopulateBase_CheckEnd(enc_id):
	EncEnd(enc_id, 'end', [])

def Enc_GetBaseHeroes():
	hero_list = []
	for index in range(20,1000):
		template='custom_template_%d'%(index)
		if (js.Campaign_IsRecruited(template)!=0):
			hero_list.append(template)
	#sort them into trait lists
	hero_dict = {}
	for t in ffxTraits:
		hero_dict[t[0]] = []
	for h in hero_list:
		speech = js.Template_GetSpeechID(h)
		for traitlist in ffxTraits:
			if speech in traitlist[1:]:
				hero_dict[traitlist[0]].append(h)
	hero_dict['hero'] = hero_list
	return hero_dict
	#return hero_list

# def Enc_SetHeroAI(name):
# 	massfactors = [1.0,3.0, 2.5, 1.25, 0.75, 1.0, 0, 1.5, 2.5, 1.5, 1.0]
# 	chdata = chardata.GetCharacterData(name)
# 	if chdata.has_key('energy'):
# 		Object_SetAttr(name,'energy',chdata['energy'])
# 	if chdata.has_key('material') and chdata.has_key('mass'):
# 		Object_SetAttr(name,'mass',chdata['mass']*massfactors[int(chdata['material'])])
# 	m25ai.AIDisableCustom(name)
# 	cshelper.regDeath(name,'MiniMod_HeroKO')
# 	m25ai.SetupAI(name)
# 	m25ai.AddToGroup('team'+`m25ai.TEAM_GOODGUYS`,name)
# 	m25ai.Group_PromoteLeader('team'+`m25ai.TEAM_GOODGUYS`)				
				
#####################################################################################
#
# Banter
# 
# Description: play the banter cutscenes in missions at random intervals
#
#####################################################################################
	
def Enc_Banter(event):
	#startup
	if type(event) == types.IntType:
		m25event.SetTimer(randint(20,40),'Enc_Banter',user=event)
	else:
		enc_id = int(event.user)
		if not IsEncounterComplete(enc_id):
			if Enc_EncounterStarted(enc_id) and not m25cutscene.Cutscene_IsPlaying():
				PlayEncCutscene(enc_id, 'bantercs', in_order=0)
			m25event.SetTimer(randint(20,40),'Enc_Banter',user=enc_id)

def Enc_EncounterStarted(enc_id):
	return GetEncVar(enc_id, 'encounter_started', default=0)

def Enc_BuddyBanter(event):
	if type(event) == types.IntType:
		m25event.SetTimer(randint(20,40), 'Enc_BuddyBanter', user=event)
	else:
		enc_id = int(event.user)
		if not IsEncounterComplete(enc_id) and not Enc_EncounterStarted(enc_id) and not m25cutscene.Cutscene_IsPlaying():
			PlayEncCutscene(enc_id, 'enroutebantercs', in_order=0)
			m25event.SetTimer(randint(20,40),'Enc_BuddyBanter',user=enc_id)	
	
#####################################################################################
#
# Event Setup
# 
# Description: add appropriate game events
#
#####################################################################################	
	
def EachAttackTarget(enc_id, groupname,targetname):
	chars = GetEncVar(enc_id, groupname, default=[])
	targets = GetEncVar(enc_id, targetname, default=[])[:]
	targets.append('enemy')
	for c in chars:
		m25ai.AddGoal_Attack(c, target=targets)

def EachFightForTeam(enc_id, group, team):
	chars = GetEncVar(enc_id, group)
	for c in chars:
		m25ai.SetTeam(c, team)	
		m25ai.SetBrawler(c)		

def EachMoveTo(enc_id,object,target,mindist=200):
	objs = GetEncVar(enc_id,object)
	targs = GetEncVar(enc_id,target)
	for obj in objs:
		m25ai.AddGoal_MoveToTarget(obj,targs,mindist=mindist, priority=m25ai.TACTIC_PRIORITY_LOW, pct=100)

def EachMoveToHero(enc_id,object,mindist=200):
	objs = GetEncVar(enc_id,object,default=[])
	for obj in objs:
		m25ai.AddGoal_MoveToHero(obj,followdist=mindist, priority=m25ai.TACTIC_PRIORITY_LOW)

def Enc_EachFleeToMarker(enc_id,objclass,marker_id,return_function=''):
	marker = GetEncVar(enc_id, marker_id)
	chars = GetEncVar(enc_id, objclass)
	for c in chars:
		m25ai.AddGoal_MoveToMarker(c,marker=marker,priority=m25ai.TACTIC_PRIORITY_LOW,pct=80,fn=return_function)
		
# def EachMoveToMarker(enc_id, groupname, markername):
# 	chars = GetEncVar(enc_id, groupname)
# 	end_num = int(GetEncVar(enc_id, markertype))
# 	start_num = int(GetEncVar(enc_id, 'start'))
# 	marker = GetEncVar(enc_id,markername)
# 	for c in chars:
# 		if not m25ai.IsAIAuto(c):
# 			m25ai.AddGoal_MoveToMarker(c,marker=marker, mindist=50, priority="DHealthWatch(low=50,high=100)", pct="DHealthWatch(low=50,high=100)")
# 		else:
# 			m25ai.AddGoal_MoveToMarker(c,marker=marker,mindist=50,priority="DHealthWatch(low=50,high=100)",pct=100)

def Enc_EachRampageObjects(enc_id, objclass):
	chars = GetEncVar(enc_id, objclass)
	for c in chars:
		m25ai.AI_SetAttacksObjects(c)

def Enc_EachRampageBuildings(enc_id, objclass):
	chars = GetEncVar(enc_id, objclass)
	for c in chars:
		m25ai.AI_SetAttacksBuildings(c)



	
#####################################################################################
#
# Event Sinks
# 
# Description: handle game events
#
#####################################################################################	
	
def EncMinionKO(event, minion=None):
	start = time.clock()
	enc_id = 0
	char = ''
	# sort out how the function is being called
	# if minion is none, this was a game event
	if minion == None:
		enc_id = int(event.user)
		char = event.object
	# otherwise, this was triggered by the group callbacks in the m25AI
	else:
		groupname = event
		enc_id = m25ai.GetGroupVar(groupname,'enc_id')
		char = minion
	#if the encounter is already over, don't do anything
	if IsEncounterComplete(enc_id):
		return
	#increment the encounter variable - another x has fallen
	gp = GetEncClassPrefix(char)
	SetEncVar(enc_id,gp+'s_koed', GetEncVar(enc_id,gp+'s_koed',default=0,integer=1)+1)	
	#try to play a cutscene for the fallen
	charlist = [[char,'fallen']]  #force the cutscene to be played for the character that got ko'd
	#try charnamekocs
	if not PlayEncCutscene(enc_id, char+'kocs', charlist=charlist):
		#try chartemplatekocs
		if not PlayEncCutscene(enc_id, m25template.GetComplexTemplate(char)+'kocs', charlist=charlist):
			#try charclasskocs
			if not PlayEncCutscene(enc_id, gp+'kocs',charlist=charlist):
				pass
	#run the end function for the encounter to see if it is done
	EncCheckEnd(enc_id)
	m25report.Report('TIME','EncMinionKO %s took %f'%(char,time.clock()-start))


#####################################################################################
#
# Checks and helpers
# 
# Description: 
#
#####################################################################################	
		
def Enc_IsAllKO(enc_id, object_class):		
# 	if type(object_class) == types.ListType:
# 		for obj in object_class:
# 			all = GetEncVar(enc_id, obj)
# 			for a in all:
# 				if js.Object_Exists(a) and js.Object_IsAlive(a): return 0
# 		return 1
# 	else:
# 		all = GetEncVar(enc_id, object_class)
# 		for a in all:
# 			if js.Object_Exists(a) and js.Object_IsAlive(a): return 0
# 		return 1
	if type(object_class) == types.ListType:
		for obj in object_class:
			if GetEncVar(enc_id, obj + 's_koed',default=0) < GetEncVar(enc_id,'total'+obj+'s', default=0):
				return 0
		return 1
	else:			
		return GetEncVar(enc_id, object_class + 's_koed',default=0) >= GetEncVar(enc_id,'total'+object_class+'s', default=0)
	
def Enc_IsAllGone(enc_id, object_class):
	if type(object_class) == types.ListType:
		for obj in object_class:
			if GetEncVar(enc_id, obj + 's_koed',default=0) + GetEncVar(enc_id, obj + 's_escaped',default=0) < GetEncVar(enc_id,'total'+obj+'s',default=0):
				return 0
		return 1
	if GetEncVar(enc_id, object_class + 's_koed',default=0) + GetEncVar(enc_id, object_class + 's_escaped', default=0) >= GetEncVar(enc_id,'total'+object_class+'s',default=0):
		return 1
	return 0	
	
def Enc_IsAllEscaped(enc_id, object_class):
	if type(object_class) == types.ListType:
		for obj in object_class:
			if GetEncVar(enc_id, obj+'s_escaped', default=0) < GetEncVar(enc_id,'total'+obj+'s'):
				return 0
		return 1
	else:
		if GetEncVar(enc_id, object_class+'s_escaped', default=0) < GetEncVar(enc_id,'total'+object_class+'s'):
			return 0
		return 1
	
def Enc_IsSomeEscaped(enc_id, object_class):
	if not Enc_IsAllGone(enc_id, object_class): return 0
	if type(object_class) == types.ListType:
		for obj in object_class:
			if GetEncVar(enc_id, obj+'s_escaped', default=0) <= 0:
				return 0
		return 1
	else:
		if GetEncVar(enc_id, object_class+'s_escaped', default=0) <= 0:
			return 0
		return 1		
		
def Enc_ValidHero(char):
	val = ffx.FFX_HasPortrait(char)
	return val

	
#####################################################################################
#
# Clean up
# 
# Description: Handles removal of allies and objects left over from an encounter
# also defines the encounter_independent variable, that prevents EZ script from
# automatically removing items
#
#####################################################################################	

def Enc_CleanUp(enc_id, return_value, extra_remove=[], complete=1, return_list=[]):
	#don't wipe things out in a base screen
	rtypes = ['ally', 'object', 'minion', 'villain', 'cohort', 'rival', 'interloper']
	extra = GetEncVar(enc_id, 'cleanup', default=[])
	for e in extra: rtypes.append(e)
	if not CS_InBase():
		for r in rtypes: Enc_RemoveType(enc_id, r)
		for e in extra_remove:
			Enc_RemoveType(enc_id, e)
	EncEnd(enc_id, return_value, return_list, complete=complete)

def Enc_RemoveType(enc_id,typename):
	typelist = GetEncVar(enc_id, typename, default=[])
	for t in typelist:
		if js.Object_Exists(t):
			if not Enc_GetEncounterIndependent(t):
				#send them towards the 'allyleave' marker and destroy them after a number of seconds
				if js.Object_IsAlive(t):
					m25ai.AddGoal_MoveToMarker(t,marker=GetEncVar(enc_id,'allyleave'),mindist=5,priority=100,pct=100)
				m25event.SetTimer(10, 'Enc_ExtraDestroy', t, user=enc_id)
			else:
				if t in m25freeroamally.GetFreeroamAllies():
				#set the encounter independent character so as not to trigger new encounters
					Enc_SetDoNotStart(t)

def Enc_ExtraDestroy(event):
	enc_id = int(event.user)
	ally = event.object
	js.Object_Destroy(ally)				
				
def Enc_SetEncounterIndependent(char,val=1):
	m25getset.SetObjectVar(char, 'encounter_independent', val)
	
def Enc_GetEncounterIndependent(char):
	return m25getset.GetObjectVar(char, 'encounter_independent', default=0)	

Global_RegisterAttr('enc_nostart', 0)
	
def Enc_SetDoNotStart(char):
	js.Object_SetAttr(char, 'enc_nostart',1)
	
def Enc_GetDoNotStart(char):
	return js.Object_GetAttr(char, 'enc_nostart')	

def Enc_FilterNonStarters(char):
	if js.Object_Exists(char):
		return js.Object_GetAttr(char, 'enc_nostart')==0
	return 0
	
	
#####################################################################################
#
# decide how many minions to spawn
#
#####################################################################################	


def Enc_MakeMinionList(minions):
	return m25randomenc.RandomEnc_MakeMinionList(minions)

def Enc_MakeCohortList(minions):
	return m25randomenc.RandomEnc_MakeMinionList(minions)
	
#####################################################################################
#
# Determine when an Encounter should start
# 
# Description: 
#
#####################################################################################	

#def Enc_AlreadyStarted(enc_id):
#	return GetEncVar(enc_id,'encounter_started',default=0)

def Enc_CheckStartProximity(event):
	enc_id = int(event.user)
	if not Enc_EncounterStarted(enc_id):
		start_function = GetEncVar(enc_id,'start_function')
		if start_function:
			eval('ff.%s'%(start_function))(event)

def Enc_CheckStart(event):
	#start = time.clock()
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return	
	if Enc_CheckStartHeroes(enc_id):
		start_function = GetEncVar(enc_id, 'start_function')
		m25report.Report('ENCOUNTER', 'Enc_CheckStart starting %i'%(enc_id))
		for fn in start_function:
			exec 'ff.'+fn + '(event)'
		return
	#not starting yet, reset timer
	m25event.SetTimer(1, 'Enc_CheckStart', user=enc_id)
	#m25report.Report('TIME','Enc_Checkstart for %i took %f'%(enc_id,time.clock()-start))
	
def Enc_CheckStartHeroes(enc_id):
	heroes = ffx.GetAllHeroes()
	checklist = []
	groups = ['villain','minion','ally','object','rival','cohort']
	for g in GetEncVar(enc_id, 'check_start_groups', default=[]): groups.append(g)
	for e in groups:
		checklist = m25ai.AddToList(checklist,GetEncVar(enc_id,e,default=[]))
	item = GetEncVar(enc_id, 'loot', default=None)
	if item: checklist.append(item)
	for h in heroes:
		nearest = skXMapInfo.MapInfo_GetNearest(h, tiDistance = m25distance.DISTANCE_LONG, taObjects = checklist, tsFilter = Enc_FilterNonStarters)
		if nearest: return 1
	return 0

	
#####################################################################################
#
# Create Encounter Markers
# 
# Description: if encounter1, etc. does not exist on a map, create them
#
# 
#
#####################################################################################	

def Enc_CreateEncounterMarkers():
	location_index = []
	for marker_num in range(1,10):
		marker_name = 'encounter' + `marker_num`
		if js.Marker_Exists(marker_name):
			location_index.append(js.Get_MarkerPos(marker_name))
		else:
			#create a new marker
			marker_location = Enc_GetSpacedMarkerLocation(location_index)
			if marker_location == None:
				#cannot create any more markers
				return
			else:
				js.Marker_Spawn(marker_name, marker_location)
				location_index.append(marker_location)

def Enc_GetSpacedMarkerLocation(taken):
	maxx, maxy, maxz, minx, miny, minz = EncFindMapBounds()
	#generate a random location
	location = (randint(int(minx),int(maxx)),randint(int(miny),int(maxy)),0)
	for i in range(1,50):
		retry = 0
		for t in taken:
			if m25distance.GetTupleDistance(location, t) < m25distance.DISTANCE_VISIBLE:
				retry = 1
		if not retry:
			return location
	return None
	

#####################################################################################
#
# Objectives
# 
# Description: mark an objective as succeeded or failed
#
#####################################################################################					
	
def Enc_CompleteObjective(enc_id, objective_name=''):
# 	encounter = GetEncVar(enc_id, '_encounter', default=None)
# 	if encounter == None: 
# 		return
# 	if encounter.has_key('objectives'):
	objectives = GetEncVar(enc_id,'objectives',default=[])
	heroes = ffx.GetAllHeroes()
	num = 1
	for e in objectives:
		if (e[0] > -1) and (e[1] != ''):
			#js.Objective_Complete(`enc_id`+'objective'+`num`)
			m25objective.Objective_CompleteObjective(e[1])
		if m25gametype.GameType_IsFreeRoam():
			ff.FreeRoam_PrestigeBonus(e[2])
			#for h in heroes:
			ff.FreeRoam_XPBonus(e[3])
		else:
		#distribute the xp to all standing heroes
			for h in heroes:
				if js.Object_IsAlive(h):
					js.Campaign_AddCP(js.Object_GetTemplate(h),e[3])
		num = num + 1

def Enc_FailObjective(enc_id, objective_name=''):
# 	encounter = GetEncVar(enc_id, '_encounter', default=None)
# 	if encounter == None: 
# 		return
# 	if encounter.has_key('objectives'):
	objectives = GetEncVar(enc_id,'objectives',default=[])
	num = 1
	for e in objectives:
		if (e[0] > -1) and (e[1] != ''):
			#js.Objective_Fail(`enc_id`+'objective'+`num`)
			m25objective.Objective_FailObjective(e[1])
		num = num + 1
		
		
#####################################################################################
#
# Starting Encounters
# 
# Description: play the start cutscene automatically, and enable any and all participants
#
#####################################################################################			
			
def Enc_StandardStart(event):
	enable_list = ['minion', 'villain', 'ally', 'rival', 'cohort']
	enc_id = int(event.user)
	if GetEncVar(enc_id,'encounter_started',default=0): return
	SetEncVar(enc_id,'encounter_started',1)
	for c in enable_list: 
		if not GetEncVar(enc_id, '%s_disabled'%c, default=0):
			EncEnableAllChars(enc_id, c)
	PlayEncCutscene(enc_id, 'startcs', return_function='Enc_StandardStartEnable('+`enc_id`+')')
	
def Enc_StandardStartEnable(enc_id):
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	#check to see if the encounter doesn't end immediately!
	EncCheckEnd(enc_id)			
	
	

#####################################################################################
#
# Encounter Setup
# 
# Description:  common setup routines for encounters
#
#####################################################################################

#---------------------------------------------------------------------
#
# Standard entries for marker, cutscenes, return function and objectives
#
#---------------------------------------------------------------------	

def Enc_StandardSetup(marker='encounter', cutscenes=[], return_function='', objectives=[]):
	encounter = {}
	encounter['start'] = marker
	encounter['cutscenes'] = cutscenes
	encounter['return_function'] = return_function
	encounter['objectives'] = objectives
	#this is the marker allies will move towards after the encounter is finished
	encounter['markers'] = [ ['allyleave','allyleave','dist500'], ]
	return encounter
	
#---------------------------------------------------------------------
#
# Standard spawn of all villains/minions/allies/objects
#
#---------------------------------------------------------------------	

def Enc_StandardSpawn(villains=[], minions=[], allies=[], objects=[]):
	encounter = {}
	encounter['goals'] = [
		'EachFightForTeam("minion",2)',
		'EachFightForTeam("villain",2)',
		'EachFightForTeam("ally",1)',
	]
	encounter = CombineEncounters(encounter, EncSpawn(group_id='minion',marker_prefix='minion',var=minions))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='villain',marker_prefix='villain',var=villains,tag='minion'))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='ally',marker_prefix='ally',var=allies))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='object',marker_prefix='object',var=objects))
	return encounter


#####################################################################################
#
# Encounters
# 
# Description:  Usable encounters.
#
#####################################################################################		
	
#---------------------------------------------------------------------
#
# Joined by an ally
#
#---------------------------------------------------------------------	

def DoEncAlliance(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncAlliance(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncAlliance(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = {}
	encounter['type'] = 'alliance'
	encounter['setup'] = ['Alliance_Setup']
	encounter['goals'] = [
		'EachMoveToHero("ally")',
	]
	encounter['state'] = [
		['allies_ended', 0],
		['villains_ended', 0],
		['allies_saved', 0],
	]
	encounter['check_end'] = ['Alliance_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter
	
def Alliance_CheckEnd(enc_id):
	if not GetEncVar(enc_id,'allies_ended'):
		if Enc_IsAllKO(enc_id, ['ally']):
			SetEncVar(enc_id, 'allies_ended', 1)
			PlayEncCutscene(enc_id, 'allieslostcs', return_function='AllyFight_Over(%i,"allies lost")'%(enc_id))
			return
	if not GetEncVar(enc_id, 'villains_ended'):
		if Enc_IsAllKO(enc_id, ['villain','minion']):
			SetEncVar(enc_id, 'villains_ended', 1)
			Enc_CompleteObjective(enc_id)
			if not GetEncVar(enc_id,'allies_ended'):
				SetEncVar(enc_id, 'allies_saved',1)
				PlayEncCutscene(enc_id, 'alliessavedcs', return_function='AllyFight_Over(%i,["end","allies saved"])'%(enc_id))
			else:
				AllyFight_Over(enc_id, "end")

def AllyFight_Over(enc_id, return_value):
	for a in GetEncVar(enc_id, 'ally'):
		if js.Object_Exists(a) and js.Object_IsAlive(a):
			m25freeroamally.AddFreeroamAlly(a)
	complete = (GetEncVar(enc_id, "allies_ended") and GetEncVar(enc_id, "villains_ended"))
	if GetEncVar(enc_id, 'villains_ended'):
		Enc_CompleteObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'endcs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',' + `return_value` + ', complete='+`complete`+')')
	else:
		Enc_FailObjective(enc_id)
		EncEnd(enc_id, return_value, complete=complete)

def Alliance_Setup(enc_id):
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		#make sure the allies don't get whisked away by another encounter
		m25getset.SetObjectVar(a, 'ai_spawn_teleport',0)
		#set them encounter independent so they don't get auto-removed
		Enc_SetEncounterIndependent(a)
			


#---------------------------------------------------------------------
#
# Chat
#
#---------------------------------------------------------------------				

def DoEncChat(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='chat', return_function='', objectives=[]):
	encounter = EncChat(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)	
	
def EncChat(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='chat', return_function='', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['Chat_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='ally', command='TALK_TO', cutscene='chatcs'))
	#override the standard goals
	encounter['goals'] = [
		'EachFightForTeam("minion", 2)',
		'EachFightForTeam("villain", 2)',
		'EachFightForTeam("ally",0)',
	]
	return encounter

def Chat_CheckEnd(enc_id):
	if GetEncVar(enc_id,'examine_complete',default=0):
		#the chat cutscene will have already played
		Enc_CompleteObjective(enc_id)
		Enc_CleanUp(enc_id, ['chat won','end'])
	elif Enc_IsAllKO(enc_id, 'ally'):
		Enc_FailObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'chatlostcs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',["chat lost","end"])')
	else:
		pass
	

#---------------------------------------------------------------------
#
# Play a cutscene and then on to the next encounter
#
#---------------------------------------------------------------------			
	

def DoEncCutscene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncSingleCutscene(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncSingleCutscene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = {}
	encounter['setup'] = ['SingleCutscene_Setup']
	encounter['check_end'] = ['SingleCutscene_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter

def SingleCutscene_Setup(enc_id):
	PlayEncCutscene(enc_id, 'startcs',return_function='SingleCutscene_End('+`enc_id`+',"end")')
	
def SingleCutscene_End(enc_id, value):
	Enc_CompleteObjective(enc_id)
	Enc_CleanUp(enc_id, value)	
	
	
def SingleCutscene_CheckEnd(enc_id):
	pass	
		
#---------------------------------------------------------------------
#
# Destroy Object
#
#---------------------------------------------------------------------			
			
def DoEncDestroyobject(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncDestroyObject(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncDestroyObject(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['DestroyObject_CheckEnd']
	encounter['goals'] = [
		'EachFightForTeam("object", 2)',
		'EachAttackTarget("ally","object")',
	]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter
	
def DestroyObject_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['villain','minion','object']):
		Enc_CompleteObjective(enc_id)
		return_value = 'end'
		PlayEncCutscene(enc_id, 'endcs', return_function='Enc_CleanUp('+`enc_id`+','+`return_value`+')')


#---------------------------------------------------------------------
#
# Disarm Bomb
#
#---------------------------------------------------------------------			
			
def DoEncDisarmbomb(enc_id, villains=[], minions=[], allies=[], objects=[], bomb='cub_military_crate', time=300, cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncDisarmBomb(enc_id, villains, minions, allies, objects, bomb, time, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncDisarmBomb(enc_id, villains=[], minions=[], allies=[], objects=[], bomb='cub_military_crate', time=300, cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['DisarmBomb_CheckEnd']
	encounter['setup'] = ['DisarmBomb_Setup']
	encounter['on_examined'] = ['DisarmBomb_Disarmed']
	encounter['state'] = [
		['bomb_state','ticking'],
		['bomb_timer', int(time)],
	]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='bomb',marker_prefix='bomb',var=[bomb]))
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='bomb', command='DISARM', cutscene='disarmbombcs'))
	return encounter
		
def DisarmBomb_Setup(enc_id):
	bomb = GetEncVar(enc_id,'examine_object')
	timer = GetEncVar(enc_id,'bomb_timer')
	js.Mission_DisplayTimer('Bomb will explode in:', timer)
	m25event.SetTimer(timer,'DisarmBomb_Explode',user=enc_id)
	
def DisarmBomb_Disarmed(enc_id, vals=[]):
	#remove the timer
	js.Mission_DisplayTimer('Bomb will explode in:', -1.0)
	SetEncVar(enc_id, 'bomb_state','disarmed')
	EncCheckEnd(enc_id)
	
def DisarmBomb_Explode(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	bomb_state = GetEncVar(enc_id,'bomb_state')
	if bomb_state == 'disarmed': return
	PlayEncCutscene(enc_id, 'bombexplodingcs', return_function='DisarmBomb_Explode2('+`enc_id`+')')
	
def DisarmBomb_Explode2(enc_id):	
	obj = GetEncVar(enc_id, 'examine_object')
	js.Trigger_Explosion(obj, 'ffx_ex_fire_large')
	SetEncVar(enc_id,'bomb_state','exploded')
	m25event.SetTimer(1,'DisarmBomb_Detonated',user=enc_id)
	
def DisarmBomb_Detonated(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Object_Destroy(obj)
	EncCheckEnd(enc_id)		
	
def DisarmBomb_CheckEnd(enc_id, vals=[]):
	bomb_state = GetEncVar(enc_id,'bomb_state')
	if bomb_state == 'ticking': return
	if bomb_state == 'disarmed':
		if Enc_IsAllGone(enc_id, ['villain','minion']):
			Enc_CompleteObjective(enc_id)
			PlayEncCutscene(enc_id, 'bombdisarmedcs', return_function='Enc_CleanUp('+`enc_id`+',["bomb disarmed","end"], ["bomb"])')
	elif bomb_state == 'exploded':
		Enc_FailObjective(enc_id)
		PlayEncCutscene(enc_id, 'bombexplodedcs', return_function='Enc_CleanUp('+`enc_id`+',["bomb exploded","end"], ["bomb"])')
	else:
		pass


#---------------------------------------------------------------------
#
# Examine object and fight off foes
#
#---------------------------------------------------------------------		
		
def DoEncExamine(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = EncExamineFight(enc_id, villains,minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)
	
def EncExamineFight(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['ExamineFight_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='object', command='EXAMINE', cutscene='examinecs'))
	return encounter
	
def ExamineFight_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'examine_complete', default=0):
		Enc_CompleteObjective(enc_id)
		Enc_CleanUp(enc_id, ['examine_won', 'end'])


#---------------------------------------------------------------------
#
# Escort
#
#---------------------------------------------------------------------				

def DoEncEscort(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='escort', return_function='', objectives=[]):
	encounter = EncEscort(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)
	
def EncEscort(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='escort', return_function='', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['Escort_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Escort_Start'))
	encounter = CombineEncounters(encounter, EncRescueally(group_id='ally', marker_id='end', follow=1))
	return encounter

def Escort_Start(event):
	enc_id = int(event.user)
	for v in GetEncVar(enc_id, 'ally'):
		RescueAlly_Freed(enc_id, v)
	Enc_StandardStart(event)
	
def Escort_CheckEnd(enc_id):
	if not Enc_IsAllGone(enc_id,'ally'):
		return
	try: js.Mission_RemoveArrow(`enc_id` + '_rescue_arrow')
	except: pass
	if GetEncVar(enc_id,'allys_escaped',default=0) > 0:
		Enc_CompleteObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'escortwoncs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',["escort won","end"])')
	else:
		Enc_FailObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'escortlostcs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',["escort lost","end"])')

		
#---------------------------------------------------------------------
#
# Fight
#
#---------------------------------------------------------------------			
			
def DoEncFight(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncFight(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['Fight_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter
	
def Fight_CheckEnd(enc_id):
	if Enc_IsAllKO(enc_id, ['villain','minion']):
		return_value = 'allies saved'
		if GetEncVar(enc_id, 'totalallys',default=0) > 0 and Enc_IsAllKO(enc_id, 'ally'):
			return_value = 'allies lost'
			PlayEncCutscene(enc_id, 'allieslostcs', return_function='Fight_Over('+`enc_id`+','+`return_value`+')')
		else:
			PlayEncCutscene(enc_id, 'alliessavedcs', return_function='Fight_Over('+`enc_id`+','+`return_value`+')')

def Fight_Over(enc_id, return_value):
	Enc_CompleteObjective(enc_id)
	PlayCutsceneFromList(GetEncCutscene(enc_id,'endcs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',' + `[return_value,"end"]` + ')')

#---------------------------------------------------------------------
#
# Flee the scene
#
#---------------------------------------------------------------------	

def DoEncFleethescene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncFleeTheScene(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def DoEncImmediatefleethescene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncFleeTheScene(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives, immediate=1)
	return SetupEncounter(enc_id, encounter)	
	
def EncFleeTheScene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[],immediate=0,mass_flee=1):
	encounter = {}
	encounter['check_end'] = ['FleeTheScene_CheckEnd']
	encounter['on_fled'] = ['FleeTheScene_CheckEnd']
	encounter['state'] = [['fleethescene_done',0]]
	if immediate:
		encounter['setup'] = [
			'FleeTheScene_Setup',
		]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='minion',marker_id='end',marker_prefix='end',mass_flee=mass_flee))
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='villain',marker_id='end',marker_prefix='end',mass_flee=mass_flee))
	return encounter

def FleeTheScene_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	EncEnableAllChars(enc_id, 'villain')
	EncEnableAllChars(enc_id, 'ally')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))	
	
def FleeTheScene_CheckEnd(enc_id, val=[]):
	if GetEncVar(enc_id,'fleethescene_done',default=0): return
	if Enc_IsAllGone(enc_id, ['villain','minion']):
		SetEncVar(enc_id,'fleethescene_done',1)
		if Enc_IsAllEscaped(enc_id, ['villain','minion']):
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'allescapedcs', return_function='Enc_CleanUp('+`enc_id`+',["all escaped","end"])')
		elif Enc_IsSomeEscaped(enc_id, ['villain','minion']):
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'someescapedcs', return_function='Enc_CleanUp('+`enc_id`+',["some escaped","end"])')
		else:
			Enc_CompleteObjective(enc_id)
			PlayEncCutscene(enc_id, 'noneescapedcs', return_function='Enc_CleanUp('+`enc_id`+',["none escaped","end"])')	

#---------------------------------------------------------------------
#
# Guard Object
#
#---------------------------------------------------------------------			
			
def DoEncGuardobject(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncHeroGuardObject(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncHeroGuardObject(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['HeroGuardObject_CheckEnd']
	encounter['goals'] = [
		'EachFightForTeam("object", 1)',
		'EachAttackTarget("minion","object")',
		'EachAttackTarget("villain","object")',
		'EachFightForTeam("minion",2)',
		'EachFightForTeam("villain",2)',
		'EachFightForTeam("ally",1)',
		'EachMoveTo("villain","object")',
		'EachMoveTo("minion","object")',
	]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='minion',marker_prefix='minionout',var=minions,marker_type='dist500'))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='villain',marker_prefix='villainout',var=villains,tag='minion',marker_type='dist500'))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='ally',marker_prefix='ally',var=allies))
	encounter = CombineEncounters(encounter, EncSpawn(group_id='object',marker_prefix='object',var=objects))
	return encounter
	
def HeroGuardObject_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'object'):
		Enc_FailObjective(enc_id)
		PlayEncCutscene(enc_id, 'objectslostcs', return_function='Enc_CleanUp('+`enc_id`+',["objects lost","end"])')	
	elif Enc_IsAllGone(enc_id, ['villain','minion']):
		Enc_CompleteObjective(enc_id)
		PlayEncCutscene(enc_id, 'objectssavedcs', return_function='Enc_CleanUp('+`enc_id`+',["objects saved","end"])')
	
#---------------------------------------------------------------------
#
# Hunt the Hero
#
#---------------------------------------------------------------------			
			
def DoEncHunt(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', hunt_type='enemy', marker='fight', objectives=[]):
	encounter = EncHunt(enc_id, villains, minions, allies, objects, cutscenes, return_function, hunt_type,marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncHunt(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='',hunt_type='enemy',marker='fight', objectives=[]):
	encounter = {}
	encounter['setup'] = ['Hunt_Setup']
	encounter['check_end'] = ['Hunt_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Hunt_Start'))
	if hunt_type == 'enemy':
		encounter = CombineEncounters(encounter, EncHuntEnemy('minion'))
		encounter = CombineEncounters(encounter, EncHuntEnemy('villain'))
	elif hunt_type == 'hero':
		encounter = CombineEncounters(encounter, EncHuntHero('minion'))
		encounter = CombineEncounters(encounter, EncHuntHero('villain'))
	else:
		pass
	return encounter

def Hunt_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	EncEnableAllChars(enc_id, 'villain')
	EncEnableAllChars(enc_id, 'ally')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))

def Hunt_Start(event):
	enc_id = int(event.user)
	if GetEncVar(enc_id,'encounter_started',default=0): return
	SetEncVar(enc_id,'encounter_started',1)
	PlayEncCutscene(enc_id,'startcs')	
		
def Hunt_CheckEnd(enc_id):
	if Enc_IsAllKO(enc_id, ['villain','minion']):
		Enc_CompleteObjective(enc_id)
		PlayEncCutscene(enc_id, 'endcs', return_function ='Enc_CleanUp('+`enc_id`+',"end")')
			
	

		
#---------------------------------------------------------------------
#
# Interrogation
#
#---------------------------------------------------------------------			

def DoEncInterrogation(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncInterrogateFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncInterrogateFight(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='',marker='fight', objectives=[]):
	encounter = {}
	encounter = CombineEncounters(encounter, EncFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives))
	encounter = CombineEncounters(encounter, EncInterrogate('villain'))
	return encounter
	
#---------------------------------------------------------------------
#
# Mugging/ Save Civilian
#
#---------------------------------------------------------------------			

def DoEncMugging(enc_id, villains=[], minions=[], allies=['civilian_female'], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = EncMugging(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)

def DoEncSavecivilian(enc_id, villains=[], minions=[], allies=['civilian_female'], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = EncMugging(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)	
	
def EncMugging(enc_id, villains=[], minions=[], allies=['civilian_female'], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['Mugging_CheckEnd']
	encounter['state'] = [
		['allythank', 0],
	]
	encounter['goals'] = [
		'EachAttackTarget("minion","ally")',
		'EachAttackTarget("villain","ally")',
	]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter
	
def Mugging_TryThank(enc_id):
 	ally = findCharFromList(chartype='ally',enc_id=enc_id,live=1)
 	hero = m25ai.GetClosestHero(ally)
 	cshelper.addCustomCommand('talk',ally,'Mugging_Thank',1,4,'')
 	cshelper.addArrow(ally+'arrow', ally, colour = js.ARROW_YELLOW)
 	m25getset.SetObjectVar(ally,'encounter',enc_id)
 	m25ai.AddGoal_MoveToTarget(ally,hero)
 	PlayEncCutscene(enc_id, 'enemiesdowncs')
  	
def Mugging_Thank(target,char):
 	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
 	SetEncVar(enc_id,'allythank',1)
 	cshelper.removeAttention(target)
 	cshelper.removeArrow(target+'arrow')
 	PlayEncCutscene(enc_id, 'allythanksherocs',charlist=[[target,'ally'],[char,'hero']],return_function='EncCheckEnd('+`enc_id`+')')
	
def Mugging_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['minion','villain']):
		if not Enc_IsAllKO(enc_id, 'ally'):
			if GetEncVar(enc_id, 'allythank', default=0) == 0:
				Mugging_TryThank(enc_id)
			else:
				Enc_CompleteObjective(enc_id)
				EncEnd(enc_id, ['ally survives','end'])
		else:
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'allylostcs', return_function='Enc_CleanUp('+`enc_id`+',["ally lost","end"])')


#---------------------------------------------------------------------
#
# Rampage
#
#---------------------------------------------------------------------			
			
def DoEncRampage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncRampage(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
	return SetupEncounter(enc_id, encounter)

def DoEncImmediaterampage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncRampage(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives, immediate=1)
	return SetupEncounter(enc_id, encounter)	
	
def DoEncBuildingrampage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncRampage(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives, ontype='buildings')
	return SetupEncounter(enc_id, encounter)	

def DoEncImmediatebuildingrampage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[]):
	encounter = EncRampage(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives, ontype='buildings', immediate=1)
	return SetupEncounter(enc_id, encounter)	
		
def EncRampage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='fight', objectives=[], ontype='objects', immediate=0):
	encounter = {}
	if ontype == 'objects':
		encounter['goals'] = [
			'Enc_EachRampageObjects("villain")',
			'Enc_EachRampageObjects("minion")',
		]
	else:  # ontype == 'buildings':
		encounter['goals'] = [
			'Enc_EachRampageBuildings("villain")',
			'Enc_EachRampageBuildings("minion")',
		]
	if immediate:
		encounter['setup'] = ['Rampage_Setup']
		
	encounter = CombineEncounters(encounter, EncFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives))
	return encounter
	
def Rampage_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	EncEnableAllChars(enc_id, 'villain')
	EncEnableAllChars(enc_id, 'ally')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	m25ai.AddGoal_MoveToBuildings(m25ai.Group_GetLeader(EncGetGroupName(enc_id,'minion')))	
	


#---------------------------------------------------------------------
#
# sabotage
#
#---------------------------------------------------------------------		
		
def DoEncSabotage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='sabotage', return_function='', objectives=[]):
	encounter = EncSabotage(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)

def EncSabotage(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='sabotage', return_function='', objectives=[]):
	encounter = {}
	encounter['start'] = marker
	encounter['cutscenes'] = cutscenes
	encounter['return_function'] = return_function
	encounter['check_end'] = ['Sabotage_CheckEnd']
	encounter['state'] = [['sabotage','not started']]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Sabotage_Start'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter['on_examined'] = ['Sabotage_Complete']
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='object', command='sabotage', cutscene='sabotagecs'))
	return encounter
	
def Sabotage_Start(event):
	#if not Enc_ValidHero(event.object): return
	enc_id = int(event.user)
	if GetEncVar(enc_id,'encounter_started',default=0): return
	SetEncVar(enc_id,'encounter_started',1)
	EncEnableAllChars(enc_id, 'minion')	
	EncEnableAllChars(enc_id, 'villain')
	EncEnableAllChars(enc_id, 'ally')	
	PlayEncCutscene(enc_id, 'startcs')
	obj = GetEncVar(enc_id, 'examine_object')
	name = obj + '_exarrow'
	js.Mission_AddArrowObject(name,obj,js.ARROW_YELLOW,1)
	
def Sabotage_Complete(enc_id, object):
	#PlayCutsceneFromList(GetEncCutscene(enc_id,'sabotagecs'),enc_id,charlist=[[object,'hero']], return_function='Sabotage_StartTimer('+`enc_id`+')')
	Sabotage_StartTimer(enc_id)
	
def Sabotage_StartTimer(enc_id):	
	js.Mission_DisplayTimer('Detonation in...',10.0)
	m25event.SetTimer(10,'Sabotage_Explode',user=enc_id)
	arrow_name = GetEncVar(enc_id, 'examine_object') + '_exarrow'
	try:
		js.Mission_RemoveArrow(arrow_name)
	except:
		pass
	
def Sabotage_Explode(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Trigger_Explosion(obj, 'ffx_ex_fire_large')
	#js.Object_Destroy(obj)
	SetEncVar(enc_id, 'sabotage','complete')
	m25event.SetTimer(1,'Sabotage_Done',user=enc_id)
	
def Sabotage_Done(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Object_Destroy(obj)
	EncCheckEnd(enc_id)
	
def Sabotage_CheckEnd(enc_id):
	if GetEncVar(enc_id,'sabotage',default=0) == 'complete':
		Enc_CompleteObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'endcs'),enc_id, return_function='Enc_CleanUp('+`enc_id`+',"end")')
			
			
#---------------------------------------------------------------------
#
# Dastardly purse snatchers!
#
#---------------------------------------------------------------------

def DoEncSnatch(enc_id, villains=[], minions=[], allies=['civilian_female'], objects=[], item='cub_market_crate2', cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = EncSnatch(enc_id, villains, minions, allies, objects, item, cutscenes, marker, return_function, objectives)
	return SetupEncounter(enc_id, encounter)

def EncSnatch(enc_id, villains=[], minions=[], allies=['civilian_female'], objects=[], item='cub_marked_crate2', cutscenes=[], marker='encounter', return_function='', objectives=[]):
	encounter = {}
	encounter['state'] = [
		['pursereturned', 0],
	]
	encounter['check_end'] = ['Snatch_CheckEnd']
	encounter['hero_pickup'] = ['Snatch_HeroPickup']
	encounter['on_fled'] = ['Snatch_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='minion',marker_id='end',marker_prefix='end'))
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='villain',marker_id='end',marker_prefix='end'))
	encounter = CombineEncounters(encounter, EncLeaderDropObject(object=item,retrieve=0,group='minion',flee_marker='end'))
	return encounter

def Snatch_HeroPickup(enc_id, values):
	object, hero = values
 	ally = findCharFromList(chartype='ally',enc_id=enc_id,live=1)
 	if ally not in [None,'']:
 		cshelper.addCustomCommand('return',ally,'SnatchReturn',1,4,hero)
 		cshelper.addArrow(ally+'arrow', ally, colour = js.ARROW_YELLOW)
 		m25getset.SetObjectVar(ally,'encounter',enc_id)
 		m25ai.AddGoal_MoveToTarget(ally,hero,mindist=100)
 	else:
 		m25report.Report(ENC_CHANNEL, 'Snatch_HeroPickup: No ally to return stolen item to.') 
 	PlayCutsceneFromList(GetEncCutscene(enc_id,'herogetsitemcs'),enc_id,charlist=[[ally,'ally',],[hero,'hero']])
 	EncCheckEnd(enc_id)
  	
def SnatchReturn(target,char):
 	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
 	SetEncVar(enc_id,'pursereturned',1)
 	SetEncVar(enc_id,'object_state','returned')
 	cshelper.removeAttention(target)
 	cshelper.removeArrow(target+'arrow')
 	PlayCutsceneFromList(GetEncCutscene(enc_id,'itemreturnedcs'),enc_id,charlist=[[target,'ally',],[char,'hero']])
 	EncCheckEnd(enc_id)
	
def Snatch_CheckEnd(enc_id, values=['','']):
	if Enc_IsAllGone(enc_id, ['minion','villain']):
		if GetEncVar(enc_id,'object_state') == 'returned':
			Enc_CompleteObjective(enc_id)
			PlayEncCutscene(enc_id, 'itemrecoveredcs',return_function='Enc_CleanUp('+`enc_id`+',["item recovered","end"])')
		elif GetEncVar(enc_id,'object_state') == 'lost':
			return_val = ['item lost', 'end', 'allies saved']
			if Enc_IsAllGone(enc_id,'ally'):
				return_val = ['item lost', 'end', 'allies lost']
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'itemlostcs',return_function='Enc_CleanUp('+`enc_id`+',' + `return_val` + ')')
		elif GetEncVar(enc_id,'object_state') == 'hero carried' and Enc_IsAllGone(enc_id,'ally'):
			Enc_FailObjective(enc_id)
			PlayEncCutscene(enc_id, 'allieslostcs',return_function='Enc_CleanUp('+`enc_id`+',["item recovered","allies lost","end"])')
		else:
			pass
			
 
#---------------------------------------------------------------------
#
# War
#
#---------------------------------------------------------------------				

# def DoEncWar(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
# 	encounter = EncInterrogateFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives)
# 	return SetupEncounter(enc_id, encounter)

# def EncWar(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='',marker='encounter', objectives=[]):
# 	encounter = {}
# 	encounter = CombineEncounters(encounter, EncFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives))
# 	encounter['check_end'] = ['War_CheckEnd']
# 	encounter['goals'] = [
# 		'EachFightForTeam("villain",2)',
# 		'EachFightForTeam("minion",2)',
# 		'EachFightForTeam("ally",3)',
# 	]
# 	return encounter
# 		
# def War_CheckEnd(enc_id):
# 	if Enc_IsAllKO(enc_id, ['villain','minion','ally']):
# 		PlayEncCutscene(enc_id, 'endcs', return_function='Enc_CleanUp('+`enc_id`+',"end")')

		
#---------------------------------------------------------------------
#
# Turn a mission into a base screen  (Not working very well.)
#
#---------------------------------------------------------------------				

def DoEncCampaignbase(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncCampaignBase(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncCampaignBase(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = {}
	encounter['setup'] = ['CampaignBase_Setup']
	encounter['check_end'] = ['CampaignBase_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter

def CampaignBase_Setup(enc_id):
	Mission_SetAttr('gametype_inbase',1)
	ff.OnBriefing = CampaignBase_PlayBriefing()
	PlayEncCutscene(enc_id, 'startcs',return_function='Enc_CleanUp('+`enc_id`+',"end")')		
	
def CampaignBase_PlayBriefing():
	m25cutscene.CS_SetInBriefing()
	has_briefing = ff.Story_DoBriefing()
	if not has_briefing:
		js.UI_Briefing_Done()	
	
def CampaignBase_CheckEnd(enc_id):
	pass
	

#---------------------------------------------------------------------
#
# Run a generic function and branch based on the function results
#
#---------------------------------------------------------------------			
			
def DoEncUserfunction(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], function=None):
	encounter = EncUserFunction(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives, function)
	return SetupEncounter(enc_id, encounter)

def EncUserFunction(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], function=None):
	encounter = {}
	encounter['setup'] = ['UserFunction_Setup']
	encounter['check_end'] = ['UserFunction_CheckEnd']
	encounter['state'] = [
		['userfunction', function],
	]
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter

def UserFunction_Setup(enc_id):
	function = GetEncVar(enc_id, 'userfunction')
	return_value, return_list = "end",[]
	val = ["end", []]
	if function:
		try:
			val = eval(function)
		except:
			m25report.Report('ERROR', 'EncUserFunction error in %s'%(repr(function)))
			return
		try:
			return_value, return_list = val
		except:
			pass
	PlayEncCutscene(enc_id, 'startcs',return_function='UserFunction_End(%i,"%s",%s)'%(enc_id,return_value,repr(return_list)))
		
	
def UserFunction_End(enc_id, value, return_list):
	Enc_CompleteObjective(enc_id)
	Enc_CleanUp(enc_id, value, return_list=return_list)

	
def UserFunction_CheckEnd(enc_id):
	pass

	
#---------------------------------------------------------------------
#
# Remove an ally
#
#---------------------------------------------------------------------			
			
def DoEncRemoveally(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], function=None):
	encounter = EncRemoveAlly(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives, function)
	return SetupEncounter(enc_id, encounter)

def EncRemoveAlly(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], function=None):
	encounter = {}
	encounter['setup'] = ['RemoveAlly_Setup']
	encounter['check_end'] = ['UserFunction_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter

def RemoveAlly_Setup(enc_id):
	#turn off the encounter independent variable, so that any allies to be removed will actually disappear
	for a in GetEncVar(enc_id, 'ally', default=[]):
		Enc_SetEncounterIndependent(a,val=0)
		m25freeroamally.RemoveFreeroamAlly(a)
	PlayEncCutscene(enc_id, 'startcs',return_function='RemoveAlly_End(%i,"%s")'%(enc_id,"end"))
	
def RemoveAlly_End(enc_id, value):
	Enc_CompleteObjective(enc_id)
	Enc_CleanUp(enc_id, value)


#---------------------------------------------------------------------
#
# Do Nothing except end
#
#---------------------------------------------------------------------			
		
def DoEncNone(enc_id,villains=[],minions=[],allies=[],objects=[],cutscenes=[],return_function='',marker='encounter',objectives=[]):
	SetEncounterReturnFunction(enc_id, return_function)
	CallEncounterReturnFunction(enc_id, 'end', [])
		
	
#---------------------------------------------------------------------
#
# Wait a random amount of time
#
#---------------------------------------------------------------------			

def DoEncWait(enc_id,villains=[],minions=[],allies=[],objects=[],cutscenes=[],return_function='',marker='encounter',objectives=[], minimumtime=30, maximumtime=120):
	wait_time = randint(int(minimumtime),int(maximumtime))
	SetEncounterReturnFunction(enc_id, return_function)
	m25event.SetTimer(wait_time, 'EncWaitEnd', user=enc_id)

def EncWaitEnd(event):
	CallEncounterReturnFunction(int(event.user), 'end', [])

	
#---------------------------------------------------------------------
#
# Play a cutscene when the hero gets to a location
#
#---------------------------------------------------------------------			
	

def DoEncLocalcutscene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncLocalCutscene(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncLocalCutscene(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = {}
	encounter['check_end'] = ['LocalCutscene_CheckEnd']
	encounter = CombineEncounters(encounter, Enc_StandardSetup(marker, cutscenes, return_function, objectives))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = CombineEncounters(encounter, Enc_StandardSpawn(villains,minions,allies,objects))
	return encounter
	
def LocalCutscene_CheckEnd(enc_id):
	Enc_CompleteObjective(enc_id)
	PlayEncCutscene(enc_id, 'endcs', return_function='Enc_CleanUp(%i,"end")'%(enc_id))

	
#---------------------------------------------------------------------
#
# Have Allies join the player's squad
#
#---------------------------------------------------------------------			
	

def DoEncJoinedbyhero(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = EncJoinedByHero(enc_id, villains, minions, allies, objects, cutscenes, return_function,marker, objectives)
	return SetupEncounter(enc_id, encounter)

def EncJoinedByHero(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[]):
	encounter = {}
	#encounter['start_function'] = ['JoinedByHero_Start']
	encounter = CombineEncounters(encounter, EncFight(enc_id, villains,minions,allies,objects,cutscenes,return_function,marker,objectives))
	encounter['start_function'] = ['Enc_StandardStart', 'JoinedByHero_Start']
	return encounter
	
def JoinedByHero_Start(event):
	enc_id = int(event.user)
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		ffx.FFX_AddPortrait(a)
		#js.AI_MakeIntoHero(a)
		Enc_SetEncounterIndependent(a)
		m25ai.AIDisableCustom(a)
		m25ai.SetAIStandardOnly(a)
		m25ai.AIEnable(a)
		m25ai.AIEnableStandard(a)

		
		
#####################################################################
#
# CUSTOM ENCOUNTER
#
#
#####################################################################		



#=========================================================================
#
#  Custom Encounters
#
#=========================================================================

def DoEncCustom(enc_id, **parameters):
	encounter = EncCustom(enc_id, parameters)
	return SetupEncounter(enc_id, encounter)

def EncCustom(enc_id, parameters):
	encounter = {}
	encounter['start'] = EncCustom_GetParameter(parameters, 'marker', default='encounter')
	encounter['cutscenes'] = EncCustom_GetParameter(parameters, 'cutscenes', default=[])
	encounter['return_function'] = EncCustom_GetParameter(parameters, 'return_function', default='')
	encounter['objectives'] = EncCustom_GetParameter(parameters, 'objectives', default=[])
	#this is the marker allies will move towards after the encounter is finished
	encounter['markers'] = [ ['allyleave','allyleave','dist500'], ]
	encounter['check_end'] = ['EncCustom_CheckEnd']
	encounter['state'] = [
		['objective_status', 1],   #default to objective complete when encounter ends
	]
	for a in EncCustom_AddDependencies(m25string.CommaToList(EncCustom_GetParameter(parameters, 'actions', default=[]))):
		encounter = CombineEncounters(encounter, EncGetCustomValues(enc_id, a, parameters))
	encounter = CombineEncounters(encounter, EncStartOnEnterMarker('Enc_StandardStart'))
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'minion', 'minion', 'minions')
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'villain', 'villain', 'villains', ai_group='minion')
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'ally', 'ally', 'allies')
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'object', 'object', 'objects')
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'cohort', 'cohort', 'cohorts')
	encounter = EncCustom_SetupSpawn(enc_id, parameters, encounter, 'rival', 'rival', 'rivals', ai_group='cohort')
	return encounter
	
def EncCustom_AddDependencies(action_list):
	new_list = []
	for a in action_list:
		new_list.append(a)
		if m25enccustom.ENC_DEPENDENCIES.has_key(a):
			new_list = m25ai.AddToList(new_list, m25enccustom.ENC_DEPENDENCIES[a])
	return EncRemoveDuplicates(new_list)
	
def EncCustom_SetupSpawn(enc_id, parameters, encounter, group_id, marker_prefix, varname, marker_type='', ai_group=''):
	spawnmarker = EncCustom_GetParameter(encounter, '%s_marker_loc'%group_id, default=[marker_prefix, marker_type])
	enc = CombineEncounters(encounter, EncSpawn(group_id=group_id, marker_prefix=spawnmarker[0], var=EncCustom_GetParameter(parameters, varname, default=[]), marker_type=spawnmarker[1], tag=ai_group))
	return enc
	
def EncCustom_CheckEnd(enc_id):
	done = 1
	value_list = ['end']
	for e in GetEncVar(enc_id, 'custom_check_end', default=[]):
		value = eval('%s(%i)'%(e, enc_id))
		m25report.Report('ENCOUNTER', '-- custom check end %s(%i) - %s'%(e, enc_id, repr(value)))
		if not value:
			done = 0
		else:
			value_list = m25ai.AddToList(value_list,value)
	if done or GetEncVar(enc_id, 'enc_immediate_end', default=0):
		PlayEncCutscene(enc_id, 'endcs', return_function = 'EncCustom_CleanUp(%i, %s)'%(enc_id, EncRemoveDuplicates(value_list)))
		
def EncCustom_CleanUp(enc_id, return_value, extra_remove=[], complete=1, return_list=[]):
	#call clean up functions
	clean_list = GetEncVar(enc_id, 'clean_up', default=[])
	for c in clean_list:
		eval('%s(%i)'%(c, enc_id))
	# check objectives
	if GetEncVar(enc_id, 'objective_status', default=1):
		Enc_CompleteObjective(enc_id)
	else:
		Enc_FailObjective(enc_id)
	extra_remove = GetEncVar(enc_id, 'extra_remove', default=[])
	Enc_CleanUp(enc_id, return_value, extra_remove, complete, return_list)

	
def EncCustom_CompleteObjective(enc_id):
	return  #do not do anything
	
def EncCustom_FailObjective(enc_id):
	SetEncVar(enc_id, 'objective_status', 0)
	
def EncCustom_GetParameter(dict, name, default=None):
	if dict.has_key(name): return dict[name]
	else: return default
	
def EncGetCustomValues(enc_id, name, parameters):
	value = {}
	try:
		value = eval('ff.EncCustom_%s(%i,%s)'%(m25string.String_ConvertToTemplate(name), enc_id, parameters))
		return value
	except:
		raise
		return {}
	
def EncCustom_SetImmediateEnd(enc_id):
	SetEncVar(enc_id, 'enc_immediate_end', 1)		
	
def EncCustom_ClearImmediateEnd(enc_id):
	SetEncVar(enc_id, 'enc_immediate_end', 0)

# TO DO
# Map Wide Fire
# Allies In Cages


def EncEvent_GetEncId(event):
	enc_id = event
	if type(event) != types.IntType: enc_id = int(event.user)
	return enc_id	

	
#=========================================================================
#
#  Custom Encounters Pieces: General
#
#=========================================================================		
	
#-------------------------------------------------------------------------
#
# Custom Item Has Power (Traps)
#
#-------------------------------------------------------------------------	
	
def EncCustom_item_has_power(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['item_power' , EncCustom_GetParameter(parameters,'trappower',default='bomber Explosion')],
	]
	encounter['start_function'] = ['EncCustom_ItemUsesPower_Start']
	#encounter['custom_check_end'] = ['EncCustom_ItemUsesPower_CheckEnd']
	return encounter

#-------------------------------------------------------------------------
#
# Custom Item uses power once only
#
#-------------------------------------------------------------------------		
	
def EncCustom_item_uses_power_once(enc_id, parameters):
	encounter = {}
	encounter['item_uses_power_once'] = 1
	encounter['custom_check_end'] = ['EncCustom_ItemPowerOnce_CheckEnd']
	return encounter
	
def EncCustom_ItemPowerOnce_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'item_used_power', default=0): return ['item used power']
	else: return None
	
#-------------------------------------------------------------------------
#
# Custom Item uses power on <character type>
#
#-------------------------------------------------------------------------		
	
def EncCustom_item_uses_power_on_heroes(enc_id, parameters):
	encounter = {}
	encounter['item_target'] = ['hero']
	return encounter
	
def EncCustom_item_uses_power_on_everyone(enc_id, parameters):
	encounter = {}
	encounter['item_target'] = ['all']
	return encounter
	
def EncCustom_item_uses_power_on_villains(enc_id, parameters):
	encounter = {}
	encounter['item_target'] = ['villain']
	return encounter
	
def EncCustom_ItemUsesPower_Start(event):
	enc_id = int(event.user)
	m25event.SetTimer(0.5, 'EncCustom_ItemUsesPower_TryPower', user=enc_id)
	
def EncCustom_ItemUsesPower_TryPower(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	used = 0
	if not m25ai.InPeace() and not GetEncVar(enc_id, 'item_switched_off',default=0):
		target_groups = GetEncVar(enc_id, 'item_target', default=['all'])
		targets = []
		for tg in target_groups:
			if tg == 'all':
				targets = m25ai.AddToList(targets, list(filter(m25ai.ST_char,js.Mission_GetDynamicObjects())))
			elif tg == 'villain':
				targets = m25ai.AddToList(targets, GetEncVar(enc_id, 'villain', default=[]))
			elif tg == 'hero':
				targets = m25ai.AddToList(targets, ffx.GetAllHeroes())
			else:
				pass
		item = GetEncVar(enc_id, 'loot', default=None)
		if item: 
			close_targets = []
			for t in targets:
				if m25distance.GetDistance(item, t) <= 100: close_targets.append(t)
			if close_targets:
				chosen = close_targets[random.randint(0,len(close_targets)-1)]
				name = m25name.GetNextObjectName()
				js.Object_SpawnAt('ffx_tiggot', js.Get_ObjectPos(item), name)
				js.Trigger_Power(name, chosen, GetEncVar(enc_id, 'item_power', default=''), '')
				m25event.SetTimer(5.0, 'EncCustom_RemoveTiggot', name)
				used = 1
				m25event.SetTimer(4, 'EncCustom_ItemUsesPower_PowerUsed', user=enc_id)
	if GetEncVar(enc_id, 'item_uses_power_once', default=0) and used:
		pass
	else:
		if used:
			m25event.SetTimer(5, 'EncCustom_ItemUsesPower_TryPower', user=enc_id)
		else:
			m25event.SetTimer(1, 'EncCustom_ItemUsesPower_TryPower', user=enc_id)
		
def EncCustom_RemoveTiggot(event):
	js.Object_Destroy(event.object)
		
def EncCustom_ItemUsesPower_PowerUsed(event):
	enc_id = int(event.user)
	SetEncVar(enc_id, 'item_used_power', 1)
	EncCheckEnd(enc_id)
		
#-------------------------------------------------------------------------
#
# Custom Item can be switched off
#
#-------------------------------------------------------------------------		
	
def EncCustom_item_can_be_switched_off(enc_id, parameters):
	encounter = {}
	encounter['on_switch_on'] = ['EncCustom_ItemSwitch_On']
	encounter['on_switch_off'] = ['EncCustom_ItemSwitch_Off']
	return encounter		
		
def EncCustom_ItemSwitch_On(enc_id, item=[]):
	SetEncVar(enc_id, 'item_switched_off', 0)
	
def EncCustom_ItemSwitch_Off(enc_id, item=[]):
	SetEncVar(enc_id, 'item_switched_off', 1)
	SetEncVar(enc_id, 'item_state', 'switched_off')
	
	
#-------------------------------------------------------------------------
#
# Custom Switches
#
#-------------------------------------------------------------------------
	
def EncCustom_switches_exist(enc_id, parameters):
	encounter = {}
	encounter = CombineEncounters(encounter, EncSpawn(group_id='switch',marker_prefix='switch',var=EncCustom_GetParameter(parameters, 'switches', default=['switch']),marker_type='dist300'))
	encounter['setup'] = ['EncCustom_SwitchOff_Setup']
	encounter['cleanup'] = ['switch']
	encounter['check_start_groups'] = ['switch']
	encounter['extra_remove'] = ['switch']
	encounter['custom_check_end'] = ['EncCustom_SwitchOff_CheckEnd']
	return encounter		
	
def EncCustom_SwitchOff_Setup(enc_id):
	switches = GetEncVar(enc_id, 'switch', default = [])
	for s in switches:
		m25getset.SetObjectVar(s, 'enc_switch_status', 0)
		cshelper.addCustomCommand('switch', s, 'EncCustom_SwitchOff_Switch', 1, 4)
		#cshelper.addArrow(s+'arrow', s, colour = js.ARROW_YELLOW)
	for e in GetEncVar(enc_id, 'on_switch_on',default=[]): EncAddReg(enc_id, e, '_switch_on')
	for e in GetEncVar(enc_id, 'on_switch_off',default=[]): EncAddReg(enc_id, e, '_switch_off')

def EncCustom_SwitchOff_Switch(item, hero):
	enc_id = Enc_GetCharEnc(item)
	s = item
	status = m25getset.GetObjectVar(s, 'enc_switch_status', default=0)
	if not status:
		m25getset.SetObjectVar(s, 'enc_switch_status', 1)
		try:
			js.Object_NextVersion(s)
		except:
			pass
		if EncCustom_SwitchOff_AllOff(enc_id): 
			PlayEncCutscene(enc_id, 'allswitchesturnedoffcs')
		else: PlayEncCutscene(enc_id, 'switchturnedoffcs')
		EncEventCallback(enc_id, '_switch_off', item)
		EncCheckEnd(enc_id)
	else:
		m25getset.SetObjectVar(s, 'enc_switch_status', 0)
		PlayEncCutscene(enc_id, 'switchturnedoncs')
		EncEventCallback(enc_id, '_switch_on', item)

def EncCustom_SwitchOff_AllOff(enc_id):
	for s in GetEncVar(enc_id, 'switch', default = []):
		if not m25getset.GetObjectVar(s, 'enc_switch_status', default=0): return 0
	return 1
		
def EncCustom_SwitchOff_CheckEnd(enc_id):
	if EncCustom_SwitchOff_AllOff(enc_id): return ['all switches off']
	else: return None
	
#-------------------------------------------------------------------------
#
# Custom Allies Held By Objects
#
#-------------------------------------------------------------------------	

def EncCustom_allies_held_by_objects(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['heldally_effect', EncCustom_GetParameter(parameters,'holdeffect',default='ffx_energyshield')],
		['allies_held', 1],
	]
	encounter['setup'] = ['EncCustom_AlliesHeldByObjects_Setup']
	encounter['custom_check_end'] = ['EncCustom_AlliesHeldByObjects_CheckEnd']
	return encounter

def EncCustom_AlliesHeldByObjects_Setup(enc_id):
	effect_name = GetEncVar(enc_id, 'heldally_effect')
	for v in GetEncVar(enc_id, 'ally', default=[]):
		js.Object_SetAttr(v, 'minHealth', js.Object_GetAttr(v, 'health'))
		invuleffect = js.Object_PlayEffect(v, effect_name, '', js.FX_LOOP|js.FX_TRACK_OBJECT_FULL)
		m25getset.SetObjectVar(v, 'enc_invuleffect', invuleffect)

def EncCustom_AlliesHeldByObjects_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'allies_held', default=1):
		if Enc_IsAllGone(enc_id, 'object'):
		#remove invulnerability
			for v in GetEncVar(enc_id, 'ally', default=[]):
				js.Object_SetAttr(v, 'minHealth', 0)
				js.Object_StopEffect(v, m25getset.GetObjectVar(v, 'enc_invuleffect', default=''))
			SetEncVar(enc_id, 'allies_held', 0)
			if not GetEncVar(enc_id, 'objects_destroyed', default=0):
				PlayEncCutscene(enc_id, 'objectsdestroyedcs')
				SetEncVar(enc_id, 'objects_destroyed', 1)
		return None	
	else:
		return ['allies saved']

#-------------------------------------------------------------------------
#
# Custom Heroes Sabotage Object
#
#-------------------------------------------------------------------------	

def EncCustom_heroes_sabotage_object(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_Sabotage_CheckEnd']
	encounter['state'] = [['sabotage','not started']]
	encounter['on_examined'] = ['Sabotage_Complete']
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='object', command='sabotage', cutscene='sabotagecs'))
	return encounter
	
def EncCustom_Sabotage_Complete(enc_id, object):
	js.Mission_DisplayTimer('Detonation in...',10.0)
	m25event.SetTimer(10,'EncCustom_Sabotage_Explode',user=enc_id)
	
def EncCustom_Sabotage_Explode(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Trigger_Explosion(obj, 'ffx_ex_fire_large')
	SetEncVar(enc_id, 'sabotage','complete')
	m25event.SetTimer(1,'EncCustom_Sabotage_Done',user=enc_id)
	
def EncCustom_Sabotage_Done(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Object_Destroy(obj)
	EncCheckEnd(enc_id)
	
def EncCustom_Sabotage_CheckEnd(enc_id):
	if GetEncVar(enc_id,'sabotage',default=0) == 'complete':
		return ['sabotage complete']
	return None


#-------------------------------------------------------------------------
#
# Custom Invulnerability can end
#
#-------------------------------------------------------------------------	
	
def EncCustom_invulnerability_can_end(enc_id, parameters):
	encounter = {}
	encounter['state'] = [['invulnerability_ended', 0]]
	encounter['setup'] = ['EncCustom_InvulnerabilityCanEnd_Setup']
	return encounter
	
def EncCustom_InvulnerabilityCanEnd_Setup(enc_id):
	for e in GetEncVar(enc_id, 'on_invulnerability_over',default=[]): EncAddReg(enc_id, e, '_invulnerability_over')
	

def EncCustom_invulnerability_ends_when_objects_destroyed(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_InvulnerableEndObjects_CheckEnd']
	return encounter

def EncCustom_InvulnerableEndObjects_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'object') and not GetEncVar(enc_id, 'invulnerability_ended', default=0):
		EncEventCallback(enc_id, '_invulnerability_over', enc_id)
	if GetEncVar(enc_id, 'invulnerability_ended', default=0):
		return ['invulnerability ended']

def EncCustom_invulnerability_ends_when_minions_defeated(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_InvulnerableEndMinions_CheckEnd']
	return encounter

def EncCustom_InvulnerableEndMinions_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'minion') and not GetEncVar(enc_id, 'invulnerability_ended', default=0):
		EncEventCallback(enc_id, '_invulnerability_over', enc_id)
		
def EncCustom_invulnerability_ends_when_switches_turned_off(enc_id, parameters):
	encounter = {}
	encounter['on_switch_off'] = ['EncCustom_InvulnerableEndSwitches_Off']
	return encounter

def EncCustom_InvulnerableEndSwitches_Off(enc_id, var=[]):
	if EncCustom_SwitchOff_AllOff(enc_id) and not GetEncVar(enc_id, 'invulnerability_ended', default=0):
		EncEventCallback(enc_id, '_invulnerability_over', enc_id)
		
#-------------------------------------------------------------------------
#
# Custom Objects Make Villain Invulnerable
#
#-------------------------------------------------------------------------

def EncCustom_villains_are_invulnerable(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['invulvill_effect', EncCustom_GetParameter(parameters,'villaininvulnerabiltyeffect',default='effect_ffx_energyshield')],
	]
	encounter['setup'] = ['EncCustom_InvulnerableVillain_Setup']
	encounter['on_invulnerability_over'] = ['EncCustom_InvulnerableVillain_Over']
	return encounter

def EncCustom_InvulnerableVillain_Setup(enc_id):
	effect_name = GetEncVar(enc_id, 'invulvill_effect')
	for v in GetEncVar(enc_id, 'villain', default=[]):
		js.Object_SetAttr(v, 'minHealth', js.Object_GetAttr(v, 'health'))
		invuleffect = js.Object_PlayEffect(v, effect_name, '', js.FX_LOOP|js.FX_TRACK_OBJECT_FULL, 0, 'centre')
		m25getset.SetObjectVar(v, 'enc_invuleffect', invuleffect)	
	
def EncCustom_InvulnerableVillain_Over(enc_id, vars=[]):
	for v in GetEncVar(enc_id, 'villain', default=[]):
		js.Object_SetAttr(v, 'minHealth', 0)
		js.Object_StopEffect(v, m25getset.GetObjectVar(v, 'enc_invuleffect', default=''))
	SetEncVar(enc_id, 'invulnerability_ended', 1)


#-------------------------------------------------------------------------
#
# Custom Invulnerable Item
#
#-------------------------------------------------------------------------		
	
def EncCustom_item_is_invulnerable(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['invulobject_effect', EncCustom_GetParameter(parameters,'iteminvulnerabiltyeffect',default='effect_ffx_energyshield')],
		#['itempower', EncCustom_GetParameter(parameters,'power',default=None)],
	]
	encounter['on_item_on_ground'] = ['EncCustom_InvulnerableItem_Setup']
	encounter['on_invulnerability_over'] = ['EncCustom_InvulnerableItem_Stop']
	return encounter

def EncCustom_InvulnerableItem_Setup(enc_id, vals=[]):
	if GetEncVar(enc_id, 'invulitem_over', default=0): return
	effect_name = GetEncVar(enc_id, 'invulobject_effect', default='effect_ffx_energyshield')
	v = GetEncVar(enc_id, 'loot', default=None)
	if v:
	#for v in GetEncVar(enc_id, 'item', default=[]):
		js.Object_SetAttr(v, 'minHealth', js.Object_GetAttr(v, 'health'))
		invuleffect = js.Object_PlayEffect(v, effect_name, '', js.FX_LOOP|js.FX_TRACK_OBJECT_FULL, 0, 'centre')
		m25getset.SetObjectVar(v, 'enc_invuleffect', invuleffect)

def EncCustom_InvulnerableItem_Stop(enc_id, vars=[]):
	if GetEncVar(enc_id, 'invulitem_over', default=0): return
	v = GetEncVar(enc_id, 'loot', default=None)
	if v:
	#for v in GetEncVar(enc_id, 'item', default=[]):
		js.Object_SetAttr(v, 'minHealth', 0)
		js.Object_StopEffect(v, m25getset.GetObjectVar(v, 'enc_invuleffect', default=''))
	SetEncVar(enc_id, 'invulitem_over', 1)
	SetEncVar(enc_id, 'invulnerability_ended', 1)
		
# def InvulnerableItem_CheckEnd(enc_id):
# 	if Enc_IsAllGone(enc_id, 'item'):
# 		PlayEncCutscene(enc_id, 'itemdestroyedcs')
# 		return ['item destroyed']

#-------------------------------------------------------------------------
#
# Custom Item Can Be Destroyed
#
#-------------------------------------------------------------------------		
		
def EncCustom_item_can_be_destroyed(enc_id, parameters):
	encounter = {}
	encounter['on_item_destroyed'] = ['EncCustom_ItemDestroyed_Destroyed']
	encounter['custom_check_end'] = ['EncCustom_ItemDestroyed_Checkend']
	return encounter

def EncCustom_ItemDestroyed_Destroyed(enc_id, vals=[]):
	if not GetEncVar(enc_id, 'item_destroyed', default=0):
		PlayEncCutscene(enc_id, 'itemdestroyedcs')
		SetEncVar(enc_id, 'item_destroyed', 1)	
	
def EncCustom_ItemDestroyed_Checkend(enc_id):
	if GetEncVar(enc_id, 'item_destroyed', default=0):
		return ['item destroyed']
	return None
		
#-------------------------------------------------------------------------
#
# Custom Building Collapsing
#
#-------------------------------------------------------------------------

def EncCustom_buildings_collapsing(enc_id, parameters):
	encounter = {}
	encounter['state'] = [ ['building', m25string.CommaToList(EncCustom_GetParameter(parameters, 'buildings', default=[]))] ]
	encounter['start_function'] = ['EncCustom_BuildingsCollapse_Start']
	return encounter

def EncCustom_BuildingsCollapse_Start(event):
	enc_id = EncEvent_GetEncId(event)
	m25event.SetTimer(1, 'EncCustom_CollapseBuilding', user=enc_id)
	for e in GetEncVar(enc_id, 'on_building_collapsed',default=[]): EncAddReg(enc_id, e, '_building_collapsed')	
	
def EncCustom_CollapseBuilding(event):
	enc_id = EncEvent_GetEncId(event)
	if IsEncounterComplete(enc_id): return
	for b in GetEncVar(enc_id, 'building'):
		if js.Object_GetAttr(b, 'health') > 0 and randint(1,2) == 1:
			js.Trigger_Damage(b, 5)
			if js.Object_GetAttr(b, 'health') <= 0: 
				PlayEncCutscene(enc_id, 'buildingcollapsescs')
				EncEventCallback(enc_id, '_building_collapsed', b)
	m25event.SetTimer(1, 'EncCustom_CollapseBuilding', user=enc_id)


#-------------------------------------------------------------------------
#
# Custom Watch Buildings
#
#-------------------------------------------------------------------------
	
def EncCustom_watch_buildings(enc_id, parameters):
	pass


#-------------------------------------------------------------------------
#
# Custom Do Not KO Villains
#
#-------------------------------------------------------------------------

def EncCustom_do_not_ko_villains(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_NoKOVillains_CheckEnd']
	return encounter
	
def EncCustom_NoKOVillains_CheckEnd(enc_id):
	for v in m25ai.AddToList(GetEncVar(enc_id, 'villain', default = []), GetEncVar(enc_id, 'minion', default=[])):
		if (js.Object_Exists(v) and not js.Object_IsAlive(v)) or not js.Object_Exists(v): 
			EncCustom_FailObjective(enc_id)
			EncCustom_SetImmediateEnd(enc_id)
			return ['villain koed']
	return None
	



#-------------------------------------------------------------------------
#
# Custom Item Up For Grabs: Villains Flee With Item
#
#-------------------------------------------------------------------------

def EncCustom_villains_only_flee_with_item(enc_id, parameters):
	encounter = {}
	encounter['on_item_pickup'] = ['EncCustom_VillainsHaveItem']
	encounter['on_item_dropped'] = ['EncCustom_VillainsLoseItem']
	encounter['on_item_on_ground'] = ['EncCustom_VillainsLoseItem']
	encounter['on_fled'] = ['EncCustom_VillainsFleeWithItem']
	encounter['custom_check_end'] = ['EncCustom_VillainsFleeWithItem_CheckEnd']
	return encounter
	
def EncCustom_VillainsHaveItem(enc_id, vals=[]):
	if vals == '%i_minion'%enc_id:
		FleeToMarker_ResumeMoveGoal(enc_id, '%i_minion'%enc_id)

def EncCustom_VillainsLoseItem(enc_id, vals=[]):
	FleeToMarker_PauseMoveGoal(enc_id, '%i_minion'%enc_id)
	
def EncCustom_VillainsFleeWithItem(enc_id, vals=[]):
	char, chartype = vals
	if chartype == 'leader':
		EncEventCallback(enc_id, '_item_lost')

def EncCustom_VillainsFleeWithItem_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['villain', 'minion']): return ['villains gone']
	return None
		
#-------------------------------------------------------------------------
#
# Custom Item Up For Grabs: Villains Want Item
#
#-------------------------------------------------------------------------

def EncCustom_villains_want_item(enc_id, parameters):
	encounter = {}
	encounter['on_item_pickup'] = ['EncCustom_VillainsWantItem_Pickup']
	#encounter['on_item_dropped'] = ['EncCustom_VillainsWantItem_Dropped']
	encounter['on_item_on_ground'] = ['EncCustom_VillainsWantItem_OnGround']
	encounter['on_hero_gets_item'] = ['EncCustom_VillainsWantItem_HeroHasIt']
	encounter['triggers'] = ['OnGroupLeaderKO("minion", "EncCustom_VillainsWantItem_LeaderKO")']
	encounter['custom_check_end'] = ['EncCustom_VillainsWantItem_CheckEnd']
	return encounter	

	
def EncCustom_VillainsWantItem_Pickup(enc_id, vals=[]):
	groupname = vals
	if groupname == '%i_minion'%enc_id:
		PlayEncCutscene(enc_id, 'villainsgetitemcs')
		#cshelper.regDeath(m25ai.Group_GetLeader('%i_minion'%enc_id), 'EncCustom_VillainsWantItem_LeaderKO')
	else:
		m25ai.GroupCommand_Attack('%i_minion'%enc_id, m25ai.Group_GetLeader(groupname))
		m25ai.GroupCommand_MoveToTarget('%i_minion'%enc_id, m25ai.Group_GetLeader(groupname), priority=TACTIC_PRIORITY_HIGH)

def EncCustom_VillainsWantItem_OnGround(enc_id, vals=[]):
	m25ai.GroupCommand_RetrieveItem('%i_minion'%enc_id, GetEncVar(enc_id,'loot'), priority=TACTIC_PRIORITY_HIGH, function='EncCustom_ItemForGrabs_Retrieved("%i_minion")'%enc_id)
	
def EncCustom_VillainsWantItem_HeroHasIt(enc_id, vals=[]):
	hero = vals
	m25ai.GroupCommand_Attack('%i_minion'%enc_id, hero)
	m25ai.GroupCommand_MoveToTarget('%i_minion'%enc_id, hero, priority=TACTIC_PRIORITY_HIGH)
		
def EncCustom_VillainsWantItem_LeaderKO(groupname, leader):
	enc_id = Enc_GetCharEnc(leader)
	#leader = m25ai.Group_GetLeader('%i_minion'%enc_id)
	#m25ai.Group_PromoteLeader('%i_minion'%enc_id)
	if GetEncVar(enc_id, 'item_group') == '%i_minion'%enc_id:
		PlayEncCutscene(enc_id, 'villainsdropitemcs')
		EncEventCallback(enc_id, '_item_dropped', leader)
	
def EncCustom_VillainsWantItem_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'item_group', default=None) == '%i_minion'%(enc_id):
		return ['villains have item']
	else:
		if Enc_IsAllGone(enc_id, ['villain','minion']):
			return ['villains do not have item']
	return None
		
#-------------------------------------------------------------------------
#
# Custom Item Up For Grabs: Villains Start With Item
#
#-------------------------------------------------------------------------	
	
def EncCustom_villains_start_with_item(enc_id, parameters):
	encounter = {}
	encounter['item_group'] = '%i_minion'%enc_id
	encounter['start_function'] = ['EncCustom_VillainsStartWithItem_Start']
	return encounter
		
def EncCustom_VillainsStartWithItem_Start(event):
	enc_id = int(event.user)
	SetEncVar(enc_id,'item_state','carried')
	SetEncVar(enc_id, 'item_group', '%i_minion'%enc_id)
	EncEventCallback(enc_id, '_item_pickup', '%i_minion'%enc_id)

def EncCustom_rivals_start_with_item(enc_id, parameters):
	encounter = {}
	encounter['item_group'] = '%i_cohort'%enc_id
	encounter['start_function'] = ['EncCustom_RivalsStartWithItem_Start']
	return encounter
		
def EncCustom_RivalsStartWithItem_Start(event):
	enc_id = int(event.user)
	SetEncVar(enc_id,'item_state','carried')
	SetEncVar(enc_id, 'item_group', '%i_cohort'%enc_id)
	EncEventCallback(enc_id, '_item_pickup', '%i_cohort'%enc_id)	
		
#-------------------------------------------------------------------------
#
# Custom Item Up For Grabs
#
#-------------------------------------------------------------------------	

def EncCustom_item_exists(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_ItemForGrabs_Setup']
	#encounter['loot'] = EncCustom_GetParameter(parameters, 'item', default=None)
	template = EncCustom_GetParameter(parameters, 'item', default='cub_market_crate2')
	encounter['state'] = [
		['item_template', template],
	]
	#encounter['on_item_pickup'] = ['EncCustom_ItemForGrabs_Retrieved']
	encounter['on_item_dropped'] = ['EncCustom_ItemForGrabs_Dropped']
	encounter['on_item_on_ground'] = ['EncCustom_ItemForGrabs_OnGround']
	encounter['on_hero_gets_item'] = ['EncCustom_ItemForGrabs_HeroHasIt']
	encounter['on_item_lost'] = ['EncCustom_ItemForGrabs_ItemLost']
	encounter['custom_check_end'] = ['EncCustom_ItemForGrabs_CheckEnd']
	return encounter	
	
def EncCustom_ItemForGrabs_Setup(enc_id):
	for e in GetEncVar(enc_id, 'on_item_pickup',default=[]): EncAddReg(enc_id, e, '_item_pickup')
	for e in GetEncVar(enc_id, 'on_item_dropped',default=[]): EncAddReg(enc_id, e, '_item_dropped')
	for e in GetEncVar(enc_id, 'on_item_on_ground',default=[]): EncAddReg(enc_id, e, '_item_on_ground')
	for e in GetEncVar(enc_id, 'on_item_lost',default=[]): EncAddReg(enc_id, e, '_item_lost')
	for e in GetEncVar(enc_id, 'on_item_destroyed',default=[]): EncAddReg(enc_id, e, '_item_destroyed')
	for e in GetEncVar(enc_id, 'on_hero_gets_item',default=[]): EncAddReg(enc_id, e, '_hero_gets_item')
	
def EncCustom_ItemForGrabs_Retrieved(char, groupname):
	enc_id = Enc_GetCharEnc(char)
	SetEncVar(enc_id,'item_state','carried')
	SetEncVar(enc_id, 'item_group', groupname)
	money = GetEncVar(enc_id,'loot', default=None)
	if money: js.Object_Destroy(money)
	PlayEncCutscene(enc_id, 'itemretrievedcs',in_order=0)
	EncEventCallback(enc_id, '_item_pickup', groupname)

def EncCustom_ItemForGrabs_Dropped(enc_id, char):
	pos = js.Get_ObjectPos(char)
	item = GetEncVar(enc_id, 'item_template', default='cub_market_crate2')
	name = m25name.GetNextObjectName()
	js.Object_SpawnAt(item, pos, name)
	SetEncVar(enc_id, 'loot', name)
	Enc_SetCharEnc(name, enc_id)
	cshelper.regDeath(name,'EncCustom_ItemForGrabs_ItemDestroyed',user=enc_id)
	SetEncVar(enc_id, 'item_group', None)
	PlayEncCutscene(enc_id, 'itemdroppedcs',in_order=0)
	EncEventCallback(enc_id, '_item_on_ground', char)
	
def EncCustom_ItemForGrabs_OnGround(enc_id, vals=[]):
	item = GetEncVar(enc_id, 'loot', default=None)
	SetEncVar(enc_id, 'item_state', 'on_ground')
# 	if item:
# 		cshelper.addCustomCommand('get item',item,'EncCustom_ItemForGrabs_HeroPickup',1,4,'')	
		
def EncCustom_ItemForGrabs_HeroHasIt(enc_id, vals=[]):
	hero = vals
	SetEncVar(enc_id, 'item_state', 'hero_has_it')
	SetEncVar(enc_id, 'loot', None)
	
def EncCustom_ItemForGrabs_HeroPickup(item, hero):
	enc_id = Enc_GetCharEnc(item)
	js.Object_Destroy(item)
	cshelper.regDeath(hero, 'EncCustom_ItemForGrabs_HeroKOed', user=enc_id)
	SetEncVar(enc_id, 'item_group', 'hero')
	PlayEncCutscene(enc_id, 'herogetsitemcs')
	EncEventCallback(enc_id, '_hero_gets_item', hero)
	EncCheckEnd(enc_id)
		
def EncCustom_ItemForGrabs_HeroKOed(event):
	enc_id = int(event.user)
	EncCustom_ItemForGrabs_Dropped(enc_id, event.object)
	
def EncCustom_ItemForGrabs_ItemLost(enc_id, vals=[]):
	SetEncVar(enc_id, 'item_state', 'lost')
	SetEncVar(enc_id, 'loot', None)	
	PlayEncCutscene(enc_id, 'itemlostcs')
	
def EncCustom_ItemForGrabs_ItemDestroyed(event):
	enc_id = int(event.user)
	SetEncVar(enc_id, 'item_state', 'destroyed')
	SetEncVar(enc_id, 'loot', None)
	PlayEncCutscene(enc_id, 'itemdestroyedcs')
	EncEventCallback(enc_id, '_item_destroyed')
	EncCheckEnd(enc_id)
	
def EncCustom_ItemForGrabs_CheckEnd(enc_id):
	state = GetEncVar(enc_id, 'item_state',default='on_ground')
	if state == 'hero_has_it': return ['hero has item', 'item recovered']
	if state == 'lost': return ['item lost']
	if state == 'destroyed': return ['item destroyed']
	if state == 'switched_off': return ['item switched off']
	return None	

#-------------------------------------------------------------------------
#
# Custom Heroes can pickup the item
#
#-------------------------------------------------------------------------	

def EncCustom_heroes_pickup_item(enc_id, parameters):
	encounter = {}
	encounter['on_item_on_ground'] = ['EncCustom_HeroesPickupItem_OnGround']
	return encounter

def EncCustom_HeroesPickupItem_OnGround(enc_id, vals=[]):
	item = GetEncVar(enc_id, 'loot', default=None)
	if item:
		cshelper.addCustomCommand('get item',item,'EncCustom_ItemForGrabs_HeroPickup',1,4,'')	
#-------------------------------------------------------------------------
#
# Custom Item Starts On The Ground
#
#-------------------------------------------------------------------------		
	
def EncCustom_item_starts_on_ground(enc_id, parameters):
	encounter = {}
	encounter['markers'] = [['itemstart','item', '']]
	encounter['setup'] = ['EncCustom_ItemStartsOnGround_Setup']
	encounter['state'] = [
		['item_state', 'on_ground'],
	]
	encounter['start_function'] = ['EncCustom_ItemStartsOnGround_Start']
	return encounter
	
def EncCustom_ItemStartsOnGround_Setup(enc_id):
	item = GetEncVar(enc_id, 'item_template', default='cub_market_crate2')
	marker = GetEncVar(enc_id, 'itemstart')
	name = m25name.GetNextObjectName()
	js.Object_Spawn(name, item, marker)
	SetEncVar(enc_id, 'loot', name)
	js.Object_SetAttr(name, 'health', js.Object_GetAttr(name, 'maxHealth'))
	Enc_SetCharEnc(name, enc_id)
	cshelper.regDeath(name,'EncCustom_ItemForGrabs_ItemDestroyed',user=enc_id)
	SetEncVar(enc_id, 'item_group', None)

def EncCustom_ItemStartsOnGround_Start(event):
	enc_id = int(event.user)
	EncEventCallback(enc_id, '_item_on_ground')


	
		
#-------------------------------------------------------------------------
#
# Custom Start at a distance
#
#-------------------------------------------------------------------------

def EncCustom_villains_start_at_distance(enc_id, parameters):
	encounter = {}
	encounter['villain_marker_loc'] = ['villainout', 'dist500']
	return encounter
	
def EncCustom_minions_start_at_distance(enc_id, parameters):
	encounter = {}
	encounter['minion_marker_loc'] = ['minionout', 'dist500']
	return encounter

def EncCustom_rivals_start_at_distance(enc_id, parameters):
	encounter = {}
	encounter['rival_marker_loc'] = ['rivalout', 'dist500']
	return encounter

def EncCustom_cohorts_start_at_distance(enc_id, parameters):
	encounter = {}
	encounter['cohort_marker_loc'] = ['cohortout', 'dist500']
	return encounter	
	
def EncCustom_allies_start_at_distance(enc_id, parameters):
	encounter = {}
	encounter['ally_marker_loc'] = ['allyout', 'dist500']
	return encounter		

	
	
	
#-------------------------------------------------------------------------
#
# Custom Allies Become Controllable
#
#-------------------------------------------------------------------------

def EncCustom_allies_become_controllable(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_AlliesControllable_CheckEnd']	
	return encounter
	
def EncCustom_AlliesControllable_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id,['villain','minion','rival','cohort']):
		allies = GetEncVar(enc_id, 'ally')
		for a in allies:
			ffx.FFX_AddPortrait(a)
			#js.AI_MakeIntoHero(a)
			Enc_SetEncounterIndependent(a)
			m25ai.AIDisableCustom(a)
			m25ai.SetAIStandardOnly(a)
			m25ai.AIEnable(a)
			m25ai.AIEnableStandard(a)
		PlayEncCutscene(enc_id, 'alliesjoinheroescs')
		return ['allies join team']
	if Enc_IsAllGone(enc_id, 'ally'):
		return ['allies defeated']
	return None
				
	
#-------------------------------------------------------------------------
#
# Custom Allies Congregate
#
#-------------------------------------------------------------------------

def EncCustom_allies_congregate(enc_id, parameters):
	encounter = {}
	encounter['markers'] = [ ['congregate', 'congregate', 'dist300'], ]
	encounter['state'] = [['allies_congregated',0]]
	encounter['setup'] = ['EncCustom_Allies_Congregate_Setup']
	encounter['custom_check_end'] = ['EncCustom_Allies_Congregate_CheckEnd']
	return encounter
	
def EncCustom_Allies_Congregate_Setup(enc_id):
	for char in GetEncVar(enc_id, 'ally', default=[]):
		m25ai.AddGoal_MoveToMarker(char,marker=GetEncVar(enc_id,'congregate'), priority=m25ai.TACTIC_PRIORITY_MEDIUM, pct=100, fn='EncCustom_Allies_Congregate_AtMarker(%i)'%enc_id)
		
def EncCustom_Allies_Congregate_AtMarker(char, enc_id):
	#disable the character and mark them as done
	m25ai.AIDisable(char)
	SetEncVar(enc_id, 'allies_congregated', GetEncVar(enc_id, 'allies_congregated', default=0)+1)
	PlayEncCutscene(enc_id, 'allycongregatescs', charlist = [[char,'ally']])
	EncCheckEnd(enc_id)

def EncCustom_Allies_Congregate_CheckEnd(enc_id):
	total = GetEncVar(enc_id, 'totalallys', default=0)
	congregated = GetEncVar(enc_id, 'allies_congregated',default=0)
	koed = GetEncVar(enc_id, 'allys_koed',default=0)
	if total <= congregated + koed:
		if congregated >= total:
			return ['all congregated']
		elif koed >= total:
			EncCustom_FailObjective(enc_id)
			return ['none congregated']
		else:
			return ['some congregated']
	return None


#-------------------------------------------------------------------------
#
# Custom Allies Fight along side Hero
#
#-------------------------------------------------------------------------	

def EncCustom_allies_fight_villains(enc_id, parameters):
	encounter = {}
	encounter['state'] = [['allies_fv_complete',0]]
	encounter['goals'] = [
		'EachFightForTeam("ally", 1)',
	]
	encounter['custom_check_end'] = ['EncCustom_AlliesFightVillains_CheckEnd']
	return encounter	
	
def EncCustom_AlliesFightVillains_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'totalallys',default=0) > 0:
		if Enc_IsAllKO(enc_id, 'ally'):
			if not GetEncVar(enc_id, 'allies_fv_complete',default=0):
				PlayEncCutscene(enc_id, 'allieslostcs')
				SetEncVar(enc_id, 'allies_fv_complete', 1)
			return ['allies lost']
		else:
			if Enc_IsAllGone(enc_id, ['villain','minion']):
				if not GetEncVar(enc_id, 'allies_fv_complete',default=0):
					PlayEncCutscene(enc_id, 'alliessavedcs')
					SetEncVar(enc_id, 'allies_fv_complete', 1)
				return ['allies saved']
			else:
				return None
	return ['no allies']
	
def EncCustom_allies_do_not_fight(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachFightForTeam("ally", 0)',
	]
	return encounter		
	
#-------------------------------------------------------------------------
#
# Custom Allies Fight Heroes
#
#-------------------------------------------------------------------------		

def EncCustom_allies_fight_heroes(enc_id, parameters):
	encounter = {}
	encounter['goals'] =[
		'EachFightForTeam("ally",2)',
	]
	encounter['custom_check_end'] = ['EncCustom_AlliesFightHeroes_CheckEnd']	
	return encounter

def EncCustom_AlliesFightHeroes_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'ally'):
		return ['allies defeated']
	return None	
	
#-------------------------------------------------------------------------
#
# Custom Allies Flee Map
#
#-------------------------------------------------------------------------

def EncCustom_allies_flee_map(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_AlliesFlee_CheckEnd']
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='ally',marker_id='end',marker_prefix='end',mass_flee=1))
	return encounter

def EncCustom_AlliesFlee_CheckEnd(enc_id, val=[]):
	return EncCustom_Flee_CheckEnd(enc_id, name='ally')

#-------------------------------------------------------------------------
#
# Custom Allies Follow Heroes
#
#-------------------------------------------------------------------------

def EncCustom_allies_follow_heroes(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachMoveToHero("ally",mindist=200)',
	]
	return encounter
	
#-------------------------------------------------------------------------
#
# Custom Allies Remain After Encounter Ends
#
#-------------------------------------------------------------------------

def EncCustom_allies_remain_after_encounter_ends(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesRemain_Setup']
	return encounter
	
def EncCustom_AlliesRemain_Setup(enc_id):
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		#make sure the allies don't get whisked away by another encounter
		m25getset.SetObjectVar(a, 'ai_spawn_teleport',0)
		#set them encounter independent so they don't get auto-removed
		Enc_SetEncounterIndependent(a)
			


#-------------------------------------------------------------------------
#
# Custom Allies Thank Heroes
#
#-------------------------------------------------------------------------	

def EncCustom_allies_thank_heroes(enc_id, parameters):
	encounter = {}
	encounter['state'] =[
		['ally_thank_complete', 0],
		['ally_thank_setup', 0],
	]
	encounter['custom_check_end'] = ['EncCustom_AlliesThankHeroes_CheckEnd']
	encounter['clean_up'] = ['EncCustom_AlliesThankHeroes_CleanUp']
	return encounter

def EncCustom_TryThank(enc_id):
 	ally = findCharFromList(chartype='ally',enc_id=enc_id,live=1)
 	hero = m25ai.GetClosestHero(ally)
 	cshelper.addCustomCommand('talk',ally,'EncCustom_Thank',1,4,'')
 	cshelper.addArrow(ally+'arrow', ally, colour = js.ARROW_YELLOW)
 	m25getset.SetObjectVar(ally,'encounter',enc_id)
 	m25ai.AddGoal_MoveToTarget(ally,hero)
 	SetEncVar(enc_id, 'ally_thank_setup', 1)
 	PlayEncCutscene(enc_id, 'allywantstothankheroescs')
 	PlayEncCutscene(enc_id, 'enemiesdowncs')
  	
def EncCustom_Thank(target,char):
 	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
 	SetEncVar(enc_id,'ally_thank_complete',1)
 	cshelper.removeAttention(target)
 	cshelper.removeArrow(target+'arrow')
 	PlayEncCutscene(enc_id, 'allythanksherocs',charlist=[[target,'ally'],[char,'hero']],return_function='EncCheckEnd('+`enc_id`+')')	
	
def EncCustom_AlliesThankHeroes_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'ally_thank_complete', default=0):
		return ['ally thanked hero', 'ally survives']
	if Enc_IsAllGone(enc_id, ['ally']):
		return ['ally did not thank hero', 'ally lost']
	if Enc_IsAllGone(enc_id, ['villain', 'minion']) and not GetEncVar(enc_id, 'ally_thank_setup', default=0):
		EncCustom_TryThank(enc_id)
	return None
			
def EncCustom_AlliesThankHeroes_CleanUp(enc_id):
	if GetEncVar(enc_id, 'ally_thank_complete', default=0):
		PlayEncCutscene(enc_id, 'itemrecoveredcs')


	
#-------------------------------------------------------------------------
#
# Custom Allies Frozen
#
#-------------------------------------------------------------------------	
	
def EncCustom_allies_frozen(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesFrozen_Setup']
	encounter['start_function'] = ['EncCustom_AlliesFrozen_Start']
	return encounter

def EncCustom_AlliesFrozen_Setup(enc_id):
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		#set minhealth so characters don't die while frozen, until the hero gets in range
		js.Object_SetAttr(a,'minHealth', js.Object_GetAttr(a, 'maxHealth'))
		#freeze the character
		js.Object_SetPrimaryState(a, js.PCSTATE_FROZEN, 10, js.PERMANENT_STATE)
		#set up the callback when the character becomes unfrozen
		js.Event_RegisterSink(js.EVENT_CHARACTER_STATE_CHANGED, 'EncCustom_AlliesFrozen_CheckFreed', a, '', enc_id, js.PCSTATE_NORMAL, 0)
		
def EncCustom_AlliesFrozen_Start(event):
	enc_id = int(event.user)
	#if GetEncVar(enc_id,'encounter_started',default=0): return
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		# set minhealth back to zero, so that chars can die while frozen
		js.Object_SetAttr(a, 'minHealth', 0)
	
def EncCustom_AlliesFrozen_CheckFreed(event):
	char = event.object
	enc_id = Enc_GetCharEnc(char)
	EncEventCallback(enc_id, '_freed', char)
	
#-------------------------------------------------------------------------
#
# Custom Allies In Stasis
#
#-------------------------------------------------------------------------	
	
def EncCustom_allies_in_stasis(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesStasis_Setup']
	return encounter

def EncCustom_AlliesStasis_Setup(enc_id):
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		js.Object_SetPrimaryState(a, js.PCSTATE_STATIC, 10, js.PERMANENT_STATE)

#-------------------------------------------------------------------------
#
# Custom Allies Cower
#
#-------------------------------------------------------------------------	
	
def EncCustom_allies_cower(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesCower_Setup']
	return encounter

def EncCustom_AlliesCower_Setup(enc_id):
	allies = GetEncVar(enc_id, 'ally')
	for a in allies:
		js.Object_SetPrimaryState(a, js.PCSTATE_INTEROGATE, 0, js.PERMANENT_STATE)


#-------------------------------------------------------------------------
#
# Custom Allies In Cages
#
#-------------------------------------------------------------------------			

def EncCustom_allies_in_cages(enc_id, parameters):
	encounter = {}
	encounter['state'] =[
		['cage', []],
		['cage_type', EncCustom_GetParameter(parameters, 'cage', default='nazihq_killarilla_cage')],
	]
	encounter['setup'] = ['EncCustom_RescueCaged_Setup']
	encounter['clean_up'] = ['EncCustom_RescueCaged_CleanUp']
	return encounter

def EncCustom_RescueCaged_Setup(enc_id):
	allies = GetEncVar(enc_id,'ally')
	cages = []
	for a in allies:
		name = m25name.GetNextObjectName()
		m25_spawn(name,'nazihq_killarilla_cage',js.Get_ObjectPos(a))
		cages.append(name)
		m25getset.SetObjectVar(name, '_enclosed_ally', a)
		cshelper.regDeath(name,'EncCustom_RescueCaged_AllyFreed',user=enc_id)		
	SetEncVar(enc_id,'cage',cages)
	
def EncCustom_RescueCaged_AllyFreed(event):
	enc_id = int(event.user)
	cage = event.object
	ally = m25getset.GetObjectVar(cage, '_enclosed_ally', default=None)
	if ally == None: return
	EncEventCallback(enc_id, '_freed', ally)	

def EncCustom_RescueCaged_CleanUp(enc_id):
	cages = GetEncVar(enc_id, 'cage', default=[])
	for c in cages:
		js.Object_Destroy(c)	
	
#-------------------------------------------------------------------------
#
# Custom Chat with Ally
#
#-------------------------------------------------------------------------

def EncCustom_allies_want_to_chat(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_Chat_CheckEnd']
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='ally', command=EncCustom_GetParameter(parameters, 'chatcommand',default='TALK_TO'), cutscene='chatcs'))
	return encounter
				
def EncCustom_Chat_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'examine_complete', default=0):
		PlayEncCutscene(enc_id, 'chatwoncs')
		return ['chat complete', 'chat won']
	if Enc_IsAllGone(enc_id, 'ally'):
		PlayEncCutscene(enc_id, 'chatlostcs')
		return ['chat lost']
	return None
	
#-------------------------------------------------------------------------
#
# Custom Heroes Destroy Objects
#
#-------------------------------------------------------------------------	
	
def EncCustom_heroes_destroy_objects(enc_id, parameters):	
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_DestroyObjects_CheckEnd']
	encounter['goals'] = [
		'EachFightForTeam("object", 2)',
	]
	return encounter
	
def EncCustom_DestroyObjects_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['object']): 
		if not GetEncVar(enc_id, 'objects_destroyed', default=0):
			PlayEncCutscene(enc_id, 'objectsdestroyedcs')
			SetEncVar(enc_id, 'objects_destroyed', 1)
		return ['objects destroyed']
	else: return None

	
def EncCustom_allies_destroy_objects(enc_id, parameters):	
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_DestroyObjects_CheckEnd']
	encounter['goals'] = [
		'EachAttackTarget("ally","object")',
	]
	return encounter
	
	
#-------------------------------------------------------------------------
#
# Custom Disarm Bomb
#
#-------------------------------------------------------------------------
	
def EncCustom_disarm_bomb(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_DisarmBomb_CheckEnd']
	encounter['on_custom_timer_up'] = ['EncCustom_DisarmBomb_Explode']
	encounter['on_examined'] = ['EncCustom_DisarmBomb_Disarmed']
	encounter['state'] = [
		['bomb_state','ticking'],
	]
	encounter = CombineEncounters(encounter, EncSpawn(group_id='bomb',marker_prefix='bomb',var=[EncCustom_GetParameter(parameters, 'bomb',default='cub_market_crate2')]))
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='bomb', command=EncCustom_GetParameter(parameters, 'disarmcommand',default='DISARM'), cutscene='disarmbombcs'))
	return encounter
	
def EncCustom_DisarmBomb_Disarmed(enc_id, vals=[]):
	EncCustom_Timer_StopEvent(enc_id)
	SetEncVar(enc_id, 'bomb_state','disarmed')
	PlayEncCutscene(enc_id, 'bombdisarmedcs', return_function='EncCheckEnd(%i)'%enc_id)

def EncCustom_DisarmBomb_Explode(enc_id, vals=[]):
	PlayEncCutscene(enc_id, 'bombexplodingcs', return_function='EncCustom_DisarmBomb_Explode2('+`enc_id`+')')
	
def EncCustom_DisarmBomb_Explode2(enc_id):	
	obj = GetEncVar(enc_id, 'examine_object')
	js.Trigger_Explosion(obj, 'ffx_ex_fire_large')
	SetEncVar(enc_id,'bomb_state','exploded')
	m25event.SetTimer(1,'EncCustom_DisarmBomb_Detonated',user=enc_id)
	
def EncCustom_DisarmBomb_Detonated(event):
	enc_id = int(event.user)
	obj = GetEncVar(enc_id, 'examine_object')
	js.Object_Destroy(obj)
	EncCheckEnd(enc_id)

def EncCustom_DisarmBomb_CheckEnd(enc_id, vals=[]):
	bomb_state = GetEncVar(enc_id,'bomb_state')
	if bomb_state == 'ticking': return None
	if bomb_state == 'disarmed': return ['bomb disarmed']
	elif bomb_state == 'exploded':
		EncCustom_FailObjective(enc_id)
		return ['bomb exploded']
	else:
		return None
	
#-------------------------------------------------------------------------
#
# Custom Examine Object
#
#-------------------------------------------------------------------------

def EncCustom_object_can_be_examined(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_Examine_CheckEnd']
	encounter = CombineEncounters(encounter, EncExamine(objectgroup='object', command=EncCustom_GetParameter(parameters, 'examinecommand',default='EXAMINE'), cutscene='examinecs'))	
	return encounter
				
def EncCustom_Examine_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'examine_complete', default=0):
		return ['examine complete']
	return None
	
#-------------------------------------------------------------------------
#
# Custom Escort Allies
#
#-------------------------------------------------------------------------			
		
def EncCustom_heroes_escort_allies(enc_id, parameters):
	encounter = {}
	encounter['start_function'] = ['EncCustom_Escort_Start']
	encounter['custom_check_end'] = ['EncCustom_Escort_CheckEnd']
	encounter = CombineEncounters(encounter, EncRescueally(group_id='ally', marker_id='end', follow=1))
	return encounter
	
def EncCustom_Escort_Start(event):
	enc_id = int(event.user)
	for v in GetEncVar(enc_id, 'ally'):
		RescueAlly_Freed(enc_id, v)
	
def EncCustom_Escort_CheckEnd(enc_id):
	if not Enc_IsAllGone(enc_id,'ally'):
		return None
	try: js.Mission_RemoveArrow(`enc_id` + '_rescue_arrow')
	except: pass
	if GetEncVar(enc_id,'allys_escaped',default=0) > 0:
		PlayCutsceneFromList(GetEncCutscene(enc_id,'escortwoncs'),enc_id)
		return ['escort won']
	else:
		EncCustom_FailObjective(enc_id)
		PlayCutsceneFromList(GetEncCutscene(enc_id,'escortlostcs'),enc_id)
		return ['escort lost']


		
#-------------------------------------------------------------------------
#
# Custom Fire
#
#-------------------------------------------------------------------------		
		
def EncCustom_fire(enc_id, parameters):	
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_Fire_CheckEnd']
	encounter['setup'] = ['EncCustom_Fire_Setup']
	return encounter
		
def EncCustom_Fire_Setup(enc_id):
	num_flames = 5
	for i in range(0,num_flames):
		m25event.SetTimer(0.5*i, 'EncCustom_Fire_FlameUp', user=enc_id)

def EncCustom_Fire_FlameUp(event):
	enc_id = int(event.user)
	m25event.SetTimer(1, 'EncCustom_Fire_FlameUp', user=enc_id)
	name = m25name.GetNextObjectName()
	pos = EncCustom_Fire_GetFlamePosition(enc_id)
	pos2 = (pos[0]+random.randint(-50,50), pos[1]+random.randint(-50,50), pos[2]) 
	js.Object_SpawnAt('ffx_flames', pos2, name)
	js.Object_PlayEffect(name,'effect_ffx_flames2','',js.FX_LOOP)
	if GetEncVar(enc_id,'encounter_started',default=0) and random.randint(1,3)==1:
		js.Trigger_Explosion(name,'ffx_flames',-1)
	m25event.SetTimer(2, 'EncCustom_Fire_DieOff', name)
	
def EncCustom_Fire_DieOff(event):
	name = event.object
	js.Object_Destroy(name)		

def EncCustom_Fire_GetFlamePosition(enc_id):
	startloc = js.Get_MarkerPos(GetEncVar(enc_id, 'start'))
	newloc = (startloc[0]+random.randint(-100,100), startloc[1]+random.randint(-100,100), startloc[2])
	return newloc		
		
def EncCustom_Fire_CheckEnd(enc_id):	
	return ['fire over']	
	

#-------------------------------------------------------------------------
#
# Custom Allies remain still
#
#-------------------------------------------------------------------------	
	
def EncCustom_allies_do_not_move(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['ally_disabled', 1],
	]	
	return encounter

#-------------------------------------------------------------------------
#
# Custom Allies can be freed
#
#-------------------------------------------------------------------------		
	
def EncCustom_allies_can_be_freed(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['allies_freed', 0],
	]
	encounter['setup'] = ['EncCustom_AlliesCanBeFreed_Setup']
	encounter['on_freed'] = ['EncCustom_AlliesCanBeFreed_Freed']
	encounter['custom_check_end'] = ['EncCustom_AlliesCanBeFreed_CheckEnd']
	return encounter
	
def EncCustom_AlliesCanBeFreed_Setup(enc_id):
	for e in GetEncVar(enc_id, 'on_freed',default=[]): EncAddReg(enc_id, e, '_freed')
	
def EncCustom_AlliesCanBeFreed_Freed(enc_id, char, vars=None):
	enc_id = Enc_GetCharEnc(char)
	hero = vars
	SetEncVar(enc_id, 'allies_freed', GetEncVar(enc_id, 'allies_freed', default=0)+1)
	if GetEncVar(enc_id, 'allies_freed', default=0) >= GetEncVar(enc_id, 'totalallys', default=0):
		PlayEncCutscene(enc_id, 'allalliesfreedcs', return_function="EncCheckEnd(%i)"%enc_id)
	else:
		PlayEncCutscene(enc_id, 'allyfreedcs', charlist = [[char,'ally'], [hero, 'hero']], return_function="EncCheckEnd(%i)"%enc_id)
		
def EncCustom_AlliesCanBeFreed_CheckEnd(enc_id):
	freed = GetEncVar(enc_id, 'allies_freed', default=0)
	total = GetEncVar(enc_id, 'totalallys', default=0)
	ko = GetEncVar(enc_id, 'allys_koed', default=0)
	print 'alliescanbefreed', total, freed, ko
	if freed + ko >= total:
		if freed >= total: return ['all allies freed']
		if freed <= 0: return ['no allies freed']
		return ['some allies freed']
	return None

	
#-------------------------------------------------------------------------
#
# Custom Allies move when freed
#
#-------------------------------------------------------------------------	
	
def EncCustom_allies_move_when_freed(enc_id, parameters):
	encounter = {}
	encounter['on_freed'] = ['EncCustom_AlliesMoveFreed_Freed']
	return encounter
	
def EncCustom_AlliesMoveFreed_Freed(enc_id, char):
	m25ai.AIEnable(char)


#-------------------------------------------------------------------------
#
# Custom Heroes Rescue Allies
#
#-------------------------------------------------------------------------

def EncCustom_heroes_rescue_allies(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesRescued_Setup']
	return encounter

def EncCustom_AlliesRescued_Setup(enc_id):
	for a in GetEncVar(enc_id, 'ally', default=[]):
		js.Mission_CustomAction('rescue', '', a, 'EncCustom_AlliesRescued_Rescue', 5, 1)
	
def EncCustom_AlliesRescued_Rescue(char, hero):
	enc_id = Enc_GetCharEnc(char)
	js.Object_SetPrimaryState(char, js.PCSTATE_NORMAL, 0)
	PlayEncCutscene(enc_id, 'allyrescuedcs', charlist = [[char,'ally'], [hero, 'hero']], return_function="EncEventCallback(%i, '_freed', '%s')"%(enc_id, char))
	#EncEventCallback(enc_id, '_freed', ally)

#-------------------------------------------------------------------------
#
# Custom Heroes Carry Allies
#
#-------------------------------------------------------------------------		

def EncCustom_heroes_carry_allies_when_freed(enc_id, parameters):
	encounter = {}
	encounter['markers'] = [ ['carryto', 'carryto', 'dist500'], ]
	encounter['state'] = [['allies_carried',0]]
	encounter['on_freed'] = ['EncCustom_Allies_Carried_Freed']
	encounter['custom_check_end'] = ['EncCustom_Allies_Carried_CheckEnd']
	return encounter
	
def EncCustom_heroes_carry_allies(enc_id, parameters):
	encounter = {}
	encounter['markers'] = [ ['carryto', 'carryto', 'dist500'], ]
	encounter['state'] = [['allies_carried',0]]
	encounter['setup'] = ['EncCustom_AlliesCarried_Setup']
	encounter['custom_check_end'] = ['EncCustom_Allies_Carried_CheckEnd']
	return encounter	
	
def EncCustom_AlliesCarried_Setup(enc_id):
	for a in GetEncVar(enc_id, 'ally', default=[]):
		js.Mission_CustomAction('carry', '', a, 'RescueBase_CarryStart', 5, 1)
	
def EncCustom_Allies_Carried_Freed(char, vars=[]):
	js.Mission_CustomAction('carry', '', char, 'RescueBase_CarryStart', 5, 1)
	js.Object_SetPrimaryState(char, js.PCSTATE_INTEROGATE, 0, js.PERMANENT_STATE)
	
def RescueBase_CarryStart(char, hero):
	enc_id = Enc_GetCharEnc(char)
	num_carrying = m25getset.GetObjectVar(hero, '_rescue_carrying', default=0)
	if num_carrying+1 < js.Object_GetAttr(hero, 'strength')/3:
		js.Mission_StatusText('Hero cannot carry any more!')
		RescueBase_SetupCarry(enc_id, char)
		return
	m25ai.AIDisable(char)
	#js.AI_Animate(char, 'land_back', '', enc_id, 1)
	js.Trigger_Move(char, 'hover')
	js.AI_Animate(hero, 'pick_up_object', '', enc_id, 0)
	#make them buoyant, so that they don't fall to the ground when pasted to flying heroes
	js.Object_SetSecondaryState(char, js.SCSTATE_BUOYANT, 100000, js.PERMANENT_STATE)
	m25event.SetTimer(2, 'RescueBase_CarryMoveWith', char, hero, user=enc_id)
	m25event.SetTimer(3, 'RescueBase_CarryCheckStop', char, hero, user=enc_id)
	if not js.Marker_Exists('%i_carrymarker'%enc_id): js.Marker_Spawn('%i_carrymarker'%enc_id, js.Get_ObjectPos(hero))
	js.Mission_AddArrowPos('%s_carryarrow'%char, js.Get_MarkerPos(GetEncVar(enc_id, 'congregate')), js.ARROW_RED)
	m25getset.SetObjectVar(hero, '_rescue_carrying', num_carrying+1)
	PlayEncCutscene(enc_id, 'carryallycs', charlist = [[char,'ally'], [hero, 'hero']], in_order=0)
	
def RescueBase_CarryMoveWith(event):
	enc_id = int(event.user)
	object = event.object
	h = event.string
	if m25getset.GetObjectVar(object, 'enc_carrydone', default=0): return
	pos = js.Get_ObjectPos(h)
	heading = js.Object_GetOrientation(h)[1] * 5
	pos = (pos[0]+heading, pos[1]+heading, pos[2])
	js.Marker_SetPos('%i_carrymarker'%enc_id, pos)
	js.Object_MoveToMarker(object, '%i_carrymarker'%enc_id)
	m25event.SetTimer(0.01, 'RescueBase_CarryMoveWith', object, h, user=enc_id)

def RescueBase_CarryCheckStop(event):
	enc_id = int(event.user)
	marker = GetEncVar(enc_id, 'carryto')
	h = event.string
	object = event.object
	if m25distance.GetDistanceToMarker(h, marker) < 50:
		m25getset.SetObjectVar(object, 'enc_carrydone', 1)
		js.AI_Animate(object, 'get_up_back', '', enc_id, 0)
		js.AI_Animate(h, 'drop_pole', '', enc_id, 0)
		RescueBase_AllyCongregates(object, enc_id)
		js.Mission_RemoveArrow('%s_carryarrow'%object)
		js.Object_SetPrimaryState(object, js.PCSTATE_NORMAL, 0)
		js.Object_SetSecondaryState(object, js.SCSTATE_BUOYANT, 0, js.REMOVE_STATE)
		num_carrying = m25getset.GetObjectVar(h, '_rescue_carrying', default=0)
		if num_carrying < 1: num_carrying = 1
		m25getset.SetObjectVar(h, '_rescue_carrying', num_carrying-1)
		PlayEncCutscene(enc_id, 'carriedtosafetycs', charlist = [[object,'ally'], [h, 'hero']],in_order=0)
	else:
		m25event.SetTimer(1, 'RescueBase_CarryCheckStop', object, h, user=enc_id)

	
	
#-------------------------------------------------------------------------
#
# Custom Heroes Return Item To Allies
#
#-------------------------------------------------------------------------		
	
def EncCustom_heroes_return_item_to_allies(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['item_return_setup', 0],
		['item_return_complete', 0],
	]
	encounter['custom_check_end'] = ['EncCustom_ReturnItem_CheckEnd']
	encounter['on_hero_gets_item'] = ['EncCustom_ReturnItem_HeroPickup']
	encounter['clean_up'] = ['EncCustom_ReturnItem_CleanUp']
	return encounter

def EncCustom_ReturnItem_HeroPickup(enc_id, vals=[]):
	hero = vals
 	ally = findCharFromList(chartype='ally',enc_id=enc_id,live=1)
 	if ally not in [None,'']:
 		cshelper.addCustomCommand('return',ally,'EncCustom_ReturnItem_Return',1,4,hero)
 		cshelper.addArrow(ally+'arrow', ally, colour = js.ARROW_YELLOW)
 		m25getset.SetObjectVar(ally,'encounter',enc_id)
 		m25ai.AddGoal_MoveToTarget(ally,hero,mindist=100)
 	else:
 		m25report.Report(ENC_CHANNEL, 'HeroPickup: No ally to return stolen item to.') 
 	PlayEncCutscene(enc_id,'herogetsitemcs',charlist=[[ally,'ally',],[hero,'hero']])
 	EncCheckEnd(enc_id)
  	
def EncCustom_ReturnItem_Return(target,char):
 	enc_id = m25getset.GetObjectVar(target,'encounter',integer=1)
 	SetEncVar(enc_id,'item_return_complete',1)
 	SetEncVar(enc_id,'object_state','returned')
 	cshelper.removeAttention(target)
 	cshelper.removeArrow(target+'arrow')
 	PlayEncCutscene(enc_id,'itemreturnedcs',charlist=[[target,'ally',],[char,'hero']])
 	EncCheckEnd(enc_id)
	
def EncCustom_ReturnItem_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'item_return_complete', default = 0):
		return ['item returned', 'item recovered']
	if Enc_IsAllGone(enc_id, ['ally']):
		return ['item not returned']
	return None
				
def EncCustom_ReturnItem_CleanUp(enc_id):
	if GetEncVar(enc_id, 'item_return_complete', default = 0):
		PlayEncCutscene(enc_id, 'itemrecoveredcs')
	else:
		PlayEncCutscene(enc_id, 'itemnotreturnedcs')
	
#-------------------------------------------------------------------------
#
# Custom Interrogate Villain
#
#-------------------------------------------------------------------------		
				
def EncCustom_villains_can_be_interrogated(enc_id, parameters):
	encounter = {}
	encounter = CombineEncounters(encounter, EncInterrogate('villain'))
	return encounter		

	


#-------------------------------------------------------------------------
#
# Custom Rivals Fight Villains
#
#-------------------------------------------------------------------------	

def EncCustom_rivals_fight_villains(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachFightForTeam("rival",3)',
		'EachFightForTeam("cohort",3)',
	]
	encounter['custom_check_end'] = ['EncCustom_RivalsFight_CheckEnd']
	return encounter
	
def EncCustom_RivalsFight_CheckEnd(enc_id):
	if Enc_IsAllKO(enc_id, ['rival','cohort']):
		return ['rivals gone']
	return None		


#-------------------------------------------------------------------------
#
# Custom Timer
#
#-------------------------------------------------------------------------		
	
# Timers cause the immediate end flag to be set so the encounter will end... immediately
# To register for the timer up event: add encounter['on_custom_timer_up'] = ['My function'] to your encounter


def EncCustom_TimerBase(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['custom_timer', EncCustom_GetParameter(parameters, 'time', default=300)],
		['custom_timer_message', EncCustom_GetParameter(parameters, 'timermessage', default='Timer:')],
		['custom_time_up', 0],
		['custom_time_stopped', 0],
	]
	#encounter ['start'] = ['EncCustom_Timer_Start']
	encounter['custom_check_end'] = ['EncCustom_Timer_CheckEnd']
	return encounter
	
def EncCustom_countdown(enc_id, parameters):
	encounter = EncCustom_TimerBase(enc_id, parameters)
	encounter ['start_function'] = ['EncCustom_Timer_Start']
	return encounter
	
def EncCustom_immediate_countdown(enc_id, parameters):
	encounter = EncCustom_TimerBase(enc_id, parameters)
	encounter ['setup'] = ['EncCustom_Timer_Start']
	return encounter

def EncCustom_timed_encounter(enc_id, parameters):
	encounter = EncCustom_TimerBase(enc_id, parameters)
	encounter['state'] = [
		['custom_time_must_expire', 1],
		['custom_timer_display', 0],
	]
	encounter ['start_function'] = ['EncCustom_Timer_Start']
	return encounter
	
def EncCustom_immediate_timed_encounter(enc_id, parameters):
	encounter = EncCustom_TimerBase(enc_id, parameters)
	encounter['state'] = [
		['custom_time_must_expire', 1],
		['custom_timer_display', 0],
	]
	encounter ['setup'] = ['EncCustom_Timer_Start']
	return encounter	
	
def EncCustom_Timer_Start(event):
	enc_id = 0
	if type(event) == types.IntType: enc_id = event
	else: enc_id = int(event.user)
	timer = float(GetEncVar(enc_id, 'custom_timer',default=60))
	if GetEncVar(enc_id, 'custom_timer_display', default=1):
		js.Mission_DisplayTimer(GetEncVar(enc_id, 'custom_timer_message',default='Timer:'), timer)
	for e in GetEncVar(enc_id, 'on_custom_timer_up',default=[]): EncAddReg(enc_id, e, '_customtimerup')
	for e in GetEncVar(enc_id, 'on_custom_timer_stopped',default=[]): EncAddReg(enc_id, e, '_customtimerstopped')
	m25event.SetTimer(timer,'EncCustom_Timer_Done',user=enc_id)
	
def EncCustom_Timer_StopEvent(enc_id):
	SetEncVar(enc_id, 'custom_time_stopped', 1)
	js.Mission_DisplayTimer(GetEncVar(enc_id, 'custom_timer_message',default='Timer:'), -1.0)
	EncEventCallback(enc_id, '_customtimerstopped', [])	
	
def EncCustom_Timer_Done(event):
	enc_id = int(event.user)
	#EncCustom_SetImmediateEnd(enc_id)
	SetEncVar(enc_id, 'custom_time_up', 1)
	EncEventCallback(enc_id, '_customtimerup', [])
	EncCheckEnd(enc_id)

def EncCustom_Timer_CheckEnd(enc_id):
	if GetEncVar(enc_id, 'custom_time_up', default=0):
		PlayEncCutscene(enc_id, 'timerexpiredcs')
		EncCustom_SetImmediateEnd(enc_id)
		return ['timer expired']
	if not GetEncVar(enc_id, 'custom_time_must_expire', default=0):
		return ['timer still going']
	return None

#-------------------------------------------------------------------------
#
# Custom Villains Mug Innocent Allies
#
#-------------------------------------------------------------------------		
	
def EncCustom_villains_attack_innocent_allies(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachAttackTarget("minion","ally")',
		'EachAttackTarget("villain","ally")',
	]	
	return encounter

#-------------------------------------------------------------------------
#
# Custom Villains Attack Object
#
#-------------------------------------------------------------------------		
			
def EncCustom_villains_attack_object(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachFightForTeam("object", 1)',
		'EachAttackTarget("minion","object")',
		'EachAttackTarget("villain","object")',
		'EachMoveTo("villain","object")',
		'EachMoveTo("minion","object")',
	]
	encounter['custom_check_end'] = ['EncCustom_VillainsAttackObject_CheckEnd']
	return encounter

def EncCustom_VillainsAttackObject_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'object'):
		EncCustom_FailObjective(enc_id)
		PlayEncCutscene(enc_id, 'objectslostcs')	
		return ['objects lost']
	else:
		if Enc_IsAllGone(enc_id, ['villain','minion']):
			PlayEncCutscene(enc_id, 'objectssavedcs')
			return ['objects saved']
		else:
			return None

#-------------------------------------------------------------------------
#
# Custom Villains Fight
#
#-------------------------------------------------------------------------	

def EncCustom_villains_fight_heroes(enc_id, parameters):
	encounter = {}
	encounter['goals'] = [
		'EachFightForTeam("minion",2)',
		'EachFightForTeam("villain",2)',
	]
	encounter['custom_check_end'] = ['EncCustom_VillainsFight_CheckEnd']
	return encounter
	
def EncCustom_VillainsFight_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, ['villain','minion']):
		return ['villains gone']
	return None

#-------------------------------------------------------------------------
#
# Custom Villains Flee
#
#-------------------------------------------------------------------------	

def EncCustom_villains_flee(enc_id, parameters):
	encounter = {}
	#encounter['markers'] = [['end', 'end', 'edge'],]
	encounter['custom_check_end'] = ['EncCustom_VillainsFlee_CheckEnd']
	encounter['clean_up'] = ['EncCustom_VillainsFlee_CleanUp']
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='villain',marker_id='end',marker_prefix='end',mass_flee=1))
	return encounter

def EncCustom_VillainsFlee_CheckEnd(enc_id, val=[]):
	return EncCustom_Flee_CheckEnd(enc_id, name='villain')
	
def EncCustom_VillainsFlee_CleanUp(enc_id):
	EncCustom_Flee_CleanUp(enc_id, 'villain')	
	
def EncCustom_minions_flee(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_MinionsFlee_CheckEnd']
	encounter['clean_up'] = ['EncCustom_MinionsFlee_CleanUp']
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='minion',marker_id='end',marker_prefix='end',mass_flee=1))
	return encounter
	
def EncCustom_MinionsFlee_CheckEnd(enc_id, val=[]):
	return EncCustom_Flee_CheckEnd(enc_id, name=['minion'])
	
def EncCustom_MinionsFlee_CleanUp(enc_id):
	EncCustom_Flee_CleanUp(enc_id, ['minion'])		
	
	
def EncCustom_villains_and_minions_flee(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_VillainsMinionsFlee_CheckEnd']
	encounter['clean_up'] = ['EncCustom_VillainsMinionsFlee_CleanUp']
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='minion',marker_id='end',marker_prefix='end',mass_flee=1))
	encounter = CombineEncounters(encounter, EncFleeToMarker(group='villain',marker_id='end',marker_prefix='end',mass_flee=1))
	return encounter
	
def EncCustom_VillainsMinionsFlee_CheckEnd(enc_id, val=[]):
	return EncCustom_Flee_CheckEnd(enc_id, name=['minion','villain'])
	
def EncCustom_VillainsMinionsFlee_CleanUp(enc_id):
	EncCustom_Flee_CleanUp(enc_id, ['minion', 'villain'])		
	
def EncCustom_Flee_CheckEnd(enc_id, name='villain'):
	if Enc_IsAllGone(enc_id, name):
		cs, rv = None, None
		if Enc_IsAllEscaped(enc_id, name):
			EncCustom_FailObjective(enc_id)
			cs, rv = 'allescapedcs', ["all escaped"]
		elif Enc_IsSomeEscaped(enc_id, name):
			EncCustom_FailObjective(enc_id)
			cs, rv = 'someescapedcs', ['some escaped']
		else:
			cs, rv = 'noneescapedcs', ['none escaped']	
		if not GetEncVar(enc_id, '%ssflee_done'%name, default=0):
			#PlayEncCutscene(enc_id, cs)
			SetEncVar(enc_id,'%ssflee_done'%name,1)
		return rv
	return None

def EncCustom_Flee_CleanUp(enc_id, name='villain'):
	if Enc_IsAllEscaped(enc_id, name):
		cs = 'allescapedcs'
	elif Enc_IsSomeEscaped(enc_id, name):
		cs='someescapedcs'
	else:
		cs = 'noneescapedcs'
	PlayEncCutscene(enc_id, cs)
	
#-------------------------------------------------------------------------
#
# Custom Villains Hunt 
#
#-------------------------------------------------------------------------	
	
def EncCustom_villains_hunt_enemy(enc_id, parameters):
	encounter = {}
	#encounter['setup'] = ['EncCustom_Hunt_Setup']
	encounter = CombineEncounters(encounter, EncHuntEnemy('minion'))
	encounter = CombineEncounters(encounter, EncHuntEnemy('villain'))
	return encounter

def EncCustom_villains_hunt_heroes(enc_id, parameters):
	encounter = {}
	#encounter['setup'] = ['EncCustom_Hunt_Setup']
	encounter = CombineEncounters(encounter, EncHuntHero('minion'))
	encounter = CombineEncounters(encounter, EncHuntHero('villain'))
	return encounter

def EncCustom_Hunt_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	EncEnableAllChars(enc_id, 'villain')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	


#-------------------------------------------------------------------------
#
# Custom Villains Rampage
#
#-------------------------------------------------------------------------		
		
def EncCustom_villains_rampage(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'objects', group=['villain'])
	
def EncCustom_minions_rampage(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'objects',group=['minion'])	
	
def EncCustom_villains_and_minions_rampage(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'objects', group=['villain','minion'])
	
#def EncCustom_villains_rampage_immediately(enc_id, parameters):
#	return EncCustom_Rampage(enc_id, 'objects', 1)
	
def EncCustom_villains_attack_buildings(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'buildings', group=['villain'])
	
def EncCustom_minions_attack_buildings(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'buildings', group=['minion'])

def EncCustom_villains_and_minions_attack_buildings(enc_id, parameters):
	return EncCustom_Rampage(enc_id, 'buildings', group=['villain','minion'])
	
#def EncCustom_villains_immediately_attack_buildings(enc_id, parameters):
#	return EncCustom_Rampage(enc_id, 'buildings', 1)
	
def EncCustom_Rampage(enc_id, ontype='objects', group=['villain','minion'], immediate=0):
	encounter = {}
	if ontype == 'objects':
		encounter['goals'] = []
		for g in group:
			encounter['goals'].append('Enc_EachRampageObjects("%s")'%g)
	else:  # ontype == 'buildings':
		encounter['goals'] = []
		for g in group:
			encounter['goals'].append('Enc_EachRampageBuildings("%s")'%g)
		#encounter['goals'] = [
		#	'Enc_EachRampageBuildings("villain")',
		#	'Enc_EachRampageBuildings("minion")',
		#]
	if immediate:
		encounter['setup'] = ['EncCustom_Rampage_Setup']
	encounter['custom_check_end'] = ['EncCustom_VillainsRampage_CheckEnd']
	return encounter
	
def EncCustom_Rampage_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	EncEnableAllChars(enc_id, 'villain')
	EncEnableAllChars(enc_id, 'ally')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	m25ai.AddGoal_MoveToBuildings(m25ai.Group_GetLeader(EncGetGroupName(enc_id,'minion')))

def EncCustom_VillainsRampage_CheckEnd(enc_id):
	if Enc_IsAllKO(enc_id, ['villain','minion']):
		return ['rampage over']
	return None

#-------------------------------------------------------------------------
#
# Custom Villains Start Immediately
#
#-------------------------------------------------------------------------		
		
def EncCustom_villains_start_immediately(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_VillainsStartImmediately_Setup']
	return encounter
	
def EncCustom_VillainsStartImmediately_Setup(enc_id):
	EncEnableAllChars(enc_id, 'villain')	
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	
def EncCustom_minions_start_immediately(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_MinionsStartImmediately_Setup']
	return encounter
	
def EncCustom_MinionsStartImmediately_Setup(enc_id):
	EncEnableAllChars(enc_id, 'minion')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'minion'))
	
def EncCustom_rivals_start_immediately(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_RivalsStartImmediately_Setup']
	return encounter
	
def EncCustom_RivalsStartImmediately_Setup(enc_id):
	EncEnableAllChars(enc_id, 'rival')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'cohort'))
	
def EncCustom_cohorts_start_immediately(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_CohortsStartImmediately_Setup']
	return encounter
	
def EncCustom_CohortsStartImmediately_Setup(enc_id):
	EncEnableAllChars(enc_id, 'cohort')
	m25ai.Group_PromoteLeader(EncGetGroupName(enc_id,'cohort'))

def EncCustom_allies_start_immediately(enc_id, parameters):
	encounter = {}
	encounter['setup'] = ['EncCustom_AlliesStartImmediately_Setup']
	return encounter
	
def EncCustom_AlliesStartImmediately_Setup(enc_id):
	EncEnableAllChars(enc_id, 'ally')
	
#-------------------------------------------------------------------------
#
# Custom Villains Stops at Health Level
#
#-------------------------------------------------------------------------	

def EncCustom_villain_stops_at_half_health(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['villain_stop_health', 0.5]
	]
	encounter['goals'] = ["Enc_OnEachHurt('villain', 'EncCheckEnd')",]
	encounter['custom_check_end'] = ['EncCustom_VillainHealthStop']

def EncCustom_villain_stops_at_one_quarter_health(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['villain_stop_health', 0.25]
	]
	encounter['goals'] = ["Enc_OnEachHurt('villain', 'EncCheckEnd')",]
	encounter['custom_check_end'] = ['EncCustom_VillainHealthStop']
	
def EncCustom_villain_stops_at_three_quarters_health(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['villain_stop_health', 0.75]
	]
	encounter['goals'] = ["Enc_OnEachHurt('villain', 'EncCheckEnd')",]
	encounter['custom_check_end'] = ['EncCustom_VillainHealthStop']	
	
def EncCustom_villain_stops_at_one_third_health(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['villain_stop_health', 0.33]
	]
	encounter['goals'] = ["Enc_OnEachHurt('villain', 'EncCheckEnd')",]
	encounter['custom_check_end'] = ['EncCustom_VillainHealthStop']
	
def EncCustom_villain_stops_at_two_thirds_health(enc_id, parameters):
	encounter = {}
	encounter['state'] = [
		['villain_stop_health', 0.66]
	]
	encounter['goals'] = ["Enc_OnEachHurt('villain', 'EncCheckEnd')",]
	encounter['custom_check_end'] = ['EncCustom_VillainHealthStop']
	
def EncCustom_VillainHealthStop(enc_id):
	for v in GetEncVar(enc_id, 'villain', default=[]):
		if js.Object_Exists(v) and ( js.Object_GetAttr(v,'health') / js.Object_GetAttr(v,'maxHealth') <= GetEncVar(enc_id, 'villain_stop_health', default=0) ):
			return None
	EncCustom_SetImmediateEnd(enc_id)
	PlayEncCutscene(enc_id, 'villainsstopcs')
	return ['villains hurt']


			
#=========================================================================
#
#  Custom Encounter Pieces: Storms
#
#=========================================================================

def Storm_Filter(char):
	return m25ai.ST_char(char) and js.Object_IsAlive(char) and (not m25ai.IsAIDisabled(char) or m25ai.ST_civilian(char))
	
def Storm_NoBuildingFilter(object):
	return not (js.Object_GetClass(object) & js.OC_BUILDING) and not (js.Object_GetClass(object) & js.OC_TERRAIN)
	
def Storm_BuildingFilter(object):
	return (js.Object_GetClass(object) & js.OC_BUILDING)	

def Storm_GroundFilter(char):
	return m25ai.ST_char(char) and js.Object_IsAlive(char) and not m25ai.IsAIDisabled(char) and not m25ai.ST_flying(char)

def EncCustom_Storm(enc_id, parameters, name='EncCustom_Hurricane', immediate=0):
	encounter = {}
	encounter['state'] = [
		['storm_intensity', int(EncCustom_GetParameter(parameters, 'stormintensity', default=5))],
	]
	if immediate:
		encounter['setup'] = ['%s_Start'%name]
	else:
		encounter['start_function'] = ['%s_Start'%name]
	encounter['custom_check_end'] = ['EncCustom_Storm_CheckEnd']
	return encounter
	
def EncCustom_Storm_CheckEnd(enc_id):	
	return ['storm over']
	
def Storm_DestroyProxy(event):
	proxy = event.object
	js.Object_Destroy(proxy)			

def Storm_GetStrikePosition(jitter=1):
	pool = filter(Storm_Filter, js.Mission_GetDynamicObjects())
	if not pool: return None
	chosen = pool[random.randint(0,len(pool)-1)]
	position = js.Get_ObjectPos(chosen)
	if jitter:
		position = (position[0]+random.randint(-100,100), position[1]+random.randint(-100,100), position[2])
	return position	
	
def Storm_SetupStrikes(enc_id, name='Hurricane'):
	intensity = GetEncVar(enc_id, 'storm_intensity', default=1)
	for i in range(0, intensity):
		m25event.SetTimer(1*i, '%s_Strike'%name, user=enc_id)
	duration = GetEncVar(enc_id, 'storm_duration', default=1)
	if duration > 0:
		m25event.SetTimer(duration, 'Storm_Stop', user=enc_id)	
	
def Storm_GetEncId(event):
	enc_id = event
	if type(event) != types.IntType: enc_id = int(event.user)
	return enc_id		
	

#---------------------------------------------------------------------
#
# Earthquake
#
#---------------------------------------------------------------------

def EncCustom_earthquake(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_Earthquake', immediate=0)
	
def EncCustom_immediate_earthquake(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_Earthquake', immediate=1)

def EncCustom_Earthquake_Start(event):
	enc_id = Storm_GetEncId(event)
	m25event.SetTimer(0.1, 'Earthquake_Shake', user=enc_id)
	
def Earthquake_Shake(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	intensity = GetEncVar(enc_id, 'storm_intensity', default=1)
	js.Camera_Shake(random.randint(10,100)*intensity, js.Get_MarkerPos(GetEncVar(enc_id, 'start')))
	buildings = filter(Storm_BuildingFilter, js.Mission_GetObjects())
	for b in buildings:
		if random.randint(1,5)==5:
			js.Trigger_Damage(b, intensity)
	m25event.SetTimer(1, 'Earthquake_Shake', user=enc_id)
	for c in filter(Storm_GroundFilter, js.Mission_GetDynamicObjects()):
		if random.randint(1,5) == 1: ffx.fallOver(c)
		
#---------------------------------------------------------------------
#
# Hurricane
#
#---------------------------------------------------------------------

def EncCustom_hurricane(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_Hurricane', immediate=0)
	
def EncCustom_immediate_hurricane(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_Hurricane', immediate=1)

def EncCustom_Hurricane_Start(event):
	enc_id = Storm_GetEncId(event)
	intensity = GetEncVar(enc_id, 'storm_intensity', default=5)
	for i in range(0, intensity):
		m25event.SetTimer(1*i, 'EncCustom_Hurricane_FlingObject', user=enc_id)
	
def EncCustom_Hurricane_FlingObject(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	m25event.SetTimer(random.randint(1,3), 'EncCustom_Hurricane_FlingObject', user=enc_id)
	pool = filter(Storm_Filter, js.Mission_GetDynamicObjects())
	if not pool: return
	chosen = pool[random.randint(0,len(pool)-1)]
	js.Sound_PlaySound('fx_bullet_whirlwind_2', js.Get_ObjectPos(chosen), 5.0)
	#get an object not too far from one of the characters
	objects = filter(Storm_NoBuildingFilter, m25ai.AI_GetNearObjects(chosen, default=[]))
	if not objects: return
	object = objects[random.randint(0,len(objects)-1)]
	js.Object_PlayEffect(object,'effect_ffx_tornado', '', js.FX_TRACK_OBJECT_POS)
	direction=js.Vector_Sub(js.Get_ObjectPos(chosen),js.Get_ObjectPos(object))
	js.Trigger_Force(object,direction[0]*random.randint(-1,1),direction[1]*random.randint(-1,1),direction[2],100000,js.TF_ABSOLUTE)


#---------------------------------------------------------------------
#
# Ice Storm
#
#---------------------------------------------------------------------		

def EncCustom_ice_storm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_IceStorm', immediate=0)
	
def EncCustom_immediate_ice_storm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_IceStorm', immediate=1)
	
def EncCustom_IceStorm_Start(event):
	Storm_SetupStrikes(Storm_GetEncId(event), name='IceStorm')
	
def IceStorm_Strike(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	position = Storm_GetStrikePosition()
	if not position: return
	m25event.SetTimer(random.randint(1,10), 'IceStorm_Strike', user=enc_id)
	name = m25name.GetNextObjectName()
	js.Object_SpawnAt('ffx_icepatch', position, name)
	#--- OnIcePatch2 is in ffx.py ---#
	m25event.SetTimer(1,'OnIcePatch2',name,user=6)
	js.Trigger_Explosion(name,'ffx_icecloud')
	

#---------------------------------------------------------------------
#
# Meteor Storm
#
#---------------------------------------------------------------------

def EncCustom_meteor_storm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_MeteorStorm', immediate=0)
	
def EncCustom_immediate_meteor_storm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_MeteorStorm', immediate=1)
	
def EncCustom_MeteorStorm_Start(event):
	Storm_SetupStrikes(Storm_GetEncId(event), name='MeteorStorm')
	
def MeteorStorm_Strike(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	position = Storm_GetStrikePosition()
	if not position: return
	m25event.SetTimer(random.randint(3,6), 'MeteorStorm_Strike', user=enc_id)
	name = m25name.GetNextObjectName()
	js.Object_SpawnAt('ffx_fallobject', (position[0], position[1], position[2]+400), name)
	js.Object_SetAttr(name, 'minHealth', 1)
	js.Object_SetAttr(name, 'health', 2)
	js.Object_PlayEffect(name, 'effect_ffx_fieryrain', '', js.FX_LOOP | js.FX_TRACK_OBJECT_POS)
	js.Object_SetAttr(name,'minForce',1)
	js.Event_RegisterSink(js.EVENT_OBJ_HEALTHCHANGED, 'MeteorStorm_Explode', name, '', 2, enc_id, 0)
	m25event.SetTimer(1, 'MeteorStorm_Fall', name)
	
def MeteorStorm_Fall(event):
	meteor = event.object
	js.Trigger_Force(meteor,0,0,-1,js.Object_GetAttr(meteor,'minForce')*1000,js.TF_ABSOLUTE)
	
def MeteorStorm_Explode(event):
	meteor = event.object
	js.Trigger_Explosion(meteor, 'ffx_firestorm')
	m25event.SetTimer(1, 'MeteorStorm_DestroyMeteor', meteor)
	
def MeteorStorm_DestroyMeteor(event):
	meteor = event.object
	js.Object_Destroy(meteor)	
	
#---------------------------------------------------------------------
#
# Sandstorm
#
#---------------------------------------------------------------------

def EncCustom_sandstorm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_SandStorm', immediate=0)
	
def EncCustom_immediate_sandstorm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_SandStorm', immediate=1)
			
def EncCustom_Sandstorm_Start(event):
	Storm_SetupStrikes(Storm_GetEncId(event), name='Sandstorm')	
	
def Sandstorm_Strike(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	position = Storm_GetStrikePosition()
	if not position: return
	m25event.SetTimer(random.randint(1,3), 'Sandstorm_Strike', user=enc_id)
	name = m25name.GetNextObjectName()
	js.Object_SpawnAt('portal_start_proxy', position, name)
	js.Trigger_Explosion(name,'ffx_sandstorm')
	m25event.SetTimer(5, 'Storm_DestroyProxy', name)
	#strong winds may knock nearby characters over
	for c in filter(Storm_GroundFilter, m25ai.AI_GetGrid(position[0], position[2], distance=1)):
		if random.randint(1,2) == 1:
			ffx.fallOver(c)	
	
#---------------------------------------------------------------------
#
# Thunderstorm
#
#---------------------------------------------------------------------

def EncCustom_thunderstorm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_ThunderStorm', immediate=0)
	
def EncCustom_immediate_thunderstorm(enc_id, parameters):
	return EncCustom_Storm(enc_id, parameters, name='EncCustom_ThunderStorm', immediate=1)
			
def EncCustom_ThunderStorm_Start(event):
	Storm_SetupStrikes(Storm_GetEncId(event), name='ThunderStorm')
	
def ThunderStorm_Strike(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	position = Storm_GetStrikePosition(jitter=0)
	if not position: return
	m25event.SetTimer(random.randint(1,3), 'ThunderStorm_Strike', user=enc_id)
	name = m25name.GetNextObjectName()
	js.Object_SpawnAt('portal_start_proxy', position, name)
	js.Trigger_Explosion(name,'ffx_lightning')
	js.Object_PlayEffect(name,'effect_ffx_thunder','')
	m25event.SetTimer(5, 'Storm_DestroyProxy', name)	

#---------------------------------------------------------------------
#
# Tornado
#
#---------------------------------------------------------------------

def DoEncTornado(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], duration='-1', intensity='5'):
	encounter = EncTornado(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives, duration, intensity)
	return SetupEncounter(enc_id, encounter)
	
def EncCustom_tornado(enc_id, parameters):
	encounter = {}
	encounter['markers'] = [ ['tornado', 'tornado', 'dist300'], ['end', 'end', 'edge'],]
	encounter = CombineEncounters(encounter, EncCustom_Storm(enc_id, parameters, 'EncCustom_Tornado', 0))
	return encounter
	
def EncCustom_Tornado_Start(event):
	enc_id = int(event.user)
	location = GetEncVar(enc_id, 'tornado')  #start marker
	destination = GetEncVar(enc_id, 'end')  #destination marker
	js.Object_Spawn('%i_tornado'%enc_id, 'ffx_tiggot', location)
	js.Object_PlayEffect('%i_tornado'%enc_id,'effect_ffx_tornado', '', js.FX_TRACK_OBJECT_POS | js.FX_LOOP)
	js.Object_SetAttr('%i_tornado'%enc_id, 'speed', 1)
	m25event.SetTimer(1, 'Tornado_Strike', user=enc_id)
	m25ai.AddGoal_MoveToMarker('%i_tornado'%enc_id, destination, fn='Tornado_BlowOut(%i)'%enc_id)
	m25team.SetTeam('%i_tornado'%enc_id, 0)
	m25ai.SetupAI('%i_tornado'%enc_id, force=1)

def Tornado_Strike(event):
	enc_id = int(event.user)
	if IsEncounterComplete(enc_id): return
	js.Object_SetAttr('%i_tornado'%enc_id, 'speed', random.randint(1,3))
	m25event.SetTimer(1, 'Tornado_Strike', user=enc_id)
	js.Trigger_Explosion('%i_tornado'%enc_id, 'ffx_cold_explosion')

def Tornado_BlowOut(tornado, enc_id):
	js.Object_Destroy(tornado)
	PlayEncCutscene(enc_id, 'stormendedcs', return_function='EncCheckEnd(%i)'%enc_id)	


	
	
#---------------------------------------------------------------------
#
# Waves of Minions
#
#---------------------------------------------------------------------

def EncCustom_waves_of_minions(enc_id, parameters):
	encounter = {}
	minions = EncCustom_GetParameter(parameters,'minions',default=[])
	minion_templates = EncCustom_Waves_GetMinionTemplates(minions)
	encounter['minion_templates'] = minion_templates
	encounter['number_of_waves'] = EncCustom_GetParameter(parameters,'numberofwaves',default=-1)
	encounter['time_between_waves'] = EncCustom_GetParameter(parameters,'timebetweenwaves',default=-1)
	encounter['start_function'] = ['EncCustom_Waves_Start']
	encounter['custom_check_end'] = ['EncCustom_Waves_CheckEnd']
	return encounter
	
def EncCustom_Waves_Start(event):
	enc_id = int(event.user)
	timer = GetEncVar(enc_id, 'time_between_waves', default=0)
	if timer > 0:
		m25event.SetTimer(timer, 'EncCustom_Waves_Spawn', user=enc_id)
		
def EncCustom_Waves_Spawn(event):
	enc_id = int(event.user)	
	waves = GetEncVar(enc_id, 'number_of_waves', default=-1)
	if waves == 0: return  #no more waves to spawn
	if waves > 0:
		SetEncVar(enc_id, 'number_of_waves', waves-1)
	EncSpawnGroup(enc_id, 'minion', GetEncVar(enc_id, 'minion_templates',default=[]), disable=0)
	PlayEncCutscene(enc_id, 'wavespawnscs')
	timer = GetEncVar(enc_id, 'time_between_waves', default=0)
	if timer > 0:
		m25event.SetTimer(timer, 'EncCustom_Waves_Spawn', user=enc_id)
	
def EncCustom_Waves_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'minion'):
		waves = GetEncVar(enc_id, 'number_of_waves', default=-1)
		if waves > 0:
			if GetEncVar(enc_id, 'time_between_waves', default=0) <= 0:		
				m25event.SetTimer(0.5, 'EncCustom_Waves_Spawn', user=enc_id)
				return None
		elif waves == 0:
			PlayEncCutscene(enc_id, 'wavescompletecs')
			return ['waves complete']
		else: #infinite number of waves
			return None

#---------------------------------------------------------------------
#
# Villains Stop When Half Are Defeated
#
#---------------------------------------------------------------------

def EncCustom_villains_stop_when_half_are_defeated(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_VillainsStopHalf_CheckEnd']
	return encounter
	
def EncCustom_VillainsStopHalf_CheckEnd(enc_id):
	total = GetEncVar(enc_id, 'totalvillains')
	koed = GetEncVar(enc_id, 'villains_koed')
	if koed/total < 0.5:
		PlayEncCutscene(enc_id, 'villainshalfdefeatedcs')
		EncCustom_SetImmediateEnd(enc_id)
		return ['villains defeated']

#---------------------------------------------------------------------
#
# Villains Stop When One Is Defeated
#
#---------------------------------------------------------------------

def EncCustom_villains_stop_when_one_is_defeated(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_VillainsStopOne_CheckEnd']
	return encounter
	
def EncCustom_VillainsStopOne_CheckEnd(enc_id):
	koed = GetEncVar(enc_id, 'villains_koed')
	if koed >= 1:
		PlayEncCutscene(enc_id, 'villaindowncs')
		EncCustom_SetImmediateEnd(enc_id)
		return ['villains defeated']		
		
#---------------------------------------------------------------------
#
# Villains Stop When Minions Defeated
#
#---------------------------------------------------------------------

def EncCustom_villains_stop_when_minions_are_defeated(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_VillainsStopMinions_CheckEnd']
	return encounter
	
def EncCustom_VillainsStopMinions_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'minion'):
		PlayEncCutscene(enc_id, 'minionsdefeatedcs')
		EncCustom_SetImmediateEnd(enc_id)
		return ['villains defeated']

#---------------------------------------------------------------------
#
# Minions Stop When Villains Defeated
#
#---------------------------------------------------------------------

def EncCustom_minions_stop_when_villains_are_defeated(enc_id, parameters):
	encounter = {}
	encounter['custom_check_end'] = ['EncCustom_MinionsStopVillains_CheckEnd']
	return encounter
	
def EncCustom_MinionsStopVillains_CheckEnd(enc_id):
	if Enc_IsAllGone(enc_id, 'villain'):
		PlayEncCutscene(enc_id, 'villainsdefeatedcs')
		EncCustom_SetImmediateEnd(enc_id)
		return ['villains defeated']
